#include "stdafx.h"
#include <malloc.h>
#include <memory.h>

#include <Log/MmcLog/MmcLog.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterSet.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterStack.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/BaseClasses/Command_DCS.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/Command_DCS_Epos4.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenDriveDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsEpos4Def.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/ProtocolStack/Command_PS_CANopen.h>
#include <CommunicationModel/ProtocolStack/BaseClasses/ProtocolStackManagerBase.h>

#include <CommunicationModel/DeviceCommandSet/Classes/Gateway/Epos4/GatewayEpos4ToCANopen.h>

CGatewayEpos4ToCANopen::CGatewayEpos4ToCANopen()
{
    m_pCommand_InitiateSDODownload = NULL;
    m_pCommand_DownloadSDOSegment = NULL;
    m_pCommand_InitiateSDOUpload = NULL;
    m_pCommand_UploadSDOSegment = NULL;
    m_pCommand_SendNMTService = NULL;
    m_pCommand_SendCANFrame = NULL;
    m_pCommand_ReadCANFrame = NULL;
    m_pCommand_RequestCANFrame = NULL;
    m_pCommand_AbortSDOTransfer = NULL;
    m_pCommand_SendLSSFrame = NULL;
    m_pCommand_ReadLSSFrame = NULL;

    InitCommands();
    InitErrorHandling();
}

CGatewayEpos4ToCANopen::~CGatewayEpos4ToCANopen()
{
    DeleteCommands();
}

CGateway* CGatewayEpos4ToCANopen::Clone()
{
    CGatewayEpos4ToCANopen* pClonedGateway;

    pClonedGateway = new CGatewayEpos4ToCANopen();
    *pClonedGateway = *this;

    return pClonedGateway;
}

CGatewayEpos4ToCANopen& CGatewayEpos4ToCANopen::operator=(CGatewayEpos4ToCANopen& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayDCStoPS*)this) = *((CGatewayDCStoPS*)&p_rOther);
    }

    return *this;
}

BOOL CGatewayEpos4ToCANopen::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CCommand_DCS* pCommand_DCS(0);
    CProtocolStackManagerBase* pProtocolStackManager(0);

    if(CheckLayers(p_pCommand, p_pManager))
    {
        pCommand_DCS = (CCommand_DCS*)p_pCommand;
        pProtocolStackManager = (CProtocolStackManagerBase*)p_pManager;

        if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_DCS)))
        {
            CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_DCS), p_pCommand->GetLogMessage(_T("CGatewayEpos4ToCANopen")));
        }

        switch(p_pCommand->GetCommandId())
        {
            //Object dictionary access commands
            case EPOS4_READ_OBJECT: oResult = Process_ReadObject(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_WRITE_OBJECT: oResult = Process_WriteObject(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_INITIATE_SEGMENT_READ: oResult = Process_InitiateSegmentedRead(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_INITIATE_SEGMENT_WRITE: oResult = Process_InitiateSegmentedWrite(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_SEGMENT_READ: oResult = Process_SegmentRead(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_SEGMENT_WRITE: oResult = Process_SegmentWrite(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_ABORT_TRANSFER: oResult = Process_AbortSegmentedTransfer(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;

            //Network management commands
            case EPOS4_SEND_NMT_SERVICE: oResult = Process_SendNMTService(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;

            //General gateway
            case EPOS4_SEND_CAN_FRAME: oResult = Process_SendCANFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_READ_CAN_FRAME: oResult = Process_ReadCANFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_REQUEST_CAN_FRAME: oResult = Process_RequestCANFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;

            //Layer setting services
            case EPOS4_SEND_LSS_FRAME: oResult = Process_SendLSSFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            case EPOS4_READ_LSS_FRAME: oResult = Process_ReadLSSFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
        }

        if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_DCS)))
        {
            CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_DCS), p_pCommand->GetLogMessage(_T("CGatewayEpos4ToCANopen"), TRUE));
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_ReadObject(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_OBJECT_LENGTH = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //Constants PS
    const BYTE MAX_OBJECT_LENGTH = 4;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    DWORD ulObjectLength = 0;
    //Return parameter
    BYTE maxDataBuffer[MAX_OBJECT_LENGTH];

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    //Return parameter
    BOOL oExpeditedTransfer;
    BOOL oSizeIndicated;
    BYTE ubNonValidNbOfBytes;
    DWORD ulReturnDataLength;
    DWORD ulAbortCode;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_OBJECT_LENGTH, &ulObjectLength, sizeof(ulObjectLength));

        //Check max objectLength
        if(ulObjectLength > MAX_OBJECT_LENGTH) ulObjectLength = MAX_OBJECT_LENGTH;

        //Prepare data buffer
        memset(maxDataBuffer, 0, sizeof(maxDataBuffer));

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);

       //Execute command
        oResult = PS_InitiateSDOUpload(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, usIndex, ubSubIndex, &oExpeditedTransfer, &oSizeIndicated, &ubNonValidNbOfBytes, maxDataBuffer, MAX_OBJECT_LENGTH, &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Restore length
        if(oSizeIndicated)
        {
            ulReturnDataLength = MAX_OBJECT_LENGTH - ubNonValidNbOfBytes;
        }
        else
        {
            ulReturnDataLength = MAX_OBJECT_LENGTH;
        }

        //Evaluate PS return parameter
        if(oResult && !oExpeditedTransfer)
        {
            //Abort segmented transfer
            ulAbortCode = ERROR_DEVICE_CAN_GENERAL;
            PS_AbortSDOTransfer(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, usIndex, ubSubIndex, ulAbortCode);

            //Error handling
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_EXPEDITED_TRANSFER_EXPECTED, &cmdErrorInfo);
            oResult = FALSE;
        }

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxDataBuffer, ulReturnDataLength);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_WriteObject(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_DATA = 2;
    const int PARAMETER_INDEX_OBJECT_LENGTH = 3;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Constants PS
    const BYTE MAX_OBJECT_LENGTH = 4;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE maxDataBuffer[MAX_OBJECT_LENGTH];
    DWORD ulObjectLength = 0;

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oExpeditedTransfer;
    BOOL oSizeIndicated;
    BYTE ubNonValidNbOfBytes;
    //Return parameter
    DWORD ulAbortCode;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_OBJECT_LENGTH, &ulObjectLength, sizeof(ulObjectLength));

        //Check max objectLength
        if(ulObjectLength > MAX_OBJECT_LENGTH) ulObjectLength = MAX_OBJECT_LENGTH;
        memset(maxDataBuffer, 0, MAX_OBJECT_LENGTH);
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, maxDataBuffer, ulObjectLength);

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);
        oExpeditedTransfer = TRUE;
        oSizeIndicated = FALSE;
        ubNonValidNbOfBytes = 0;

       //Execute command
        oResult = PS_InitiateSDODownload(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, oExpeditedTransfer, oSizeIndicated, ubNonValidNbOfBytes, usIndex, ubSubIndex, maxDataBuffer, MAX_OBJECT_LENGTH, &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_InitiateSegmentedRead(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_COMPLETE_SIZE = 1;

    //Variables DCS
    //Parameter
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE ubNodeId = 0;
    //Return parameter

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    //Return parameter
    BOOL oExpeditedTransfer;
    BOOL oSizeIndicated;
    BYTE ubNonValidNbOfBytes;
    DWORD ulAbortCode;
    DWORD ulCompleteSize = 0;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);

        //Execute command
        oResult = PS_InitiateSDOUpload(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, usIndex, ubSubIndex, &oExpeditedTransfer, &oSizeIndicated, &ubNonValidNbOfBytes, &ulCompleteSize, sizeof(ulCompleteSize), &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Evaluate PS return parameter
        if(oResult && oExpeditedTransfer)
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_DEVICE_CAN_ACCESS, &cmdErrorInfo);
            oResult = FALSE;
        }

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COMPLETE_SIZE, &ulCompleteSize, sizeof(ulCompleteSize));

        //Unlock critical section
        if(oExpeditedTransfer || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_InitiateSegmentedWrite(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_OBJECT_LENGTH = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Variables DCS
    //Parameter
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE ubNodeId = 0;
    DWORD ulObjectLength = 0;

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oExpeditedTransfer;
    BOOL oSizeIndicated;
    BYTE ubNonValidNbOfBytes;

    //Return parameter
    DWORD ulAbortCode;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_OBJECT_LENGTH, &ulObjectLength, sizeof(ulObjectLength));

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);
        oExpeditedTransfer = FALSE;
        oSizeIndicated = TRUE;
        ubNonValidNbOfBytes = 0;

       //Execute command
        oResult = PS_InitiateSDODownload(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, oExpeditedTransfer, oSizeIndicated, ubNonValidNbOfBytes, usIndex, ubSubIndex, &ulObjectLength, sizeof(ulObjectLength), &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Unlock critical section
        if(oExpeditedTransfer || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_SegmentRead(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_CONTROL_BYTE = 0;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_CONTROL_BYTE = 1;
    const int RETURN_PARAMETER_INDEX_DATA_READ = 2;
    const int RETURN_PARAMETER_INDEX_DATA = 3;

    //Constants PS
    const BYTE MAX_SEGMENT_LENGTH = 7;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    UEpos4ControlByte controlByte;
    controlByte.bValue = 0;
    //Return parameter
    UEpos4ControlByte retControlByte;
    retControlByte.bValue = 0;
    BYTE maxDataBuffer[MAX_SEGMENT_LENGTH];

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oToggle;
    //Return parameter
    BYTE ubRetDataSize = 0;
    BOOL oRetToggle;
    BYTE ubNonValidNbOfBytes;
    BOOL oNoMoreSegments;
    DWORD ulAbortCode;
    CErrorInfo comErrorInfo;
    //Abort
    WORD usIndex(0);
    BYTE ubSubIndex(0);

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Check critical section
    if(p_pCommand && IsLocked(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_CONTROL_BYTE, &controlByte.bValue, sizeof(controlByte.bValue));

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);
        oToggle = controlByte.structure.bToggle;

       //Execute command
        oResult = PS_UploadSDOSegment(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, oToggle, &oRetToggle, &ubNonValidNbOfBytes, &oNoMoreSegments, &maxDataBuffer, MAX_SEGMENT_LENGTH, &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Restore data length
        ubRetDataSize = MAX_SEGMENT_LENGTH - ubNonValidNbOfBytes;

        //Restore control byte
        retControlByte.structure.bToggle = oRetToggle;
        retControlByte.structure.bLastSegment = (BYTE)oNoMoreSegments;

        //Evaluate PS return parameter
        if(oResult && (controlByte.structure.bToggle != retControlByte.structure.bToggle))
        {
            //Abort segmented transfer
            GetSDOUploadParameters(usIndex, ubSubIndex);
            ulAbortCode = ERROR_DEVICE_CAN_TOGGLE;
            PS_AbortSDOTransfer(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, usIndex, ubSubIndex, ulAbortCode);

            //Error handling
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_DEVICE_CAN_TOGGLE, &cmdErrorInfo);
            oResult = FALSE;
        }

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_CONTROL_BYTE, &retControlByte, sizeof(retControlByte));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA_READ, &ubRetDataSize, sizeof(ubRetDataSize));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxDataBuffer, ubRetDataSize);

        //Unlock critical section
        if(oNoMoreSegments || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_SegmentWrite(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_CONTROL_BYTE = 0;
    const int PARAMETER_INDEX_DATA = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_CONTROL_BYTE = 1;
    const int RETURN_PARAMETER_INDEX_DATA_WRITTEN = 2;

    //Constants PS
    const BYTE MAX_SEGMENT_LENGTH = 7;

    //Variables DCS
    //Parameter
    BYTE ubDataSize = 0;
    BYTE ubNodeId = 0;
    UEpos4ControlByte controlByte;
    controlByte.bValue = 0;
    BYTE maxDataBuffer[MAX_SEGMENT_LENGTH];
    //Return parameter
    UEpos4ControlByte retControlByte;
    retControlByte.bValue = 0;
    DWORD ulDataWritten = 0;

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oToggle;
    BYTE ubNonValidNbOfBytes;
    BOOL oNoMoreSegments;
    //Return parameter
    BOOL oRetToggle;
    DWORD ulAbortCode;
    CErrorInfo comErrorInfo;
    //Abort
    WORD usIndex(0);
    BYTE ubSubIndex(0);

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Check critical section
    if(p_pCommand && IsLocked(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_CONTROL_BYTE, &controlByte.bValue, sizeof(controlByte.bValue));

        //Check max segment length
        ubDataSize = (BYTE)p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
        if(ubDataSize > MAX_SEGMENT_LENGTH)
        {
            controlByte.structure.bLastSegment = FALSE;
            ubDataSize = MAX_SEGMENT_LENGTH;
        }
        memset(maxDataBuffer, 0, MAX_SEGMENT_LENGTH);
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, maxDataBuffer, ubDataSize);

        //Prepare PS parameter
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, &ulCobIdServerClient);
        oToggle = controlByte.structure.bToggle;
        ubNonValidNbOfBytes = MAX_SEGMENT_LENGTH - ubDataSize;
        oNoMoreSegments = controlByte.structure.bLastSegment;

       //Execute command
        oResult = PS_DownloadSDOSegment(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, ulCobIdServerClient, oToggle, ubNonValidNbOfBytes, oNoMoreSegments, maxDataBuffer, MAX_SEGMENT_LENGTH, &oRetToggle, &ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Restore control byte
        retControlByte.structure.bToggle = oRetToggle;
        retControlByte.structure.bLastSegment = controlByte.structure.bLastSegment;
        ulDataWritten = ubDataSize;

        //Evaluate PS return parameter
        if(oResult && (controlByte.structure.bToggle != retControlByte.structure.bToggle))
        {
            //Abort segmented transfer
            GetSDODownloadParameters(usIndex, ubSubIndex);
            ulAbortCode = ERROR_DEVICE_CAN_TOGGLE;
            PS_AbortSDOTransfer(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, usIndex, ubSubIndex, ulAbortCode);

            //Error handling
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_DEVICE_CAN_TOGGLE, &cmdErrorInfo);
            oResult = FALSE;
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_CONTROL_BYTE, &retControlByte, sizeof(retControlByte));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA_WRITTEN, &ulDataWritten, sizeof(ulDataWritten));

        //Unlock critical section
        if(oNoMoreSegments || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_AbortSegmentedTransfer(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_ABORT_CODE = 2;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    DWORD ulAbortCode = 0;

    //Variables PS
    //Parameter
    DWORD ulCobIdClientServer;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    if(p_pCommand)
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Prepare PS parameter
        ulAbortCode = ERROR_DEVICE_CAN_GENERAL;
        CalculateDefaultSdoCobIds(ubNodeId, &ulCobIdClientServer, NULL);

       //Execute command
        oResult = PS_AbortSDOTransfer(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobIdClientServer, usIndex, ubSubIndex, ulAbortCode, &comErrorInfo);
        oResult = EvaluateErrorCode(oResult, ulAbortCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_SendNMTService(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_NODE_SPECIFIER = 0;
    const int PARAMETER_INDEX_CMD_SPECIFIER = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Variables DCS
    //Parameter
    WORD usNodeIdentifier = 0;
    WORD usCmdSpecifier = 0;

    //Variables PS
    //Parameter
    BYTE ubCommandSpecifier;
    BYTE ubNodeId;
    //Return parameter
    DWORD ulAbortCode = 0;
    CErrorInfo comErrorInfo;

    BOOL oResult(FALSE);

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_NODE_SPECIFIER, &usNodeIdentifier, sizeof(usNodeIdentifier));
        p_pCommand->GetParameterData(PARAMETER_INDEX_CMD_SPECIFIER, &usCmdSpecifier, sizeof(usCmdSpecifier));

        //Prepare PS parameter
        ubCommandSpecifier = (BYTE)usCmdSpecifier;
        ubNodeId = (BYTE)usNodeIdentifier;

        //Execute command
        oResult = PS_SendNMTService(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ubCommandSpecifier, ubNodeId, &comErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_SendCANFrame(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_COB_ID = 0;
    const int PARAMETER_INDEX_LENGTH = 1;
    const int PARAMETER_INDEX_DATA = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Constants PS
    const BYTE MAX_FRAME_LENGTH = 8;

    //Variables DCS
    //Parameter
    WORD usCobId = 0;
    WORD usLength = 0;
    BYTE maxFrameBuffer[MAX_FRAME_LENGTH];

    //Variables PS
    //Parameter
    DWORD ulCobId;
    BYTE ubLength;
    //Return parameter
    DWORD ulAbortCode = 0;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COB_ID, &usCobId, sizeof(usCobId));
        p_pCommand->GetParameterData(PARAMETER_INDEX_LENGTH, &usLength, sizeof(usLength));

        //Limit Length
        if(usLength > MAX_FRAME_LENGTH) usLength = MAX_FRAME_LENGTH;
        memset(maxFrameBuffer, 0, MAX_FRAME_LENGTH);
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, maxFrameBuffer, usLength);

        //Prepare PS parameter
        ulCobId = usCobId;
        ubLength = (BYTE)usLength;

       //Execute command
        oResult = PS_SendCANFrame(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobId, ubLength, maxFrameBuffer, ubLength, &comErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_ReadCANFrame(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_COB_ID = 0;
    const int PARAMETER_INDEX_LENGTH = 1;
    const int PARAMETER_INDEX_TIMEOUT = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //Constants PS
    const BYTE MAX_FRAME_LENGTH = 8;

    //Variables DCS
    //Parameter
    WORD usCobId = 0;
    WORD usLength = 0;
    DWORD ulTimeout = 0;
    BYTE maxFrameBuffer[MAX_FRAME_LENGTH];
    //Return parameter
    DWORD ulDeviceErrorCode = 0;

    //Variables PS
    //Parameter
    DWORD ulCobId;
    BYTE ubLength;
    //Return parameter
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COB_ID, &usCobId, sizeof(usCobId));
        p_pCommand->GetParameterData(PARAMETER_INDEX_LENGTH, &usLength, sizeof(usLength));
        p_pCommand->GetParameterData(PARAMETER_INDEX_TIMEOUT, &ulTimeout, sizeof(ulTimeout));

        //Limit Length
        if(usLength > MAX_FRAME_LENGTH) usLength = MAX_FRAME_LENGTH;
        memset(maxFrameBuffer, 0, MAX_FRAME_LENGTH);

        //Prepare PS parameter
        ulCobId = usCobId;
        ubLength = (BYTE)usLength;

       //Execute command
        oResult = PS_ReadCANFrame(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobId, ubLength, maxFrameBuffer, ubLength, ulTimeout, &comErrorInfo);
        ulDeviceErrorCode = comErrorInfo.GetErrorCode();

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxFrameBuffer, usLength);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_RequestCANFrame(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_COB_ID = 0;
    const int PARAMETER_INDEX_LENGTH = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //Constants PS
    const BYTE MAX_FRAME_LENGTH = 8;

    //Variables DCS
    //Parameter
    WORD usCobId = 0;
    WORD usLength = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;
    BYTE maxFrameBuffer[MAX_FRAME_LENGTH];

    //Variables PS
    //Parameter
    DWORD ulCobId;
    BYTE ubLength;
    //Return parameter
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;
    CErrorInfo cmdErrorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COB_ID, &usCobId, sizeof(usCobId));
        p_pCommand->GetParameterData(PARAMETER_INDEX_LENGTH, &usLength, sizeof(usLength));

        //Limit Length
        if(usLength > MAX_FRAME_LENGTH) usLength = MAX_FRAME_LENGTH;
        memset(maxFrameBuffer, 0, MAX_FRAME_LENGTH);

        //Prepare PS parameter
        ulCobId = usCobId;
        ubLength = (BYTE)usLength;

       //Execute command
        oResult = PS_RequestCANFrame(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, ulCobId, ubLength, maxFrameBuffer, ubLength, &comErrorInfo);
        ulDeviceErrorCode = comErrorInfo.GetErrorCode();

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxFrameBuffer, ubLength);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_SendLSSFrame(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_DATA = 0;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Constants PS
    const BYTE MAX_FRAME_LENGTH = 8;

    //Variables DCS
    //Parameter
    DWORD ulDataBufferLength = 0;
    void* pDataBuffer = NULL;

    //Variables PS
    //Return parameter
    DWORD ulErrorCode = 0;
    CErrorInfo comErrorInfo;

    BOOL oResult = FALSE;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Prepare data buffer
        ulDataBufferLength = p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
        if(ulDataBufferLength > MAX_FRAME_LENGTH) ulDataBufferLength = MAX_FRAME_LENGTH;
        if(ulDataBufferLength > 0)
        {
            pDataBuffer = malloc(ulDataBufferLength);
            memset(pDataBuffer, 0, ulDataBufferLength);
        }

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, pDataBuffer, ulDataBufferLength);

       //Execute command
        oResult = PS_SendLSSFrame(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, pDataBuffer, ulDataBufferLength, &comErrorInfo);
        if(!oResult) ulErrorCode = comErrorInfo.GetErrorCode();

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulErrorCode, sizeof(ulErrorCode));

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::Process_ReadLSSFrame(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_TIMEOUT = 0;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //Constants PS
    const BYTE MAX_FRAME_LENGTH = 8;

    //Variables DCS
    //Parameter
    WORD usTimeout;
    //Return parameter
    DWORD ulErrorCode = 0;
    DWORD ulDataBufferLength = 0;
    void* pDataBuffer = NULL;

    CErrorInfo comErrorInfo;
    BOOL oResult = FALSE;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Prepare data buffer
        ulDataBufferLength = MAX_FRAME_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);
        memset(pDataBuffer, 0, ulDataBufferLength);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_TIMEOUT, &usTimeout, sizeof(usTimeout));

       //Execute command
        oResult = PS_ReadLSSFrame(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, usTimeout, pDataBuffer, ulDataBufferLength, &comErrorInfo);
        if(!oResult) ulErrorCode = comErrorInfo.GetErrorCode();

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &comErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulErrorCode, sizeof(ulErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pDataBuffer, ulDataBufferLength);

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

void CGatewayEpos4ToCANopen::InitCommands()
{
    DeleteCommands();

    //Init InitiateSDODownload
    m_pCommand_InitiateSDODownload = new CCommand_PS_CANopen();
    m_pCommand_InitiateSDODownload->InitCommand(CANOPEN_INITIATE_SDO_DOWNLOAD);
    //Init DownloadSDOSegment
    m_pCommand_DownloadSDOSegment = new CCommand_PS_CANopen();
    m_pCommand_DownloadSDOSegment->InitCommand(CANOPEN_DOWNLOAD_SDO_SEGMENT);
    //Init InitiateSDOUpload
    m_pCommand_InitiateSDOUpload = new CCommand_PS_CANopen();
    m_pCommand_InitiateSDOUpload->InitCommand(CANOPEN_INITIATE_SDO_UPLOAD);
    //Init UploadSDOSegment
    m_pCommand_UploadSDOSegment = new CCommand_PS_CANopen();
    m_pCommand_UploadSDOSegment->InitCommand(CANOPEN_UPLOAD_SDO_SEGMENT);
    //Init SendNMTService
    m_pCommand_SendNMTService = new CCommand_PS_CANopen();
    m_pCommand_SendNMTService->InitCommand(CANOPEN_SEND_NMT_SERVICE);
    //Init SendCANFrame
    m_pCommand_SendCANFrame = new CCommand_PS_CANopen();
    m_pCommand_SendCANFrame->InitCommand(CANOPEN_SEND_CAN_FRAME);
    //Init ReadCANFrame
    m_pCommand_ReadCANFrame = new CCommand_PS_CANopen();
    m_pCommand_ReadCANFrame->InitCommand(CANOPEN_READ_CAN_FRAME);
    //Init RequestCANFrame
    m_pCommand_RequestCANFrame = new CCommand_PS_CANopen();
    m_pCommand_RequestCANFrame->InitCommand(CANOPEN_REQUEST_CAN_FRAME);
    //Init AbortSDOTransfer
    m_pCommand_AbortSDOTransfer = new CCommand_PS_CANopen();
    m_pCommand_AbortSDOTransfer->InitCommand(CANOPEN_ABORT_SDO_TRANSFER);
    //Init SendLSSFrame
    m_pCommand_SendLSSFrame = new CCommand_PS_CANopen();
    m_pCommand_SendLSSFrame->InitCommand(CANOPEN_SEND_LSS_FRAME);
    //Init ReadLSSFrame
    m_pCommand_ReadLSSFrame = new CCommand_PS_CANopen();
    m_pCommand_ReadLSSFrame->InitCommand(CANOPEN_READ_LSS_FRAME);
}

void CGatewayEpos4ToCANopen::DeleteCommands()
{
    if(m_pCommand_InitiateSDODownload)
    {
        delete m_pCommand_InitiateSDODownload;
        m_pCommand_InitiateSDODownload = NULL;
    }
    if(m_pCommand_DownloadSDOSegment)
    {
        delete m_pCommand_DownloadSDOSegment;
        m_pCommand_DownloadSDOSegment = NULL;
    }
    if(m_pCommand_InitiateSDOUpload)
    {
        delete m_pCommand_InitiateSDOUpload;
        m_pCommand_InitiateSDOUpload = NULL;
    }
    if(m_pCommand_UploadSDOSegment)
    {
        delete m_pCommand_UploadSDOSegment;
        m_pCommand_UploadSDOSegment = NULL;
    }
    if(m_pCommand_SendNMTService)
    {
        delete m_pCommand_SendNMTService;
        m_pCommand_SendNMTService = NULL;
    }
    if(m_pCommand_SendCANFrame)
    {
        delete m_pCommand_SendCANFrame;
        m_pCommand_SendCANFrame = NULL;
    }
    if(m_pCommand_ReadCANFrame)
    {
        delete m_pCommand_ReadCANFrame;
        m_pCommand_ReadCANFrame = NULL;
    }
    if(m_pCommand_RequestCANFrame)
    {
        delete m_pCommand_RequestCANFrame;
        m_pCommand_RequestCANFrame = NULL;
    }
    if(m_pCommand_AbortSDOTransfer)
    {
        delete m_pCommand_AbortSDOTransfer;
        m_pCommand_AbortSDOTransfer = NULL;
    }
    if(m_pCommand_SendLSSFrame)
    {
        delete m_pCommand_SendLSSFrame;
        m_pCommand_SendLSSFrame = NULL;
    }
    if(m_pCommand_ReadLSSFrame)
    {
        delete m_pCommand_ReadLSSFrame;
        m_pCommand_ReadLSSFrame = NULL;
    }
}

BOOL CGatewayEpos4ToCANopen::PS_InitiateSDODownload(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobIdClientServer, DWORD p_ulCobIdServerClient, BOOL p_oExpeditedTransfer, BOOL p_oSizeIndicated, BYTE p_ubNonValidNbOfBytes, WORD p_usIndex, BYTE p_ubSubIndex, void* p_pDataBuffer, DWORD p_ulDataBufferLength, DWORD* p_pulAbortCode, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 4;

    BOOL oResult = FALSE;
    DWORD ulRetCobId;
    WORD usRetIndex;
    BYTE uRetSubIndex;

    if(m_pCommand_InitiateSDODownload && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_InitiateSDODownload->ResetStatus();
        m_pCommand_InitiateSDODownload->SetParameterData(0, &p_ulCobIdClientServer, sizeof(p_ulCobIdClientServer));
        m_pCommand_InitiateSDODownload->SetParameterData(1, &p_ulCobIdServerClient, sizeof(p_ulCobIdServerClient));
        m_pCommand_InitiateSDODownload->SetParameterData(2, &p_oExpeditedTransfer, sizeof(p_oExpeditedTransfer));
        m_pCommand_InitiateSDODownload->SetParameterData(3, &p_oSizeIndicated, sizeof(p_oSizeIndicated));
        m_pCommand_InitiateSDODownload->SetParameterData(4, &p_ubNonValidNbOfBytes, sizeof(p_ubNonValidNbOfBytes));
        m_pCommand_InitiateSDODownload->SetParameterData(5, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSDODownload->SetParameterData(6, &p_ubSubIndex, sizeof(p_ubSubIndex));
        m_pCommand_InitiateSDODownload->SetParameterData(7, p_pDataBuffer, p_ulDataBufferLength);

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_InitiateSDODownload, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSDODownload->GetReturnParameterData(0, &ulRetCobId, sizeof(ulRetCobId));
        m_pCommand_InitiateSDODownload->GetReturnParameterData(1, &usRetIndex, sizeof(usRetIndex));
        m_pCommand_InitiateSDODownload->GetReturnParameterData(2, &uRetSubIndex, sizeof(uRetSubIndex));
        m_pCommand_InitiateSDODownload->GetReturnParameterData(3, p_pulAbortCode, sizeof(*p_pulAbortCode));

        //Get error code
        m_pCommand_InitiateSDODownload->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_DownloadSDOSegment(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobIdClientServer, DWORD p_ulCobIdServerClient, BOOL p_oToggle, BYTE p_ubNonValidNbOfBytes, BOOL p_oNoMoreSegments, void* p_pSegDataBuffer, DWORD p_ulSegDataBufferLength, BOOL* p_poRetToggle, DWORD* p_pulAbortCode, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 7;

    BOOL oResult = FALSE;
    DWORD ulRetCobId;

    if(m_pCommand_DownloadSDOSegment && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulSegDataBufferLength > MAX_CAN_DATA) p_ulSegDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_DownloadSDOSegment->ResetStatus();
        m_pCommand_DownloadSDOSegment->SetParameterData(0, &p_ulCobIdClientServer, sizeof(p_ulCobIdClientServer));
        m_pCommand_DownloadSDOSegment->SetParameterData(1, &p_ulCobIdServerClient, sizeof(p_ulCobIdServerClient));
        m_pCommand_DownloadSDOSegment->SetParameterData(2, &p_oToggle, sizeof(p_oToggle));
        m_pCommand_DownloadSDOSegment->SetParameterData(3, &p_ubNonValidNbOfBytes, sizeof(p_ubNonValidNbOfBytes));
        m_pCommand_DownloadSDOSegment->SetParameterData(4, &p_oNoMoreSegments, sizeof(p_oNoMoreSegments));
        m_pCommand_DownloadSDOSegment->SetParameterData(5, p_pSegDataBuffer, p_ulSegDataBufferLength);

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_DownloadSDOSegment, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_DownloadSDOSegment->GetReturnParameterData(0, &ulRetCobId, sizeof(ulRetCobId));
        m_pCommand_DownloadSDOSegment->GetReturnParameterData(1, p_poRetToggle, sizeof(*p_poRetToggle));
        m_pCommand_DownloadSDOSegment->GetReturnParameterData(2, p_pulAbortCode, sizeof(*p_pulAbortCode));

        //Get error code
        m_pCommand_DownloadSDOSegment->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_AbortSDOTransfer(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobIdClientServer, WORD p_usIndex, BYTE p_ubSubIndex, DWORD p_ulAbortCode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;

    if(m_pCommand_AbortSDOTransfer && p_pProtocolStackManager)
    {
        //Set parameter data
        m_pCommand_AbortSDOTransfer->ResetStatus();
        m_pCommand_AbortSDOTransfer->SetParameterData(0, &p_ulCobIdClientServer, sizeof(p_ulCobIdClientServer));
        m_pCommand_AbortSDOTransfer->SetParameterData(1, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_AbortSDOTransfer->SetParameterData(2, &p_ubSubIndex, sizeof(p_ubSubIndex));
        m_pCommand_AbortSDOTransfer->SetParameterData(3, &p_ulAbortCode, sizeof(p_ulAbortCode));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_AbortSDOTransfer, p_hPS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_AbortSDOTransfer->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_InitiateSDOUpload(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobIdClientServer, DWORD p_ulCobIdServerClient, WORD p_usIndex, BYTE p_ubSubIndex, BOOL* p_poExpeditedTransfer, BOOL* p_poSizeIndicated, BYTE* p_pubNonValidNbOfBytes, void* p_pDataBuffer, DWORD p_ulDataBufferLength, DWORD* p_pulAbortCode, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 4;

    BOOL oResult = FALSE;
    DWORD ulRetCobId;
    WORD usRetIndex;
    BYTE uRetSubIndex;

    if(m_pCommand_InitiateSDOUpload && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_InitiateSDOUpload->ResetStatus();
        m_pCommand_InitiateSDOUpload->SetParameterData(0, &p_ulCobIdClientServer, sizeof(p_ulCobIdClientServer));
        m_pCommand_InitiateSDOUpload->SetParameterData(1, &p_ulCobIdServerClient, sizeof(p_ulCobIdServerClient));
        m_pCommand_InitiateSDOUpload->SetParameterData(2, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSDOUpload->SetParameterData(3, &p_ubSubIndex, sizeof(p_ubSubIndex));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_InitiateSDOUpload, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(0, &ulRetCobId, sizeof(ulRetCobId));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(1, p_poExpeditedTransfer, sizeof(*p_poExpeditedTransfer));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(2, p_poSizeIndicated, sizeof(*p_poSizeIndicated));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(3, p_pubNonValidNbOfBytes, sizeof(*p_pubNonValidNbOfBytes));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(4, &usRetIndex, sizeof(usRetIndex));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(5, &uRetSubIndex, sizeof(uRetSubIndex));
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(6, p_pDataBuffer, p_ulDataBufferLength);
        m_pCommand_InitiateSDOUpload->GetReturnParameterData(7, p_pulAbortCode, sizeof(*p_pulAbortCode));

        //Get error code
        m_pCommand_InitiateSDOUpload->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_UploadSDOSegment(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobIdClientServer, DWORD p_ulCobIdServerClient, BOOL p_oToggle, BOOL* p_poRetToggle, BYTE* p_pubNonValidNbOfBytes, BOOL* p_poNoMoreSegments, void* p_pSegDataBuffer, DWORD p_ulSegDataBufferLength, DWORD* p_pulAbortCode, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 7;

    BOOL oResult = FALSE;
    DWORD ulRetCobId;

    if(m_pCommand_UploadSDOSegment && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulSegDataBufferLength > MAX_CAN_DATA) p_ulSegDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_UploadSDOSegment->ResetStatus();
        m_pCommand_UploadSDOSegment->SetParameterData(0, &p_ulCobIdClientServer, sizeof(p_ulCobIdClientServer));
        m_pCommand_UploadSDOSegment->SetParameterData(1, &p_ulCobIdServerClient, sizeof(p_ulCobIdServerClient));
        m_pCommand_UploadSDOSegment->SetParameterData(2, &p_oToggle, sizeof(p_oToggle));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_UploadSDOSegment, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_UploadSDOSegment->GetReturnParameterData(0, &ulRetCobId, sizeof(ulRetCobId));
        m_pCommand_UploadSDOSegment->GetReturnParameterData(1, p_poRetToggle, sizeof(*p_poRetToggle));
        m_pCommand_UploadSDOSegment->GetReturnParameterData(2, p_pubNonValidNbOfBytes, sizeof(*p_pubNonValidNbOfBytes));
        m_pCommand_UploadSDOSegment->GetReturnParameterData(3, p_poNoMoreSegments, sizeof(*p_poNoMoreSegments));
        m_pCommand_UploadSDOSegment->GetReturnParameterData(4, p_pSegDataBuffer, p_ulSegDataBufferLength);
        m_pCommand_UploadSDOSegment->GetReturnParameterData(5, p_pulAbortCode, sizeof(*p_pulAbortCode));

        //Get error code
        m_pCommand_UploadSDOSegment->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_SendNMTService(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubCommandSpecifier, BYTE p_ubNodeId, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;

    if(m_pCommand_SendNMTService && p_pProtocolStackManager)
    {
        //Set parameter data
        m_pCommand_SendNMTService->ResetStatus();
        m_pCommand_SendNMTService->SetParameterData(0, &p_ubCommandSpecifier, sizeof(p_ubCommandSpecifier));
        m_pCommand_SendNMTService->SetParameterData(1, &p_ubNodeId, sizeof(p_ubNodeId));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_SendNMTService, p_hPS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_SendNMTService->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_SendCANFrame(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobId, BYTE p_ubLength, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 8;

    BOOL oResult = FALSE;

    if(m_pCommand_SendCANFrame && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_SendCANFrame->ResetStatus();
        m_pCommand_SendCANFrame->SetParameterData(0, &p_ulCobId, sizeof(p_ulCobId));
        m_pCommand_SendCANFrame->SetParameterData(1, &p_ubLength, sizeof(p_ubLength));
        m_pCommand_SendCANFrame->SetParameterData(2, p_pDataBuffer, p_ulDataBufferLength);

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_SendCANFrame, p_hPS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_SendCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_ReadCANFrame(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobId, BYTE p_ubLength, void* p_pDataBuffer, DWORD p_ulDataBufferLength, DWORD p_ulTimeout, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 8;

    BOOL oResult = FALSE;

    if(m_pCommand_ReadCANFrame && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_ReadCANFrame->ResetStatus();
        m_pCommand_ReadCANFrame->SetParameterData(0, &p_ulCobId, sizeof(p_ulCobId));
        m_pCommand_ReadCANFrame->SetParameterData(1, &p_ubLength, sizeof(p_ubLength));
        m_pCommand_ReadCANFrame->SetParameterData(2, &p_ulTimeout, sizeof(p_ulTimeout));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_ReadCANFrame, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadCANFrame->GetReturnParameterData(0, p_pDataBuffer, p_ulDataBufferLength);

        //Get error code
        m_pCommand_ReadCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_RequestCANFrame(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulCobId, BYTE p_ubLength, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 8;

    BOOL oResult = FALSE;

    if(m_pCommand_RequestCANFrame && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_RequestCANFrame->ResetStatus();
        m_pCommand_RequestCANFrame->SetParameterData(0, &p_ulCobId, sizeof(p_ulCobId));
        m_pCommand_RequestCANFrame->SetParameterData(1, &p_ubLength, sizeof(p_ubLength));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_RequestCANFrame, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_RequestCANFrame->GetReturnParameterData(0, p_pDataBuffer, p_ulDataBufferLength);

        //Get error code
        m_pCommand_RequestCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_SendLSSFrame(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 8;

    BOOL oResult = FALSE;

    if(m_pCommand_SendLSSFrame && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_SendLSSFrame->ResetStatus();
        m_pCommand_SendLSSFrame->SetParameterData(0, p_pDataBuffer, p_ulDataBufferLength);

        //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_SendLSSFrame, p_hPS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_SendLSSFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::PS_ReadLSSFrame(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, WORD p_usTimeout, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    const DWORD MAX_CAN_DATA = 8;

    BOOL oResult = FALSE;

    if(m_pCommand_ReadLSSFrame && p_pProtocolStackManager)
    {
        //Limit data length
        if(p_ulDataBufferLength > MAX_CAN_DATA) p_ulDataBufferLength = MAX_CAN_DATA;

        //Set parameter data
        m_pCommand_ReadLSSFrame->ResetStatus();
        m_pCommand_ReadLSSFrame->SetParameterData(0, &p_usTimeout, sizeof(p_usTimeout));

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_ReadLSSFrame, p_hPS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadLSSFrame->GetReturnParameterData(0, p_pDataBuffer, p_ulDataBufferLength);

        //Get error code
        m_pCommand_ReadLSSFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::InitLayerParameterStack(CCommandRoot* p_pCommand)
{
    BOOL oResult = FALSE;
    CLayerParameterSet layerParameterSet;
    CLayerParameterStack layerParameterStack;

    if(p_pCommand && p_pCommand->GetLayerParameterStack(layerParameterStack))
    {
        //Pop DCS Layer Parameter Set
        layerParameterStack.PopLayer(DEVICE_COMMAND_SET_LAYER, layerParameterSet);

        //Set PS Layer Commands
        if(m_pCommand_InitiateSDODownload) m_pCommand_InitiateSDODownload->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_DownloadSDOSegment) m_pCommand_DownloadSDOSegment->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_InitiateSDOUpload) m_pCommand_InitiateSDOUpload->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_UploadSDOSegment) m_pCommand_UploadSDOSegment->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_AbortSDOTransfer) m_pCommand_AbortSDOTransfer->SetLayerParameterStack(layerParameterStack);

        if(m_pCommand_SendCANFrame) m_pCommand_SendCANFrame->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_RequestCANFrame) m_pCommand_RequestCANFrame->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_ReadCANFrame) m_pCommand_ReadCANFrame->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_SendNMTService) m_pCommand_SendNMTService->SetLayerParameterStack(layerParameterStack);

        if(m_pCommand_SendLSSFrame) m_pCommand_SendLSSFrame->SetLayerParameterStack(layerParameterStack);
        if(m_pCommand_ReadLSSFrame) m_pCommand_ReadLSSFrame->SetLayerParameterStack(layerParameterStack);

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::GetNodeId(CCommand_DCS* p_pCommand, BYTE& p_rubNodeId, CErrorInfo* p_pErrorInfo)
{
    const CStdString NODE_ID(_T("NodeId"));

    BOOL oResult(FALSE);

    //Get node id from command
    if(!oResult && p_pCommand)
    {
        oResult = p_pCommand->GetLayerParameter(DEVICE_COMMAND_SET_LAYER, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
    }

    //Get node id from device
    if(!oResult)
    {
        oResult = GetParameter(NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
    }

    //Check for undefined value
    if(oResult && (p_rubNodeId == NODE_ID_UNDEFINED))
    {
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className(_T("GatewayEpos4ToCANopen"));

    if(m_pErrorHandling)
    {
        // Init error producer
        errorProducer.Init(DEVICE_COMMAND_SET_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::EvaluateErrorCode(BOOL p_oResult, DWORD p_ulDeviceErrorCode, CErrorInfo* p_pComErrorInfo, CErrorInfo* p_pCmdErrorInfo)
{
    if(p_pCmdErrorInfo && p_pComErrorInfo)
    {
        if(!p_oResult)
        {
            *p_pCmdErrorInfo = *p_pComErrorInfo;
        }

        if(p_oResult && (p_ulDeviceErrorCode != 0))
        {
            p_pCmdErrorInfo->Init(p_ulDeviceErrorCode);
            p_oResult = FALSE;
        }
    }

    return p_oResult;
}

BOOL CGatewayEpos4ToCANopen::CalculateDefaultSdoCobIds(BYTE p_ubNodeId, DWORD* p_pulCobIdClientServer, DWORD* p_pulCobIdServerClient)
{
    BOOL oResult(TRUE);

    if(p_pulCobIdClientServer)
    {
        *p_pulCobIdClientServer = p_ubNodeId + TX_SDO_OFFSET;
    }

    if(p_pulCobIdServerClient)
    {
        *p_pulCobIdServerClient = p_ubNodeId + RX_SDO_OFFSET;
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::GetSDODownloadParameters(WORD& p_rusIndex, BYTE& p_rubSubIndex)
{
    BOOL oResult = FALSE;

    if(m_pCommand_InitiateSDODownload)
    {
        //Get parameter data
        m_pCommand_InitiateSDODownload->GetParameterData(5, &p_rusIndex, sizeof(p_rusIndex));
        m_pCommand_InitiateSDODownload->GetParameterData(6, &p_rubSubIndex, sizeof(p_rubSubIndex));
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToCANopen::GetSDOUploadParameters(WORD& p_rusIndex, BYTE& p_rubSubIndex)
{
    BOOL oResult = FALSE;

    if(m_pCommand_InitiateSDOUpload)
    {
        //Get parameter data
        m_pCommand_InitiateSDOUpload->GetParameterData(2, &p_rusIndex, sizeof(p_rusIndex));
        m_pCommand_InitiateSDOUpload->GetParameterData(3, &p_rubSubIndex, sizeof(p_rubSubIndex));
        oResult = TRUE;
    }

    return oResult;
}
