#include "stdafx.h"
#include <malloc.h>
#include <memory.h>

#include <Log/MmcLog/MmcLog.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterSet.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterStack.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/BaseClasses/Command_DCS.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/Command_DCS_Epos4.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsEpos4Def.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/ProtocolStack/Command_PS_MaxonSerialV2.h>
#include <CommunicationModel/ProtocolStack/BaseClasses/ProtocolStackManagerBase.h>

#include <CommunicationModel/DeviceCommandSet/Classes/Gateway/Epos4/GatewayEpos4ToMaxonSerialV2.h>

//Constants
const WORD  MAX_LENGTH_NON_SEGMENTED_TRANSFER   = 4;
const BYTE  MAX_LENGTH_SEGMENT                  = 255;

const BYTE OP_CODE_READ_OBJECT                  = 0x60;
const BYTE OP_CODE_INITIATE_SEGMENTED_READ      = 0x81;
const BYTE OP_CODE_SEGMENT_READ                 = 0x62;
const BYTE OP_CODE_WRITE_OBJECT                 = 0x68;
const BYTE OP_CODE_INITIATE_SEGMENTED_WRITE     = 0x69;
const BYTE OP_CODE_SEGMENT_WRITE                = 0x6A;
const BYTE OP_CODE_SEND_NMT_SERVICE             = 0x70;
const BYTE OP_CODE_SEND_CAN_FRAME               = 0x71;
const BYTE OP_CODE_READ_CAN_FRAME               = 0x72;
const BYTE OP_CODE_REQUEST_CAN_FRAME            = 0x72;
const BYTE OP_CODE_SEND_LSS_FRAME               = 0x7A;
const BYTE OP_CODE_READ_LSS_FRAME               = 0x7B;

CGatewayEpos4ToMaxonSerialV2::CGatewayEpos4ToMaxonSerialV2()
{
    m_pCommand_ProcessProtocol = 0;
    m_pCommand_AbortProtocol = 0;

    InitCommands();
    InitErrorHandling();
}

CGatewayEpos4ToMaxonSerialV2::~CGatewayEpos4ToMaxonSerialV2()
{
    DeleteCommands();
}

CGateway* CGatewayEpos4ToMaxonSerialV2::Clone()
{
    CGatewayEpos4ToMaxonSerialV2* pClonedGateway;

    pClonedGateway = new CGatewayEpos4ToMaxonSerialV2();
    *pClonedGateway = *this;

    return pClonedGateway;
}

CGatewayEpos4ToMaxonSerialV2& CGatewayEpos4ToMaxonSerialV2::operator=(CGatewayEpos4ToMaxonSerialV2& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayDCStoPS*)this) = *((CGatewayDCStoPS*)&p_rOther);
    }

    return *this;
}

BOOL CGatewayEpos4ToMaxonSerialV2::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CCommand_DCS* pCommand_DCS(0);
    CProtocolStackManagerBase* pProtocolStackManager(0);

    if(CGateway::ProcessCommand(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle))
    {
        if(CheckLayers(p_pCommand, p_pManager))
        {
            pCommand_DCS = (CCommand_DCS*)p_pCommand;
            pProtocolStackManager = (CProtocolStackManagerBase*)p_pManager;

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_DCS)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_DCS), p_pCommand->GetLogMessage(_T("CGatewayEpos4ToMaxonSerialV2")));
            }

            switch(p_pCommand->GetCommandId())
            {
                //Object dictionary access commands
                case EPOS4_READ_OBJECT: oResult = Process_ReadObject(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_WRITE_OBJECT: oResult = Process_WriteObject(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_INITIATE_SEGMENT_READ: oResult = Process_InitiateSegmentedRead(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_INITIATE_SEGMENT_WRITE: oResult = Process_InitiateSegmentedWrite(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_SEGMENT_READ: oResult = Process_SegmentRead(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_SEGMENT_WRITE: oResult = Process_SegmentWrite(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_ABORT_TRANSFER: oResult = Process_AbortSegmentedTransfer(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;

                //Network management commands
                case EPOS4_SEND_NMT_SERVICE: oResult = Process_SendNMTService(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;

                //Layer setting services
                case EPOS4_SEND_LSS_FRAME: oResult = Process_SendLSSFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
                case EPOS4_READ_LSS_FRAME: oResult = Process_ReadLSSFrame(pCommand_DCS, pProtocolStackManager, p_hHandle, p_hTransactionHandle); break;
            }

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_DCS)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_DCS), p_pCommand->GetLogMessage(_T("CGatewayEpos4ToMaxonSerialV2"), TRUE));
            }
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_ReadObject(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_DATA_SIZE = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //Constants PS
    const BYTE DATA_LENGTH = 4;
    const BYTE RETURN_DATA_LENGTH = 8;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    DWORD ulDataSize = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;
    BYTE maxDataBuffer[MAX_LENGTH_NON_SEGMENTED_TRANSFER];

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 0;

    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA_SIZE, &ulDataSize, sizeof(ulDataSize));

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &ubNodeId, sizeof(ubNodeId));
        CopyData(pDest, &usIndex, sizeof(usIndex));
        CopyData(pDest, &ubSubIndex, sizeof(ubSubIndex));

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_READ_OBJECT, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && (ulRetDataBufferLength != RETURN_DATA_LENGTH))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if(oResult) oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        if(oResult) oResult = CopyReturnData(maxDataBuffer, MAX_LENGTH_NON_SEGMENTED_TRANSFER, pSource, ulRetDataBufferLength);

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxDataBuffer, MAX_LENGTH_NON_SEGMENTED_TRANSFER);

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_WriteObject(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_DATA = 2;
    const int PARAMETER_INDEX_DATA_SIZE = 3;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Constants PS
    const BYTE DATA_LENGTH = 8;
    const BYTE RETURN_DATA_LENGTH = 4;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE maxDataBuffer[MAX_LENGTH_NON_SEGMENTED_TRANSFER];
    DWORD ulDataSize = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 0;
    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA_SIZE, &ulDataSize, sizeof(ulDataSize));

        //Check max objectLength
        if(ulDataSize > MAX_LENGTH_NON_SEGMENTED_TRANSFER) ulDataSize = MAX_LENGTH_NON_SEGMENTED_TRANSFER;
        memset(maxDataBuffer, 0, MAX_LENGTH_NON_SEGMENTED_TRANSFER);
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, maxDataBuffer, ulDataSize);

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &ubNodeId, sizeof(ubNodeId));
        CopyData(pDest, &usIndex, sizeof(usIndex));
        CopyData(pDest, &ubSubIndex, sizeof(ubSubIndex));
        CopyData(pDest, &maxDataBuffer, MAX_LENGTH_NON_SEGMENTED_TRANSFER);

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_WRITE_OBJECT, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && (ulRetDataBufferLength != RETURN_DATA_LENGTH))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if(oResult)
        {
            oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        }

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_InitiateSegmentedRead(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_COMPLETE_SIZE = 1;
    const int RETURN_PARAMETER_INDEX_DATA_SIZE_READ = 2;
    const int RETURN_PARAMETER_INDEX_DATA = 3;

    //Constants PS
    const BYTE DATA_LENGTH = 4;
    const BYTE MIN_RETURN_DATA_LENGTH = 8;

    //Variables DCS
    //Parameter
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE ubNodeId = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;
    void* pData = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 1;
    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;
    DWORD ulCompleteSize = 0;
    BYTE ubDataSizeRead = 0;
    DWORD ulDataSizeRead = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &ubNodeId, sizeof(ubNodeId));
        CopyData(pDest, &usIndex, sizeof(usIndex));
        CopyData(pDest, &ubSubIndex, sizeof(ubSubIndex));

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_INITIATE_SEGMENTED_READ, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && (ulRetDataBufferLength < MIN_RETURN_DATA_LENGTH))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if(oResult) oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        if(oResult) oResult = CopyReturnData(&ulCompleteSize, sizeof(ulCompleteSize), pSource, ulRetDataBufferLength);
        if(oResult) oResult = CopyReturnData(&ubDataSizeRead, sizeof(ubDataSizeRead), pSource, ulRetDataBufferLength);
        if(oResult) ulDataSizeRead = ubDataSizeRead;
        if(oResult) pData = pSource;

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COMPLETE_SIZE, &ulCompleteSize, sizeof(ulCompleteSize));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA_SIZE_READ, &ulDataSizeRead, sizeof(ulDataSizeRead));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pData, ulDataSizeRead);

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        if((!oResult) || (ulCompleteSize <= ulDataSizeRead))
        {
            PS_AbortProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle);
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_InitiateSegmentedWrite(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_INDEX = 0;
    const int PARAMETER_INDEX_SUB_INDEX = 1;
    const int PARAMETER_INDEX_DATA_SIZE = 2;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //Constants PS
    const BYTE DATA_LENGTH = 8;
    const BYTE RETURN_DATA_LENGTH = 4;

    //Variables DCS
    //Parameter
    WORD usIndex = 0;
    BYTE ubSubIndex = 0;
    BYTE ubNodeId = 0;
    DWORD ulDataSize = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 1;

    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA_SIZE, &ulDataSize, sizeof(ulDataSize));

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &ubNodeId, sizeof(ubNodeId));
        CopyData(pDest, &usIndex, sizeof(usIndex));
        CopyData(pDest, &ubSubIndex, sizeof(ubSubIndex));
        CopyData(pDest, &ulDataSize, sizeof(ulDataSize));

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_INITIATE_SEGMENTED_WRITE, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && (ulRetDataBufferLength != RETURN_DATA_LENGTH))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if(oResult)
        {
            oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        }

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        if(!oResult)
        {
            PS_AbortProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle);
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_SegmentRead(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_CONTROL_BYTE = 0;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_CONTROL_BYTE = 1;
    const int RETURN_PARAMETER_INDEX_DATA_READ = 2;
    const int RETURN_PARAMETER_INDEX_DATA = 3;

    //Constants PS
    const BYTE DATA_LENGTH = 2;
    const DWORD MIN_RETURN_DATA_LENGTH = 6;
    const DWORD MAX_RETURN_DATA_LENGTH = 262;

    //Variables DCS
    //Parameter
    BYTE ubNodeId = 0;
    UEpos4ControlByte controlByte;
    controlByte.bValue = 0;
    BYTE ubDummyByte = 0;
    //Return parameter
    DWORD ulDeviceErrorCode = 0;
    UEpos4ControlByte retControlByte;
    retControlByte.bValue = 0;
    void* pData = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 1;
    //Return parameter
    BYTE ubRetDataSize = 0;
    DWORD ulRetDataSize = 0;
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Check critical section
    if(p_pCommand && IsLocked(p_pCommand))
    {
        GetNodeId(p_pCommand, ubNodeId);

        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_CONTROL_BYTE, &controlByte.bValue, sizeof(controlByte.bValue));

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &controlByte.bValue, sizeof(controlByte.bValue));
        CopyData(pDest, &ubDummyByte, sizeof(ubDummyByte));

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_SEGMENT_READ, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && ((ulRetDataBufferLength < MIN_RETURN_DATA_LENGTH) || (ulRetDataBufferLength > MAX_RETURN_DATA_LENGTH)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if(oResult) oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        if(oResult) oResult = CopyReturnData(&ubRetDataSize, sizeof(ubRetDataSize), pSource, ulRetDataBufferLength);
        if(oResult) oResult = CopyReturnData(&retControlByte.bValue, sizeof(retControlByte.bValue), pSource, ulRetDataBufferLength);
        if(oResult) pData = pSource;

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        if (oResult)
        {
            //Restore control byte
            if(controlByte.structure.bToggle != retControlByte.structure.bToggle)
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_DEVICE_CAN_TOGGLE, &cmdErrorInfo);
                }

                oResult = FALSE;
            }
        }

        //Prepare return data
        ulRetDataSize = (BYTE)ubRetDataSize;

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_CONTROL_BYTE, &retControlByte, sizeof(retControlByte));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA_READ, &ulRetDataSize, sizeof(ulRetDataSize));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pData, ulRetDataSize);

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        if(retControlByte.structure.bLastSegment || !oResult)
        {
            PS_AbortProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle);
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_SegmentWrite(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    //Constants DCS
    const int PARAMETER_INDEX_CONTROL_BYTE = 0;
    const int PARAMETER_INDEX_DATA = 1;

    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_CONTROL_BYTE = 1;
    const int RETURN_PARAMETER_INDEX_DATA_WRITTEN = 2;

    //Constants PS
    const BYTE DATA_LENGTH = 2;
    const BYTE RETURN_DATA_LENGTH = 6;
    const BYTE RETURN_DATA_LENGTH_ERROR = 4;

    //Variables DCS
    //Parameter
    BYTE ubDataSize = 0;
    UEpos4ControlByte controlByte;
    controlByte.bValue = 0;
    void* pData = 0;
    //Return parameter
    UEpos4ControlByte retControlByte;
    retControlByte.bValue = 0;
    DWORD ulDeviceErrorCode = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength;
    BYTE ubKeepLock = 1;
    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;
    BYTE ubRetDataSize = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    //Check critical section
    if(p_pCommand && IsLocked(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_CONTROL_BYTE, &controlByte.bValue, sizeof(controlByte.bValue));

        //Check max segment length
        ubDataSize = (BYTE)p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
        if(ubDataSize > MAX_LENGTH_SEGMENT) ubDataSize = MAX_LENGTH_SEGMENT;
        pData = malloc(ubDataSize);
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, pData, ubDataSize);

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH + ubDataSize;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &ubDataSize, sizeof(ubDataSize));
        CopyData(pDest, &controlByte.bValue, sizeof(controlByte.bValue));
        CopyData(pDest, pData, ubDataSize);

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_SEGMENT_WRITE, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //Check return data size
        if(oResult && (ulRetDataBufferLength != RETURN_DATA_LENGTH) && (ulRetDataBufferLength != RETURN_DATA_LENGTH_ERROR))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_MAXON_SERIAL_V2_BAD_DATA_SIZE_RECEIVED, &comErrorInfo);
            oResult = FALSE;
        }

        //Return data
        pSource = pRetDataBuffer;
        if (ulRetDataBufferLength == RETURN_DATA_LENGTH)
        {
            if (oResult) oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
            if (oResult) oResult = CopyReturnData(&ubRetDataSize, sizeof(ubRetDataSize), pSource, ulRetDataBufferLength);
            if (oResult) oResult = CopyReturnData(&retControlByte.bValue, sizeof(retControlByte.bValue), pSource, ulRetDataBufferLength);
        }
        else if (ulRetDataBufferLength == RETURN_DATA_LENGTH_ERROR)
        {
            if (oResult) oResult = CopyReturnData(&ulDeviceErrorCode, sizeof(ulDeviceErrorCode), pSource, ulRetDataBufferLength);
        }

        //Evaluate error code
        oResult = EvaluateErrorCode(oResult, ulDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Restore control byte
        if(oResult && (controlByte.structure.bToggle != retControlByte.structure.bToggle))
        {
            if(m_pErrorHandling)
            {
                m_pErrorHandling->GetError(ERROR_DEVICE_CAN_TOGGLE, &cmdErrorInfo);
            }

            oResult = FALSE;
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_CONTROL_BYTE, &retControlByte, sizeof(retControlByte));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA_WRITTEN, &ubRetDataSize, sizeof(ubRetDataSize));

        //Free data buffer
        if(pData) free(pData);
        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        if(controlByte.structure.bLastSegment || !oResult)
        {
            PS_AbortProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle);
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_AbortSegmentedTransfer(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);

    //Abort protocol
    oResult = PS_AbortProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle);

    //Unlock critical section
    Unlock();

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_SendNMTService(CCommand_DCS* p_pCommand, CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD RESET_NODE = 129;
    const WORD RESET_COMMUNICATION = 130;

    //Constants DCS
    const int PARAMETER_INDEX_NODE_SPECIFIER = 0;
    const int PARAMETER_INDEX_CMD_SPECIFIER = 1;

    //Constants PS
    const BYTE DATA_LENGTH = 4;

    //Variables DCS
    //Parameter
    WORD usNodeIdentifier = 0;
    WORD usCmdSpecifier = 0;

    //Variables PS
    //Parameter
    void* pDataBuffer = 0;
    DWORD ulDataBufferLength = 0;
    BYTE ubKeepLock = 0;

    //Return parameter
    void* pRetDataBuffer = 0;
    DWORD ulRetDataBufferLength = 0;

    BOOL oResult(FALSE);
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest = 0;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get DCS parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_NODE_SPECIFIER, &usNodeIdentifier, sizeof(usNodeIdentifier));
        p_pCommand->GetParameterData(PARAMETER_INDEX_CMD_SPECIFIER, &usCmdSpecifier, sizeof(usCmdSpecifier));

        //Prepare data buffer
        ulDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(ulDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &usNodeIdentifier, sizeof(usNodeIdentifier));
        CopyData(pDest, &usCmdSpecifier, sizeof(usCmdSpecifier));

       //Execute command
        oResult = PS_ProcessProtocol(p_pProtocolStackManager, p_hPS_Handle, p_hTransactionHandle, OP_CODE_SEND_NMT_SERVICE, pDataBuffer, ulDataBufferLength, ubKeepLock, &pRetDataBuffer, &ulRetDataBufferLength, &comErrorInfo);

        //No EndAck when ResetNode or ResetCommunication
        if((usCmdSpecifier == RESET_NODE) || (usCmdSpecifier == RESET_COMMUNICATION))
        {
            oResult = TRUE;
            comErrorInfo.Reset();
        }

        //Set DCS return parameter data
        p_pCommand->SetStatus(oResult, &cmdErrorInfo);

        if(pDataBuffer) free(pDataBuffer);
        if(pRetDataBuffer) free(pRetDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::PS_ProcessProtocol(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubOpCode, void* p_pDataBuffer, DWORD p_ulDataBufferLength, BYTE p_ubKeepLock, void** p_ppRetDataBuffer, DWORD* p_pulRetDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    BYTE ubLen = 0; //Calculated internally
    WORD usCrc = 0; //Calculated internally

    BYTE ubRetLen;
    BYTE ubRetOpCode;
    WORD usRetCrc;

    BOOL oResult(FALSE);
    CStdString msg;

    if(m_pCommand_ProcessProtocol && p_pProtocolStackManager && p_ppRetDataBuffer && p_pulRetDataBufferLength)
    {
        //Init return parameter data
        m_pCommand_ProcessProtocol->GetReturnParameterData(2, *p_ppRetDataBuffer, *p_pulRetDataBufferLength);

        //Set parameter data
        m_pCommand_ProcessProtocol->ResetStatus();
        m_pCommand_ProcessProtocol->SetParameterData(0, &p_ubOpCode, sizeof(p_ubOpCode));
        m_pCommand_ProcessProtocol->SetParameterData(1, &ubLen, sizeof(ubLen));
        m_pCommand_ProcessProtocol->SetParameterData(2, p_pDataBuffer, p_ulDataBufferLength);
        m_pCommand_ProcessProtocol->SetParameterData(3, &usCrc, sizeof(usCrc));
        m_pCommand_ProcessProtocol->SetParameterData(4, &p_ubKeepLock, sizeof(p_ubKeepLock));

        msg.Format(_T("CGatewayEpos4ToMaxonSerialV2, process protocol opcode = %d, data length = %ld"), p_ubOpCode, p_ulDataBufferLength);
        CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_DCS), msg);

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_ProcessProtocol, p_hPS_Handle, p_hTransactionHandle);

        //Prepare return data buffer
        if(*p_ppRetDataBuffer == 0)
        {
            *p_pulRetDataBufferLength = m_pCommand_ProcessProtocol->GetReturnParameterLength(2);
            if(*p_pulRetDataBufferLength > 0) *p_ppRetDataBuffer = malloc(*p_pulRetDataBufferLength);
        }

        //Get return parameter data
        m_pCommand_ProcessProtocol->GetReturnParameterData(0, &ubRetOpCode, sizeof(ubRetOpCode));
        m_pCommand_ProcessProtocol->GetReturnParameterData(1, &ubRetLen, sizeof(ubRetLen));
        m_pCommand_ProcessProtocol->GetReturnParameterData(2, *p_ppRetDataBuffer, *p_pulRetDataBufferLength);
        m_pCommand_ProcessProtocol->GetReturnParameterData(3, &usRetCrc, sizeof(usRetCrc));

        //Get error code
        m_pCommand_ProcessProtocol->GetErrorInfo(p_pErrorInfo);

        msg.Format(_T("CGatewayEpos4ToMaxonSerialV2, result = %s, process protocol retopcode = %d, ret data length = %d"), oResult ? _T("Success") : _T("Failure"), ubRetOpCode, ubRetLen);
        CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_DCS), msg);
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::PS_AbortProtocol(CProtocolStackManagerBase* p_pProtocolStackManager, HANDLE p_hPS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pCommand_AbortProtocol && p_pProtocolStackManager)
    {
        //Set parameter data
        m_pCommand_AbortProtocol->ResetStatus();

       //Execute command
        oResult = p_pProtocolStackManager->ExecuteCommand(m_pCommand_AbortProtocol, p_hPS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_AbortProtocol->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

void CGatewayEpos4ToMaxonSerialV2::InitCommands()
{
    DeleteCommands();

    //Process Protocol
    m_pCommand_ProcessProtocol = new CCommand_PS_MaxonSerialV2();
    m_pCommand_ProcessProtocol->InitCommand(MAXON_SERIAL_V2_PROCESS_PROTOCOL);

    //Abort Protocol
    m_pCommand_AbortProtocol = new CCommand_PS_MaxonSerialV2();
    m_pCommand_AbortProtocol->InitCommand(MAXON_SERIAL_V2_ABORT_PROTOCOL);
    m_pCommand_AbortProtocol->EnableTransactionEvent(FALSE);
}

void CGatewayEpos4ToMaxonSerialV2::DeleteCommands()
{
    if(m_pCommand_ProcessProtocol)
    {
        delete m_pCommand_ProcessProtocol;
        m_pCommand_ProcessProtocol = 0;
    }

    if(m_pCommand_AbortProtocol)
    {
        delete m_pCommand_AbortProtocol;
        m_pCommand_AbortProtocol = 0;
    }
}

BOOL CGatewayEpos4ToMaxonSerialV2::InitLayerParameterStack(CCommandRoot* p_pCommand)
{
    BOOL oResult(FALSE);
    CLayerParameterSet layerParameterSet;
    CLayerParameterStack layerParameterStack;

    if(p_pCommand && p_pCommand->GetLayerParameterStack(layerParameterStack))
    {
        //Pop DCS layer Parameter Set
        layerParameterStack.PopLayer(DEVICE_COMMAND_SET_LAYER, layerParameterSet);

        //Set PS layer Commands
        if(m_pCommand_ProcessProtocol)
        {
            m_pCommand_ProcessProtocol->SetLayerParameterStack(layerParameterStack);
        }

        if(m_pCommand_AbortProtocol)
        {
            m_pCommand_AbortProtocol->SetLayerParameterStack(layerParameterStack);
        }

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::GetNodeId(CCommand_DCS* p_pCommand, BYTE& p_rubNodeId, CErrorInfo* p_pErrorInfo)
{
    const CStdString NODE_ID(_T("NodeId"));

    BOOL oResult(FALSE);

    //Get node id from command
    if(!oResult && p_pCommand)
    {
        oResult = p_pCommand->GetLayerParameter(DEVICE_COMMAND_SET_LAYER, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
    }

    //Get node id from device
    if(!oResult)
    {
        oResult = GetParameter(NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
    }

    //Check for undefined value
    if(oResult && (p_rubNodeId == NODE_ID_UNDEFINED))
    {
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className(_T("GatewayEpos4ToMaxonSerialV2"));

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(DEVICE_COMMAND_SET_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::EvaluateErrorCode(BOOL p_oResult, DWORD p_ulDeviceErrorCode, CErrorInfo* p_pComErrorInfo, CErrorInfo* p_pCmdErrorInfo)
{
    if(p_pCmdErrorInfo && p_pComErrorInfo)
    {
        if(!p_oResult)
        {
            *p_pCmdErrorInfo = *p_pComErrorInfo;
        }

        if(p_oResult && (p_ulDeviceErrorCode != 0))
        {
            p_pCmdErrorInfo->Init(p_ulDeviceErrorCode);
            p_oResult = FALSE;
        }
    }

    return p_oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::CopyData(void*& p_pDest, void* p_pSource, DWORD p_ulSourceSize)
{
    BOOL oResult(FALSE);
    BYTE* p;

    if(p_pDest && p_pSource)
    {
        //Copy data
        memcpy(p_pDest, p_pSource, p_ulSourceSize);

        //Move destination pointer
        p = (BYTE*)p_pDest;
        p += p_ulSourceSize;
        p_pDest = p;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::CopyReturnData(void* p_pDest, DWORD p_ulDestSize, void*& p_pSource, DWORD& p_rulSourceSize)
{
    BOOL oResult(FALSE);
    BYTE* p = NULL;

    if(p_pDest && p_pSource)
    {
        if(p_rulSourceSize >= p_ulDestSize)
        {
            //Copy data
            memcpy(p_pDest, p_pSource, p_ulDestSize);

            //Move source pointer
            p = (BYTE*)p_pSource;
            p += p_ulDestSize;
            p_pSource = p;

            //Reduce source size
            p_rulSourceSize -= p_ulDestSize;

            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_SendLSSFrame(CCommand_DCS* pCommand, CProtocolStackManagerBase* pProtocolStackManager, HANDLE hPS_Handle, HANDLE hTransactionHandle)
{
    const BYTE MAX_FRAME_LENGTH = 8;

    //*Constants DCS*
    const int PARAMETER_INDEX_DATA = 0;
    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;

    //*Constants PS*
    const BYTE OP_CODE = 0x7A;
    const BYTE DATA_LENGTH = 8;

    //*Variables DCS*
    //Parameter
    BYTE maxFrameBuffer[MAX_FRAME_LENGTH];
    //Return Parameter
    DWORD dDeviceErrorCode = 0;

    //*Variables PS*
    //Parameter
    BYTE uOpCode = OP_CODE;
    void* pDataBuffer = NULL;
    DWORD dDataBufferLength;
    BYTE ubKeepLock(0);

    //ReturnParameter
    void* pRetDataBuffer = NULL;
    DWORD dRetDataBufferLength = 0;

    BOOL oResult = FALSE;
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest = NULL;
    void* pSource = NULL;

    if (pCommand)
    {
        //Lock CriticalSection
        if (!Lock(pCommand)) return FALSE;

        //Get DCS Parameter Data
        dDataBufferLength = pCommand->GetParameterLength(PARAMETER_INDEX_DATA);

        //Limit Length
        if (dDataBufferLength > MAX_FRAME_LENGTH) dDataBufferLength = MAX_FRAME_LENGTH;
        memset(maxFrameBuffer, 0, MAX_FRAME_LENGTH);
        pCommand->GetParameterData(PARAMETER_INDEX_DATA, maxFrameBuffer, dDataBufferLength);

        //Prepare DataBuffer
        dDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(dDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, maxFrameBuffer, MAX_FRAME_LENGTH);

        //Execute Command
        oResult = PS_ProcessProtocol(pProtocolStackManager, hPS_Handle, hTransactionHandle, uOpCode, pDataBuffer, dDataBufferLength, ubKeepLock, &pRetDataBuffer, &dRetDataBufferLength, &comErrorInfo);

        //ReturnData
        pSource = pRetDataBuffer;
        if (oResult) oResult = CopyReturnData(&dDeviceErrorCode, sizeof(dDeviceErrorCode), pSource, dRetDataBufferLength);

        //Evaluate ErrorCode
        oResult = EvaluateErrorCode(oResult, dDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);

        //Set DCS ReturnParameter Data
        pCommand->SetStatus(oResult, &cmdErrorInfo);
        pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &dDeviceErrorCode, sizeof(dDeviceErrorCode));

        if (pRetDataBuffer) free(pRetDataBuffer);
        if (pDataBuffer) free(pDataBuffer);

        //Unlock CriticalSection
        Unlock();
    }

    return oResult;
}

BOOL CGatewayEpos4ToMaxonSerialV2::Process_ReadLSSFrame(CCommand_DCS* pCommand, CProtocolStackManagerBase* pProtocolStackManager, HANDLE hPS_Handle, HANDLE hTransactionHandle)
{
    const BYTE MAX_FRAME_LENGTH = 8;

    //*Constants DCS*
    const int PARAMETER_INDEX_TIMEOUT = 0;
    const int RETURN_PARAMETER_INDEX_ERROR_CODE = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 1;

    //*Constants PS*
    const BYTE OP_CODE = 0x7B;
    const BYTE DATA_LENGTH = 2;

    //*Variables DCS*
    //Parameter
    WORD wTimeout = 0;
    //Return Parameter
    DWORD dDeviceErrorCode = 0;
    BYTE maxFrameBuffer[MAX_FRAME_LENGTH];

    //*Variables PS*
    //Parameter
    BYTE uOpCode = OP_CODE;
    void* pDataBuffer = NULL;
    DWORD dDataBufferLength;
    BYTE ubKeepLock(0);

    //ReturnParameter
    void* pRetDataBuffer = NULL;
    DWORD dRetDataBufferLength = 0;

    BOOL oResult = FALSE;
    CErrorInfo comErrorInfo;
    CErrorInfo cmdErrorInfo;

    void* pDest;
    void* pSource;

    if (pCommand)
    {
        //Lock CriticalSection
        if (!Lock(pCommand)) return FALSE;

        //Get DCS Parameter Data
        pCommand->GetParameterData(PARAMETER_INDEX_TIMEOUT, &wTimeout, sizeof(wTimeout));

        //Limit Length
        memset(maxFrameBuffer, 0, MAX_FRAME_LENGTH);

        //Prepare DataBuffer
        dDataBufferLength = DATA_LENGTH;
        pDataBuffer = malloc(dDataBufferLength);

        //Data
        pDest = pDataBuffer;
        CopyData(pDest, &wTimeout, sizeof(wTimeout));

        //Execute Command
        oResult = PS_ProcessProtocol(pProtocolStackManager, hPS_Handle, hTransactionHandle, uOpCode, pDataBuffer, dDataBufferLength, ubKeepLock, &pRetDataBuffer, &dRetDataBufferLength, &comErrorInfo);

        //ReturnData
        pSource = pRetDataBuffer;
        if (oResult) oResult = CopyReturnData(&dDeviceErrorCode, sizeof(dDeviceErrorCode), pSource, dRetDataBufferLength);

        if (oResult)
        {
            oResult = EvaluateErrorCode(oResult, dDeviceErrorCode, &comErrorInfo, &cmdErrorInfo);
        }

        if (oResult) 
        {
            oResult = CopyReturnData(maxFrameBuffer, MAX_FRAME_LENGTH, pSource, dRetDataBufferLength);
        }

        //Set DCS ReturnParameter Data
        pCommand->SetStatus(oResult, &cmdErrorInfo);
        pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ERROR_CODE, &dDeviceErrorCode, sizeof(dDeviceErrorCode));
        pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, maxFrameBuffer, MAX_FRAME_LENGTH);

        if (pDataBuffer) free(pDataBuffer);
        if (pRetDataBuffer) free(pRetDataBuffer);

        //Unlock CriticalSection
        Unlock();
    }

    return oResult;
}