#include "stdafx.h"

#include <CommunicationModel/CommonLayer/Classes/Commands/Interface/BaseClasses/Command_I.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Interface/Command_I_RS232.h>
#include <CommunicationModel/Interface/BaseClasses/InterfaceManagerBase.h>
#include <CommunicationModel/Interface/Classes/Gateway/RS232/Classes/Rs232DeviceInfoHandling.h>
#include <Log/MmcLog/MmcLog.h>
#include <Thread/MmcSingleLock.h>

#include <CommunicationModel/Interface/Classes/Gateway/RS232/GatewayRS232.h>

CGatewayRS232ToWin32::CGatewayRS232ToWin32():
    m_pInfoHandling(0)
{
    InitErrorHandling();
    InitBaudrateSelection();
    InitDefaultPortSettings();
}

CGatewayRS232ToWin32::~CGatewayRS232ToWin32()
{
}

BOOL CGatewayRS232ToWin32::InitInfoHandling(CDeviceInfoHandlingBase* p_pInfoHandling)
{
    BOOL oResult(TRUE);

    m_pInfoHandling = dynamic_cast<CRs232DeviceInfoHandling*>(p_pInfoHandling);

    return oResult;
}

BOOL CGatewayRS232ToWin32::InitBaudrateSelection()
{
    m_BaudrateSelection.clear();
    m_BaudrateSelection.push_back(9600);
#ifdef WINVER
    m_BaudrateSelection.push_back(14400);
#endif
    m_BaudrateSelection.push_back(19200);
    m_BaudrateSelection.push_back(38400);
    m_BaudrateSelection.push_back(57600);
    m_BaudrateSelection.push_back(115200);

    return TRUE;
}

BOOL CGatewayRS232ToWin32::InitDefaultPortSettings()
{
    m_dDefaultBaudrate = 115200;
    m_dDefaultTimeout = 500;

    GetDefaultPortSettings(&m_dBaudrate, &m_dTimeout);

    return TRUE;
}

BOOL CGatewayRS232ToWin32::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult = FALSE;
    CCommand_I* pCommand_I;

    if(CGateway::ProcessCommand(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle))
    {
        if(CheckLayers(p_pCommand, p_pManager))
        {
            pCommand_I = (CCommand_I*)p_pCommand;

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayRS232ToWin32")));
            }

            switch(p_pCommand->GetCommandId())
            {
                case RS232_WRITE_DATA:
                    {
                        oResult = Process_WriteData(pCommand_I);
                        break;
                    }
                case RS232_READ_DATA:
                    {
                        oResult = Process_ReadData(pCommand_I);
                        break;
                    }
            }

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayRS232ToWin32"), TRUE));
            }
        }
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::Process_WriteData(CCommand_I* p_pCommand)
{
    //*Constants I*
    const int PARAMETER_INDEX_DATA = 0;
    const int RETURN_PARAMETER_INDEX_NB_OF_BYTES_WRITTEN = 0;

    //*Variables I*
    //Parameter
    void* pDataBuffer = NULL;
    DWORD ulDataBufferLength;
    //ReturnParameter
    DWORD ulNbOfBytesWritten = 0;

    BOOL oResult = FALSE;
    CErrorInfo errorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand->GetTimeout()))
    {
        //Prepare data buffer
        ulDataBufferLength = p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
        if (ulDataBufferLength > 0)
        {
            pDataBuffer = malloc(ulDataBufferLength);
        }

        //Get I Parameter Data
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, pDataBuffer, ulDataBufferLength);

        //Execute command
        oResult = WriteData(pDataBuffer, ulDataBufferLength, &ulNbOfBytesWritten, &errorInfo);

        //Set PS ReturnParameter Data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_NB_OF_BYTES_WRITTEN, &ulNbOfBytesWritten, sizeof(ulNbOfBytesWritten));

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::Process_ReadData(CCommand_I* p_pCommand)
{
    //*Constants I*
    const int PARAMETER_INDEX_NB_OF_BYTES_TO_READ = 0;
    const int RETURN_PARAMETER_INDEX_DATA = 0;

    //*Variables I*
    //Parameter
    DWORD ulNbOfBytesToRead;

    //ReturnParameter
    void* pDataBuffer = NULL;
    DWORD ulDataBufferLength;
    DWORD ulNbOfBytesRead = 0;

    BOOL oResult = FALSE;
    CErrorInfo errorInfo;

    //Lock critical section
    if (p_pCommand && Lock(p_pCommand->GetTimeout()))
    {
        //Get I Parameter Data
        p_pCommand->GetParameterData(PARAMETER_INDEX_NB_OF_BYTES_TO_READ, &ulNbOfBytesToRead, sizeof(ulNbOfBytesToRead));

        //Prepare data buffer
        ulDataBufferLength = ulNbOfBytesToRead;
        if (ulDataBufferLength > 0)
        {
            pDataBuffer = malloc(ulDataBufferLength);
        }

        //Execute command
        oResult = ReadData(pDataBuffer, ulNbOfBytesToRead, &ulNbOfBytesRead, &errorInfo);

        //Set PS ReturnParameter Data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pDataBuffer, ulNbOfBytesRead);

        //Free data buffer
        if(pDataBuffer) free(pDataBuffer);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

CGateway* CGatewayRS232ToWin32::Clone()
{
    CGatewayRS232ToWin32* pClonedGateway;

    pClonedGateway = new CGatewayRS232ToWin32();
    *pClonedGateway = *this;

    return pClonedGateway;
}

CGatewayRS232ToWin32& CGatewayRS232ToWin32::operator=(CGatewayRS232ToWin32& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayIToDrv*)this) = *((CGatewayIToDrv*)&p_rOther);
        m_pInfoHandling = p_rOther.m_pInfoHandling;
    }

    return *this;
}

BOOL CGatewayRS232ToWin32::OpenPort(CStdString p_portName, CErrorInfo* p_pErrorInfo)
{
    const BOOL CHANGE_ONLY = FALSE;

    BOOL oResult(FALSE);

    if(m_pInfoHandling)
    {
        //Init
        oResult = TRUE;
        CRs232DeviceInfo* pDeviceInfo = m_pInfoHandling->FindDeviceInfo(p_portName);
        if(pDeviceInfo)
        {
            if(oResult && !m_Rs232Hndl.OpenPort(*pDeviceInfo))
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_OPENING_PORT, p_pErrorInfo);
                oResult = FALSE;
            }

            //Configure default settings for port
            if(oResult && !SetPortSettings(m_dBaudrate, m_dTimeout, CHANGE_ONLY, p_pErrorInfo))
            {
                ClosePort();
                oResult = FALSE;
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_OPENING_PORT, p_pErrorInfo);
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::SetPortSettings(DWORD p_ulBaudrate, DWORD p_ulTimeout, BOOL p_oChangeOnly, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = TRUE;

    if(!p_oChangeOnly || (p_ulTimeout != m_dTimeout) || (p_ulBaudrate != m_dBaudrate))
    {
        if(m_Rs232Hndl.SetPortSettings(p_ulBaudrate, p_ulTimeout, p_oChangeOnly == TRUE))
        {
            m_dBaudrate = p_ulBaudrate;
            m_dTimeout = p_ulTimeout;
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_SETTINGS, p_pErrorInfo);
            oResult = FALSE;
        }
    }

    Trace_Settings(p_ulBaudrate, p_ulTimeout);

    return oResult;
}

BOOL CGatewayRS232ToWin32::GetPortModeSelection(CStdStringArray* p_pPortModeSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(p_pPortModeSel)
    {
        //Init
        oResult = TRUE;
        p_pPortModeSel->clear();

        //NoParity (Index 0)
        p_pPortModeSel->push_back(_T("Standard"));

        //EvenParity (Index 1)
        p_pPortModeSel->push_back(_T("EvenParity"));
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::SetPortMode(WORD p_usPortMode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_Rs232Hndl.SetPortMode(p_usPortMode))
    {
        m_wPortMode = p_usPortMode;
        oResult = TRUE;
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_MODE, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::SetTimeout(DWORD p_ulTimeout)
{
    m_dTimeout = p_ulTimeout;

    return m_Rs232Hndl.SetTimeout(p_ulTimeout)  ? TRUE : FALSE;
}

BOOL CGatewayRS232ToWin32::PurgeBuffer()
{
    return m_Rs232Hndl.Purge() ? TRUE : FALSE;
}

BOOL CGatewayRS232ToWin32::WriteData(void* p_pDataBuffer, DWORD p_ulNbOfBytesToWrite, DWORD* p_pulNbOfBytesWritten, CErrorInfo* p_pErrorInfo)
{
    DWORD ulNumberOfBytesWritten = 0;
    CStdString msg;

    //ResetErrorInfo
    if(p_pErrorInfo) p_pErrorInfo->Reset();

    msg.Format(_T("WriteData Buffer Before: NbOfBytesToWrite %i\n"), p_ulNbOfBytesToWrite);

    CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);

    //Execute command
    if(p_pDataBuffer && (p_ulNbOfBytesToWrite > 0))
    {
        m_Rs232Hndl.Write((unsigned char*)p_pDataBuffer, p_ulNbOfBytesToWrite, (unsigned int*)&ulNumberOfBytesWritten);

        if(ulNumberOfBytesWritten == p_ulNbOfBytesToWrite)
        {
            Trace_WriteData(TRUE, p_pDataBuffer, p_ulNbOfBytesToWrite, &ulNumberOfBytesWritten);
            if(p_pulNbOfBytesWritten) *p_pulNbOfBytesWritten = ulNumberOfBytesWritten;

            msg.Format(_T("WriteData Buffer End: NbOfBytesToWrite %i NbOfBytesWritten %i\n"), p_ulNbOfBytesToWrite, ulNumberOfBytesWritten);

            CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);

            return TRUE;
        }

        Trace_WriteData(FALSE, p_pDataBuffer, p_ulNbOfBytesToWrite, &ulNumberOfBytesWritten);
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_RS232_WRITE_DATA, p_pErrorInfo);
        return FALSE;
    }

    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    return FALSE;
}

BOOL CGatewayRS232ToWin32::ReadData(void* p_pDataBuffer, DWORD p_ulNbOfBytesToRead, DWORD* p_pulNbOfBytesRead, CErrorInfo* p_pErrorInfo)
{
    DWORD ulNumberOfBytesRead;

    //ResetErrorInfo
    if(p_pErrorInfo) p_pErrorInfo->Reset();

    //Execute command
    if(p_pDataBuffer && (p_ulNbOfBytesToRead > 0))
    {
        if(m_Rs232Hndl.Read((unsigned char*)p_pDataBuffer, p_ulNbOfBytesToRead, (unsigned int*)&ulNumberOfBytesRead))
        {
            *p_pulNbOfBytesRead = ulNumberOfBytesRead;

            Trace_ReadData(TRUE, p_pDataBuffer, p_ulNbOfBytesToRead, &ulNumberOfBytesRead);

            return TRUE;
        }

        Trace_ReadData(FALSE, p_pDataBuffer, p_ulNbOfBytesToRead, &ulNumberOfBytesRead);

        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_RS232_READ_DATA, p_pErrorInfo);

        return FALSE;
    }

    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    return FALSE;
}

BOOL CGatewayRS232ToWin32::GetPortSettings(DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    if(p_pulBaudrate) *p_pulBaudrate = m_dBaudrate;
    if(p_pulTimeout) *p_pulTimeout = m_dTimeout;
    return TRUE;
}

BOOL CGatewayRS232ToWin32::GetPortMode(WORD* p_pusPortMode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(p_pusPortMode)
    {
        *p_pusPortMode = m_wPortMode;
        oResult = TRUE;
    }
    else
    {
        if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::ClosePort(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);

    if(!m_Rs232Hndl.ClosePort())
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_CLOSING_PORT, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::ResetPort(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);

    if(!PurgeBuffer())
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_CLOSING_PORT, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::InitPortList()
{
    CMmcSingleLock lock(CRs232DeviceInfoHandling::GetSyncInstance(), TRUE);

    tRs232DeviceInfoList deviceInfoList;
    BOOL oResult(FALSE);

    if(m_pInfoHandling)
    {
        //Get all physically available ports
        if(m_Rs232Hndl.GetDeviceInfos(deviceInfoList))
        {
            if(m_pInfoHandling->InitDeviceInfos(deviceInfoList))
            {
                oResult = UpdatePortList(m_pInfoHandling);
            }

            DeleteDeviceInfoList(deviceInfoList);
        }
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::UpdatePortList(CRs232DeviceInfoHandling* p_pDeviceInfoHandling)
{
    CStdStringArray portNames;
    CStdString portName(_T(""));
    BOOL oResult(FALSE);

    //Reset PortList
    DeletePortList();

    if(p_pDeviceInfoHandling)
    {
        //Add Ports
        if(p_pDeviceInfoHandling->GetPortNames(portNames))
        {
            oResult = TRUE;
            for(std::size_t i = 0; i < portNames.size(); i++)
            {
                portName = portNames.at(i);
                AddPortList(portName);
            }
        }
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className = _T("GatewayRS232ToWin32");

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(INTERFACE_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CGatewayRS232ToWin32::DeleteDeviceInfoList(tRs232DeviceInfoList& p_rDeviceInfoList)
{
    BOOL oResult(FALSE);

    for(tRs232DeviceInfoList::iterator it=p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CRs232DeviceInfo* pDeviceInfo = (*it);
        delete pDeviceInfo;
    }

    p_rDeviceInfoList.clear();

    return oResult;
}
