#include "stdafx.h"
#include <stdio.h>
#include <memory.h>

#include "../../../../Factory/DeviceInfoHandlingFactory.h"
#include "../Factory/UsbDeviceInfoFactory.h"

#include <Log/MmcLog/MmcLog.h>
#include <Process/MmcProcess.h>
#include <Storage/MmcFileInfo.h>
#include <Thread/MmcCriticalSection.h>
#include <Thread/MmcSingleLock.h>
#include <Thread/MmcSemaphore.h>

#include <CommunicationModel/Interface/Classes/Gateway/USB/Classes/UsbDeviceInfoHandling.h>

using namespace std;

//USB definitions
const TCHAR REGISTRY_MAXON[]        = _T("Software\\maxon motor ag");
const TCHAR REGISTRY_USB_HANDLING[] = _T("UsbDeviceInfoHandling");

list<CUsbDeviceInfoHandling::SVendorProductIDs> CUsbDeviceInfoHandling::m_VendorProductIdList;

CUsbDeviceInfoHandling::CUsbDeviceInfoHandling(void)
    : m_pRegistrySemaphore(0)
    , m_oGlobalRegistryKey(FALSE)
{
    CreateRegistrySemaphore();
}

CUsbDeviceInfoHandling::CUsbDeviceInfoHandling(int p_lInstanceValue):
    CDeviceInfoHandlingBase(p_lInstanceValue)
    , m_pRegistrySemaphore(0)
{
    CreateRegistrySemaphore();
}

CUsbDeviceInfoHandling::~CUsbDeviceInfoHandling(void)
{
    ResetDeviceInfoList();
    DeleteRegistrySemaphore();
}

BOOL CUsbDeviceInfoHandling::InitRegistryKeyType(BOOL p_oGlobal)
{
    m_oGlobalRegistryKey = p_oGlobal;

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::UpdateRegistryKeyType(CStdString p_applicationName)
{
    const CStdString KEY_EPOS = _T("epos");
    const CStdString KEY_STUDIO = _T("studio");
    const CStdString KEY_PROXY = _T("proxy");

    //Non case sensitive
    p_applicationName.MakeLower();

    //Exceptions
    if((p_applicationName.Find(KEY_EPOS) != -1) && (p_applicationName.Find(KEY_STUDIO) != -1))
    {
        //Exception for EPOS Studio -> Use Global Key
        m_oGlobalRegistryKey = TRUE;
    }
    else if((p_applicationName.Find(KEY_EPOS) != -1) && (p_applicationName.Find(KEY_PROXY) != -1))
    {
        //Exception for EPOS Studio -> Use Global Key
        m_oGlobalRegistryKey = TRUE;
    }

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::CreateRegistrySemaphore()
{
    const long NB_OF_ALLOWED_CLIENTS = 1;

    CStdString strName = _T("Maxon-UsbDeviceInfoHandling");
    BOOL oResult = FALSE;

    if(!m_pRegistrySemaphore)
    {
        m_pRegistrySemaphore = new CMmcSemaphore(NB_OF_ALLOWED_CLIENTS, NB_OF_ALLOWED_CLIENTS, strName);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::DeleteRegistrySemaphore()
{
    BOOL oResult(FALSE);

    if(m_pRegistrySemaphore)
    {
        delete m_pRegistrySemaphore;
        m_pRegistrySemaphore = 0;

        oResult = TRUE;
    }

    return oResult;
}

CUsbDeviceInfoHandling& CUsbDeviceInfoHandling::operator=(CUsbDeviceInfoHandling& p_rOther)
{
    if(this != &p_rOther)
    {
        CopyDeviceInfoList(p_rOther.m_DeviceInfoList, m_DeviceInfoList);
    }

    return *this;
}

BOOL CUsbDeviceInfoHandling::CopyDeviceInfos(CUsbDeviceInfoHandling* p_pOther)
{
    BOOL oResult = FALSE;

    if(p_pOther)
    {
        oResult = CopyDeviceInfoList(p_pOther->m_DeviceInfoList, m_DeviceInfoList);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::CopyDeviceInfoList(tUsbDeviceInfoList& p_rSourceDeviceInfoList, tUsbDeviceInfoList& p_rTargetDeviceInfoList)
{
    BOOL oResult = FALSE;

    if(DeleteDeviceInfoList(p_rTargetDeviceInfoList))
    {
        oResult = TRUE;
        for(tUsbDeviceInfoList::iterator it = p_rSourceDeviceInfoList.begin(); it != p_rSourceDeviceInfoList.end(); it++)
        {
            CUsbDeviceInfo* pDeviceInfo = (*it);
            if(pDeviceInfo != 0)
            {
                CUsbDeviceInfo* pClonedDeviceInfo = pDeviceInfo->Clone();
                if(pClonedDeviceInfo)
                {
                    p_rTargetDeviceInfoList.push_back(pClonedDeviceInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::DeleteDeviceInfoList(tUsbDeviceInfoList& p_rDeviceInfoList)
{
    BOOL oResult = TRUE;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo != 0)
        {
            delete pDeviceInfo;
        }
    }

    p_rDeviceInfoList.clear();

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ResetDeviceInfoList()
{
    BOOL oResult = FALSE;

    oResult = DeleteDeviceInfoList(m_DeviceInfoList);

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ResetDeviceInfos()
{
    CMmcSingleLock lock(m_pRegistrySemaphore, TRUE);
    BOOL oResult = FALSE;

    //Reset
    if(ResetDeviceInfoList())
    {
        oResult = WriteToRegistry();
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::IsDeviceInfoInList(tUsbDeviceInfoList& p_deviceInfoList, CUsbDeviceInfo* p_pUsbDeviceInfo)
{
    BOOL oResult = FALSE;
    CUsbDeviceInfo* pUsbDeviceInfo = 0;

    for(tUsbDeviceInfoList::iterator it=p_deviceInfoList.begin(); it!=p_deviceInfoList.end(); it++)
    {
        pUsbDeviceInfo = (*it);

        if(pUsbDeviceInfo == p_pUsbDeviceInfo)
        {
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::SearchOrphanDevices(tUsbDeviceInfoList p_deviceInfoList, CUsbDeviceInfo* pUsbDeviceInfo, tUsbDeviceInfoList& p_OrphanDevices)
{
    BOOL oResult = FALSE;
    CUsbDeviceInfo* pOrphanUsbDeviceInfo = 0;

    for(tUsbDeviceInfoList::iterator it=p_deviceInfoList.begin(); it!=p_deviceInfoList.end(); it++)
    {
        pOrphanUsbDeviceInfo = (*it);

        if(pOrphanUsbDeviceInfo != pUsbDeviceInfo)
        {
            if(pOrphanUsbDeviceInfo->GetPortName() == pUsbDeviceInfo->GetPortName() ||
               pOrphanUsbDeviceInfo->GetPortChain() == pUsbDeviceInfo->GetPortChain() ||
               pOrphanUsbDeviceInfo->GetDevicePath() == pUsbDeviceInfo->GetDevicePath())
            {
                if(!IsDeviceInfoInList(p_OrphanDevices, pOrphanUsbDeviceInfo))
                {
                    p_OrphanDevices.push_back(pOrphanUsbDeviceInfo);
                    oResult = TRUE;
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::RemoveOrphanDevices(tUsbDeviceInfoList& p_rDeviceInfoList)
{
    BOOL oResult = TRUE;
    CUsbDeviceInfo* pUsbDeviceInfo = 0;
    tUsbDeviceInfoList orphanDevices;

    for(tUsbDeviceInfoList::iterator it=p_rDeviceInfoList.begin(); it!=p_rDeviceInfoList.end(); it++)
    {
        pUsbDeviceInfo = (*it);

        if(SearchOrphanDevices(p_rDeviceInfoList, pUsbDeviceInfo, orphanDevices))
        {
            orphanDevices.push_back(pUsbDeviceInfo);
        }
    }

    for(tUsbDeviceInfoList::iterator it=orphanDevices.begin(); it!=orphanDevices.end(); it++)
    {
        DeleteFromDeviceInfoList(*it);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::InitDeviceInfos(tUsbDeviceInfoList& p_rDeviceInfoList, CStdString p_deviceClassType)
{
    CMmcSingleLock lock(m_pRegistrySemaphore, TRUE);

    tPortList openPortList;
    BOOL oResult = FALSE;

    //Read DeviceInfo List from Registry
    ReadFromRegistry();

    //Logging
    WriteToLog(_T("I-ReadFromUsbDrv"), p_rDeviceInfoList);
    WriteToLog(_T("I-ReadFromReg"));

    //Update Removed Devices
    if(UpdateRemovedDevices(p_rDeviceInfoList, p_deviceClassType))
    {
        //Update Added Devices
        if(UpdateAddedDevices(p_rDeviceInfoList, openPortList))
        {
            if(UpdateNewAddedDevices(p_rDeviceInfoList, openPortList))
            {
                //Write DeviceInfo List to Registry
                WriteToRegistry();

                //Logging
                WriteToLog(_T("I-WriteToReg"));
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdateDeviceInfos(tUsbDeviceInfoList& p_rDeviceInfoList, CStdString p_deviceClassType, tPortList& p_rOpenPortList)
{
    CMmcSingleLock lock(m_pRegistrySemaphore, TRUE);

    BOOL oResult = FALSE;

    //Read DeviceInfo List from Registry
    ReadFromRegistry();

    //Logging
    WriteToLog(_T("U-ReadFromUsbDrv"), p_rDeviceInfoList);
    WriteToLog(_T("U-ReadFromReg"));

    //Update Removed Devices
    if(UpdateRemovedDevices(p_rDeviceInfoList, p_deviceClassType))
    {
        //Update Added Devices
        if(UpdateAddedDevices(p_rDeviceInfoList, p_rOpenPortList))
        {
            if(UpdateNewAddedDevices(p_rDeviceInfoList, p_rOpenPortList))
            {
                //Write DeviceInfo List to Registry
                WriteToRegistry();

                //Logging
                WriteToLog(_T("U-WriteToReg"));
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdatePortOpened(CStdString p_PortName, BOOL p_oOpened)
{
    CMmcSingleLock lock(m_pRegistrySemaphore, TRUE);

    BOOL oResult = FALSE;

    if(UpdateRegistry(p_PortName, p_oOpened))
    {
        oResult = ReadFromRegistry();
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetPortNames(CStdStringArray& p_rPortNames)
{
    BOOL oResult = TRUE;
    //Reset
    p_rPortNames.clear();

    //Sort List by PortName
    SortDeviceInfoListByPortName(m_DeviceInfoList);

    //Add Port Names
    for(tUsbDeviceInfoList::iterator it = m_DeviceInfoList.begin(); it != m_DeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo != 0)
        {
            p_rPortNames.push_back(pDeviceInfo->GetPortName());
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetPortNames(CUsbDeviceInfo& p_rDeviceType, CStdStringArray& p_rPortNames)
{
    BOOL oResult = TRUE;
    //Reset
    p_rPortNames.clear();

    //Sort List by PortName
    SortDeviceInfoListByPortName(m_DeviceInfoList);

    //Add Port Names
    for(tUsbDeviceInfoList::iterator it = m_DeviceInfoList.begin(); it != m_DeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo != 0)
        {
            //Check Device Type
            if(p_rDeviceType.GetClassType() == pDeviceInfo->GetClassType())
            {
                p_rPortNames.push_back(pDeviceInfo->GetPortName());
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::WriteToLog(CStdString p_Action, tUsbDeviceInfoList& p_rDeviceInfoList)
{
    if(CMmcLog::IsActive(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum"))))
    {
        CStdString infos(_T(""));
        CStdString message(_T(""));

        for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
        {
            CUsbDeviceInfo* pDeviceInfo = (*it);
            if(pDeviceInfo)
            {
                //Init
                CStdString info;
                if(pDeviceInfo->GetLogInfos(info, CUsbDeviceInfo::LIT_IDENTIFIER))
                {
                    infos += info;
                }
            }
        }

        message.Format(_T("%s: %s"), p_Action, infos);
        CMmcLog::Log(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum")), message);
    }

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::WriteToLog(CStdString p_Action)
{
    if(CMmcLog::IsActive(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum"))))
    {
        CStdString infos(_T(""));
        CStdString message(_T(""));

        //DeviceInfos
        for(tUsbDeviceInfoList::iterator it = m_DeviceInfoList.begin(); it != m_DeviceInfoList.end(); it++)
        {
            CUsbDeviceInfo* pDeviceInfo = (*it);
            if(pDeviceInfo)
            {
                //Init
                CStdString portName;
                CStdString info;
                if(pDeviceInfo->GetLogInfos(info, CUsbDeviceInfo::LIT_STATE))
                {
                    infos += info;
                }
            }
        }

        message.Format(_T("%s: %s"), p_Action, infos);
        CMmcLog::Log(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum")), message);
    }

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::WriteToLog(CStdString p_Action, CStdString p_PortName)
{
    if(CMmcLog::IsActive(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum"))))
    {
        CStdString message(_T(""));

        message.Format(_T("%s: Port %s"), p_Action, p_PortName);
        CMmcLog::Log(LMT_INFO, CUserDefinedMessageLevel(_T("UsbEnum")), message);
    }

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::DeleteFromRegistry()
{
    CStdString subKey = _T("");
    CMmcUserSettings userSettings;

    //Init
    GetRegistryKey(subKey);

    //Delete
    return userSettings.Delete(subKey);
}

BOOL CUsbDeviceInfoHandling::WriteToRegistry()
{
    BOOL oResult = FALSE;
    CStdString subKey = _T("");
    CStdString entryName = _T("");
    int lIndex(0);

    //Init
    DeleteFromRegistry();

    GetRegistryKey(subKey);

    CMmcUserSettings userSettings(subKey, TRUE);

    //Init
    oResult = TRUE;

    //DeviceInfo List
    for(tUsbDeviceInfoList::iterator it = m_DeviceInfoList.begin(); it != m_DeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(!WriteDeviceInfoToRegistry(&userSettings, lIndex, pDeviceInfo))
        {
            oResult = FALSE;
        }
        lIndex++;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ReadFromRegistry()
{
    BOOL oResult = TRUE;
    CStdString subKey = _T("");

    // Init
    GetRegistryKey(subKey);
    CMmcUserSettings userSettings(subKey, FALSE);

    // Clear list
    ResetDeviceInfoList();
    // Get children (sorted alphabetically)
    list<CMmcUserSettings*> children = userSettings.GetChildren();

    for(list<CMmcUserSettings*>::iterator it = children.begin(); it != children.end(); it++)
    {
        CMmcUserSettings* pChild = (*it);
        CUsbDeviceInfo* pDeviceInfo(0);

        if(oResult)
        {
            if(ReadDeviceInfoFromRegistry(pChild, pDeviceInfo) && pDeviceInfo)
            {
                // Device info list
                m_DeviceInfoList.push_back(pDeviceInfo);
            }
            else
            {
                // Delete device info
                if(pDeviceInfo)
                {
                    delete pDeviceInfo;
                }

                oResult = FALSE;
            }
        }

        // Delete child
        delete pChild;
    }

    if(oResult)
    {
#ifdef __linux__
        RemoveOrphanDevices(m_DeviceInfoList);
#endif        
        SortDeviceInfoListByPortName(m_DeviceInfoList);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetClassTypeName(CUsbDeviceInfo* p_pDeviceInfo, int p_lIndex, CStdString& p_rClassType)
{
    BOOL oResult = FALSE;

    const char CLASS_NAME_PREFIX = 'C';
    const int FIRST_CHAR_INDEX = 0;
    const int SINGLE_CHAR_LEN = 1;

    if(p_pDeviceInfo)
    {
        CStdString classType = p_pDeviceInfo->GetClassType();

        if(classType.GetLength() > 0 && classType[FIRST_CHAR_INDEX] == CLASS_NAME_PREFIX)
        {
            classType = classType.Mid(SINGLE_CHAR_LEN);
        }

        p_rClassType.Format(_T("%i %s"), p_lIndex, classType);

        oResult = TRUE;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::WriteDeviceInfoToRegistry(CMmcUserSettings* p_pUserSettings, int p_lIndex, CUsbDeviceInfo* p_pDeviceInfo)
{
    BOOL oResult = FALSE;
    CStdString classType;
    CMmcUserSettings* pUserSettings = 0;

    if(p_pDeviceInfo)
    {
        int lIndex(-1);

        if(p_pDeviceInfo->GetPortNameIndex(lIndex))
        {
            p_lIndex = lIndex;
        }

        if(GetClassTypeName(p_pDeviceInfo, p_lIndex, classType))
        {
            pUserSettings = new CMmcUserSettings(p_pUserSettings, classType, TRUE);

            //Write
            oResult = p_pDeviceInfo->WriteTo(pUserSettings);
        }
    }

    if(pUserSettings)
    {
        delete pUserSettings;
        pUserSettings = 0;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::WriteOpenedToRegistry(CMmcUserSettings* p_pUserSettings, BOOL p_oOpened)
{
    BOOL oResult = FALSE;

    if(p_pUserSettings != 0)
    {
        DWORD openedByProcess = 0;
        DWORD openedByInstance = 0;
        CStdString openedBy = _T("");

        //Init
        oResult = TRUE;

        //Open Key
        if(p_oOpened)
        {
            openedByProcess = MmcGetCurrentProcessId();
            openedByInstance = m_lInstanceValue;
            openedBy = MmcGetModuleName();
        }
        else
        {
            openedByProcess = 0;
            openedByInstance = 0;
            openedBy = "";
        }

        //Write
        if(oResult && !p_pUserSettings->Write(_T("OpenedByProcess"), openedByProcess)) oResult = FALSE;
        if(oResult && !p_pUserSettings->Write(_T("OpenedByInstance"), openedByInstance)) oResult = FALSE;
        if(oResult && !p_pUserSettings->Write(_T("OpenedBy"), openedBy)) oResult = FALSE;
    }

    return oResult;
}

CStdString CUsbDeviceInfoHandling::ConvertKeyNameToClassType(CStdString p_KeyName)
{
    CStdString classType = p_KeyName;
    int lIndex = classType.ReverseFind(_T(" "));

    if(p_KeyName.Find(_T("_")) == 0 && lIndex == -1)
    {
        //Format: _0_XXXInfo
        classType.Replace(_T("_"), _T(" "));

        lIndex = classType.ReverseFind(_T(" "));

        classType = classType.Right(classType.GetLength() - (lIndex + 1));
    }
    else if(lIndex != -1)
    {
        if(p_KeyName.Find(_T("Info ")) != -1)
        {
            // Format: UsbDeviceInfo 0
            classType = classType.Left(lIndex);
        }
        else
        {
            // Format: 0 UsbDeviceInfo
            classType = classType.Right(classType.GetLength() - (lIndex + 1));
        }
    }

    classType = _T("C") + classType;

    return classType;
}

BOOL CUsbDeviceInfoHandling::ReadDeviceInfoFromRegistry(CMmcUserSettings* p_pUserSettings, CUsbDeviceInfo*& p_rpDeviceInfo)
{
    BOOL oResult = FALSE;

    if(p_pUserSettings)
    {
        CStdString classType = ConvertKeyNameToClassType(p_pUserSettings->GetKeyName());

        list<CStdString> aIdentifiers;
        CUsbDeviceInfoFactory* pDeviceInfoFactory = new CUsbDeviceInfoFactory();
        if(pDeviceInfoFactory != NULL)
        {
            if(pDeviceInfoFactory->GetAllIdentifiers(aIdentifiers))
            {
                for(list<CStdString>::iterator it = aIdentifiers.begin(); it != aIdentifiers.end() && !oResult; it++)
                {
                    CUsbDeviceInfo* pDeviceInfo;

                    if(pDeviceInfoFactory->CreateInstance((*it), pDeviceInfo) && pDeviceInfo != NULL)
                    {
                        if(pDeviceInfo->GetClassType() == classType)
                        {
                            p_rpDeviceInfo = pDeviceInfo->Clone();

                            oResult = p_rpDeviceInfo->ReadFrom(p_pUserSettings);

                            delete pDeviceInfo;

                            break;
                        }
                        else
                        {
                            delete pDeviceInfo;
                        }
                    }
                }
            }

            delete pDeviceInfoFactory;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetRegistryKey(CStdString& p_rRegistryKey)
{
    CStdString subKey = _T("");
    CStdString application = _T("");
    CStdString version = _T("");
    BOOL oResult = FALSE;

    //Application Infos
    application = RemoveFileExtension(MmcGetApplicationName()) + _T(" ") + MmcGetApplicationVersion();
    application = application.Trim();

    //RegistryKey Type
    UpdateRegistryKeyType(application);

    if(m_oGlobalRegistryKey)
    {
        //Global Key
        subKey.Format(_T("%s\\%s"), REGISTRY_MAXON, REGISTRY_USB_HANDLING);
    }
    else
    {
        //Local Key
        subKey.Format(_T("%s\\%s\\%s"), REGISTRY_MAXON, application, REGISTRY_USB_HANDLING);
    }

    p_rRegistryKey = subKey;
    oResult = TRUE;

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdateRemovedDevices(tUsbDeviceInfoList& p_rUsbDrvDeviceInfoList, CStdString p_deviceClassType)
{
    BOOL oResult = TRUE;
    CUsbDeviceInfo* pExistingDeviceInfo = 0;
    CUsbDeviceInfo* pDeviceInfo = 0;

    //Check Ports
    for(tUsbDeviceInfoList::iterator it = m_DeviceInfoList.begin(); it != m_DeviceInfoList.end(); it++)
    {
        pDeviceInfo = (*it);

        if(pDeviceInfo != 0)
        {
            //Reset OpenedBy if process is not alive
            if(!MmcIsProcessAvailable(pDeviceInfo->GetOpenedByProcess()))
            {
                pDeviceInfo->SetOpenedByProcess(0);
                pDeviceInfo->SetOpenedByInstance(0);
                pDeviceInfo->SetOpenedBy(_T(""));
            }

            //Update List
            if(!pDeviceInfo->IsOpened() || pDeviceInfo->IsOpenedBy(MmcGetCurrentProcessId(), m_lInstanceValue))
            {
                if(IsClassTypeEqual(p_deviceClassType, *pDeviceInfo))
                {
                    if(pDeviceInfo->IsPortChainEmpty())
                    {
                        // Workaround for FTDI devices when serial number can't be read via device descriptor in usb tree (Win10 USB 3.0)
                        if(!IsPortExistingInList(*pDeviceInfo, p_rUsbDrvDeviceInfoList, pExistingDeviceInfo, CUsbDeviceInfo::CM_DRV_PORT_INFOS))
                        {
                            // Reset PluggedIn Flag
                            pDeviceInfo->SetPluggedIn(0);
                        }
                    }
                    else
                    {
                        // Standard behaviour
                        if(!IsPortExistingInList(*pDeviceInfo, p_rUsbDrvDeviceInfoList, pExistingDeviceInfo, CUsbDeviceInfo::CM_PORT_CHAIN))
                        {
                            // Reset PluggedIn Flag
                            pDeviceInfo->SetPluggedIn(0);
                        }
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdateAddedDevices(tUsbDeviceInfoList& p_rUsbDrvDeviceInfoList, tPortList& p_rOpenPortList)
{
    BOOL oResult = TRUE;
    BOOL oRestarted(FALSE);
    BOOL oIsPortExistingInList(FALSE);
    CUsbDeviceInfo* pNewDeviceInfo(0);

    //Check Ports
    tUsbDeviceInfoList::iterator it = p_rUsbDrvDeviceInfoList.begin();
    while(it != p_rUsbDrvDeviceInfoList.end())
    {
        //Init
        oRestarted = FALSE;

        CUsbDeviceInfo* pUsbDrvDeviceInfo = (*it);
        if(pUsbDrvDeviceInfo)
        {
            CUsbDeviceInfo* pExistingDeviceInfo(0);

            if(pUsbDrvDeviceInfo->IsPortChainEmpty())
            {
                // Workaround for FTDI devices when serial number can't be read via device descriptor in usb tree (Win10 USB 3.0)
                oIsPortExistingInList = IsPortExistingInList(*pUsbDrvDeviceInfo, m_DeviceInfoList, pExistingDeviceInfo, CUsbDeviceInfo::CM_DRV_PORT_INFOS);
            }
            else
            {
                // Standard behaviour
                oIsPortExistingInList = IsPortExistingInList(*pUsbDrvDeviceInfo, m_DeviceInfoList, pExistingDeviceInfo, CUsbDeviceInfo::CM_PORT_CHAIN);
            }

            if(oIsPortExistingInList && pExistingDeviceInfo)
            {
                if(pUsbDrvDeviceInfo->GetClassType() != _T("CUsbDeviceInfo"))
                {
                    pNewDeviceInfo = pUsbDrvDeviceInfo->Clone();

                    pNewDeviceInfo->InitState(pExistingDeviceInfo);
                    pNewDeviceInfo->InitPortName(pExistingDeviceInfo);

                    if(ReplaceInList(m_DeviceInfoList, pExistingDeviceInfo, pNewDeviceInfo))
                    {
                        //Port is plugged in
                        if(!pNewDeviceInfo->IsOpened())
                        {
                            //Port is not opened by any process
                            pNewDeviceInfo->Update(*pUsbDrvDeviceInfo);
                            pNewDeviceInfo->SetPluggedIn(1);
                        }
                        else if(pNewDeviceInfo->IsOpenedBy(MmcGetCurrentProcessId(), m_lInstanceValue))
                        {
                            if(!pNewDeviceInfo->IsPluggedIn())
                            {
                                pNewDeviceInfo->Update(*pUsbDrvDeviceInfo);
                                pNewDeviceInfo->SetPluggedIn(1);

                                //Reopen Port
                                SetReopenFlag(pNewDeviceInfo->GetPortName(), p_rOpenPortList);
                            }
                            else if(pNewDeviceInfo->IsKindOf(_T("CHidDeviceInfo")))
                            {
                                //ODR: Always reopen port in case of fast Remove/Add sequence after Escon Firmware Update
                                pNewDeviceInfo->Update(*pUsbDrvDeviceInfo);
                                pNewDeviceInfo->SetPluggedIn(1);

                                //Reopen Port
                                SetReopenFlag(pNewDeviceInfo->GetPortName(), p_rOpenPortList);
                            }
                        }
                    }
                    else
                    {
                        //Release
                        delete pNewDeviceInfo;
                    }

                    //Remove from List
                    p_rUsbDrvDeviceInfoList.remove(pUsbDrvDeviceInfo);
                    delete pUsbDrvDeviceInfo;

                    //restart check
                    if(p_rUsbDrvDeviceInfoList.size() > 0)
                    {
                        it = p_rUsbDrvDeviceInfoList.begin();
                        oRestarted = TRUE;
                    }
                    else
                    {
                        //Last Element removed
                        break;
                    }
                }
            }
        }

        //Next DeviceInfo
        if(!oRestarted) it++;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdateNewAddedDevices(tUsbDeviceInfoList& p_rUsbDrvDeviceInfoList, tPortList& p_rOpenPortList)
{
    CUsbDeviceInfo* pUsbDrvDeviceInfo = 0;
    CUsbDeviceInfo* pFirstNonPluggedInDeviceInfo = 0;
    BOOL oResult = TRUE;

    //Check Ports
    for(tUsbDeviceInfoList::iterator it = p_rUsbDrvDeviceInfoList.begin(); it != p_rUsbDrvDeviceInfoList.end(); it++)
    {
        pUsbDrvDeviceInfo = (*it);
        if(pUsbDrvDeviceInfo)
        {
            if(pUsbDrvDeviceInfo->GetClassType() != _T("CUsbDeviceInfo"))
            {
                if(FindNonPluggedInDeviceInfoInList(m_DeviceInfoList, pUsbDrvDeviceInfo->GetClassType(), pFirstNonPluggedInDeviceInfo) && pFirstNonPluggedInDeviceInfo)
                {
                    //Port is plugged in
                    if(!pFirstNonPluggedInDeviceInfo->IsOpened())
                    {
                        //Update with new DeviceInfos
                        CUsbDeviceInfo* pNewDeviceInfo = pUsbDrvDeviceInfo->Clone();
                        pNewDeviceInfo->InitState(pFirstNonPluggedInDeviceInfo);
                        pNewDeviceInfo->InitPortName(pFirstNonPluggedInDeviceInfo);

                        if(ReplaceInList(m_DeviceInfoList, pFirstNonPluggedInDeviceInfo, pNewDeviceInfo))
                        {
                            pNewDeviceInfo->SetPluggedIn(1);
                        }
                        else
                        {
                            //Release
                            delete pNewDeviceInfo;
                        }
                    }
                    else if(pFirstNonPluggedInDeviceInfo->IsOpenedBy(MmcGetCurrentProcessId(), m_lInstanceValue))
                    {
                        if(!pFirstNonPluggedInDeviceInfo->IsPluggedIn())
                        {
                            //Update with new DeviceInfos
                            CUsbDeviceInfo* pNewDeviceInfo = pUsbDrvDeviceInfo->Clone();
                            pNewDeviceInfo->InitState(pFirstNonPluggedInDeviceInfo);
                            pNewDeviceInfo->InitPortName(pFirstNonPluggedInDeviceInfo);

                            if(ReplaceInList(m_DeviceInfoList, pFirstNonPluggedInDeviceInfo, pNewDeviceInfo))
                            {
                                pNewDeviceInfo->SetPluggedIn(1);

                                //Reopen Port
                                SetReopenFlag(pNewDeviceInfo->GetPortName(), p_rOpenPortList);
                            }
                            else
                            {
                                //Release
                                delete pNewDeviceInfo;
                            }
                        }
                    }
                }
                else
                {
                    //Add new entry
                    pUsbDrvDeviceInfo->SetOpenedBy(_T(""));
                    pUsbDrvDeviceInfo->SetOpenedByProcess(0);
                    pUsbDrvDeviceInfo->SetOpenedByInstance(0);
                    pUsbDrvDeviceInfo->SetPluggedIn(1);
                    AddToDeviceInfoList(*pUsbDrvDeviceInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::SetReopenFlag(CStdString p_PortName, tPortList& p_rOpenPortList)
{
    BOOL oResult = FALSE;

    for(tPortList::iterator it = p_rOpenPortList.begin(); it != p_rOpenPortList.end(); it++)
    {
        CPortBase* pPort = (*it);
        if(pPort)
        {
            //Get PortName
            CStdString portName = _T("");
            if(pPort->GetPortName(&portName))
            {
                //Check PortName
                if(p_PortName.Compare(portName) == 0)
                {
                    //Set Reopen Flag to force reopen of port
                    oResult = pPort->SetReopenFlag();
                    break;
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::IsPortNameUsed(CStdString p_PortName, tUsbDeviceInfoList& p_rDeviceInfoList)
{
    CStdString portName = _T("");
    BOOL oResult = FALSE;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo->GetPortName(portName))
        {
            if(portName.CompareNoCase(p_PortName) == 0)
            {
                oResult = TRUE;
                break;
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetNextFreePortName(CStdString& p_rPortName)
{
    int lPortIndex = 0;
    TCHAR szPortName[32];
    BOOL oResult = FALSE;

    memset(szPortName, 0, 32);

    while(!oResult)
    {
        _stprintf(szPortName, _T("USB%i"), lPortIndex);

        if(!IsPortNameUsed(szPortName, m_DeviceInfoList))
        {
            p_rPortName = szPortName;
            oResult = TRUE;
        }

        lPortIndex++;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::IsPortExistingInList(CUsbDeviceInfo& p_rDeviceInfo, tUsbDeviceInfoList& p_rDeviceInfoList, CUsbDeviceInfo*& p_rpExistingDeviceInfo, CUsbDeviceInfo::ECompareMode p_eCompareMode)
{
    BOOL oResult = FALSE;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo->IsEqual(p_rDeviceInfo, p_eCompareMode))
        {
            p_rpExistingDeviceInfo = pDeviceInfo;
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::FindNonPluggedInDeviceInfoInList(tUsbDeviceInfoList& p_rDeviceInfoList, CStdString p_deviceClassType, CUsbDeviceInfo*& p_rpNonPluggedInDeviceInfo)
{
    BOOL oResult = FALSE;
    CUsbDeviceInfo* pDeviceInfo = 0;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        pDeviceInfo = (*it);

        if(IsClassTypeEqual(p_deviceClassType, *pDeviceInfo))
        {
            if(!pDeviceInfo->IsPluggedIn())
            {
                p_rpNonPluggedInDeviceInfo = pDeviceInfo;
                oResult = TRUE;
                break;
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::AddToDeviceInfoList(CUsbDeviceInfo& p_rDeviceInfo)
{
    CUsbDeviceInfo* pDeviceInfo = 0;
    CStdString portName = _T("");
    BOOL oResult = FALSE;

    if(GetNextFreePortName(portName))
    {
        pDeviceInfo = p_rDeviceInfo.Clone();
        pDeviceInfo->Init(p_rDeviceInfo);
        pDeviceInfo->SetPortName(portName);
        pDeviceInfo->SetPluggedIn(p_rDeviceInfo.GetPluggedIn());
        pDeviceInfo->SetOpenedByProcess(p_rDeviceInfo.GetOpenedByProcess());
        pDeviceInfo->SetOpenedByInstance(p_rDeviceInfo.GetOpenedByInstance());
        m_DeviceInfoList.push_back(pDeviceInfo);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::DeleteFromDeviceInfoList(CUsbDeviceInfo* p_pDeviceInfo)
{
    BOOL oResult = FALSE;
    CUsbDeviceInfo* pDeviceInfo = 0;

    for(tUsbDeviceInfoList::iterator it=m_DeviceInfoList.begin(); it!=m_DeviceInfoList.end(); it++)
    {
        pDeviceInfo = (*it);

        if(pDeviceInfo == p_pDeviceInfo)
        {
            m_DeviceInfoList.erase(it);
            delete pDeviceInfo;
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::SortPortNames(CStdStringArray& p_rPortNames)
{
    size_t n = 0;
    CStdString portName = _T("");
    CStdString nextPortName = _T("");
    BOOL oExchanged = FALSE;
    BOOL oResult = TRUE;

    //Bubble Sort Algorithm
    n = p_rPortNames.size() - 1;

    do
    {
        oExchanged = FALSE;
        for(int i = 0; i < (int)n; i++)
        {
            //DeviceInfo
            portName = p_rPortNames.at(i);

            //Next DeviceInfo
            nextPortName = p_rPortNames.at(i+1);

            //Compare PortName
            if(portName > nextPortName)
            {
                //Interchange Values
                p_rPortNames[i]=nextPortName;
                p_rPortNames[i+1]=portName;
                oExchanged = TRUE;
            }
        }
        n = n - 1;
    }
    while(oExchanged);

    return oResult;
}

BOOL CUsbDeviceInfoHandling::SortDeviceInfoListByPortName(tUsbDeviceInfoList& p_rDeviceInfoList)
{
    BOOL oResult = FALSE;
    size_t n = 0;
    CUsbDeviceInfo* pDeviceInfo = NULL;
    CUsbDeviceInfo* pNextDeviceInfo = NULL;
    BOOL oExchanged = FALSE;

    if(p_rDeviceInfoList.size() > 0)
    {
        oResult = TRUE;

        //Bubble Sort Algorithm
        n = p_rDeviceInfoList.size() - 1;
        do
        {
            oExchanged = FALSE;
            for(int i = 0; i < (int)n; i++)
            {
                tUsbDeviceInfoList::iterator itDeviceInfo;
                tUsbDeviceInfoList::iterator itNextDeviceInfo;
                int lIndex = 0;
                for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
                {
                    if(lIndex == i)
                    {
                        pDeviceInfo = (*it);
                        itDeviceInfo = it;
                    }

                    if(lIndex == i + 1)
                    {
                        pNextDeviceInfo = (*it);
                        itNextDeviceInfo = it;
                        break;
                    }

                    lIndex++;
                }

                //Compare DeviceInfos
                if(pDeviceInfo && pNextDeviceInfo)
                {
                    //Compare PortName
                    if(pDeviceInfo->GetPortName() > pNextDeviceInfo->GetPortName())
                    {
                        //Interchange Values
                        (*itDeviceInfo) = pNextDeviceInfo;
                        (*itNextDeviceInfo) = pDeviceInfo;
                        oExchanged = TRUE;
                    }
                }
            }

            n = n - 1;
        }
        while(oExchanged);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::FindDeviceInfoByPortName(CStdString p_PortName, tUsbDeviceInfoList& p_rDeviceInfoList, CUsbDeviceInfo*& p_rpDeviceInfo)
{
    BOOL oResult = FALSE;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo->IsPortNameEqual(p_PortName))
        {
            p_rpDeviceInfo = pDeviceInfo;
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::FindDeviceInfoBySerialNumber(CUsbDeviceInfo& p_rDeviceInfo, tUsbDeviceInfoList& p_rDeviceInfoList, CUsbDeviceInfo*& p_rpFoundDeviceInfo)
{
    BOOL oResult = FALSE;

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        CUsbDeviceInfo* pDeviceInfo = (*it);
        if(pDeviceInfo->IsSerialNumberEqual(p_rDeviceInfo.GetSerialNumber()))
        {
            p_rpFoundDeviceInfo = pDeviceInfo;
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::UpdateRegistry(CStdString p_PortName, BOOL p_oOpened)
{
    BOOL oResult = FALSE;
    CStdString subKey = _T("");
    CUsbDeviceInfo* pDeviceInfo(0);
    BOOL oPortFound(FALSE);

    //Init
    GetRegistryKey(subKey);
    CMmcUserSettings userSettings(subKey, FALSE);

    list<CMmcUserSettings*> children = userSettings.GetChildren();
    for(list<CMmcUserSettings*>::iterator it = children.begin(); it != children.end(); it++)
    {
        CMmcUserSettings* pChild = (*it);
        if(!oPortFound)
        {
            if(ReadDeviceInfoFromRegistry(pChild, pDeviceInfo) && pDeviceInfo)
            {
                if(pDeviceInfo->GetPortName() == p_PortName)
                {
                    //Update Opened Flag
                    oResult = WriteOpenedToRegistry(pChild, p_oOpened);
                    oPortFound = TRUE;
                }
            }

            //Release
            if(pDeviceInfo) delete pDeviceInfo;
        }

        //Delete Child
        delete pChild;
    }

    return oResult;
}

CUsbDeviceInfo* CUsbDeviceInfoHandling::GetDeviceInfo(CStdString p_PortName)
{
    CUsbDeviceInfo* pResultDeviceInfo(0);
    CUsbDeviceInfo* pDeviceInfo(0);

    if(FindDeviceInfoByPortName(p_PortName, m_DeviceInfoList, pDeviceInfo) && pDeviceInfo)
    {
        pResultDeviceInfo = pDeviceInfo;
    }

    return pResultDeviceInfo;
}

BOOL CUsbDeviceInfoHandling::UpdateDeviceInfo(tUsbDeviceInfoList& p_rDeviceInfoList, CStdString p_deviceClassType, CStdString p_PortName)
{
    CUsbDeviceInfo* pDeviceInfo(0);
    CUsbDeviceInfo* pDeviceInfoToBeUpdated(0);
    BOOL oResult = FALSE;

    if(FindDeviceInfoByPortName(p_PortName, m_DeviceInfoList, pDeviceInfoToBeUpdated) && pDeviceInfoToBeUpdated)
    {
        if(FindDeviceInfoBySerialNumber(*pDeviceInfoToBeUpdated, p_rDeviceInfoList, pDeviceInfo) && pDeviceInfo)
        {
            pDeviceInfoToBeUpdated->Init(*pDeviceInfo);
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::IsClassTypeEqual(CStdString deviceClassType, CUsbDeviceInfo& p_rDeviceInfo)
{
    BOOL oResult(FALSE);

    oResult = (deviceClassType == p_rDeviceInfo.GetClassType());

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ReplaceInList(tUsbDeviceInfoList&p_rUsbDrvDeviceInfoList, CUsbDeviceInfo* p_pDeviceInfo, CUsbDeviceInfo* p_pNewDeviceInfo)
{
    BOOL oResult(TRUE);
    CUsbDeviceInfo* pDeviceInfo(0);

    for(tUsbDeviceInfoList::iterator it = p_rUsbDrvDeviceInfoList.begin(); it != p_rUsbDrvDeviceInfoList.end(); it++)
    {
        pDeviceInfo = (*it);

        if(pDeviceInfo == p_pDeviceInfo)
        {
            (*it) = p_pNewDeviceInfo;
            delete pDeviceInfo;
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ResetVendorProductIds()
{
    m_VendorProductIdList.clear();

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::AddVendorProductIds(CStdString p_DriverName, WORD p_VendorId, WORD p_ProductId)
{
    SVendorProductIDs vendorProductIDs;

    //Init
    vendorProductIDs.driverName = p_DriverName;
    vendorProductIDs.vendorId = p_VendorId;
    vendorProductIDs.productId = p_ProductId;

    //Add
    if(!IsVendorProductIdExisting(p_DriverName, p_VendorId, p_ProductId))
    {
        m_VendorProductIdList.push_back(vendorProductIDs);
    }

    return TRUE;
}

BOOL CUsbDeviceInfoHandling::AddVendorProductIds(CStdString p_DriverName, CStdString p_DriverParameters)
{
    BOOL oResult(TRUE);
    WORD vendorId(0);
    WORD productId(0);

    //Driver Parameter IDs (Format: "VID_XXXX&PIDYYYY VID_ZZZZ&PID_WWWW ...")
    while(ExtractVendorProductIDs(p_DriverParameters, vendorId, productId))
    {
        oResult = AddVendorProductIds(p_DriverName, vendorId, productId);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetVendorProductIds(CStdString p_DriverName, list<SVendorProductIDs>& p_rIdsList)
{
    BOOL oResult(TRUE);

    for(list<SVendorProductIDs>::iterator it = m_VendorProductIdList.begin(); it != m_VendorProductIdList.end(); it++)
    {
        SVendorProductIDs vendorProductIds = (*it);
        if(vendorProductIds.driverName == p_DriverName)
        {
            p_rIdsList.push_back(vendorProductIds);
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::GetAllVendorProductIds(list<SVendorProductIDs>& p_rIdsList)
{
    BOOL oResult(TRUE);

    for(list<SVendorProductIDs>::iterator it = m_VendorProductIdList.begin(); it != m_VendorProductIdList.end(); it++)
    {
        SVendorProductIDs vendorProductIds = (*it);
        p_rIdsList.push_back(vendorProductIds);
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::AreVendorProductIdsSupportedByDriver(CStdString p_DriverName, WORD p_VendorId, WORD p_ProductId)
{
    return IsVendorProductIdExisting(p_DriverName, p_VendorId, p_ProductId);
}

BOOL CUsbDeviceInfoHandling::IsVendorProductIdExisting(CStdString p_DriverName, WORD p_VendorId, WORD p_ProductId)
{
    BOOL oResult(FALSE);

    for(list<SVendorProductIDs>::iterator it = m_VendorProductIdList.begin(); it != m_VendorProductIdList.end(); it++)
    {
        SVendorProductIDs vendorProductIds = (*it);
        if(vendorProductIds.driverName == p_DriverName)
        {
            if(vendorProductIds.vendorId == p_VendorId)
            {
                if(vendorProductIds.productId == p_ProductId)
                {
                    oResult = TRUE;
                    break;
                }
            }
        }
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ExtractVendorProductIDs(CStdString& p_rDriverParameters, WORD& p_rVendorId, WORD& p_rProductId)
{
    BOOL oResult(FALSE);
    int lStartIndex(0);
    int lEndIndex(0);
    CStdString vendorProductIDs(_T(""));

    //Init
    p_rDriverParameters = p_rDriverParameters.Trim();
    lStartIndex = p_rDriverParameters.Find(_T("VID_"));
    if(lStartIndex != -1)
    {
        lEndIndex = p_rDriverParameters.Find(_T(" "));
        if(lEndIndex != -1)
        {
            //Parameters following
            vendorProductIDs = p_rDriverParameters.Mid(lStartIndex, lEndIndex - lStartIndex);
            p_rDriverParameters = (p_rDriverParameters.Left(lStartIndex) + p_rDriverParameters.Right((int)p_rDriverParameters.length() - lEndIndex));
        }
        else
        {
            //Last Ids
            vendorProductIDs = p_rDriverParameters.Right((int)p_rDriverParameters.length() - lStartIndex);
            p_rDriverParameters = p_rDriverParameters.Left(lStartIndex);
        }

        //Extract Values
        oResult = TRUE;
        if(oResult && !ExtractVendorIdValue(vendorProductIDs, p_rVendorId)) oResult = FALSE;
        if(oResult && !ExtractProductIdValue(vendorProductIDs, p_rProductId)) oResult = FALSE;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ExtractVendorIdValue(CStdString p_VendorProductIDs, WORD& p_rVendorId)
{
    const CStdString VID(_T("VID_"));
    const CStdString SEPERATOR(_T("&"));

    BOOL oResult(FALSE);
    int lStartIndex(0);
    int lEndIndex(0);
    CStdString id(_T(""));

    //Init (Format: VID_XXXX&PID_YYYY)
    p_VendorProductIDs = p_VendorProductIDs.Trim();

    //VendorId
    lStartIndex = p_VendorProductIDs.Find(VID);
    lEndIndex = p_VendorProductIDs.Find(SEPERATOR);
    if((lStartIndex != -1) && (lEndIndex != -1))
    {
        id = p_VendorProductIDs.Mid(lStartIndex + (int)VID.length(), lEndIndex - lStartIndex - (int)VID.length());
        p_rVendorId = (WORD)_tcstol(id, 0, 16);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CUsbDeviceInfoHandling::ExtractProductIdValue(CStdString p_VendorProductIDs, WORD& p_rProductId)
{
    const CStdString PID(_T("PID_"));

    BOOL oResult(FALSE);
    int lStartIndex(0);
    CStdString id(_T(""));

    //Init (Format: VID_XXXX&PID_YYYY)
    p_VendorProductIDs = p_VendorProductIDs.Trim();

    //VendorId
    lStartIndex = p_VendorProductIDs.Find(PID);
    if(lStartIndex != -1)
    {
        id = p_VendorProductIDs.Mid(lStartIndex + (int)PID.length(), (int)p_VendorProductIDs.length() - lStartIndex - (int)PID.length());
        p_rProductId = (WORD)_tcstol(id, 0, 16);
        oResult = TRUE;
    }

    return oResult;
}

CStdString CUsbDeviceInfoHandling::RemoveFileExtension(CStdString p_rFileName)
{
    CStdString result(p_rFileName);

    int lIndex = result.ReverseFind('.');
    if(lIndex != -1)
    {
        result = result.Left(lIndex);
    }

    return result;
}

BOOL CUsbDeviceInfoHandling::UpdateRegistryIgnoreHwSerialNumber()
{
    BOOL oResult(FALSE);

#ifdef WINVER

    CStdString root = _T("System\\CurrentControlSet\\Control\\usbflags");
    HKEY rootKey(0);
    DWORD dDisposition(0);
    SVendorProductIDs sVendorProductIds;
    long lResult(0);

    //Usb Flags Registry Entry
    lResult = RegCreateKeyEx(HKEY_LOCAL_MACHINE, root, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &rootKey, &dDisposition);
    if(lResult == ERROR_SUCCESS)
    {
        //Init
        oResult = TRUE;

        //Add all PID/VIDs to registry
        for(list<SVendorProductIDs>::iterator it = m_VendorProductIdList.begin(); it != m_VendorProductIdList.end(); it++)
        {
            sVendorProductIds = *(it);

            if(!WriteToRegistryIgnoreHwSerialNumber(rootKey, sVendorProductIds))
            {
                oResult = FALSE;
            }
        }
    }
#else
    oResult = TRUE;
#endif

    return oResult;
}

#ifdef WINVER

BOOL CUsbDeviceInfoHandling::WriteToRegistryIgnoreHwSerialNumber(HKEY& p_rRootKey, SVendorProductIDs& p_rVendorProductIds)
{
    BOOL oResult(FALSE);
    CStdString ignoreHwSerNumEntry = _T("");
    BYTE ubValue(1);
    long lResult(0);

    //Init Entry
    ignoreHwSerNumEntry.Format(_T("IgnoreHWSerNum%.4X%.4X"), p_rVendorProductIds.vendorId, p_rVendorProductIds.productId);

    //IgnoreHwSerNum  Entry
    lResult = RegSetValueEx(p_rRootKey, ignoreHwSerNumEntry, 0, REG_BINARY, (BYTE*)&ubValue, sizeof(ubValue));
    if(lResult == ERROR_SUCCESS)
    {
        oResult = TRUE;
    }

    return oResult;
}

#endif
