#include "stdafx.h"

#include <CommunicationModel/CommonLayer/Classes/Commands/Interface/BaseClasses/Command_I.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Interface/Command_I_HID.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Interface/Command_I_USB.h>
#include <CommunicationModel/Interface/BaseClasses/InterfaceManagerBase.h>
#include <CommunicationModel/Interface/Classes/Gateway/USB/Classes/UsbDeviceInfoHandling.h>
#include <Drv/Hid/MmcHidHndl.h>
#include <Log/MmcLog/MmcLog.h>
#include <Log/MmcLogger.h>
#include <Time/MmcHiResTimer.h>
#include <Process/MmcProcess.h>
#include <Thread/MmcSingleLock.h>

#include <CommunicationModel/Interface/Classes/Gateway/USB/GatewayUSBToHidDrv.h>

using namespace std;

#ifndef ERROR_SUCCESS
    #define ERROR_SUCCESS 0
#endif

#ifndef ERROR_IO_PENDING
    #define ERROR_IO_PENDING 997
#endif

#ifndef WAIT_OBJECT_0
    #define WAIT_OBJECT_0 0
#endif

DWORD CGatewayUSBToHidDrv::BAUDRATE_1000000 = 1000000;

CGatewayUSBToHidDrv::CGatewayUSBToHidDrv():
    m_oFirstInit(0),
    m_pInputBuffer(0),
    m_ulInputBufferSize(0),
    m_ulInputBufferSizeActual(0),
    m_ulGetInputBufferIndex(0),
    m_oUpdatePortOpened(TRUE),
    m_pInfoHandling(0)
{
    m_PortName = _T("");

    m_pHidHndl = new CMmcHidHndl();

    InitErrorHandling();
    InitBaudrateSelection();
    InitDefaultPortSettings();

    if(m_pHidHndl != 0)
    {
        m_pHidHndl->ResetOutReport();
        m_pHidHndl->Init();
    }

    InitInputBuffer(1024);
}

CGatewayUSBToHidDrv::~CGatewayUSBToHidDrv()
{
    if(m_pHidHndl != 0)
    {
        delete m_pHidHndl;
    }

    DeleteInputBuffer();
}

BOOL CGatewayUSBToHidDrv::InitDriverInfos(CStdString p_DriverName, CStdString p_DriverParameters)
{
    BOOL oResult(FALSE);

    if(CGatewayIToDrv::InitDriverInfos(p_DriverName, p_DriverParameters))
    {
        oResult = InitVendorProductIds(p_DriverName, p_DriverParameters);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitBaudrateSelection()
{
    BOOL oResult(TRUE);

    m_BaudrateSelection.clear();
    m_BaudrateSelection.push_back(BAUDRATE_1000000);

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitDefaultPortSettings()
{
    BOOL oResult(TRUE);

    m_dDefaultBaudrate = BAUDRATE_1000000;
    m_dDefaultTimeout = 500;

    GetDefaultPortSettings(&m_dBaudrate, &m_dTimeout);

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CCommand_I* pCommand_I;

    if(CGateway::ProcessCommand(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle))
    {
        if(CheckLayers(p_pCommand, p_pManager))
        {
            pCommand_I = (CCommand_I*)p_pCommand;

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayUSBToHidDrv")));
            }

            switch(p_pCommand->GetCommandId())
            {
                case USB_WRITE_DATA:
                    {
                        oResult = Process_WriteData(pCommand_I);
                        break;
                    }
                case USB_READ_DATA:
                    {
                        oResult = Process_ReadData(pCommand_I);
                        break;
                    }
            }

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayUSBToHidDrv"), TRUE));
            }
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::Process_WriteData(CCommand_I* p_pCommand)
{
    //Constants
    const int MAX_RETRY(2);
    const int PARAMETER_INDEX_DATA(0);
    const int RETURN_PARAMETER_INDEX_NB_OF_BYTES_WRITTEN(0);

    //ReturnParameter
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Lock CriticalSection
        if(Lock(p_pCommand->GetTimeout()))
        {
           //Prepare DataBuffer
            DWORD ulDataBufferLength = p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
            char* pDataBuffer = 0;

            if(ulDataBufferLength > 0)
            {
                pDataBuffer = new char[ulDataBufferLength];
            }

            //Get Parameter Data
            p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, pDataBuffer, ulDataBufferLength);

            //Execute Command
            DWORD ulNbOfBytesWritten(0);

            if(CMmcLog::IsActive(LMT_DEBUG, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayUSBToHidDrv::WriteData->Before")));
            }

            oResult = WriteData(pDataBuffer, ulDataBufferLength, &ulNbOfBytesWritten, MAX_RETRY, &errorInfo);

            //Set ReturnParameter Data
            p_pCommand->SetStatus(oResult, &errorInfo);
            p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_NB_OF_BYTES_WRITTEN, &ulNbOfBytesWritten, sizeof(ulNbOfBytesWritten));

            if(pDataBuffer)
            {
                delete[] pDataBuffer;
            }

            if(CMmcLog::IsActive(LMT_DEBUG, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), _T("CGatewayUSBToHidDrv::WriteData->After"));
            }

            //Unlock CriticalSection
            Unlock();
       }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::Process_ReadData(CCommand_I* p_pCommand)
{
    //Constants
    const int MAX_RETRY(2);
    const int PARAMETER_INDEX_NB_OF_BYTES_TO_READ(0);
    const int RETURN_PARAMETER_INDEX_DATA(0);

    //Parameter
    DWORD ulNbOfBytesToRead(0);

    //ReturnParameter
    DWORD ulDataBufferLength(0);
    DWORD ulNbOfBytesRead(0);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Lock CriticalSection
        if(Lock(p_pCommand->GetTimeout()))
        {
            if(CMmcLog::IsActive(LMT_DEBUG, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayUSBToHidDrv::Process_ReadData->Begin")));
            }

            //Get Parameter Data
            p_pCommand->GetParameterData(PARAMETER_INDEX_NB_OF_BYTES_TO_READ, &ulNbOfBytesToRead, sizeof(ulNbOfBytesToRead));

            //Prepare DataBuffer
            ulDataBufferLength = ulNbOfBytesToRead;

            char* pDataBuffer = 0;

            if(ulDataBufferLength > 0)
            {
                pDataBuffer = new char[ulDataBufferLength];
            }

            //Execute Command
            oResult = ReadData(pDataBuffer, ulNbOfBytesToRead, &ulNbOfBytesRead, MAX_RETRY, &errorInfo);

            //Set ReturnParameter Data
            p_pCommand->SetStatus(oResult, &errorInfo);
            p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pDataBuffer, ulNbOfBytesRead);

            if(pDataBuffer != 0)
            {
                delete[] pDataBuffer;
            }

            if(CMmcLog::IsActive(LMT_DEBUG, CCommunicationMessageLevel(CML_I)))
            {
                CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), p_pCommand->GetLogMessage(_T("CGatewayUSBToHidDrv::Process_ReadData->After")));
            }

            //Unlock CriticalSection
            Unlock();
       }
    }

    return oResult;
}

CGateway* CGatewayUSBToHidDrv::Clone()
{
    CGatewayUSBToHidDrv* pClonedGateway;

    pClonedGateway = new CGatewayUSBToHidDrv();
    *pClonedGateway = *this;

    return pClonedGateway;
}

CGatewayUSBToHidDrv& CGatewayUSBToHidDrv::operator=(CGatewayUSBToHidDrv& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayIToDrv*)this) = *((CGatewayIToDrv*)&p_rOther);
        m_pInfoHandling = p_rOther.m_pInfoHandling;
    }

    return *this;
}

BOOL CGatewayUSBToHidDrv::OpenPort(CStdString p_PortName, CErrorInfo* p_pErrorInfo)
{
    const BOOL CHANGE_ONLY(FALSE);

    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);
    CUsbDeviceInfo* pDeviceInfo(0);

    if(m_pInfoHandling && m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        pDeviceInfo = m_pInfoHandling->GetDeviceInfo(p_PortName);
        if(pDeviceInfo && pDeviceInfo->IsKindOf(_T("CHidDeviceInfo")))
        {
            if(m_pHidHndl->OpenPort(*((CHidDeviceInfo*)pDeviceInfo)))
            {
                //Configure default settings for port
                if(ConfigurePortSettings(m_dBaudrate, m_dTimeout, CHANGE_ONLY, p_pErrorInfo, 0))
                {
                    m_PortName = p_PortName;
                    if(m_oUpdatePortOpened) UpdatePortOpened(m_PortName, TRUE);

                    oResult = TRUE;
                }
                else
                {
                    ClosePort();
                    oResult = FALSE;
                }
            }
            else
            {
                ClosePort();

                if(m_pErrorHandling) m_pErrorHandling->GetError(MmcGetLastError(), p_pErrorInfo);
                oResult = FALSE;
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_OPENING_PORT, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_LIBRARY_NOT_LOADED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ReloadDriver()
{
    //Delay only for retry
    MmcSleep(200);

    return TRUE;
}

BOOL CGatewayUSBToHidDrv::SetPortSettings(DWORD p_ulBaudrate, DWORD p_ulTimeout, BOOL p_oChangeOnly, CErrorInfo* p_pErrorInfo)
{
    return ConfigurePortSettings(p_ulBaudrate, p_ulTimeout, p_oChangeOnly, p_pErrorInfo, 1);
}

BOOL CGatewayUSBToHidDrv::ConfigurePortSettings(DWORD p_ulBaudrate, DWORD p_ulTimeout, BOOL p_oChangeOnly, CErrorInfo* p_pErrorInfo, BOOL p_oDoLock)
{
    BOOL oResult(TRUE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), FALSE);

    //Lock
    if(p_oDoLock) lock.Lock();

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            //Do only if changed
            if(!p_oChangeOnly || (p_ulTimeout != m_dTimeout) || (p_ulBaudrate != m_dBaudrate))
            {
                //Do only if changed
                if(oResult)
                {
                    if(!p_oChangeOnly || (p_ulTimeout != m_dTimeout))
                    {
                        //Set Timeout
                        if(!SetTimeout(p_ulTimeout))
                        {
                            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_SETTINGS, p_pErrorInfo);
                            oResult = FALSE;
                        }
                    }
                }

                //Do only if changed
                if(oResult)
                {
                    if(!p_oChangeOnly || (p_ulBaudrate != m_dBaudrate))
                    {
                        if(SetBaudrate(p_ulBaudrate))
                        {
                            //Baudrate is set
                            m_dBaudrate = p_ulBaudrate;
                        }
                        else
                        {
                            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_SETTINGS, p_pErrorInfo);
                            oResult = FALSE;
                        }
                    }
                }

                if(oResult)
                {
                    if(!p_oChangeOnly || !m_oFirstInit)
                    {
                        //Purge the buffer
                        if(!PurgeBuffer())
                        {
                            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_SETTINGS, p_pErrorInfo);
                            oResult = FALSE;
                        }
                        else
                        {
                            m_oFirstInit = TRUE;
                        }
                    }
                }
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_PORT_NOT_OPEN, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_LIBRARY_NOT_LOADED, p_pErrorInfo);
        oResult = FALSE;
    }

    //Tracing
    Trace_Settings(p_ulBaudrate, p_ulTimeout);

    //Unlock
    if(p_oDoLock) lock.Unlock();

    if(!oResult && m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_SET_PORT_SETTINGS, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayUSBToHidDrv::SetTimeout(DWORD p_ulTimeout)
{
    BOOL oResult(FALSE);

    if(m_pHidHndl != 0 && m_pHidHndl->IsOpened())
    {
        //Set timeout
        m_dTimeout = p_ulTimeout;

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::SetBaudrate(DWORD p_ulBaudrate)
{
    BOOL oResult(FALSE);

    if(m_pHidHndl != 0 && m_pHidHndl->IsOpened())
    {
        oResult = m_pHidHndl->SetBaudrate(p_ulBaudrate) ? true : false;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::PurgeBuffer()
{
    BOOL oResult(FALSE);

    if(m_pHidHndl != 0 && m_pHidHndl->IsOpened())
    {
        oResult = m_pHidHndl->PurgeBuffer() ? TRUE : FALSE;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::WriteData(void* p_pDataBuffer, DWORD p_ulNbOfBytesToWrite, DWORD* p_pulNbOfBytesWritten, int p_lMaxRetry, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE* pDataBuffer(0);
    DWORD ulNbOfBytesToWrite(0);
    DWORD ulNbOfBytesWritten(0);
    DWORD ulTotalNbOfBytesToWrite(0);
    DWORD ulTotalNbOfBytesWritten(0);
    DWORD ulMaxReportDataLength(0);
    int lRetryCount(0);
    CStdString msg;

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            //Init
            oResult = TRUE;
            ulMaxReportDataLength = m_pHidHndl->GetOutputReportByteLength() - 1;
            ulTotalNbOfBytesToWrite = p_ulNbOfBytesToWrite;

            msg.Format(_T("WriteData Buffer Before: NbOfBytesToWrite %i\n"), p_ulNbOfBytesToWrite);

            CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);

            while(oResult && (ulTotalNbOfBytesToWrite > 0))
            {
                //Init
                oResult = FALSE;
                lRetryCount = 0;

                //Init report buffer
                if(ulTotalNbOfBytesToWrite > ulMaxReportDataLength) ulNbOfBytesToWrite = ulMaxReportDataLength; else ulNbOfBytesToWrite = ulTotalNbOfBytesToWrite;
                pDataBuffer = (BYTE*)p_pDataBuffer + ulTotalNbOfBytesWritten;

                //Retry loop
                while(!oResult && (lRetryCount < p_lMaxRetry))
                {
                    //Init
                    ulNbOfBytesWritten = 0;

                    //Writer report
                    if(WriteReport((BYTE*)pDataBuffer, ulNbOfBytesToWrite, &ulNbOfBytesWritten, p_pErrorInfo) && (ulNbOfBytesToWrite == ulNbOfBytesWritten))
                    {
                        //Succeeded
                        ulTotalNbOfBytesWritten += ulNbOfBytesWritten;
                        ulTotalNbOfBytesToWrite -= ulNbOfBytesWritten;
                        oResult = TRUE;
                    }
                    else
                    {
                        //Retry
                        lRetryCount++;
                    }
                }
            }

            if(p_pulNbOfBytesWritten) *p_pulNbOfBytesWritten = ulTotalNbOfBytesWritten;

            msg.Format(_T("WriteData Buffer End: NbOfBytesToWrite %i NbOfBytesWritten %i\n"), p_ulNbOfBytesToWrite, ulTotalNbOfBytesWritten);

            CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::WaitForOverlappedResult(DWORD& p_rulNumberOfBytesTransferred)
{
    BOOL oResult = FALSE;
    DWORD ulError = ERROR_SUCCESS;
    DWORD ulTime = 0;
    unsigned long ulResult(0);

    //Check read status
    switch(ulError = MmcGetLastError())
    {
        case ERROR_IO_PENDING:
        {
            //Start time
            DWORD ulStart = MmcGetTickCount();

            while(!oResult && (ulTime < m_dTimeout))
            {
                ulResult = HidWaitForOverlapped(m_dTimeout);

                if(WAIT_OBJECT_0 == ulResult)
                {
                    oResult = HidGetOverlappedResult(&p_rulNumberOfBytesTransferred, FALSE);
                }

                //Update time
                ulTime = MmcGetTickCount() - ulStart;
            }

            if(!oResult)
            {
                if(m_pHidHndl->CancelIo())
                {
                    DWORD ulNumberOfBytesTransferred(0);
                    ulTime = 0;
                    ulStart = MmcGetTickCount();
                    while(ulTime < (m_dTimeout / 2))
                    {
                        MmcSleep(10);
                        HidGetOverlappedResult(&ulNumberOfBytesTransferred, FALSE);
                        ulTime = MmcGetTickCount() - ulStart;
                    }
                }
            }
        }; break;
        default:
        {
            oResult = FALSE;
            break;
        };
    }
    return oResult;
}

BOOL CGatewayUSBToHidDrv::WriteReport(BYTE* p_pDataBuffer, DWORD p_ulNbOfBytesToWrite, DWORD* p_pulNbOfBytesWritten, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    char* pbOutputReport(0);
    DWORD ulNumberOfBytesToWrite(0);
    DWORD ulNumberOfBytesWritten(0);
    DWORD ulOutputReportLength(0);
    DWORD ulTotalNumberOfBytesToWrite(0);
    int lReportNumber(0);

    //Reset error info
    if(p_pErrorInfo) p_pErrorInfo->Reset();

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            ulOutputReportLength = m_pHidHndl->GetOutputReportByteLength();
            ulTotalNumberOfBytesToWrite = p_ulNbOfBytesToWrite;

            //Execute write
            if(p_pDataBuffer && p_ulNbOfBytesToWrite > 0 && ulOutputReportLength > 0)
            {
                //Limit size
                if(ulTotalNumberOfBytesToWrite <= (ulOutputReportLength - sizeof(BYTE)))
                {
                    ulNumberOfBytesToWrite = ulTotalNumberOfBytesToWrite;
                }
                else
                {
                    ulNumberOfBytesToWrite = ulOutputReportLength - sizeof(BYTE);
                }

                //Allocate buffer
                pbOutputReport = new char[ulOutputReportLength];
                memset(pbOutputReport, 0, ulOutputReportLength);

                //Set report number
                memcpy(pbOutputReport, &lReportNumber, sizeof(BYTE));
                lReportNumber++;
                memcpy(pbOutputReport + sizeof(BYTE), p_pDataBuffer, ulNumberOfBytesToWrite);

                //InhibitTime
                WaitForWriteInhibitTimeElapsed();

                //Write to device
                if(!HidWriteFile(pbOutputReport, ulOutputReportLength, &ulNumberOfBytesWritten))
                {
                    oResult = WaitForOverlappedResult(ulNumberOfBytesWritten);
                    if(oResult)
                    {
                        if(p_pulNbOfBytesWritten) *p_pulNbOfBytesWritten = ulNumberOfBytesToWrite;
                    }
                    else
                    {
                        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HID_WRITE_DATA, p_pErrorInfo);
                    }
                }
                else
                {
                    oResult = TRUE;

                    if(p_pulNbOfBytesWritten) *p_pulNbOfBytesWritten = ulNumberOfBytesToWrite;
                }

                //InhibitTime
                StartWriteInhibitTime();

                //Free buffer
                if(pbOutputReport)
                {
                    delete []pbOutputReport;
                    pbOutputReport = 0;
                }
            }
            else
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
                oResult = FALSE;
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_PORT_NOT_OPEN, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_LIBRARY_NOT_LOADED, p_pErrorInfo);
        oResult = FALSE;
    }

    Trace_WriteData(oResult, p_pDataBuffer, p_ulNbOfBytesToWrite, p_pulNbOfBytesWritten);

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ReadData(void* p_pDataBuffer, DWORD p_ulNbOfBytesToRead, DWORD* p_pulNbOfBytesRead, int p_lMaxRetry, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CStdString msg;

    msg.Format(_T("ReadData BlockBuffer Before: NbOfBytesToRead %i\n"), p_ulNbOfBytesToRead);
    CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);

    if(GetDataFromInputBuffer((BYTE*)p_pDataBuffer, p_ulNbOfBytesToRead, p_pulNbOfBytesRead, p_pErrorInfo))
    {
        //Buffered data from input buffer
        oResult = TRUE;
    }
    else
    {
        //Update input buffer
        if(UpdateInputBuffer(p_ulNbOfBytesToRead, p_lMaxRetry, p_pErrorInfo))
        {
            oResult = GetDataFromInputBuffer((BYTE*)p_pDataBuffer, p_ulNbOfBytesToRead, p_pulNbOfBytesRead, p_pErrorInfo);
        }
    }

    if(p_pulNbOfBytesRead)
    {
        msg.Format(_T("ReadData BlockBuffer End: NbOfBytesToRead %i NbOfBytesRead %i\n"), p_ulNbOfBytesToRead, *p_pulNbOfBytesRead);
        CMmcLog::Log(LMT_DEBUG, CCommunicationMessageLevel(CML_I), msg);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ReadReport(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE* pubInputReport(0);
    BYTE* pubData(0);
    DWORD ulInputReportLength(0);
    DWORD ulInputReportRead(0);

    //Reset error info
    if(p_pErrorInfo) p_pErrorInfo->Reset();

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            ulInputReportLength = m_pHidHndl->GetInputReportByteLength();

            //Execute read
            if(m_pInputBuffer && ulInputReportLength > 0)
            {
                //Allocate buffer
                pubInputReport = new BYTE[ulInputReportLength];
                memset(pubInputReport, 0, ulInputReportLength);

                //Read frame
                oResult = HidReadFile(pubInputReport, ulInputReportLength, &ulInputReportRead);
                if(!oResult)
                {
                    oResult = WaitForOverlappedResult(ulInputReportRead);
                }

                if(oResult && ulInputReportRead != 0)
                {
                    //Init
                    pubData = pubInputReport;

                    //Get report number
                    pubData++;
                    ulInputReportRead--;

                    //Copy data bytes
                    for(DWORD ul = 0; ul < ulInputReportRead; ul++)
                    {
                        *(m_pInputBuffer + m_ulInputBufferSizeActual) = *pubData++;
                        m_ulInputBufferSizeActual++;
                    }
                }
                else
                {
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HID_READ_DATA, p_pErrorInfo);
                }

                //Free buffer
                if(pubInputReport)
                {
                    delete []pubInputReport;
                    pubInputReport = 0;
                }
            }
            else
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
                oResult = FALSE;
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_PORT_NOT_OPEN, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_LIBRARY_NOT_LOADED, p_pErrorInfo);
        oResult = FALSE;
    }
#ifdef _DEBUG
    Trace_ReadData(oResult, pubInputReport, ulInputReportLength, &ulInputReportRead);
#endif
    return oResult;
}

BOOL CGatewayUSBToHidDrv::GetPortSettings(DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);

    if(p_pulBaudrate) *p_pulBaudrate = m_dBaudrate;
    if(p_pulTimeout) *p_pulTimeout = m_dTimeout;

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ClosePort(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            if(m_pHidHndl->ClosePort())
            {
                Trace_Close();
                if(m_oUpdatePortOpened) UpdatePortOpened(m_PortName, FALSE);
                m_PortName = _T("");
                oResult = TRUE;
            }
            else
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_CLOSING_PORT, p_pErrorInfo);
                oResult = FALSE;
            }
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_PORT_NOT_OPEN, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_LIBRARY_NOT_LOADED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::ResetPort(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0 && m_pHidHndl->IsOpened())
    {
        oResult = InitInputBuffer(1024);

        if(!PurgeBuffer())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_RESET_PORT, p_pErrorInfo);
            oResult = FALSE;
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_I_PORT_NOT_OPEN, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitPortList()
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);
    tUsbDeviceInfoList deviceInfoList;
    CStdString deviceClassType(_T(""));

    if(m_pInfoHandling)
    {
        //Get all physically available ports
        if(GetDeviceInfos(deviceInfoList, deviceClassType))
        {
            if(m_pInfoHandling->InitDeviceInfos(deviceInfoList, deviceClassType))
            {
                //Update port list
                oResult = UpdatePortList(m_pInfoHandling);
            }

            DeleteDeviceInfoList(deviceInfoList);
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::UpdateDeviceInfos(tPortList& p_rOpenPortList)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);
    tUsbDeviceInfoList deviceInfoList;
    CStdString deviceClassType(_T(""));

    if(m_pInfoHandling)
    {
        //Get all physically available ports
        if(GetDeviceInfos(deviceInfoList, deviceClassType))
        {
            if(m_pInfoHandling->UpdateDeviceInfos(deviceInfoList, deviceClassType, p_rOpenPortList))
            {
                oResult = TRUE;
            }

            DeleteDeviceInfoList(deviceInfoList);
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::UpdatePortList(tPortList& p_rOpenPortList)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pInfoHandling)
    {
        //Update port list
        oResult = UpdatePortList(m_pInfoHandling);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::UpdatePortList(CUsbDeviceInfoHandling* p_pDeviceInfoHandling)
{
    BOOL oResult(FALSE);
    CStdString portName(_T(""));
    CHidDeviceInfo deviceType;
    CStdStringArray portNames;

    //Reset PortList
    DeletePortList();

    if(p_pDeviceInfoHandling)
    {
        //Add ports
        if(p_pDeviceInfoHandling->GetPortNames(deviceType, portNames))
        {
            oResult = TRUE;
            for(CStdStringArray::iterator it = portNames.begin(); it != portNames.end(); it++)
            {
                portName = (*it);
                AddPortList(portName);
            }
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className(_T("GatewayHIDToHidDrv"));

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(INTERFACE_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::UpdatePortOpened(CStdString p_PortName, BOOL p_oOpened)
{
    BOOL oResult(FALSE);

    if(m_pInfoHandling)
    {
        oResult = m_pInfoHandling->UpdatePortOpened(p_PortName, p_oOpened);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::GetDevicePath(CStdString p_PortName, CStdString& p_rDevicePath)
{
    BOOL oResult(FALSE);
    CUsbDeviceInfo* pDeviceInfo(0);

    if(m_pInfoHandling)
    {
        pDeviceInfo = m_pInfoHandling->GetDeviceInfo(p_PortName);
        if(pDeviceInfo && pDeviceInfo->IsKindOf(_T("CHidDeviceInfo")))
        {
            CHidDeviceInfo* pHidDeviceInfo = (CHidDeviceInfo*)pDeviceInfo;
            p_rDevicePath = pHidDeviceInfo->GetDevicePath();
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::GetDeviceInfos(tUsbDeviceInfoList& p_rDeviceInfoList, CStdString& p_rDeviceClassType)
{
    BOOL oResult(FALSE);
    CUsbDeviceInfoHandling::SVendorProductIDs vendorProductId;
    list<CUsbDeviceInfoHandling::SVendorProductIDs> vendorProductIdList;
    BOOL oIsHidDevice(FALSE);

    if(m_pInfoHandling && m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        //Reset
        DeleteDeviceInfoList(p_rDeviceInfoList);

        if(m_pInfoHandling->GetAllVendorProductIds(vendorProductIdList))
        {
            //Init
            oResult = TRUE;
            p_rDeviceClassType = m_pHidHndl->GetDeviceClassType();

            //Enum Devices
            for(list<CUsbDeviceInfoHandling::SVendorProductIDs>::iterator it = vendorProductIdList.begin(); it != vendorProductIdList.end(); it++)
            {
                vendorProductId = (*it);

                oIsHidDevice = m_pInfoHandling->AreVendorProductIdsSupportedByDriver(m_DriverName, vendorProductId.vendorId, vendorProductId.productId);

                if(oResult && !m_pHidHndl->GetDeviceInfos(p_rDeviceInfoList, vendorProductId.vendorId, vendorProductId.productId, oIsHidDevice)) oResult = FALSE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::DeleteDeviceInfoList(tUsbDeviceInfoList& p_rDeviceInfoList)
{
    BOOL oResult(TRUE);
    CUsbDeviceInfo* pDeviceInfo(0);

    for(tUsbDeviceInfoList::iterator it = p_rDeviceInfoList.begin(); it != p_rDeviceInfoList.end(); it++)
    {
        pDeviceInfo = (*it);
        delete pDeviceInfo;
    }

    p_rDeviceInfoList.clear();

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitUsbDeviceInfoHandling(CUsbDeviceInfoHandling* p_pHidDeviceInfoHandling)
{
    BOOL oResult(TRUE);

    m_pInfoHandling = p_pHidDeviceInfoHandling;

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitInputBuffer(DWORD p_ulReadInputSize)
{
    BOOL oResult(TRUE);

    DeleteInputBuffer();

    if(!m_pInputBuffer)
    {
        m_pInputBuffer = new BYTE[p_ulReadInputSize];
        m_ulInputBufferSize = p_ulReadInputSize;
        m_ulInputBufferSizeActual = 0;
        m_ulGetInputBufferIndex = 0;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::UpdateInputBuffer(DWORD p_ulNbOfBytesToRead, int p_lMaxRetry, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulTotalNbOfBytesToRead(0);
    int lRetryCount(0);

    if(m_pHidHndl != 0 && m_pHidHndl->IsValid())
    {
        if(m_pHidHndl->IsOpened())
        {
            //Init
            oResult = TRUE;
            ulTotalNbOfBytesToRead = p_ulNbOfBytesToRead;

            while(oResult && (ulTotalNbOfBytesToRead > (m_ulInputBufferSizeActual - m_ulGetInputBufferIndex)))
            {
                //Init
                oResult = FALSE;
                lRetryCount = 0;

                //Retry loop
                while(!oResult && (lRetryCount < p_lMaxRetry))
                {
                    //Read report
                    if(ReadReport(p_pErrorInfo))
                    {
                        //Succeeded
                        oResult = TRUE;
                    }
                    else
                    {
                        //Retry
                        lRetryCount++;
                        oResult = FALSE;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::DeleteInputBuffer()
{
    BOOL oResult(TRUE);

    if(m_pInputBuffer)
    {
        m_ulInputBufferSize = 0;
        m_ulInputBufferSizeActual = 0;
        m_ulGetInputBufferIndex = 0;
        delete []m_pInputBuffer;
        m_pInputBuffer = 0;
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::GetDataFromInputBuffer(BYTE* p_pInputBuffer, DWORD p_ulNbOfBytesToRead, DWORD* p_pulNbOfBytesRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pInputBuffer && p_pInputBuffer)
    {
        //Check buffer size
        if(p_ulNbOfBytesToRead <= (m_ulInputBufferSizeActual - m_ulGetInputBufferIndex))
        {
            for(int i = 0; i < (int)p_ulNbOfBytesToRead; i++)
            {
                *p_pInputBuffer++ = *(m_pInputBuffer + m_ulGetInputBufferIndex);
                m_ulGetInputBufferIndex++;
                if(p_pulNbOfBytesRead) *p_pulNbOfBytesRead = i + 1;

                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitInfoHandling(CDeviceInfoHandlingBase* p_pInfoHandling)
{
    BOOL oResult(TRUE);

    m_pInfoHandling = dynamic_cast<CUsbDeviceInfoHandling*>(p_pInfoHandling);

    return oResult;
}

BOOL CGatewayUSBToHidDrv::InitVendorProductIds(CStdString p_DriverName, CStdString p_DriverParameters)
{
    BOOL oResult(FALSE);

    if(m_pInfoHandling)
    {
        if(p_DriverParameters.size() == 0)
        {
            //Default IDs
            oResult = TRUE;
            if(!m_pInfoHandling->AddVendorProductIds(p_DriverName, USB_VENDOR_PRODUCT_ID_ESCON)) oResult = FALSE;
            if(!m_pInfoHandling->AddVendorProductIds(p_DriverName, USB_VENDOR_PRODUCT_ID_EPOS4)) oResult = FALSE;
            if(!m_pInfoHandling->AddVendorProductIds(p_DriverName, USB_VENDOR_PRODUCT_ID_HPSC)) oResult = FALSE;
            if(!m_pInfoHandling->AddVendorProductIds(p_DriverName, USB_VENDOR_PRODUCT_ID_UAVESC)) oResult = FALSE;
        }
        else
        {
            //Driver Parameter IDs (Format: "VID_XXXX&PIDYYYY VID_ZZZZ&PID_WWWW ...")
            oResult = m_pInfoHandling->AddVendorProductIds(p_DriverName, p_DriverParameters);
        }
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::HidWriteFile(void* p_pBuffer, DWORD p_ulNumberOfBytesToWrite, DWORD* p_pulNumberOfBytesWritten)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0)
    {
        oResult = m_pHidHndl->WriteFile(p_pBuffer, p_ulNumberOfBytesToWrite, p_pulNumberOfBytesWritten);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::HidReadFile(void* p_pBuffer, DWORD p_ulNumberOfBytesToRead, DWORD* p_pulNumberOfBytesRead)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0)
    {
        oResult = m_pHidHndl->ReadFile(p_pBuffer, p_ulNumberOfBytesToRead, p_pulNumberOfBytesRead);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::HidWaitForOverlapped(unsigned long p_ulTimeout)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0)
    {
        oResult = m_pHidHndl->WaitForOverlapped(p_ulTimeout);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::HidGetOverlappedResult(DWORD* p_pulNumberOfBytesTransferred, BOOL p_oWait)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(CGateway::GetSyncInstance(), TRUE);

    if(m_pHidHndl != 0)
    {
        oResult = m_pHidHndl->GetOverlappedResult(p_pulNumberOfBytesTransferred, p_oWait);
    }

    return oResult;
}

BOOL CGatewayUSBToHidDrv::StartWriteInhibitTime()
{
    m_WriteInhibitTimer.reset();
    m_WriteInhibitTimer.start();

    return TRUE;
}

BOOL CGatewayUSBToHidDrv::WaitForWriteInhibitTimeElapsed()
{
    double ffInhibitTime = (double)GetWriteInhibitTimeInMicroSeconds();
    double ffElapsedTime(0);
    DWORD ulWaitCounter(0);

    if (ffInhibitTime > 0)
    {
        while(TRUE)
        {
            ffElapsedTime = m_WriteInhibitTimer.getElapsedTimeInMicroSec();
            if (ffElapsedTime > 0)
            {
                if (ffElapsedTime >= ffInhibitTime)
                {
                    // Inhibit time elapsed
                    break;
                }
                else
                {
                    ulWaitCounter++;
                }
            }
            else
            {
                // Inhibit timer not started
                break;
            }
        }

    }

    return TRUE;
}

DWORD CGatewayUSBToHidDrv::GetWriteInhibitTimeInMicroSeconds()
{
    const CStdString WRITE_INHIBIT_TIME(_T("WriteInhibitTimeInMicroSeconds"));

    DWORD ulResult(0);
    DWORD ulInhibitTime(0);

    if (GetParameter(WRITE_INHIBIT_TIME, (BYTE*)&ulInhibitTime, sizeof(ulInhibitTime)))
    {
        ulResult = ulInhibitTime;
    }

    return ulResult;
}