#include "stdafx.h"
#include <algorithm>

#include <Process/MmcProcess.h>
#include <Log/MmcLog/MmcLog.h>
#include <CommunicationModel/Interface/Classes/Ports/BaseClasses/PortBase.h>
#include <CommunicationModel/Interface/Classes/Handle/HandleRegistration_I.h>
#include <CommunicationModel/Interface/Factory/InterfaceFactory.h>
#include <CommunicationModel/Interface/Factory/DeviceInfoHandlingFactory.h>
#include <ParameterSet/ParameterSet.h>
#include <Thread/MmcSingleLock.h>
#include <Thread/MmcSemaphore.h>

#include <CommunicationModel/Interface/InterfaceManager.h>

using namespace std;

CInterfaceManager::CInterfaceManager():
    m_pHandleRegistrationMap_I(0)
{
    InitErrorHandling();
    InitRegistrationMap();
    CreateDeviceInfoHandling();
}

CInterfaceManager::CInterfaceManager(int p_lInstanceValue):
    CInterfaceManagerBase(p_lInstanceValue),
    m_pHandleRegistrationMap_I(0)
{
    InitErrorHandling();
    InitRegistrationMap();
    CreateDeviceInfoHandling();
}

CInterfaceManager::CInterfaceManager(const CInterfaceManager& p_rObject):
    CInterfaceManagerBase(p_rObject),
    m_pHandleRegistrationMap_I(0)
{
    InitErrorHandling();
    InitRegistrationMap();
    CreateDeviceInfoHandling();
}

CInterfaceManagerBase* CInterfaceManager::Clone()
{
    CInterfaceManager* pManager;

    pManager = new CInterfaceManager(*this);
    return pManager;
}

CInterfaceManager::~CInterfaceManager()
{
    ReleaseRegistrationMap();
    DeleteDeviceInfoHandling();
    DeleteInterfaceList();
}

BOOL CInterfaceManager::InitRegistrationMap()
{
    BOOL oResult(FALSE);

    if(!m_pHandleRegistrationMap_I)
    {
        m_pHandleRegistrationMap_I = CHandleRegistrationMap_I::GetInstance(m_lInstanceValue, this);
        if(m_pHandleRegistrationMap_I)
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::ReleaseRegistrationMap()
{
    BOOL oResult(TRUE);

    if(m_pHandleRegistrationMap_I)
    {
        CHandleRegistrationMap_I::ReleaseInstance(m_lInstanceValue, this);
        m_pHandleRegistrationMap_I = 0;
    }

    return oResult;
}

BOOL CInterfaceManager::CreateDeviceInfoHandling()
{
    BOOL oResult(FALSE);
    list<CStdString> aIdentifiers;

    oResult = CInterfaceFactory::GetAllIdentifiers(aIdentifiers);
    if(oResult)
    {
        for(list<CStdString>::iterator it = aIdentifiers.begin(); it != aIdentifiers.end() && oResult; it++)
        {
            CDeviceInfoHandlingBase* pInstance = NULL;
            if(CDeviceInfoHandlingFactory::CreateInstance((*it), pInstance))
            {
                pInstance->InitInstanceValue(m_lInstanceValue);
                m_DeviceInfoHandlingMap[*it] = pInstance;
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::DeleteDeviceInfoHandling()
{
    BOOL oResult(FALSE);

    for(map<CStdString, CDeviceInfoHandlingBase*>::iterator it = m_DeviceInfoHandlingMap.begin(); it != m_DeviceInfoHandlingMap.end(); it++)
    {
        CDeviceInfoHandlingBase* pDeviceInfoHandlingBase = (*it).second;

        if(pDeviceInfoHandlingBase)
        {
            delete pDeviceInfoHandlingBase;
        }

        oResult = TRUE;
    }

    return oResult;
}

BOOL CInterfaceManager::InitDeviceInfoHandling(CInterfaceBase* p_pInterface)
{
    BOOL oResult(FALSE);
    CStdString interfaceName;
    CDeviceInfoHandlingBase* pInfoHandlingInstance = NULL;

    if(p_pInterface)
    {
        oResult = p_pInterface->I_GetInterfaceName(&interfaceName);

        map<CStdString, CDeviceInfoHandlingBase*>::iterator it = m_DeviceInfoHandlingMap.find(interfaceName);
        if(it != m_DeviceInfoHandlingMap.end())
        {
            pInfoHandlingInstance = (*it).second;
            oResult = TRUE;
        }

        if(pInfoHandlingInstance == NULL)
        {
            //Try to create
            oResult = CDeviceInfoHandlingFactory::CreateInstance(interfaceName, pInfoHandlingInstance);
            if(oResult)
            {
                m_DeviceInfoHandlingMap[interfaceName] = pInfoHandlingInstance;
            }
        }

        if(oResult)
        {
            oResult = p_pInterface->InitInfoHandling(pInfoHandlingInstance);
        }
    }

    return oResult;
}

BOOL CInterfaceManager::ExecuteCommand(CCommandRoot *p_pCommand, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        if(CheckLayer(p_pCommand))
        {
            //Init
            oResult = TRUE;

            //Get registration
            if((oResult && (!m_pHandleRegistrationMap_I->GetRegistration(p_hHandle, &pRegistration))) || (pRegistration == 0))
            {
                oResult = FALSE;
            }

            //Lock (Only if registration is not locked from last command)
            if(oResult && !pRegistration->IsLocked())
            {
                oResult = pRegistration->DoLock();
            }

            //Get registration entries
            if((oResult && (!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hHandle, &pInterface, &pPort))) || (pInterface == 0) || (pPort == 0))
            {
                oResult = FALSE;
            }

            //Execute command
            if(oResult && !pInterface->ExecuteCommand(p_pCommand, pPort, p_hTransactionHandle))
            {
                oResult = FALSE;
            }

            //Unlock (Only if interface does not remain locked until next command)
            if(pInterface && !pInterface->IsLocked())
            {
                //Unlock
                if(pRegistration) pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::AbortCommands(HANDLE p_hHandle, BOOL p_oActive)
{
    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        oResult = TRUE;

        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistration(p_hHandle, &pRegistration))) || (pRegistration == 0))
        {
            oResult = FALSE;
        }

        //Get registration entries
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hHandle, &pInterface, &pPort))) || (pInterface == 0) || (pPort == 0))
        {
            oResult = FALSE;
        }

        //Abort commands
        if(oResult && !pInterface->AbortCommands(p_oActive))
        {
            oResult = FALSE;
        }

        //Abort lock
        if(oResult)
        {
            pRegistration->AbortLock(p_oActive);
        }
    }

    return oResult;
}

BOOL CInterfaceManager::GetLayerManager(HANDLE p_hHandle, ELayer p_eLayer, CLayerManagerBase** p_ppLayerManager, HANDLE* p_pLayerHandle)
{
    BOOL oResult(FALSE);

    if(CheckLayer(p_eLayer))
    {
        if(p_ppLayerManager)
        {
            *p_ppLayerManager = this;
        }

        if(p_pLayerHandle)
        {
            *p_pLayerHandle = p_hHandle;
        }

        oResult = TRUE;
    }

    return oResult;
}

BOOL CInterfaceManager::GetCommands(HANDLE p_hHandle, ELayer p_eLayer, CStdString* p_pCommandInfo)
{
    BOOL oResult = FALSE;
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;

    if(m_pHandleRegistrationMap_I)
    {
        oResult = TRUE;

        //Get registration entries
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hHandle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            oResult = FALSE;
        }

        if(oResult && CheckLayer(p_eLayer))
        {
            //Get commands
            oResult = pPort->GetCommands(p_pCommandInfo);
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_InitAllInterfaces(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);

    //Reset
    DeleteInterfaceList();

    list<CStdString> aIdentifiers;
    oResult = CInterfaceFactory::GetAllIdentifiers(aIdentifiers);

    if(oResult)
    {
        for(list<CStdString>::iterator it = aIdentifiers.begin(); it != aIdentifiers.end() && oResult; it++)
        {
            if(!I_InitInterface(*it, p_pErrorInfo))
            {
                oResult = FALSE;
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_InitInterface(CStdString p_InterfaceName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);
    BOOL oNewCreated(FALSE);
    CStdStringArray interfaceHardwareNames;
    BOOL oResult(FALSE);

    //Get interface names
    if(GetInterfaceHardwareNames(p_InterfaceName, interfaceHardwareNames))
    {
        for(CStdStringArray::iterator it = interfaceHardwareNames.begin(); it != interfaceHardwareNames.end(); it++)
        {
            //Init
            CStdString interfaceHardwareName = (*it);
            pInterface = 0;

            //Get interface
            if(GetInterface(p_InterfaceName, interfaceHardwareName, pInterface, oNewCreated, p_pErrorInfo) && pInterface)
            {
                //Init interface
                if(pInterface->InitInterface(p_pErrorInfo))
                {
                    if(oNewCreated)
                    {
                        m_InterfaceList.push_back(pInterface);
                    }

                    oResult = TRUE;
                }
                else
                {
                    if(oNewCreated)
                    {
                        delete pInterface;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_InitInterface(CStdString p_InterfaceName, CStdString p_DriverName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);
    BOOL oNewCreated(FALSE);
    CStdStringArray interfaceHardwareNames;
    BOOL oResult(FALSE);

    //Get interface names
    if(GetInterfaceHardwareNames(p_InterfaceName, p_DriverName, interfaceHardwareNames))
    {
        for(CStdStringArray::iterator it = interfaceHardwareNames.begin(); it != interfaceHardwareNames.end(); it++)
        {
            //Init
            CStdString interfaceHardwareName = (*it);
            pInterface = 0;

            //Get interface
            if(GetInterface(p_InterfaceName, interfaceHardwareName, pInterface, oNewCreated, p_pErrorInfo) && pInterface)
            {
                //Init interface
                if(pInterface->InitInterface(p_DriverName, p_pErrorInfo))
                {
                    if(oNewCreated)
                    {
                        m_InterfaceList.push_back(pInterface);
                    }

                    oResult = TRUE;
                }
                else
                {
                    if(oNewCreated)
                    {
                        delete pInterface;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::GetInterface(CStdString p_InterfaceName, CStdString p_InterfaceHardwareName, CInterfaceBase*& p_rpInterface, BOOL& p_roNewCreated, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    //Init
    p_rpInterface = 0;
    p_roNewCreated = FALSE;

    //Get ProtocolStack
    if(p_InterfaceName.size() == 0)
    {
        //Don't create interface
        oResult = TRUE;
    }
    else
    {
        //Interface is already existing
        if(!FindInterface(p_InterfaceHardwareName, &p_rpInterface))
        {
            //Init
            p_roNewCreated = TRUE;

            oResult = CInterfaceFactory::CreateInstance(p_InterfaceName, p_rpInterface);

            if(!oResult)
            {
                //ProtocolStack not found
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
                }

                p_roNewCreated = FALSE;
                oResult = FALSE;
            }
        }
        else
        {
            //Init
            p_roNewCreated = FALSE;
            oResult = TRUE;
        }
    }

    //Init interface
    if(oResult && p_rpInterface)
    {
        InitDeviceInfoHandling(p_rpInterface);
        p_rpInterface->InitInterfaceName(p_InterfaceHardwareName);
        p_rpInterface->InitInstanceValue(m_lInstanceValue);
    }

    return oResult;
}

BOOL CInterfaceManager::I_UpdateInterfaceNameSelection(CStdString p_strInterfaceName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);
    BOOL oNewCreated(FALSE);
    CStdStringArray interfaceHardwareNames;

    //Get interface names
    if(GetInterfaceHardwareNames(p_strInterfaceName, interfaceHardwareNames))
    {
        //Init
        oResult = TRUE;

        //Add new
        for(CStdStringArray::iterator it = interfaceHardwareNames.begin(); it != interfaceHardwareNames.end(); it++)
        {
            //Init
            CStdString interfaceHardwareName = (*it);
            pInterface = 0;

            //Get interface
            if(GetInterface(p_strInterfaceName, interfaceHardwareName, pInterface, oNewCreated, p_pErrorInfo) && pInterface)
            {
                //Init interface
                if(pInterface->InitInterface(p_pErrorInfo))
                {
                    if(oNewCreated)
                    {
                        m_InterfaceList.push_back(pInterface);
                    }
                }
                else
                {
                    if(oNewCreated)
                    {
                        delete pInterface;
                    }
                }
            }
        }

        //Delete
        oResult = DeleteNonExistingInterfaces(interfaceHardwareNames);
    }

    return oResult;
}

BOOL CInterfaceManager::I_UpdateInterfaceNameSelection(CStdString p_strInterfaceName, CStdString p_DriverName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);
    BOOL oNewCreated(FALSE);
    CStdStringArray interfaceHardwareNames;

    //Get interface names
    if(GetInterfaceHardwareNames(p_strInterfaceName, p_DriverName, interfaceHardwareNames))
    {
        //Init
        oResult = TRUE;

        //Add new
        for(CStdStringArray::iterator it = interfaceHardwareNames.begin(); it != interfaceHardwareNames.end(); it++)
        {
            //Init
            CStdString interfaceHardwareName = (*it);
            pInterface = 0;

            //Get interface
            if(GetInterface(p_strInterfaceName, interfaceHardwareName, pInterface, oNewCreated, p_pErrorInfo) && pInterface)
            {
                //Init interface
                if(pInterface->InitInterface(p_DriverName, p_pErrorInfo))
                {
                    if(oNewCreated)
                    {
                        m_InterfaceList.push_back(pInterface);
                    }
                }
                else
                {
                    if(oNewCreated)
                    {
                        delete pInterface;
                    }
                }
            }
        }

        //Delete
        oResult = DeleteNonExistingInterfaces(interfaceHardwareNames);
    }

    return oResult;
}

BOOL CInterfaceManager::I_ResetPortEnumeration(CStdString p_strInterfaceName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface = 0;
    BOOL oResult(FALSE);

    if(!FindInterface(p_strInterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling)
        {
            m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
        }
    }
    else
    {
        //Enumeration
        oResult = pInterface->I_ResetPortEnumeration(p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_UpdatePortNameSelection(CStdString p_strInterfaceName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        list<CStdString> rTargets;

        oResult = CInterfaceFactory::GetAllIdentifiers(rTargets);

        if(oResult)
        {
            for(list<CStdString>::iterator it = rTargets.begin(); it != rTargets.end() && oResult; it++)
            {
                CInterfaceBase* pInterface = 0;

                //Interface supported
                int lPosition = (int)p_strInterfaceName.find(*it);
                if(lPosition != -1)
                {
                    //Interface is existing
                    if(FindInterface(p_strInterfaceName, &pInterface) && pInterface)
                    {
                        tPortList openPortList;
                        if(m_pHandleRegistrationMap_I->GetRegisteredPorts(pInterface, openPortList))
                        {
                            oResult = pInterface->I_UpdatePortNameSelection(openPortList);
                        }

                        ProcessPendingReopenFlags(p_strInterfaceName);
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_AddInterfacePort(CStdString p_InterfaceName, CStdString p_PortName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);

    oResult = CInterfaceFactory::CreateInstance(p_InterfaceName, pInterface);

    if(oResult && pInterface)
    {
        if(!pInterface->I_AddPort(p_PortName))
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_DeleteInterfacePort(CStdString p_InterfaceName, CStdString p_PortName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);

    oResult = CInterfaceFactory::CreateInstance(p_InterfaceName, pInterface);

    if(oResult && pInterface)
    {
        if(!pInterface->I_DeletePort(p_PortName))
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_InitBaudrateSelection(CStdString p_InterfaceName, CStdDWordArray& p_rBaudrateSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);
    BOOL oResult(TRUE);

    if(FindInterface(p_InterfaceName, &pInterface))
    {
        if(pInterface)
        {
            if(!pInterface->InitBaudrateSelection(p_rBaudrateSel))
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                }

                oResult = FALSE;
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_InitDefaultInterfaceSettings(CStdString p_InterfaceName, DWORD p_ulBaudrate, DWORD p_ulTimeout, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface = 0;
    BOOL oResult(TRUE);

    if(FindInterface(p_InterfaceName, &pInterface))
    {
        if(pInterface)
        {
            if(!pInterface->InitDefaultInterfaceSettings(p_ulBaudrate, p_ulTimeout))
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                }

                oResult = FALSE;
            }
        }
    }

    return oResult;
}

void CInterfaceManager::DeleteInterfaceList()
{
    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);
        delete pInterface;
    }

    m_InterfaceList.clear();
}

BOOL CInterfaceManager::FindInterfaceIndex(CStdString p_InterfaceName, short* p_psInterfaceIndex)
{
    BOOL oResult(FALSE);
    CStdString name;
    short sIndex = 0;

    if(p_psInterfaceIndex)
    {
        *p_psInterfaceIndex = -1;
    }

    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);
        if(pInterface->I_GetInterfaceName(&name))
        {
            if(p_InterfaceName.CompareNoCase(name) == 0)
            {
                if(p_psInterfaceIndex)
                {
                    *p_psInterfaceIndex = sIndex;
                }

                oResult = TRUE;
                break;
            }
        }

        sIndex++;
    }

    return oResult;
}

BOOL CInterfaceManager::FindInterface(CStdString p_InterfaceName, CInterfaceBase** p_ppInterface)
{
    BOOL oResult(FALSE);
    CStdString name;

    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);
        if(pInterface && pInterface->I_GetInterfaceName(&name))
        {
            if(p_InterfaceName.CompareNoCase(name) == 0)
            {
                if(p_ppInterface) *p_ppInterface = pInterface;
                oResult = TRUE;
                break;
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::DeleteInterface(CStdString p_InterfaceName, BOOL p_oCheckPrefixOnly)
{
    CStdString name(_T(""));
    BOOL oResult(FALSE);

    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);
        if(pInterface && pInterface->I_GetInterfaceName(&name))
        {
            if((p_oCheckPrefixOnly && name.find(p_InterfaceName) == 0) || (!p_oCheckPrefixOnly && name.CompareNoCase(p_InterfaceName) == 0))
            {
                m_InterfaceList.remove(pInterface);
                delete pInterface;
                oResult = TRUE;
                break;
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::DeleteNonExistingInterfaces(CStdStringArray& p_rExistingInterfaceNames)
{
    BOOL oResult(FALSE);
    CStdStringArray interfacesToDelete;
    CStdString name;

    //Check existence
    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);
        if(pInterface->I_GetInterfaceName(&name))
        {
            if(!IsInterfaceInList(name, p_rExistingInterfaceNames))
            {
                interfacesToDelete.push_back(name);
            }
        }
    }

    //Delete
    oResult = DeleteInterfaces(interfacesToDelete);

    return oResult;
}

BOOL CInterfaceManager::DeleteInterfaces(CStdStringArray& p_rExistingInterfaceNames)
{
    BOOL oResult(TRUE);

    for(CStdStringArray::iterator it = p_rExistingInterfaceNames.begin(); it != p_rExistingInterfaceNames.end(); it++)
    {
        CStdString name = *(it);

        if(!DeleteInterface(name, FALSE))
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::IsInterfaceInList(CStdString p_Name, CStdStringArray& p_rInterfaceNames)
{
    BOOL oResult(FALSE);

    for(CStdStringArray::iterator it = p_rInterfaceNames.begin(); it != p_rInterfaceNames.end(); it++)
    {
        if(p_Name == *(it))
        {
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::DeleteAllInterfaces(CStdString p_InterfaceName, BOOL p_oCheckPrefixOnly)
{
    BOOL oInterfaceDeleted(TRUE);
    BOOL oResult(TRUE);

    //Delete All interfaces
    while(oInterfaceDeleted)
    {
        oInterfaceDeleted = DeleteInterface(p_InterfaceName, p_oCheckPrefixOnly);
    }

    return oResult;
}

BOOL CInterfaceManager::I_IsInterfaceNameSupported(CStdString p_InterfaceName, CErrorInfo* p_pErrorInfo)
{
    return FindInterface(p_InterfaceName, 0);
}

BOOL CInterfaceManager::FindInterfacePortIndex(CStdString p_InterfaceName, CStdString p_PortName, short* p_puPortIndex)
{
    BOOL oResult(FALSE);
    CInterfaceBase* pInterface(0);

    if(FindInterface(p_InterfaceName, &pInterface) && pInterface)
    {
        oResult = pInterface->I_FindInterfacePortIndex(p_PortName, p_puPortIndex);
    }

    return oResult;
}

BOOL CInterfaceManager::FindPort(CStdString p_InterfaceName, CStdString p_PortName, CPortBase** p_ppPort)
{
    BOOL oResult(FALSE);
    CInterfaceBase* pInterface(0);

    if(FindInterface(p_InterfaceName, &pInterface) && pInterface)
    {
        oResult = pInterface->I_FindPort(p_InterfaceName, p_PortName, p_ppPort);
    }

    return oResult;
}

HANDLE CInterfaceManager::I_OpenInterface(CStdString p_InterfaceName, CStdString p_PortName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface = 0;
    CInterfaceBase* pExistingInterface = 0;
    CPortBase* pPort = 0;
    HANDLE hI_Handle = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Suchen Position der Schnittstellen in der Liste
        if(!FindInterface(p_InterfaceName, &pInterface))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
            return (HANDLE)0;
        }
        //Suche Position des Ports in der Liste
        if(!FindPort(p_InterfaceName, p_PortName, &pPort))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_PORT_NAME, p_pErrorInfo);
            return (HANDLE)0;
        }

        //Überprüfen ob interface & Port schon offen
        hI_Handle = m_pHandleRegistrationMap_I->IsRegistrationExisting(pInterface, p_PortName);
        if(!hI_Handle)
        {
            //Überprüfen ob interface schon offen
            hI_Handle = m_pHandleRegistrationMap_I->IsRegistrationExisting(pInterface);
            if(hI_Handle)
            {
                //Interface schon offen -> nur Port öffnen
                if(m_pHandleRegistrationMap_I->GetRegistrationValues(hI_Handle, &pExistingInterface, 0) && pExistingInterface)
                {
                    pInterface = pExistingInterface;

                    //Schnittstelle in der Mappe registrieren (-> Clone)
                    hI_Handle = m_pHandleRegistrationMap_I->RegisterHandle(pInterface, pPort);
                    if(hI_Handle)
                    {
                        if(m_pHandleRegistrationMap_I->GetRegistrationValues(hI_Handle, &pInterface, &pPort) && pInterface && pPort)
                        {
                            //Port öffnen
                            if(pInterface->I_OpenInterfacePort(pPort, p_PortName, p_pErrorInfo))
                            {
                                return hI_Handle;
                            }
                            else
                            {
                                m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                                return (HANDLE)0;
                            }
                        }

                        m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                    }

                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    return (HANDLE)0;
                }
            }
            else
            {
                //Schnittstelle in der Mappe registrieren (-> Clone)
                hI_Handle = m_pHandleRegistrationMap_I->RegisterHandle(pInterface, pPort);
                if(hI_Handle)
                {
                    if(m_pHandleRegistrationMap_I->GetRegistrationValues(hI_Handle, &pInterface, &pPort) && pInterface && pPort)
                    {
                        //Interface & Port öffnen
                        if(pInterface->I_OpenInterface(p_pErrorInfo))
                        {
                            if(pInterface->I_OpenInterfacePort(pPort, p_PortName, p_pErrorInfo))
                            {
                                return hI_Handle;
                            }
                            else
                            {
                                m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                                return (HANDLE)0;
                            }
                        }
                        else
                        {
                            m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                            return (HANDLE)0;
                        }
                    }

                    m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                }

                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                return (HANDLE)0;
            }
        }
    }

    return hI_Handle;
}

BOOL CInterfaceManager::I_CloseInterface(HANDLE p_hI_Handle, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;
    int portCount = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //Anzahl geöffnete Ports pro interface ermitteln
        portCount = m_pHandleRegistrationMap_I->GetRegistrationCount(pInterface);

        //InterfacePort schliessen
        if(pInterface->I_CloseInterfacePort(pPort, p_pErrorInfo))
        {
            //Interface schliessen
            if(portCount == 1)
            {
                if(pInterface->I_CloseInterface(p_pErrorInfo))
                {
                    m_pHandleRegistrationMap_I->DeleteMapRegistration(p_hI_Handle);
                    if(pRegistration) pRegistration->DoUnlock();
                    return TRUE;
                }
                else
                {
                    m_pHandleRegistrationMap_I->DeleteMapRegistration(p_hI_Handle);
                    if(pRegistration) pRegistration->DoUnlock();
                    return FALSE;
                }
            }
            else
            {
                //Registrierte Werte aus der Mappe löschen
                if(m_pHandleRegistrationMap_I->DeleteMapRegistration(p_hI_Handle))
                {
                    if(pRegistration) pRegistration->DoUnlock();
                    return TRUE;
                }
                else
                {
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    if(pRegistration) pRegistration->DoUnlock();
                    return FALSE;
                }
            }
        }
        else
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }
    }

    return FALSE;
}

BOOL CInterfaceManager::I_CloseAllInterfaces(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    tHandleRegistrationI_List* pRegistrationList;

    if(m_pHandleRegistrationMap_I)
    {
        if(m_pHandleRegistrationMap_I->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            oResult = TRUE;

            while(pRegistrationList->size() > 0)
            {
                tHandleRegistrationI_List::iterator it = pRegistrationList->begin();
                if(it != pRegistrationList->end())
                {
                    CHandleRegistration_I* pRegistration = (*it);
                    if(pRegistration)
                    {
                        HANDLE hIHandle = pRegistration->GetKeyHandle();
                        if(hIHandle)
                        {
                            if(!I_CloseInterface(hIHandle, p_pErrorInfo))
                            {
                                oResult = FALSE;
                                break;
                            }
                        }

                        pRegistrationList->remove(pRegistration);

                        //Delete registration
                        delete pRegistration;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_AreAllInterfacesClosed()
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        oResult = m_pHandleRegistrationMap_I->IsRegistrationMapEmpty();
    }

    return oResult;
}

BOOL CInterfaceManager::I_SetInterfaceSettings(HANDLE p_hI_Handle, DWORD p_ulBaudrate, DWORD p_ulTimeout, BOOL p_oChangeOnly, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //SetInterfaceSettings
        if(!pInterface->I_SetInterfaceSettings(pPort, p_ulBaudrate, p_ulTimeout, p_oChangeOnly, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetInterfaceSettings(HANDLE p_hI_Handle, DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetInterfaceSettings
        if(!pInterface->I_GetInterfaceSettings(pPort, p_pulBaudrate, p_pulTimeout, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetInterfaceMode(HANDLE p_hI_Handle, WORD* p_pusModeIndex, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetInterfaceMode
        if(!pInterface->I_GetInterfaceMode(p_pusModeIndex, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_SetInterfaceMode(HANDLE p_hI_Handle, WORD p_usModeIndex, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //SetInterfaceMode
        if(!pInterface->I_SetInterfaceMode(p_usModeIndex, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetPortMode(HANDLE p_hI_Handle, WORD* p_pusModeIndex, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetPortMode
        if(!pPort->GetPortMode(p_pusModeIndex, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_SetPortMode(HANDLE p_hI_Handle, WORD p_usModeIndex, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //SetPortMode
        if(!pPort->SetPortMode(p_usModeIndex, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_ResetInterface(HANDLE p_hI_Handle, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //ResetInterface
        if(!pInterface->I_ResetInterface(pPort, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::SetParameter(EParameterType p_eParameterType, HANDLE p_hI_Handle, CStdString p_Name, BYTE* p_pValue, DWORD p_ulSize, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        oResult = m_pHandleRegistrationMap_I->SetParameter(p_eParameterType, p_hI_Handle, p_Name, p_pValue, p_ulSize, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::SetParameter(EParameterType p_eParameterType, HANDLE p_hI_Handle, CStdString p_Name, CStdString p_Value, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        oResult = m_pHandleRegistrationMap_I->SetParameter(p_eParameterType, p_hI_Handle, p_Name, p_Value, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::GetParameter(EParameterType p_eParameterType, HANDLE p_hI_Handle, CStdString p_Name, BYTE* p_pValue, DWORD p_ulSize, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        oResult = m_pHandleRegistrationMap_I->GetParameter(p_eParameterType, p_hI_Handle, p_Name, p_pValue, p_ulSize, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::GetParameter(EParameterType p_eParameterType, HANDLE hVCS_Handle, CStdString p_Name, CStdString& p_rValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_I)
    {
        oResult = m_pHandleRegistrationMap_I->GetParameter(p_eParameterType, hVCS_Handle, p_Name, p_rValue, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetKeyHandle(CStdString p_InterfaceName, CStdString p_PortName, HANDLE* p_pIKeyHandle, CErrorInfo* p_pErrorInfo)
{
    HANDLE hKeyHandle = 0;
    BOOL oInterfaceNameFound = FALSE;
    BOOL oPortNameFound = FALSE;
    CHandleRegistration_I* pRegistration = 0;
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CStdString strInterfaceNameReg;
    CStdString strPortNameReg;
    tHandleRegistrationI_List* pRegistrationList = 0;

    if(m_pHandleRegistrationMap_I)
    {
        if(m_pHandleRegistrationMap_I->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            for(tHandleRegistrationI_List::iterator it = pRegistrationList->begin(); it != pRegistrationList->end(); it++)
            {
                pRegistration = (*it);

                pRegistration->GetRegistrationValues(&pInterface, &pPort);

                if(!pInterface || !pPort)
                {
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    return FALSE;
                }

                //Compare interfaceName
                if(pInterface->I_GetInterfaceName(&strInterfaceNameReg))
                {
                    if(strInterfaceNameReg.CompareNoCase(p_InterfaceName) == 0)
                    {
                        oInterfaceNameFound = TRUE;

                        //Compare PortName
                        if(pPort->GetPortName(&strPortNameReg))
                        {
                            if(strPortNameReg.CompareNoCase(p_PortName) == 0)
                            {
                                oPortNameFound = TRUE;
                                hKeyHandle = pRegistration->GetKeyHandle();
                            }
                        }
                    }
                }
            }
        }
    }

    //Error handling
    if(hKeyHandle == 0)
    {
        if(m_pErrorHandling)
        {
            if(!oInterfaceNameFound)    m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
            else if(!oPortNameFound)    m_pErrorHandling->GetError(ERROR_BAD_PORT_NAME, p_pErrorInfo);
            else if(!hKeyHandle)        m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        }
    }

    if(p_pIKeyHandle) *p_pIKeyHandle = hKeyHandle;
    return (hKeyHandle != 0);
}

BOOL CInterfaceManager::I_GetInterfaceName(HANDLE p_hI_Handle, CStdString* p_pInterfaceName, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetInterfaceName
        if(!pInterface->I_GetInterfaceName(p_pInterfaceName))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetInterfaceNameSelection(CStdStringArray* p_pInterfaceNameSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(FALSE);
    CStdString strNameSel;

    if(p_pInterfaceNameSel)
    {
        p_pInterfaceNameSel->clear();

        for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
        {
            CInterfaceBase* pInterface = (*it);
            if(pInterface->I_GetInterfaceName(&strNameSel))
            {
                p_pInterfaceNameSel->push_back(strNameSel);
            }
        }

        oResult = TRUE;
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetPortName(HANDLE p_hI_Handle, CStdString* p_pPortName, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetPortName
        if(!pPort->GetPortName(p_pPortName))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetDriverInfos(HANDLE p_hI_Handle, CStdString& p_rDriverName, CStdString& p_rDriverParameters, CErrorInfo* p_pErrorInfo)
{
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            return FALSE;
        }

        //Lock
        if(!pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            return FALSE;
        }

        //Get registration values
        if(!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        //GetDriverInfos
        if(!pPort->GetDriverInfos(p_rDriverName, p_rDriverParameters))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            return FALSE;
        }

        if(pRegistration) pRegistration->DoUnlock();
        return TRUE;
    }

    return FALSE;
}

BOOL CInterfaceManager::I_GetPortNameSelection(CStdString p_InterfaceName, CStdStringArray* p_pPortSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;

    //FindInterface
    if(!FindInterface(p_InterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
    }
    else
    {
        //GetPortNameSelection
        oResult = pInterface->I_GetPortNameSelection(p_pPortSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetBaudrateSelection(CStdString p_InterfaceName, CStdString p_PortName, CStdDWordArray* p_pulBaudrateSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;

    //FindInterface
    if(!FindInterface(p_InterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
    }
    else
    {
        //GetBaudrateSelection
        oResult = pInterface->I_GetBaudrateSelection(p_PortName, p_pulBaudrateSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetInterfaceModeSelection(CStdString p_InterfaceName, CStdStringArray* p_pInterfaceModeSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;

    //FindInterface
    if(!FindInterface(p_InterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
    }
    else
    {
        //GetInterfaceModeSelection
         pInterface->I_GetInterfaceModeSelection(p_pInterfaceModeSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetPortModeSelection(CStdString p_InterfaceName, CStdString p_PortName, CStdStringArray* p_pPortModeSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(FALSE);
    CPortBase* pPort(0);

    //FindPort
    if(!FindPort(p_InterfaceName, p_PortName, &pPort) || (pPort == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_PORT_NAME, p_pErrorInfo);
    }
    else
    {
        //GetPortModeSelection
        oResult = pPort->GetPortModeSelection(p_pPortModeSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::GetInterfaceHardwareNames(CStdString p_InterfaceName, CStdString p_DriverName, CStdStringArray& p_rNames)
{
    BOOL oResult = FALSE;
    CInterfaceBase* pInterface(0);

    oResult = CInterfaceFactory::CreateInstance(p_InterfaceName, pInterface);

    if(oResult)
    {
        if(pInterface)
        {
            //Init
            InitDeviceInfoHandling(pInterface);

            //Names
            oResult = pInterface->GetInterfaceHardwareNames(p_DriverName, p_rNames);
            delete pInterface;
        }
        else
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CInterfaceManager::GetInterfaceHardwareNames(CStdString p_InterfaceName, CStdStringArray& p_rNames)
{
    BOOL oResult = FALSE;
    CInterfaceBase* pInterface(0);

    oResult = CInterfaceFactory::CreateInstance(p_InterfaceName, pInterface);

    if(oResult)
    {
        if(pInterface)
        {
            //Init
            InitDeviceInfoHandling(pInterface);

            //Names
            oResult = pInterface->GetInterfaceHardwareNames(p_rNames);
            delete pInterface;
        }
        else
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

void CInterfaceManager::InitJournalManager(CJournalManagerBase *p_pJournalManager)
{
    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);

        pInterface->InitJournalManager(p_pJournalManager);
    }
}

void CInterfaceManager::ResetJournalManager()
{
    for(list<CInterfaceBase*>::iterator it = m_InterfaceList.begin(); it != m_InterfaceList.end(); it++)
    {
        CInterfaceBase* pInterface = (*it);

        pInterface->ResetJournalManager();
    }
}

BOOL CInterfaceManager::InitJournalManager(HANDLE p_hHandle, CJournalManagerBase* p_pJournalManager)
{
    BOOL oResult(FALSE);
    CHandleRegistration_I* pRegistration(0);
    CInterfaceBase* pInterface(0);
    CPortBase* pPort(0);

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(m_pHandleRegistrationMap_I->GetRegistration(p_hHandle, &pRegistration))
        {
            //Lock
            if(pRegistration && !pRegistration->IsLocked())
            {
                pRegistration->DoLock();
            }

            //Get registration entries
            if(m_pHandleRegistrationMap_I->GetRegistrationValues(p_hHandle, &pInterface, &pPort))
            {
                if(pInterface) pInterface->InitJournalManager(p_pJournalManager);
                if(pPort) pPort->InitJournalManager(p_hHandle, p_pJournalManager);

                oResult = TRUE;
            }

            //Unlock
            if(pRegistration)
            {
                pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::ResetJournalManager(HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    CHandleRegistration_I* pRegistration = 0;
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;

    if(m_pHandleRegistrationMap_I)
    {
        //Get registration
        if(m_pHandleRegistrationMap_I->GetRegistration(p_hHandle, &pRegistration))
        {
            //Lock
            if(pRegistration && !pRegistration->IsLocked())
            {
                pRegistration->DoLock();
            }

            //Get registration entries
            if(m_pHandleRegistrationMap_I->GetRegistrationValues(p_hHandle, &pInterface, &pPort))
            {
                if(pInterface) pInterface->ResetJournalManager();
                if(pPort) pPort->ResetJournalManager(p_hHandle);

                oResult = TRUE;
            }

            //Unlock
            if(pRegistration)
            {
                pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::I_GetDefaultInterfaceSettings(CStdString p_InterfaceName, DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);

    //FindInterface
    if(!FindInterface(p_InterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
    }
    else
    {
        //GetDefaultInterfaceSettings
        oResult = pInterface->I_GetDefaultInterfaceSettings(p_pulBaudrate, p_pulTimeout, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::I_SetDefaultInterfaceSettings(CStdString p_InterfaceName, DWORD p_ulBaudrate, DWORD p_ulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CInterfaceBase* pInterface(0);

    //FindInterface
    if(!FindInterface(p_InterfaceName, &pInterface) || (pInterface == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
    }
    else
    {
        //SetDefaultInterfaceSettings
        oResult = pInterface->I_SetDefaultInterfaceSettings(p_ulBaudrate, p_ulTimeout, p_pErrorInfo);
    }

    return oResult;
}

BOOL CInterfaceManager::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className(_T("InterfaceManager"));

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(INTERFACE_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CInterfaceManager::EnableTracing(HANDLE p_hI_Handle, CStdString p_TracingFileName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        oResult = TRUE;

        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }

        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }

        //Get registration values
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort))) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        //Enable tracing
        if(oResult && !pInterface->EnableTracing(pPort, p_TracingFileName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        //Unlock
        if(oResult && pRegistration)
        {
            pRegistration->DoUnlock();
        }
    }

    return oResult;
}

BOOL CInterfaceManager::DisableTracing(HANDLE p_hI_Handle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CInterfaceBase* pInterface = 0;
    CPortBase* pPort = 0;
    CHandleRegistration_I* pRegistration = 0;

    if(m_pHandleRegistrationMap_I)
    {
        oResult = TRUE;

        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistration(p_hI_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }

        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }

        //Get registration values
        if((oResult && (!m_pHandleRegistrationMap_I->GetRegistrationValues(p_hI_Handle, &pInterface, &pPort))) || (pInterface == 0) || (pPort == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        //Disable tracing
        if(oResult && !pInterface->DisableTracing(pPort, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        //Unlock
        if(oResult && pRegistration)
        {
            pRegistration->DoUnlock();
        }
    }

    return oResult;
}

BOOL CInterfaceManager::ProcessPendingReopenFlags(CStdString p_InterfaceName)
{
    BOOL oResult(FALSE);
    CStdStringArray portsToReopen;
    CStdString portName(_T(""));

    if(m_pHandleRegistrationMap_I != 0)
    {
        if(m_pHandleRegistrationMap_I->GetPendingPortsToReopen(p_InterfaceName, portsToReopen))
        {
            oResult = TRUE;

            for(CStdStringArray::iterator it = portsToReopen.begin(); it != portsToReopen.end(); it++)
            {
                portName = (*it);

                if(!ProcessPendingReopenFlag(p_InterfaceName, portName))
                {
                    oResult = FALSE;
                }
            }
        }
    }

    return oResult;
}

BOOL CInterfaceManager::ProcessPendingReopenFlag(CStdString p_InterfaceName, CStdString p_PortName)
{
    BOOL oResult(FALSE);
    CMmcSemaphore* pReopenSemaphore(0);

    if(m_pHandleRegistrationMap_I != 0)
    {
        //Lock
        pReopenSemaphore = CreateReopenSemaphore(p_InterfaceName, p_PortName);
        CMmcSingleLock lock(pReopenSemaphore, TRUE);

        //Reopen
        oResult = ReopenInterfacePort(p_InterfaceName, p_PortName);

        m_pHandleRegistrationMap_I->ResetReopenFlag(p_InterfaceName, p_PortName);

        //Unlock
        lock.Unlock();
        DeleteReopenSemaphore(pReopenSemaphore);
    }

    return oResult;
}

CMmcSemaphore* CInterfaceManager::CreateReopenSemaphore(CStdString p_interfaceName, CStdString p_portName)
{
    const long NB_OF_ALLOWED_CLIENTS = 1;

    CMmcSemaphore* pSemaphore(0);
    CStdString name(_T(""));

    name.Format(_T("Process-0x%.8X_%s_%s"), MmcGetCurrentProcessId(), p_interfaceName, p_portName);
    pSemaphore = new CMmcSemaphore(NB_OF_ALLOWED_CLIENTS, NB_OF_ALLOWED_CLIENTS, name);

    return pSemaphore;
}

BOOL CInterfaceManager::DeleteReopenSemaphore(CMmcSemaphore*& p_rpSemaphore)
{
    BOOL oResult(FALSE);

    if(p_rpSemaphore)
    {
        delete p_rpSemaphore;
        p_rpSemaphore = 0;

        oResult = TRUE;
    }

    return oResult;
}

BOOL CInterfaceManager::ReopenInterfacePort(CStdString p_InterfaceName, CStdString p_PortName)
{
    BOOL oResult(FALSE);
    CHandleRegistration_I* pRegistration(0);
    CInterfaceBase* pExistingInterface(0);
    CInterfaceBase* pInterface(0);
    CPortBase* pExistingPort(0);
    CPortBase* pPort(0);
    DWORD ulBaudrate(0);
    DWORD ulTimeout(0);
    HANDLE hI_Handle(0);
    WORD usPortMode(0);
    CParameterSet parameterSet;

    if(m_pHandleRegistrationMap_I)
    {
        LogHotPlugInReopenMessage(_T("ReopenInterfacePort"), p_InterfaceName, p_PortName);
        oResult = TRUE;

        if(oResult && !FindInterface(p_InterfaceName, &pInterface))
        {
            oResult = FALSE;
        }

        if(oResult && !FindPort(p_InterfaceName, p_PortName, &pPort))
        {
            oResult = FALSE;
        }

        if(oResult)
        {
            oResult = FALSE;
            hI_Handle = m_pHandleRegistrationMap_I->IsRegistrationExisting(pInterface, p_PortName);
            if(hI_Handle)
            {
                oResult = TRUE;

                //Get registration
                if(oResult && (!m_pHandleRegistrationMap_I->GetRegistration(hI_Handle, &pRegistration) || (pRegistration == 0)))
                {
                    oResult = FALSE;
                }

                //Lock
                if(oResult && !pRegistration->DoLock())
                {
                    oResult = FALSE;
                }

                if(oResult && m_pHandleRegistrationMap_I->GetRegistrationValues(hI_Handle, &pExistingInterface, &pExistingPort) && pExistingInterface && pExistingPort)
                {
                    oResult = FALSE;
                    pExistingPort->GetPortSettings(&ulBaudrate, &ulTimeout);
                    pExistingPort->GetPortMode(&usPortMode);
                    pExistingPort->GetParameterSet(parameterSet);

                    //Close port
                    if(pExistingInterface->I_CloseInterfacePort(pExistingPort, 0))
                    {
                        LogHotPlugInReopenMessage(_T("I_CloseInterfacePort"), pExistingPort, TRUE);

                        if(m_pHandleRegistrationMap_I->UpdateMapRegistration(hI_Handle, pInterface, pPort))
                        {
                            if(m_pHandleRegistrationMap_I->GetRegistrationValues(hI_Handle, &pInterface, &pPort) && pInterface && pPort)
                            {
                                //Init
                                pPort->InitDefaultPortSettings(ulBaudrate, ulTimeout);
                                pPort->InitDefaultPortMode(usPortMode);
                                pPort->SetParameterSet(parameterSet);

                                //Open port
                                oResult = pInterface->I_OpenInterfacePort(pPort, p_PortName, 0);
                                LogHotPlugInReopenMessage(_T("I_OpenInterfacePort"), pPort, oResult);

                                if(!oResult)
                                {
                                    //Retry
                                    if(pPort->ReloadDriver())
                                    {
                                        oResult = pInterface->I_OpenInterfacePort(pPort, p_PortName, 0);
                                        LogHotPlugInReopenMessage(_T("I_OpenInterfacePort - Retry"), pPort, oResult);
                                    }
                                }
                            }
                        }

                        //Registration
                        if(!oResult)
                        {
                            m_pHandleRegistrationMap_I->DeleteMapRegistration(hI_Handle);
                        }
                    }
                    else
                    {
                        LogHotPlugInReopenMessage(_T("I_CloseInterfacePort"), pExistingPort, FALSE);
                    }
                }

                //Unlock
                if(pRegistration) pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

void CInterfaceManager::LogHotPlugInReopenMessage(CStdString p_label, CStdString p_interfaceName, CStdString p_portName)
{
    if(CMmcLog::IsActive(LMT_INFO, CUserDefinedMessageLevel(_T("HotPlugIn"))))
    {
        CStdString message(_T(""));
        message.Format(_T("%s: %s:%s"), p_label, p_interfaceName, p_portName);

        CMmcLog::Log(LMT_INFO, CUserDefinedMessageLevel(_T("HotPlugIn")), message);
    }
}

void CInterfaceManager::LogHotPlugInReopenMessage(CStdString p_label, CPortBase* p_pPortBase, BOOL p_oResult)
{
    if(CMmcLog::IsActive(LMT_INFO, CUserDefinedMessageLevel(_T("HotPlugIn"))))
    {
        if(p_pPortBase != 0)
        {
            CStdString message(_T(""));
            CStdString portName(_T(""));

            p_pPortBase->GetPortName(&portName);
        
            if(p_oResult)
            {
                message.Format(_T("%s: %s - Succeeded"), p_label, portName);
            }
            else
            {
                message.Format(_T("%s: %s - Failed"), p_label, portName);
            }

            CMmcLog::Log(LMT_INFO, CUserDefinedMessageLevel(_T("HotPlugIn")), message);
        }
    }
}
