#include "stdafx.h"
#include <malloc.h>
#include <memory.h>

#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterSet.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterStack.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/Command_DCS_Epos4.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenCommunicationDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenDriveDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsEpos4Def.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/ProtocolStack/BaseClasses/Command_PS.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/ProtocolStack/Command_PS_CANopen.h>
#include <CommunicationModel/CommonLayer/DeviceErrorHandling/Epos4/DeviceErrorCode_Definitions_Epos4.h>
#include <CommunicationModel/DeviceCommandSet/BaseClasses/DeviceCommandSetManagerBase.h>
#include <Log/MmcLog/MmcLog.h>
#include <Process/MmcProcess.h>

#include <CommunicationModel/ProtocolStack/Classes/GatewayToDevice/CANopen/GatewayCANopenToEpos4.h>

// Constants
const CStdString PARAMETER_NODE_ID(_T("NodeId"));
const DWORD MAX_SERIAL_DATA_SIZE(255);

CGatewayCANopenToEpos4::CGatewayCANopenToEpos4()
{
    m_pCommand_WriteObject = NULL;
    m_pCommand_InitiateSegmentedWrite = NULL;
    m_pCommand_SegmentedWrite = NULL;
    m_pCommand_ReadObject = NULL;
    m_pCommand_InitiateSegmentedRead = NULL;
    m_pCommand_SegmentedRead = NULL;
    m_pCommand_AbortTransfer = NULL;
    m_pCommand_SendNMTService = NULL;

    m_ulDeviceTimeoutBackup = 0;

    InitCommands();

    m_pBufferedData = NULL;
    ResetBufferedData();

    m_ulWriteSegmentSize = MAX_SERIAL_DATA_SIZE;
}

CGatewayCANopenToEpos4::~CGatewayCANopenToEpos4()
{
    ResetBufferedData();
    DeleteCommands();
}

CGateway* CGatewayCANopenToEpos4::Clone()
{
    CGatewayCANopenToEpos4* pClonedGateway;

    pClonedGateway = new CGatewayCANopenToEpos4();
    *pClonedGateway = *this;

    return pClonedGateway;
}

CGatewayCANopenToEpos4& CGatewayCANopenToEpos4::operator=(CGatewayCANopenToEpos4& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayPStoDCS*)this) = *((CGatewayPStoDCS*)&p_rOther);
    }

    return *this;
}

BOOL CGatewayCANopenToEpos4::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CCommand_PS* pCommand_PS;
    CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager;

    if(CGateway::ProcessCommand(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle))
    {
        if(CheckLayers(p_pCommand, p_pManager))
        {
            pCommand_PS = (CCommand_PS*)p_pCommand;
            p_pDeviceCommandSetManager = (CDeviceCommandSetManagerBase*)p_pManager;

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_PS)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_PS), _T("CGatewayCANopenToEpos4-Before"));
            }

            switch(p_pCommand->GetCommandId())
            {
                case CANOPEN_INITIATE_SDO_DOWNLOAD: oResult = Process_InitiateSDODownload(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_DOWNLOAD_SDO_SEGMENT: oResult = Process_DownloadSDOSegment(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_INITIATE_SDO_UPLOAD: oResult = Process_InitiateSDOUpload(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_UPLOAD_SDO_SEGMENT: oResult = Process_UploadSDOSegment(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_NETWORK_INDICATION: oResult = Process_NetworkIndication(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_ABORT_SDO_TRANSFER: oResult = Process_AbortSDOTransfer(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_SEND_NMT_SERVICE: oResult = Process_SendNMTService(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
                case CANOPEN_CLEANUP_COMMUNICATION: oResult = Process_CleanupCommunication(pCommand_PS, p_pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            }

            if(CMmcLog::IsActive(LMT_INFO, CCommunicationMessageLevel(CML_PS)))
            {
                CMmcLog::Log(LMT_INFO, CCommunicationMessageLevel(CML_PS), _T("CGatewayCANopenToEpos4-After"));
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_InitiateSDODownload(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_DATA_SIZE(4);

    const int PARAMETER_INDEX_COBID_CLIENT_SERVER(0);
    const int PARAMETER_INDEX_COBID_SERVER_CLIENT(1);
    const int PARAMETER_INDEX_EXPEDITED_TRANSFER(2);
    const int PARAMETER_INDEX_SIZE_INDICATED(3);
    const int PARAMETER_INDEX_NON_VALID_NB_OF_BYTES(4);
    const int PARAMETER_INDEX_INDEX(5);
    const int PARAMETER_INDEX_SUB_INDEX(6);
    const int PARAMETER_INDEX_DATA(7);

    const int RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT(0);
    const int RETURN_PARAMETER_INDEX_INDEX(1);
    const int RETURN_PARAMETER_INDEX_SUB_INDEX(2);
    const int RETURN_PARAMETER_INDEX_ABORT_CODE(3);

    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oExpeditedTransfer;
    BOOL oSizeIndicated;
    BYTE ubNonValidNbOfBytes;
    WORD usIndex;
    BYTE ubSubIndex;
    void* pDataBuffer(NULL);
    DWORD ulDataBufferLength;
    DWORD ulObjectLength(0);

    DWORD ulAbortCode(0);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Prepare data buffer
        ulDataBufferLength = p_pCommand->GetParameterLength(PARAMETER_INDEX_DATA);
        if(ulDataBufferLength > 0)
        {
            pDataBuffer = malloc(ulDataBufferLength);
        }

        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_CLIENT_SERVER, &ulCobIdClientServer, sizeof(ulCobIdClientServer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->GetParameterData(PARAMETER_INDEX_EXPEDITED_TRANSFER, &oExpeditedTransfer, sizeof(oExpeditedTransfer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SIZE_INDICATED, &oSizeIndicated, sizeof(oSizeIndicated));
        p_pCommand->GetParameterData(PARAMETER_INDEX_NON_VALID_NB_OF_BYTES, &ubNonValidNbOfBytes, sizeof(ubNonValidNbOfBytes));
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_DATA, pDataBuffer, ulDataBufferLength);

        //Check data buffer length
        ulDataBufferLength = CheckDataBufferLength(oSizeIndicated, ubNonValidNbOfBytes, ulDataBufferLength, MAX_DATA_SIZE);

        //Translate command
        if(oExpeditedTransfer)
        {
            //Expedited transfer
            oResult = DCS_WriteObject(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), usIndex, ubSubIndex, (BYTE*)pDataBuffer, ulDataBufferLength, &errorInfo);
        }
        else
        {
            //NonExpedited transfer
            if(oSizeIndicated && (ulDataBufferLength >= MAX_DATA_SIZE))
            {
                ulObjectLength = *((DWORD*)pDataBuffer);
                oResult = DCS_InitiateSegmentedWrite(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), usIndex, ubSubIndex, ulObjectLength, &errorInfo);
                m_oToggle = FALSE;

                m_ulWriteSegmentSize = MAX_SERIAL_DATA_SIZE;
            }
            else
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_INTERNAL, &errorInfo);
                }

                oResult = FALSE;
            }
        }

        //Get abort code
        ulAbortCode = errorInfo.GetErrorCode();

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Free data buffer
        if(pDataBuffer)
        {
            free(pDataBuffer);
        }
        ResetBufferedData();

        //Unlock critical section
        if(oExpeditedTransfer || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_DownloadSDOSegment(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_CANOPEN_DATA_SIZE(7);

    const int PARAMETER_INDEX_COBID_CLIENT_SERVER(0);
    const int PARAMETER_INDEX_COBID_SERVER_CLIENT(1);
    const int PARAMETER_INDEX_TOGGLE(2);
    const int PARAMETER_INDEX_NON_VALID_NB_OF_BYTES(3);
    const int PARAMETER_INDEX_NO_MORE_SEGMENTS(4);
    const int PARAMETER_INDEX_SEGMENT_DATA(5);

    const int RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT(0);
    const int RETURN_PARAMETER_INDEX_TOGGLE(1);
    const int RETURN_PARAMETER_INDEX_ABORT_CODE(2);

    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oToggle;
    BYTE ubNonValidNbOfBytes;
    BOOL oNoMoreSegments;
    void* pSegDataBuffer(NULL);
    DWORD ulSegDataBufferLength;

    DWORD ulAbortCode(0);

    BOOL oRetToggle;
    void* pDataBuffer(NULL);
    DWORD ulDataBufferLength(0);
    DWORD ulDataLengthRead(0);
    DWORD ulDataLengthWritten;

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand && IsLocked(p_pCommand))
    {
        //Prepare data buffer
        ulSegDataBufferLength = p_pCommand->GetParameterLength(PARAMETER_INDEX_SEGMENT_DATA);
        if(ulSegDataBufferLength > 0)
        {
            pSegDataBuffer = malloc(ulSegDataBufferLength);
        }
        ulDataBufferLength = m_ulWriteSegmentSize;
        if(ulDataBufferLength > 0)
        {
            pDataBuffer = malloc(ulDataBufferLength);
        }

        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_CLIENT_SERVER, &ulCobIdClientServer, sizeof(ulCobIdClientServer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->GetParameterData(PARAMETER_INDEX_TOGGLE, &oToggle, sizeof(oToggle));
        p_pCommand->GetParameterData(PARAMETER_INDEX_NON_VALID_NB_OF_BYTES, &ubNonValidNbOfBytes, sizeof(ubNonValidNbOfBytes));
        p_pCommand->GetParameterData(PARAMETER_INDEX_NO_MORE_SEGMENTS, &oNoMoreSegments, sizeof(oNoMoreSegments));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SEGMENT_DATA, pSegDataBuffer, ulSegDataBufferLength);

        //Check data buffer length
        ulSegDataBufferLength = CheckDataBufferLength(TRUE, ubNonValidNbOfBytes, ulSegDataBufferLength, MAX_CANOPEN_DATA_SIZE);

        //Translate command
        oResult = AddBufferedData(pSegDataBuffer, ulSegDataBufferLength, oNoMoreSegments);
        if(oResult)
        {
            while(GetBufferedData(pDataBuffer, ulDataBufferLength, &ulDataLengthRead, &oNoMoreSegments))
            {
                //AbortCommand
                if(m_oAbortCommands)
                {
                    if(m_pErrorHandling)
                    {
                        m_pErrorHandling->GetError(ERROR_COMMAND_ABORTED_BY_USER, &errorInfo);
                    }
                    oResult = FALSE;
                    break;
                }

                oResult = DCS_SegmentedWrite(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), oNoMoreSegments, m_oToggle, pDataBuffer, ulDataLengthRead, &oRetToggle, &ulDataLengthWritten, &errorInfo);
                m_oToggle = !m_oToggle;

                //Update SegmentLength
                if(ulDataLengthWritten < ulDataLengthRead)
                {
                    ReverseBufferedData(ulDataLengthRead-ulDataLengthWritten);
                    if(pDataBuffer)
                    {
                        free(pDataBuffer);
                        pDataBuffer = NULL;

                        m_ulWriteSegmentSize = ulDataLengthWritten;
                        ulDataBufferLength = ulDataLengthWritten;

                        if(ulDataBufferLength > 0)
                        {
                            pDataBuffer = malloc(ulDataBufferLength);
                        }
                    }
                }

                if(oNoMoreSegments || !oResult)
                {
                    break;
                }
            }
        }

        // Get abort code
        ulAbortCode = errorInfo.GetErrorCode();

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_TOGGLE, &oToggle, sizeof(oToggle));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Free data buffer
        if(pSegDataBuffer)
        {
            free(pSegDataBuffer);
        }
        if(pDataBuffer)
        {
            free(pDataBuffer);
        }

        //Unlock critical section
        if(oNoMoreSegments || !oResult)
        {
            ResetBufferedData();
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_InitiateSDOUpload(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_DATA_SIZE(4);

    const int PARAMETER_INDEX_COBID_CLIENT_SERVER(0);
    const int PARAMETER_INDEX_COBID_SERVER_CLIENT(1);
    const int PARAMETER_INDEX_INDEX(2);
    const int PARAMETER_INDEX_SUB_INDEX(3);

    const int RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT(0);
    const int RETURN_PARAMETER_INDEX_EXPEDITED_TRANSFER(1);
    const int RETURN_PARAMETER_INDEX_SIZE_INDICATED(2);
    const int RETURN_PARAMETER_INDEX_NON_VALID_NB_OF_BYTES(3);
    const int RETURN_PARAMETER_INDEX_INDEX(4);
    const int RETURN_PARAMETER_INDEX_SUB_INDEX(5);
    const int RETURN_PARAMETER_INDEX_DATA(6);
    const int RETURN_PARAMETER_INDEX_ABORT_CODE(7);

    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    WORD usIndex;
    BYTE ubSubIndex;

    BOOL oRetExpeditedTransfer(FALSE);
    BOOL oRetSizeIndicated(FALSE);
    BYTE ubRetNonValidNbOfBytes(0);
    void* pRetDataBuffer(NULL);
    DWORD ulRetDataBufferLength(0);
    DWORD ulAbortCode(0);

    DWORD ulObjectLength;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Prepare data buffer
        ulRetDataBufferLength = MAX_DATA_SIZE;
        pRetDataBuffer = malloc(ulRetDataBufferLength);
        memset(pRetDataBuffer, 0, ulRetDataBufferLength);

        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_CLIENT_SERVER, &ulCobIdClientServer, sizeof(ulCobIdClientServer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));

        //Translate command
        if(DCS_ReadObject(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), usIndex, ubSubIndex, (BYTE*)pRetDataBuffer, ulRetDataBufferLength, &ulObjectLength, &errorInfo))
        {
            oResult = TRUE;

            //Expedited transfer
            oRetExpeditedTransfer = TRUE;
        }
        else if(IsDeviceResponding(oResult, &errorInfo) && DCS_InitiateSegmentedRead(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), usIndex, ubSubIndex, &errorInfo))
        {
            oResult = TRUE;

            //Non expedited transfer
            oRetExpeditedTransfer = FALSE;
            m_oToggle = FALSE;
            *((DWORD*)pRetDataBuffer) = 0;
        }
        else
        {
            //Failed
            oResult = FALSE;
        }

        //Get abort code
        ulAbortCode = errorInfo.GetErrorCode();

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_EXPEDITED_TRANSFER, &oRetExpeditedTransfer, sizeof(oRetExpeditedTransfer));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_SIZE_INDICATED, &oRetSizeIndicated, sizeof(oRetSizeIndicated));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_NON_VALID_NB_OF_BYTES, &ubRetNonValidNbOfBytes, sizeof(ubRetNonValidNbOfBytes));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_DATA, pRetDataBuffer, ulRetDataBufferLength);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Free data buffer
        if(pRetDataBuffer)
        {
            free(pRetDataBuffer);
        }
        ResetBufferedData();

        //Unlock critical section
        if(oRetExpeditedTransfer || !oResult)
        {
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_UploadSDOSegment(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_CANOPEN_DATA_SIZE(7);

    const int PARAMETER_INDEX_COBID_CLIENT_SERVER(0);
    const int PARAMETER_INDEX_COBID_SERVER_CLIENT(1);
    const int PARAMETER_INDEX_TOGGLE(2);

    const int RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT(0);
    const int RETURN_PARAMETER_INDEX_TOGGLE(1);
    const int RETURN_PARAMETER_INDEX_NON_VALID_NB_OF_BYTES(2);
    const int RETURN_PARAMETER_INDEX_NO_MORE_SEGMENTS(3);
    const int RETURN_PARAMETER_INDEX_SEGMENT_DATA(4);
    const int RETURN_PARAMETER_INDEX_ABORT_CODE(5);

    DWORD ulCobIdClientServer;
    DWORD ulCobIdServerClient;
    BOOL oToggle;

    BYTE ubRetNonValidNbOfBytes(0);
    BOOL oRetNoMoreSegments(FALSE);
    void* pRetSegDataBuffer(NULL);
    DWORD ulRetSegDataBufferLength(0);
    DWORD ulSegmentLengthRead(0);
    DWORD ulAbortCode(0);

    void* pDataBuffer(NULL);
    DWORD ulDataBufferLength(0);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand && IsLocked(p_pCommand))
    {
        //Prepare data buffer
        ulRetSegDataBufferLength = MAX_CANOPEN_DATA_SIZE;
        pRetSegDataBuffer = malloc(ulRetSegDataBufferLength);
        memset(pRetSegDataBuffer, 0, ulRetSegDataBufferLength);

        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_CLIENT_SERVER, &ulCobIdClientServer, sizeof(ulCobIdClientServer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->GetParameterData(PARAMETER_INDEX_TOGGLE, &oToggle, sizeof(oToggle));

        //Translate command
        oResult = GetBufferedData(pRetSegDataBuffer, ulRetSegDataBufferLength, &ulSegmentLengthRead, &oRetNoMoreSegments);
        if(!oResult)
        {
            oResult = DCS_SegmentedRead(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), m_oToggle, &oRetNoMoreSegments, &pDataBuffer, &ulDataBufferLength, &errorInfo);

            if(oResult)
            {
                m_oToggle = !m_oToggle;
                AddBufferedData(pDataBuffer, ulDataBufferLength, oRetNoMoreSegments);
                oResult = GetBufferedData(pRetSegDataBuffer, ulRetSegDataBufferLength, &ulSegmentLengthRead, &oRetNoMoreSegments);
            }
        }

        //Check size
        if(oResult && (ulSegmentLengthRead <= MAX_CANOPEN_DATA_SIZE))
        {
            ubRetNonValidNbOfBytes = BYTE(MAX_CANOPEN_DATA_SIZE - ulSegmentLengthRead);
        }
        else if(oResult)
        {
            oResult = FALSE;

            if(m_pErrorHandling)
            {
                m_pErrorHandling->GetError(ERROR_INTERNAL, &errorInfo);
            }
        }
        else
        {
            ubRetNonValidNbOfBytes = MAX_CANOPEN_DATA_SIZE;
        }

        //Get abort code
        ulAbortCode = errorInfo.GetErrorCode();

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_COBID_SERVER_CLIENT, &ulCobIdServerClient, sizeof(ulCobIdServerClient));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_TOGGLE, &oToggle, sizeof(oToggle));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_NON_VALID_NB_OF_BYTES, &ubRetNonValidNbOfBytes, sizeof(ubRetNonValidNbOfBytes));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_NO_MORE_SEGMENTS, &oRetNoMoreSegments, sizeof(oRetNoMoreSegments));
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_SEGMENT_DATA, pRetSegDataBuffer, ulSegmentLengthRead);
        p_pCommand->SetReturnParameterData(RETURN_PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Free data buffer
        if(pRetSegDataBuffer)
        {
            free(pRetSegDataBuffer);
        }
        if(pDataBuffer)
        {
            free(pDataBuffer);
        }

        //Unlock critical section
        if(oRetNoMoreSegments || !oResult)
        {
            ResetBufferedData();
            Unlock();
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_AbortSDOTransfer(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const int PARAMETER_INDEX_COBID_CLIENT_SERVER(0);
    const int PARAMETER_INDEX_INDEX(1);
    const int PARAMETER_INDEX_SUB_INDEX(2);
    const int PARAMETER_INDEX_ABORT_CODE(3);

    DWORD ulCobIdClientServer;
    WORD usIndex;
    BYTE ubSubIndex;
    DWORD ulAbortCode;

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COBID_CLIENT_SERVER, &ulCobIdClientServer, sizeof(ulCobIdClientServer));
        p_pCommand->GetParameterData(PARAMETER_INDEX_INDEX, &usIndex, sizeof(usIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        p_pCommand->GetParameterData(PARAMETER_INDEX_ABORT_CODE, &ulAbortCode, sizeof(ulAbortCode));

        //Execute command
        oResult = DCS_AbortTransfer(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, GetNodeId(ulCobIdClientServer), usIndex, (WORD)ubSubIndex, ulAbortCode, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_NetworkIndication(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    //Not Supported
    return TRUE;
}

BOOL CGatewayCANopenToEpos4::Process_SendNMTService(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const int PARAMETER_INDEX_COMMAND_SPECIFIER(0);
    const int PARAMETER_INDEX_NODE_ID(1);

    BYTE ubCommandSpecifier;
    BYTE ubNodeId;

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    //Lock critical section
    if(p_pCommand && Lock(p_pCommand))
    {
        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_INDEX_COMMAND_SPECIFIER, &ubCommandSpecifier, sizeof(ubCommandSpecifier));
        p_pCommand->GetParameterData(PARAMETER_INDEX_NODE_ID, &ubNodeId, sizeof(ubNodeId));

        //Translate command
        oResult = DCS_SendNMTService(p_pDeviceCommandSetManager, p_hDCS_Handle, p_hTransactionHandle, (WORD)ubNodeId, (WORD)ubCommandSpecifier, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);

        //Unlock critical section
        Unlock();
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::Process_CleanupCommunication(CCommand_PS* p_pCommand, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const int PARAMETER_LAST_RESULT(0);

    BOOL oResult(FALSE);
    BOOL oCleanupCanPassive(FALSE);

    if (p_pCommand != 0)
    {
        //Get parameter data
        p_pCommand->GetParameterData(PARAMETER_LAST_RESULT, &oCleanupCanPassive, sizeof(oCleanupCanPassive));

        //Cleanup
        if (oCleanupCanPassive)
        {
            ResetGatewayCommunication(p_pDeviceCommandSetManager, p_hDCS_Handle);

            if (IsGatewayInCanPassiveState(p_pDeviceCommandSetManager, p_hDCS_Handle))
            {
                ClearGatewayErrors(p_pDeviceCommandSetManager, p_hDCS_Handle);
            }
        }

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_InitiateSegmentedRead(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usIndex, BYTE p_ubSubIndex, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;

    if(m_pCommand_InitiateSegmentedRead && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_InitiateSegmentedRead, ubNodeId);

        //Set parameter data
        m_pCommand_InitiateSegmentedRead->ResetStatus();
        m_pCommand_InitiateSegmentedRead->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSegmentedRead->SetParameterData(1, &p_ubSubIndex, sizeof(p_ubSubIndex));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_InitiateSegmentedRead, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSegmentedRead->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_InitiateSegmentedRead->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_InitiateSegmentedWrite(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usIndex, BYTE p_ubSubIndex, DWORD p_ulObjectLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;

    if(m_pCommand_InitiateSegmentedWrite && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_InitiateSegmentedWrite, ubNodeId);

        //Set parameter data
        m_pCommand_InitiateSegmentedWrite->ResetStatus();
        m_pCommand_InitiateSegmentedWrite->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSegmentedWrite->SetParameterData(1, &p_ubSubIndex, sizeof(p_ubSubIndex));
        m_pCommand_InitiateSegmentedWrite->SetParameterData(2, &p_ulObjectLength, sizeof(p_ulObjectLength));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_InitiateSegmentedWrite, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSegmentedWrite->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_InitiateSegmentedWrite->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_ReadObject(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usIndex, BYTE p_ubSubIndex, BYTE* p_pObjectBuffer, DWORD p_ulObjectBufferLength, DWORD* p_pulObjectLengthRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;
    DWORD ulObjectLength;

    if(m_pCommand_ReadObject && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_ReadObject, ubNodeId);

        //Set parameter data
        m_pCommand_ReadObject->ResetStatus();
        m_pCommand_ReadObject->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_ReadObject->SetParameterData(1, &p_ubSubIndex, sizeof(p_ubSubIndex));
        m_pCommand_ReadObject->SetParameterData(2, &p_ulObjectBufferLength, sizeof(p_ulObjectBufferLength));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_ReadObject, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter length
        ulObjectLength = m_pCommand_ReadObject->GetReturnParameterLength(1);
        if(p_ulObjectBufferLength < ulObjectLength)
        {
            ulObjectLength = p_ulObjectBufferLength;
        }

        if(p_pulObjectLengthRead)
        {
            *p_pulObjectLengthRead = ulObjectLength;
        }

        //Get return parameter data
        m_pCommand_ReadObject->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_ReadObject->GetReturnParameterData(1, p_pObjectBuffer, ulObjectLength);

        //Get error code
        m_pCommand_ReadObject->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_SegmentedRead(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, BOOL p_oToggle, BOOL* p_poLastDataSegment, void** p_ppSegmentBuffer, DWORD* p_pulSegmentBufferLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;
    UEpos4ControlByte sControlByte;

    if(m_pCommand_SegmentedRead && p_pDeviceCommandSetManager && p_ppSegmentBuffer && p_pulSegmentBufferLength)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_SegmentedRead, ubNodeId);

        //Prepare control byte
        sControlByte.bValue = 0;
        sControlByte.structure.bToggle = p_oToggle;

        //Set parameter data
        m_pCommand_SegmentedRead->ResetStatus();
        m_pCommand_SegmentedRead->SetParameterData(0, &sControlByte, sizeof(sControlByte));
        m_pCommand_SegmentedRead->SetParameterData(1, p_pulSegmentBufferLength, sizeof(*p_pulSegmentBufferLength));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_SegmentedRead, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SegmentedRead->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_SegmentedRead->GetReturnParameterData(1, &sControlByte, sizeof(sControlByte));
        m_pCommand_SegmentedRead->GetReturnParameterData(2, p_pulSegmentBufferLength, sizeof(*p_pulSegmentBufferLength));

        //Prepare data buffer
        if(*p_pulSegmentBufferLength > 0)
        {
            *p_ppSegmentBuffer = malloc(*p_pulSegmentBufferLength);
            memset(*p_ppSegmentBuffer, 0, *p_pulSegmentBufferLength);
        }
        m_pCommand_SegmentedRead->GetReturnParameterData(3, *p_ppSegmentBuffer, *p_pulSegmentBufferLength);

        //Get error code
        m_pCommand_SegmentedRead->GetErrorInfo(p_pErrorInfo);

        //Restore control byte
        if(p_poLastDataSegment)
        {
            *p_poLastDataSegment = sControlByte.structure.bLastSegment;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_SegmentedWrite(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, BOOL p_oLastSegment, BOOL p_oToggle, void* p_pSegmentBuffer, DWORD p_ulSegmentLength, BOOL* p_poRetToggle, DWORD* p_pulSegmentLengthWritten, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubDataSizeWritten(0);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;
    UEpos4ControlByte sControlByte;

    if(m_pCommand_SegmentedWrite && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_SegmentedWrite, ubNodeId);

        //Prepare control byte
        sControlByte.bValue = 0;
        sControlByte.structure.bToggle = p_oToggle;
        sControlByte.structure.bLastSegment = p_oLastSegment;

        //Set parameter data
        m_pCommand_SegmentedWrite->ResetStatus();
        m_pCommand_SegmentedWrite->SetParameterData(0, &sControlByte, sizeof(sControlByte));
        m_pCommand_SegmentedWrite->SetParameterData(1, p_pSegmentBuffer, p_ulSegmentLength);

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_SegmentedWrite, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SegmentedWrite->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_SegmentedWrite->GetReturnParameterData(1, &sControlByte, sizeof(sControlByte));
        m_pCommand_SegmentedWrite->GetReturnParameterData(2, &ubDataSizeWritten, sizeof(ubDataSizeWritten));

        //Restore control byte
        if(p_poRetToggle)
        {
            *p_poRetToggle = sControlByte.structure.bToggle;
        }
        if(p_pulSegmentLengthWritten)
        {
            *p_pulSegmentLengthWritten = ubDataSizeWritten;
        }

        //Get error code
        m_pCommand_SegmentedWrite->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_WriteObject(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usIndex, BYTE p_ubSubIndex, BYTE* p_pObjectBuffer, DWORD p_ulObjectLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    DWORD ulDeviceErrorCode;

    if(m_pCommand_WriteObject && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_WriteObject, ubNodeId);

        //Set parameter data
        m_pCommand_WriteObject->ResetStatus();
        m_pCommand_WriteObject->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_WriteObject->SetParameterData(1, &p_ubSubIndex, sizeof(p_ubSubIndex));
        m_pCommand_WriteObject->SetParameterData(2, p_pObjectBuffer, p_ulObjectLength);
        m_pCommand_WriteObject->SetParameterData(3, &p_ulObjectLength, sizeof(p_ulObjectLength));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_WriteObject, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_WriteObject->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_WriteObject->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_AbortTransfer(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usIndex, WORD p_usSubIndex, DWORD p_ulAbortCode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNodeId((BYTE)p_usNodeId);
    BYTE ubSubIndex((BYTE)p_usSubIndex);

    if(m_pCommand_AbortTransfer && p_pDeviceCommandSetManager)
    {
        //Address parameter
        UpdateNodeId(m_pCommand_AbortTransfer, ubNodeId);

        //Set parameter data
        m_pCommand_AbortTransfer->ResetStatus();
        m_pCommand_AbortTransfer->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_AbortTransfer->SetParameterData(1, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_AbortTransfer->SetParameterData(2, &p_ulAbortCode, sizeof(p_ulAbortCode));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_AbortTransfer, p_hDCS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_AbortTransfer->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::DCS_SendNMTService(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeId, WORD p_usCommandSpecifier, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pCommand_SendNMTService && p_pDeviceCommandSetManager)
    {
        //Set parameter data
        m_pCommand_SendNMTService->ResetStatus();
        m_pCommand_SendNMTService->SetParameterData(0, &p_usNodeId, sizeof(p_usNodeId));
        m_pCommand_SendNMTService->SetParameterData(1, &p_usCommandSpecifier, sizeof(p_usCommandSpecifier));

        //Execute command
        oResult = p_pDeviceCommandSetManager->ExecuteCommand(m_pCommand_SendNMTService, p_hDCS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_SendNMTService->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

void CGatewayCANopenToEpos4::InitCommands()
{
    DeleteCommands();

    m_pCommand_WriteObject = new CCommand_DCS_Epos4();
    m_pCommand_WriteObject->InitCommand(EPOS4_WRITE_OBJECT);

    m_pCommand_ReadObject = new CCommand_DCS_Epos4();
    m_pCommand_ReadObject->InitCommand(EPOS4_READ_OBJECT);

    m_pCommand_InitiateSegmentedWrite = new CCommand_DCS_Epos4();
    m_pCommand_InitiateSegmentedWrite->InitCommand(EPOS4_INITIATE_SEGMENT_WRITE);

    m_pCommand_InitiateSegmentedRead = new CCommand_DCS_Epos4();
    m_pCommand_InitiateSegmentedRead->InitCommand(EPOS4_INITIATE_SEGMENT_READ);

    m_pCommand_SegmentedWrite = new CCommand_DCS_Epos4();
    m_pCommand_SegmentedWrite->InitCommand(EPOS4_SEGMENT_WRITE);

    m_pCommand_SegmentedRead = new CCommand_DCS_Epos4();
    m_pCommand_SegmentedRead->InitCommand(EPOS4_SEGMENT_READ);

    m_pCommand_AbortTransfer = new CCommand_DCS_Epos4();
    m_pCommand_AbortTransfer->InitCommand(EPOS4_ABORT_TRANSFER);

    m_pCommand_SendNMTService = new CCommand_DCS_Epos4();
    m_pCommand_SendNMTService->InitCommand(EPOS4_SEND_NMT_SERVICE);

}

void CGatewayCANopenToEpos4::DeleteCommands()
{
    if(m_pCommand_WriteObject)
    {
        delete m_pCommand_WriteObject;
        m_pCommand_WriteObject = NULL;
    }

    if(m_pCommand_InitiateSegmentedWrite)
    {
        delete m_pCommand_InitiateSegmentedWrite;
        m_pCommand_InitiateSegmentedWrite = NULL;
    }

    if(m_pCommand_SegmentedWrite)
    {
        delete m_pCommand_SegmentedWrite;
        m_pCommand_SegmentedWrite = NULL;
    }

    if(m_pCommand_ReadObject)
    {
        delete m_pCommand_ReadObject;
        m_pCommand_ReadObject = NULL;
    }

    if(m_pCommand_InitiateSegmentedRead)
    {
        delete m_pCommand_InitiateSegmentedRead;
        m_pCommand_InitiateSegmentedRead = NULL;
    }

    if(m_pCommand_SegmentedRead)
    {
        delete m_pCommand_SegmentedRead;
        m_pCommand_SegmentedRead = NULL;
    }

    if(m_pCommand_AbortTransfer)
    {
        delete m_pCommand_AbortTransfer;
        m_pCommand_AbortTransfer = NULL;
    }

    if(m_pCommand_SendNMTService)
    {
        delete m_pCommand_SendNMTService;
        m_pCommand_SendNMTService = NULL;
    }
}

WORD CGatewayCANopenToEpos4::GetNodeId(DWORD p_ulSdoCobIdClientServer)
{
    WORD usResult(0);

    if((p_ulSdoCobIdClientServer >= MIN_TX_SDO) && (p_ulSdoCobIdClientServer <= MAX_TX_SDO))
    {
        usResult = WORD(p_ulSdoCobIdClientServer - TX_SDO_OFFSET);
    }

    return usResult;
}

DWORD CGatewayCANopenToEpos4::CheckDataBufferLength(BOOL p_oSizeIndicated, BYTE p_ubNonValidNbOfBytes, DWORD p_ulDataBufferLength, DWORD p_ulMaxDataBufferLength)
{
    if(p_oSizeIndicated)
    {
        //Size indicated
        if(p_ulDataBufferLength > (p_ulMaxDataBufferLength - p_ubNonValidNbOfBytes))
        {
            p_ulDataBufferLength = (p_ulMaxDataBufferLength - p_ubNonValidNbOfBytes);
        }
    }
    else
    {
        //Size not indicated
        if(p_ulDataBufferLength > p_ulMaxDataBufferLength)
        {
            p_ulDataBufferLength = p_ulMaxDataBufferLength;
        }
    }

    return p_ulDataBufferLength;
}

BOOL CGatewayCANopenToEpos4::ResetBufferedData()
{
    if(m_pBufferedData)
    {
        free(m_pBufferedData);
        m_pBufferedData = 0;
    }

    m_ulBufferedDataSize = 0;
    m_ulBufferedDataRead = 0;
    m_oNoMoreSegments = FALSE;

    return TRUE;
}

BOOL CGatewayCANopenToEpos4::AddBufferedData(void* p_pDataBuffer, DWORD p_ulDataBufferSize, BOOL p_oNoMoreSegments)
{
    BOOL oResult(FALSE);
    BYTE* pBuffer;
    DWORD ulSize;

    if(p_ulDataBufferSize > 0)
    {
        //Allocate memory
        ulSize = m_ulBufferedDataSize + p_ulDataBufferSize;
        m_pBufferedData = realloc(m_pBufferedData, ulSize);

        pBuffer = (BYTE*)m_pBufferedData + m_ulBufferedDataSize;
        memcpy(pBuffer, p_pDataBuffer, p_ulDataBufferSize);

        m_ulBufferedDataSize += p_ulDataBufferSize;

        //No more segments
        m_oNoMoreSegments = p_oNoMoreSegments;

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::IsBufferedDataAvailable(DWORD p_ulSize, DWORD* p_pulSizeAvailable, BOOL* p_poNoMoreSegments)
{
    BOOL oResult(FALSE);
    DWORD ulTotalSizeAvailable;

    if(m_ulBufferedDataSize > m_ulBufferedDataRead)
    {
        ulTotalSizeAvailable = m_ulBufferedDataSize - m_ulBufferedDataRead;
        if(ulTotalSizeAvailable > p_ulSize)
        {
            if(p_poNoMoreSegments)
            {
                *p_poNoMoreSegments = FALSE;
            }
            if(p_pulSizeAvailable)
            {
                *p_pulSizeAvailable = p_ulSize;
            }

            oResult = TRUE;
        }
        else if(ulTotalSizeAvailable == p_ulSize)
        {
            if(p_poNoMoreSegments)
            {
                *p_poNoMoreSegments = m_oNoMoreSegments;
            }
            if(p_pulSizeAvailable)
            {
                *p_pulSizeAvailable = ulTotalSizeAvailable;
            }

            oResult = TRUE;
        }
        else if(ulTotalSizeAvailable < p_ulSize)
        {
            if(m_oNoMoreSegments)
            {
                if(p_poNoMoreSegments)
                {
                    *p_poNoMoreSegments = TRUE;
                }
                if(p_pulSizeAvailable)
                {
                    *p_pulSizeAvailable = ulTotalSizeAvailable;
                }

                oResult = TRUE;
            }
        }
    }

    if(!oResult && p_poNoMoreSegments)
    {
        *p_poNoMoreSegments = FALSE;
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::ReverseBufferedData(DWORD p_ulReverseSize)
{
    BOOL oResult(FALSE);

    if(m_ulBufferedDataRead >= p_ulReverseSize)
    {
        m_ulBufferedDataRead -= p_ulReverseSize;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::InitLayerParameterStack(CCommandRoot* p_pCommand)
{
    BOOL oResult(FALSE);
    CLayerParameterSet layerParameterSet;
    CLayerParameterStack layerParameterStack;

    if(p_pCommand)
    {
        if(p_pCommand->GetLayerParameterStack(layerParameterStack))
        {
            //Pop PS layer Parameter Set
            layerParameterStack.PopLayer(PROTOCOL_STACK_LAYER, layerParameterSet);

            //Set I layer Commands
            if(m_pCommand_WriteObject) m_pCommand_WriteObject->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_InitiateSegmentedWrite) m_pCommand_InitiateSegmentedWrite->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_SegmentedWrite) m_pCommand_SegmentedWrite->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ReadObject) m_pCommand_ReadObject->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_InitiateSegmentedRead) m_pCommand_InitiateSegmentedRead->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_SegmentedRead) m_pCommand_SegmentedRead->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_AbortTransfer) m_pCommand_AbortTransfer->SetLayerParameterStack(layerParameterStack);

            if(m_pCommand_SendNMTService) m_pCommand_SendNMTService->SetLayerParameterStack(layerParameterStack);

            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::SetDeviceTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, DWORD p_ulTimeout)
{
    WORD usTimeout(0);
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    CStdString interfaceName(_T(""));

    if(p_pManager)
    {
        if(p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            //Init
            usTimeout = (WORD)p_ulTimeout;

            if(GetAddressParameter(p_pManager, p_hHandle, ubNodeId, &errorInfo))
            {
                if(interfaceName == INTERFACE_RS232)
                {
                    //RS232 Frame Timeout
                    oResult = DCS_WriteObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_RS232_FRAME_TIMEOUT, SUBINDEX_RS232_FRAME_TIMEOUT, (BYTE*)&usTimeout, sizeof(usTimeout), &errorInfo);
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    //USB Frame Timeout
                    oResult = DCS_WriteObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_USB_FRAME_TIMEOUT, SUBINDEX_USB_FRAME_TIMEOUT, (BYTE*)&usTimeout, sizeof(usTimeout), &errorInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::GetDeviceTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, DWORD* p_pulTimeout)
{
    WORD usTimeout(0);
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    DWORD ulLengthRead(0);
    CStdString interfaceName(_T(""));

    if (p_pManager && p_pulTimeout)
    {
        if (p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            if (GetAddressParameter(p_pManager, p_hHandle, ubNodeId, &errorInfo))
            {
                if (interfaceName == INTERFACE_RS232)
                {
                    //RS232 Frame Timeout
                    if (DCS_ReadObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_RS232_FRAME_TIMEOUT, SUBINDEX_RS232_FRAME_TIMEOUT, (BYTE*)&usTimeout, sizeof(usTimeout), &ulLengthRead, &errorInfo))
                    {
                        *p_pulTimeout = usTimeout;
                        oResult = TRUE;
                    }
                }
                else if (interfaceName == INTERFACE_USB)
                {
                    //USB Frame Timeout
                    if (DCS_ReadObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_USB_FRAME_TIMEOUT, SUBINDEX_USB_FRAME_TIMEOUT, (BYTE*)&usTimeout, sizeof(usTimeout), &ulLengthRead, &errorInfo))
                    {
                        *p_pulTimeout = usTimeout;
                        oResult = TRUE;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::GetAddressParameter(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE& p_rNodeId, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(p_pManager)
    {
        //Get address parameter from device
        if (p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, PARAMETER_NODE_ID, (BYTE*)&p_rNodeId, sizeof(p_rNodeId), p_pErrorInfo))
        {
            oResult = TRUE;
        }
        else
        {
            p_rNodeId = 0;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::GetBufferedData(void* p_pDataBuffer, DWORD p_ulDataBufferSize, DWORD* p_pulDataRead, BOOL* p_poNoMoreSegments)
{
    BOOL oResult(FALSE);
    BYTE* pBuffer;
    DWORD ulSizeAvailable;

    if(p_pDataBuffer && p_pulDataRead && p_poNoMoreSegments)
    {
        if(IsBufferedDataAvailable(p_ulDataBufferSize, &ulSizeAvailable, p_poNoMoreSegments))
        {
            if(ulSizeAvailable > p_ulDataBufferSize)
            {
                ulSizeAvailable = p_ulDataBufferSize;
            }

            pBuffer = (BYTE*)m_pBufferedData + m_ulBufferedDataRead;
            memcpy(p_pDataBuffer, pBuffer, ulSizeAvailable);
            m_ulBufferedDataRead += ulSizeAvailable;
            *p_pulDataRead = ulSizeAvailable;

            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::UpdateNodeId(CCommand_DCS_Epos4* p_pCommand, BYTE p_ubNodeId)
{
    BOOL oResult(FALSE);
    CLayerParameterSet deviceParameterSet(DEVICE_COMMAND_SET_LAYER);

    if(p_pCommand)
    {
        p_pCommand->ResetLayerParameterStack();

        deviceParameterSet.InitParameter(PARAMETER_NODE_ID, &p_ubNodeId, sizeof(p_ubNodeId));
        oResult = p_pCommand->PushLayerParameterSet(deviceParameterSet);
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::IsDeviceResponding(BOOL p_oResult, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(p_oResult)
    {
        oResult = TRUE;
    }
    else
    {
        if(p_pErrorInfo != 0)
        {
            switch(p_pErrorInfo->GetErrorCode())
            {
                case NO_ERROR_DETECTED:
                case ERROR_DEVICE_CAN_CLIENT_SERVER:
                case ERROR_DEVICE_CAN_INVALID_BLOCK_SIZE:
                case ERROR_DEVICE_CAN_INVALID_SEQUENCE:
                case ERROR_DEVICE_CAN_CRC_ERROR:
                case ERROR_DEVICE_CAN_TOGGLE:
                case ERROR_DEVICE_CAN_SDO_TIME_OUT: oResult = FALSE; break;
                default: oResult = TRUE; break;
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::IsGatewayInCanPassiveState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    CErrorInfo errorInfo;
    BYTE ubNbOfErrors(0);
    DWORD ulLengthRead(0);
    DWORD ulErrorCode(0);

    if (GetAddressParameter(p_pManager, p_hHandle, ubNodeId, &errorInfo))
    {
        if (DCS_ReadObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_ERROR_HISTORY, SUBINDEX_NB_OF_ERRORS, (BYTE*)&ubNbOfErrors, sizeof(ubNbOfErrors), &ulLengthRead, &errorInfo))
        {
            for (BYTE i = 1; i <= ubNbOfErrors; i++)
            {
                if (DCS_ReadObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_ERROR_HISTORY, i, (BYTE*)&ulErrorCode, sizeof(ulErrorCode), &ulLengthRead, &errorInfo))
                {
                    if (ulErrorCode == ERROR_EPOS4_CAN_PASSIVE)
                    {
                        oResult = TRUE;
                        break;
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::ResetGatewayCommunication(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    CErrorInfo errorInfo;

    if (GetAddressParameter(p_pManager, p_hHandle, ubNodeId, &errorInfo))
    {
        if (DCS_SendNMTService(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, NSI_RESET_COMMUNICATION, &errorInfo))
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::ClearGatewayErrors(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    CErrorInfo errorInfo;
    WORD usStatusWord(0);
    DWORD ulLengthRead(0);

    if (GetAddressParameter(p_pManager, p_hHandle, ubNodeId, &errorInfo))
    {
        if (DCS_ReadObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_STATUS_WORD, SUBINDEX_STATUS_WORD, (BYTE*)&usStatusWord, sizeof(usStatusWord), &ulLengthRead, &errorInfo))
        {
            if (usStatusWord & DS_FAULT)
            {
                //Clear fault 0
                if (DCS_WriteObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_CONTROL_WORD, SUBINDEX_CONTROL_WORD, (BYTE*)&DISABLE_VOLTAGE, sizeof(DISABLE_VOLTAGE), &errorInfo))
                {
                    //Clear fault 1
                    oResult = DCS_WriteObject(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, INDEX_CONTROL_WORD, SUBINDEX_CONTROL_WORD, (BYTE*)&FAULT_RESET, sizeof(FAULT_RESET), &errorInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::SetGatewayTimeout(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle, DWORD p_ulTimeout)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceTimeout(0);

    if (GetDeviceTimeout(p_pDeviceCommandSetManager, p_hDCS_Handle, &ulDeviceTimeout))
    {
        m_ulDeviceTimeoutBackup = ulDeviceTimeout;
        if (ulDeviceTimeout != p_ulTimeout)
        {
            oResult = SetDeviceTimeout(p_pDeviceCommandSetManager, p_hDCS_Handle, p_ulTimeout);
        }
        else
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayCANopenToEpos4::RestoreGatewayTimeout(CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, HANDLE p_hDCS_Handle)
{
    BOOL oResult(FALSE);

    if (m_ulDeviceTimeoutBackup != 0)
    {
        if (SetDeviceTimeout(p_pDeviceCommandSetManager, p_hDCS_Handle, m_ulDeviceTimeoutBackup))
        {
            m_ulDeviceTimeoutBackup = 0;
            oResult = TRUE;
        }
    }

    return oResult;
}