#include "stdafx.h"
#include <math.h>
#include <memory.h>

#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterSet.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterStack.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/Command_DCS_Epos4.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenCommunicationDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenDriveDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsEpos4Def.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/BaseClasses/Command_VCS.h>
#include <CommunicationModel/CommonLayer/DeviceErrorHandling/Epos4/DeviceErrorCode_Definitions_Epos4.h>
#include <CommunicationModel/DeviceCommandSet/BaseClasses/DeviceCommandSetManagerBase.h>
#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/BaseClasses/Cia402StateMachine.h>
#include <Process/MmcProcess.h>
#include <Thread/MmcSingleLock.h>

#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/BaseClasses/GatewayVCStoEpos4.h>

using namespace std;

//Constants
const WORD MAX_LENGTH_NON_SEGMENTED_TRANSFER = 4;

auto_ptr<CMmcCriticalSection> CGatewayVCStoEpos4::m_SyncObject(new CMmcCriticalSection());

CGatewayVCStoEpos4::CGatewayVCStoEpos4()
    : m_pCia402StateMachine(0)
    , m_pCommand_AbortTransfer(0)
    , m_pCommand_InitiateSegmentedRead(0)
    , m_pCommand_InitiateSegmentedWrite(0)
    , m_pCommand_ReadCANFrame(0)
    , m_pCommand_ReadLSSFrame(0)
    , m_pCommand_ReadObject(0)
    , m_pCommand_RequestCANFrame(0)
    , m_pCommand_SegmentedRead(0)
    , m_pCommand_SegmentedWrite(0)
    , m_pCommand_SendCANFrame(0)
    , m_pCommand_SendLSSFrame(0)
    , m_pCommand_SendNMTService(0)
    , m_pCommand_WriteObject(0)
    , m_pCommand_InitEsm(0)
    , m_pCommand_ChangeEsmState(0)
    , m_pCommand_GetEsmState(0)
    , m_pCommand_WriteEscRegister(0)
    , m_pCommand_ReadEscRegister(0)
{
    InitCommands_ObjectDictionary();
    InitCommands_GeneralGateway();
    InitCommands_NetworkManagement();
    InitCommands_LayerSettingServices();
    InitCommands_Register();
    InitCommands_EtherCatStateMachine();

    CreateStateMachine();
}

CGatewayVCStoEpos4::~CGatewayVCStoEpos4()
{
    DeleteCommands_ObjectDictionary();
    DeleteCommands_NetworkManagement();
    DeleteCommands_GeneralGateway();
    DeleteCommands_LayerSettingServices();
    DeleteCommands_Register();
    DeleteCommands_EtherCatStateMachine();

    DeleteStateMachine();
}

BOOL CGatewayVCStoEpos4::CreateStateMachine()
{
    BOOL oResult(FALSE);

    if(!m_pCia402StateMachine)
    {
        m_pCia402StateMachine = new CCia402StateMachine();
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DeleteStateMachine()
{
   BOOL oResult(FALSE);

    if(m_pCia402StateMachine)
    {
        delete m_pCia402StateMachine;
        m_pCia402StateMachine = NULL;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsGatewayInCanPassiveState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    CErrorInfo errorInfo;
    BYTE ubNbOfErrors(0);
    DWORD ulErrorCode(0);

    if (ReadUInt8Object(p_pManager, p_hHandle, hTransactionHandle, INDEX_ERROR_HISTORY, SUBINDEX_NB_OF_ERRORS, &ubNbOfErrors, &errorInfo))
    {
        for (BYTE i = 1; i <= ubNbOfErrors; i++)
        {
            if (ReadUInt32Object(p_pManager, p_hHandle, hTransactionHandle, INDEX_ERROR_HISTORY, i, &ulErrorCode, &errorInfo))
            {
                if (ulErrorCode == ERROR_EPOS4_CAN_PASSIVE)
                {
                    oResult = TRUE;
                    break;
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ResetGatewayCommunication(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    BYTE ubNodeId(0);
    CErrorInfo errorInfo;

    if (GetAddressParameter(p_pManager, p_hHandle, ubNodeId, 0))
    {
        if (DCS_SendNMTService(p_pManager, p_hHandle, hTransactionHandle, ubNodeId, NSI_RESET_COMMUNICATION, &errorInfo))
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ClearGatewayErrors(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    HANDLE hTransactionHandle(NULL);
    CErrorInfo errorInfo;
    WORD usStatusWord(0);

    if (ReadUInt16Object(p_pManager, p_hHandle, hTransactionHandle, INDEX_STATUS_WORD, SUBINDEX_STATUS_WORD, &usStatusWord, &errorInfo))
    {
        if (usStatusWord & DS_FAULT)
        {
            //Clear fault 0
            if (WriteUInt16Object(p_pManager, p_hHandle, hTransactionHandle, INDEX_CONTROL_WORD, SUBINDEX_CONTROL_WORD, DISABLE_VOLTAGE, &errorInfo))
            {
                //Clear fault 1
                oResult = WriteUInt16Object(p_pManager, p_hHandle, hTransactionHandle, INDEX_CONTROL_WORD, SUBINDEX_CONTROL_WORD, FAULT_RESET, &errorInfo);
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pubDataBuffer, DWORD p_ulNbOfBytesToRead, DWORD* p_pulNbOfBytesRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pManager)
    {
        //Preparation
        if(p_pulNbOfBytesRead) *p_pulNbOfBytesRead = 0;

        if(p_ulNbOfBytesToRead <= MAX_LENGTH_NON_SEGMENTED_TRANSFER)
        {
            //Non segmented transfer
            if(DCS_ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, (BYTE)p_usSubIndex, p_pubDataBuffer, p_ulNbOfBytesToRead, &errorInfo))
            {
                if(p_pulNbOfBytesRead)
                {
                    *p_pulNbOfBytesRead = p_ulNbOfBytesToRead;
                }

                oResult = TRUE;
            }
        }

        if((p_ulNbOfBytesToRead > MAX_LENGTH_NON_SEGMENTED_TRANSFER) || (!oResult && errorInfo.GetErrorCode() == ERROR_EXPEDITED_TRANSFER_EXPECTED))
        {
            //Segmented transfer
            if(m_oAbortCommands)
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_COMMAND_ABORTED_BY_USER, &errorInfo);
                }
            }
            else
            {
                oResult = ReadSegmentedObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, p_pubDataBuffer, p_ulNbOfBytesToRead, p_pulNbOfBytesRead, &errorInfo);
            }
        }

        if(p_pErrorInfo != 0) *p_pErrorInfo = errorInfo;
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pubDataBuffer, DWORD p_ulNbOfBytesToWrite, DWORD* p_pulNbOfBytesWritten, CErrorInfo* p_pErrorInfo)
{
    const WORD MAX_SEGMENT_LENGTH = 255;

    BOOL oLastSegment = FALSE;
    BOOL oResult(FALSE);
    BOOL oToggle = FALSE;
    BYTE ubSubIndex = (BYTE)p_usSubIndex;
    BYTE* pSegmentBuffer = (BYTE*)p_pubDataBuffer;
    DWORD ulAbortCode = (DWORD)-1;
    DWORD ulSegmentLength = p_ulNbOfBytesToWrite;
    DWORD ulSegmentLengthWritten = 0;
    DWORD ulTotalNbOfBytesToWrite = p_ulNbOfBytesToWrite;
    DWORD ulTotalNbOfBytesWritten = 0;

    if(p_pManager)
    {
        //Preparation
        if(p_pulNbOfBytesWritten)
        {
            *p_pulNbOfBytesWritten = 0;
        }

        if(ulSegmentLength <= MAX_LENGTH_NON_SEGMENTED_TRANSFER)
        {
            //Non segmented transfer
            if(DCS_WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, ubSubIndex, pSegmentBuffer, ulSegmentLength, p_pErrorInfo))
            {
                ulTotalNbOfBytesWritten = p_ulNbOfBytesToWrite;
                oResult = TRUE;
            }
        }
        else
        {
            //Don't allow new segmented transfer
            if(m_oAbortCommands)
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_COMMAND_ABORTED_BY_USER, p_pErrorInfo);
                }
            }
            else
            {
                //Segmented transfer
                if(DCS_InitiateSegmentedWrite(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, ubSubIndex, p_ulNbOfBytesToWrite, p_pErrorInfo))
                {
                    //Initialize variables
                    oResult = TRUE;
                    if(p_ulNbOfBytesToWrite < MAX_SEGMENT_LENGTH)
                    {
                        ulSegmentLength = p_ulNbOfBytesToWrite;
                    }
                    else
                    {
                        ulSegmentLength = MAX_SEGMENT_LENGTH;
                    }

                    oLastSegment = (p_ulNbOfBytesToWrite <= ulSegmentLength);

                    do
                    {
                        //Abort transfer
                        if(m_oAbortCommands)
                        {
                            if(DCS_AbortTransfer(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, ulAbortCode))
                            {
                                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_COMMAND_ABORTED_BY_USER, p_pErrorInfo);
                                oResult = FALSE;
                                break;
                            }
                        }

                        //Do transfer
                        oResult = DCS_SegmentedWrite(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oToggle, oLastSegment, pSegmentBuffer, ulSegmentLength, &ulSegmentLengthWritten, p_pErrorInfo);
                        if(oResult)
                        {
                            //Prepare next segment
                            ulTotalNbOfBytesWritten += ulSegmentLengthWritten;
                            p_ulNbOfBytesToWrite -= (WORD)ulSegmentLengthWritten;
                            oToggle = !oToggle;
                            pSegmentBuffer += ulSegmentLengthWritten;

                            if(p_ulNbOfBytesToWrite < ulSegmentLengthWritten)
                            {
                                ulSegmentLength = p_ulNbOfBytesToWrite;
                            }
                            else
                            {
                                ulSegmentLength = ulSegmentLengthWritten;
                            }

                            if(ulSegmentLength > MAX_SEGMENT_LENGTH)
                            {
                                ulSegmentLength = MAX_SEGMENT_LENGTH;
                            }

                            oLastSegment = (p_ulNbOfBytesToWrite <= ulSegmentLength);
                        }

                    } while(oResult && (ulTotalNbOfBytesWritten < ulTotalNbOfBytesToWrite));

                }
            }
        }

        if(p_pulNbOfBytesWritten)
        {
            *p_pulNbOfBytesWritten = ulTotalNbOfBytesWritten;
        }
    }
    else
    {
        if(m_pErrorHandling)
        {
            m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadInt8Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, __int8* p_pbValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_pbValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pbValue, sizeof(*p_pbValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_pbValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadInt16Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, short* p_psValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_psValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_psValue, sizeof(*p_psValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_psValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadInt32Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, int* p_plValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_plValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_plValue, sizeof(*p_plValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_plValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadUInt8Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pubValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_pubValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pubValue, sizeof(*p_pubValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_pubValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadUInt16Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, WORD* p_pusValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_pusValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pusValue, sizeof(*p_pusValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_pusValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadUInt32Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, DWORD* p_pulValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_pulValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pulValue, sizeof(*p_pulValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_pulValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadUInt64Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, __uint64* p_pullValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead;

    if(p_pullValue)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pullValue, sizeof(*p_pullValue), &ulNbOfBytesRead, p_pErrorInfo);
        if(oResult && (sizeof(*p_pullValue) != ulNbOfBytesRead))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadStringObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, CStdString* p_pValue, CErrorInfo* p_pErrorInfo)
{
    const DWORD EXPEDITED_DATA_BUFFER_SIZE = 4;
    const DWORD NON_EXPEDITED_DATA_BUFFER_SIZE = 256;

    char expeditedDataBuffer[EXPEDITED_DATA_BUFFER_SIZE + 1];
    char nonExpeditedDataBuffer[NON_EXPEDITED_DATA_BUFFER_SIZE + 1];
    DWORD ulNbOfBytesRead = 0;
    BOOL oResult(FALSE);

    if(p_pValue)
    {
        memset(expeditedDataBuffer, 0, EXPEDITED_DATA_BUFFER_SIZE + 1);
        memset(nonExpeditedDataBuffer, 0, NON_EXPEDITED_DATA_BUFFER_SIZE + 1);

        if((ReadDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (void*)nonExpeditedDataBuffer, NON_EXPEDITED_DATA_BUFFER_SIZE, &ulNbOfBytesRead, p_pErrorInfo) && (NON_EXPEDITED_DATA_BUFFER_SIZE >= ulNbOfBytesRead)))
        {
            *p_pValue = nonExpeditedDataBuffer;
            oResult = TRUE;
        }
        else if((ReadDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (void*)expeditedDataBuffer, EXPEDITED_DATA_BUFFER_SIZE, &ulNbOfBytesRead, p_pErrorInfo) && (EXPEDITED_DATA_BUFFER_SIZE >= ulNbOfBytesRead)))
        {
            *p_pValue = expeditedDataBuffer;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadDataBufferObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, void* p_pData, DWORD p_ulNumberOfBytesToRead, DWORD* p_pulNbOfBytesRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesRead = 0;

    if(p_pData)
    {
        oResult = ReadObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pData, p_ulNumberOfBytesToRead, &ulNbOfBytesRead, p_pErrorInfo);

        if(p_pulNbOfBytesRead) *p_pulNbOfBytesRead = ulNbOfBytesRead;
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteUInt8Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE p_ubValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesWritten = 0;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, &p_ubValue, sizeof(p_ubValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_ubValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteUInt16Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, WORD p_usValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesWritten = 0;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_usValue, sizeof(p_usValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_usValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteUInt32Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, DWORD p_ulValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesWritten = 0;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_ulValue, sizeof(p_ulValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_ulValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteUInt64Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, __uint64 p_ullValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesWritten;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_ullValue, sizeof(p_ullValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_ullValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteInt8Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, __int8 p_bValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    DWORD ulNbOfBytesWritten;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_bValue, sizeof(p_bValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_bValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteInt16Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, short p_sValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    DWORD ulNbOfBytesWritten;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_sValue, sizeof(p_sValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_sValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteInt32Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, int p_lValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    DWORD ulNbOfBytesWritten;

    oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)&p_lValue, sizeof(p_lValue), &ulNbOfBytesWritten, p_pErrorInfo);
    if(oResult && (sizeof(p_lValue) != ulNbOfBytesWritten))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED, p_pErrorInfo);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteDataBufferObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, void* p_pData, DWORD p_ulNbOfBytesToWrite, DWORD* p_pulNbOfBytesWritten, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulNbOfBytesWritten;

    if(p_pData)
    {
        oResult = WriteObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, (BYTE*)p_pData, p_ulNbOfBytesToWrite, &ulNbOfBytesWritten, p_pErrorInfo);

        if(p_pulNbOfBytesWritten) *p_pulNbOfBytesWritten = ulNbOfBytesWritten;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::UpdateBitsInUInt32Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle,
    HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex,
    DWORD p_ulValue, DWORD p_ulMask, CErrorInfo* p_pErrorInfo)
{
    // Read parameter value
    DWORD ulParameterValue;
    BOOL oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, &ulParameterValue, p_pErrorInfo);

    ulParameterValue = (ulParameterValue & (p_ulMask ^ 0xFFFFFFFF)) | (p_ulValue & p_ulMask);

    // Write parameter value back
    if (oResult)
    {
        oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, ulParameterValue, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::UpdateBitsInUInt16Object(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle,
    HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex,
    WORD p_ulValue, WORD p_ulMask, CErrorInfo* p_pErrorInfo)
{
    // Read parameter value
    WORD ulParameterValue;
    BOOL oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, &ulParameterValue, p_pErrorInfo);

    ulParameterValue = (ulParameterValue & (p_ulMask ^ 0xFFFF)) | (p_ulValue & p_ulMask);

    // Write parameter value back
    if (oResult)
    {
        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, p_usSubIndex, ulParameterValue, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::SetDisableState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    WORD usDeviceState = 0;
    vector<EDeviceState> disabledStates(2);

    oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, p_pErrorInfo);
    if(oResult)
    {
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;
        switch(usDeviceState)
        {
            case DS_FAULT:
            case DS_SWITCH_ON_DISABLED: oResult = TRUE; break;
            case DS_OPERATION_ENABLED:
            case DS_QUICK_STOP_ACTIV:
            case DS_SWITCHED_ON:
            case DS_READY_TO_SWITCH_ON:
                if(oResult)
                    oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, DISABLE_VOLTAGE, p_pErrorInfo);
                if(oResult)
                {
                    disabledStates[0] = DS_SWITCH_ON_DISABLED;
                    disabledStates[1] = DS_FAULT;
                    oResult = CheckDeviceStates(p_pManager, p_hDCS_Handle, p_hTransactionHandle, disabledStates, p_pErrorInfo);
                }
                break;
            default:
                oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadStatusWord(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusStatusWord, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    WORD usStatusWord = 0;

    usIndex = INDEX_STATUS_WORD;
    usSubIndex = SUBINDEX_STATUS_WORD;

    if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usStatusWord, p_pErrorInfo))
    {
        p_rusStatusWord = usStatusWord;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadControlWord(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusControlWord, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    WORD usControlWord = 0;

    usIndex = INDEX_CONTROL_WORD;
    usSubIndex = SUBINDEX_CONTROL_WORD;

    if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usControlWord, p_pErrorInfo))
    {
        p_rusControlWord = usControlWord;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadDeviceState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD* p_pusDeviceState, CErrorInfo* p_pErrorInfo)
{
    const WORD STATE_BIT5_MASK = 0x0020;
    const WORD STATE_5BIT_MASK = 0x004F;
    const WORD STATE_6BIT_MASK = 0x006F;

    BOOL oResult(FALSE);
    WORD usStatusWord = 0;

    if(ReadStatusWord(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usStatusWord, p_pErrorInfo))
    {
        oResult = TRUE;

        switch(usStatusWord & STATE_5BIT_MASK)
        {
            case DS_NOT_READY_TO_SWITCH_ON:
            case DS_SWITCH_ON_DISABLED:
            case DS_FAULT_REACTION_ACTIV:
            case DS_FAULT:
                {
                    usStatusWord &= ~STATE_BIT5_MASK;
                }; break;
        }

        usStatusWord &= STATE_6BIT_MASK;
        if(p_pusDeviceState) *p_pusDeviceState = usStatusWord;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CheckDeviceStates(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, vector<EDeviceState> p_usStates, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulTimeoutDeviceState(0);

    if (GetDeviceTransitionsTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usStates, ulTimeoutDeviceState, p_pErrorInfo))
    {
        oResult = TRUE;
        DWORD ulStart = MmcGetTickCount();
        WORD usDeviceState = 0;
        DWORD ulTime = 0;

        while((ulTime < ulTimeoutDeviceState) && (!IsStateInStates((EDeviceState)usDeviceState, p_usStates)) && (oResult))
        {
            if(!ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, p_pErrorInfo))
            {
                oResult = FALSE;
                break;
            }

            ulTime = MmcGetTickCount() - ulStart;

            if(ulTime > ulTimeoutDeviceState)
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_TIMEOUT_ELAPSED, p_pErrorInfo);
                }

                oResult = FALSE;
                break;
            }

            MmcSleep(10);
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsStateInStates(EDeviceState p_usState, vector<EDeviceState> p_usStates)
{
    BOOL oResult(FALSE);

    for(int i = 0; i < (int)p_usStates.size(); ++i)
    {
        if (p_usStates[i] == p_usState)
        {
            oResult = TRUE;
            break;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CheckDeviceState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usState, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    vector<EDeviceState> deviceStates(1);
    deviceStates[0] = (EDeviceState)p_usState;

    oResult = CheckDeviceStates(p_pManager, p_hDCS_Handle, p_hTransactionHandle, deviceStates, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDeviceTransitionsTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, vector<EDeviceState> p_targetStates, DWORD& p_rulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    DWORD ulSingleTransitionTimeout(0);

    p_rulTimeout = 0;

    for(int i = 0; i < (int)p_targetStates.size(); ++i)
    {
        if(oResult)
        {
            oResult = GetDeviceTransitionTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_targetStates[0], ulSingleTransitionTimeout, p_pErrorInfo);
            if (p_rulTimeout < ulSingleTransitionTimeout)
            {
                p_rulTimeout = ulSingleTransitionTimeout;
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDeviceTransitionTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, EDeviceState p_targetState, DWORD& p_rulTimeout, CErrorInfo* p_pErrorInfo)
{
    const DWORD MINIMUM_DEVICE_STATE_TIMEOUT = 5000;

    BOOL oResult(FALSE);
    EDeviceState actualState(DS_NOT_READY_TO_SWITCH_ON);
    WORD usTransistion(0);
    DWORD ulTimeoutDeviceState(0);

    if(m_pCia402StateMachine != 0)
    {
        if(ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, (WORD *)&actualState, p_pErrorInfo))
        {
            m_pCia402StateMachine->FindTransition(actualState, p_targetState, &usTransistion);

            if(GetDeviceStateTimeoutValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usTransistion, ulTimeoutDeviceState, p_pErrorInfo))
            {
                //Limit timeout to minimum
                if (ulTimeoutDeviceState < MINIMUM_DEVICE_STATE_TIMEOUT)
                {
                    ulTimeoutDeviceState = MINIMUM_DEVICE_STATE_TIMEOUT;
                }
            }
            else
            {
                //Use minimal timeout if no timeout value available
                ulTimeoutDeviceState = MINIMUM_DEVICE_STATE_TIMEOUT;
            }

            p_rulTimeout = ulTimeoutDeviceState;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDeviceStateTimeoutValue(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTransistion, DWORD& p_rulTimeoutDeviceState, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    DWORD ulDecelerationTime(0);
    WORD usStandstillWindowTimeout(0);
    WORD usHoldingBrakeRiseTime(0);
    WORD usHoldingBrakeFallTime(0);

    //Init
    oResult = TRUE;
    p_rulTimeoutDeviceState = 0;

    //Decleration
    if(GetDecelerationTime(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usTransistion, ulDecelerationTime, p_pErrorInfo))
    {
        p_rulTimeoutDeviceState += ulDecelerationTime;
    }

    //Standstill
    if(GetStandstillTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usTransistion, usStandstillWindowTimeout, p_pErrorInfo))
    {
        p_rulTimeoutDeviceState += usStandstillWindowTimeout;
    }

    //Holding brake
    if (GetBrakeRiseTime(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usTransistion, usHoldingBrakeRiseTime, p_pErrorInfo))
    {
        p_rulTimeoutDeviceState += usHoldingBrakeRiseTime;
    }
    else if(GetBrakeFallTime(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usTransistion, usHoldingBrakeFallTime, p_pErrorInfo))
    {
        p_rulTimeoutDeviceState += usHoldingBrakeFallTime;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDecelerationTime(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTransition, DWORD& p_rulDecelerationTime, CErrorInfo* p_pErrorInfo)
{
    const DWORD MILLISECONDS_PER_SECOND = 1000;

    BOOL oResult(FALSE);
    WORD usOptionCodeIndex(0);
    WORD usOptionCodeSubIndex(0);
    WORD usOptionCode(0);
    WORD usDecelerationIndex(0);
    WORD usDecelerationSubIndex(0);
    DWORD ulDeceleration(0);
    DWORD ulActualSpeed(0);

    if (m_pCia402StateMachine)
    {
        // Init
        p_rulDecelerationTime = 0;

        if(m_pCia402StateMachine->IsDecelerationRelevantForTransition(p_usTransition))
        {
            // Option code
            m_pCia402StateMachine->FindRelevantOptionCode(p_usTransition, &usOptionCodeIndex, &usOptionCodeSubIndex);
            if (usOptionCodeIndex != 0)
            {
                if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usOptionCodeIndex, usOptionCodeSubIndex, &usOptionCode, p_pErrorInfo))
                {
                    m_pCia402StateMachine->FindRelevantDecelerationProfile(usOptionCode, &usDecelerationIndex, &usDecelerationSubIndex);
                    if (usDecelerationIndex != 0)
                    {
                        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usDecelerationIndex, usDecelerationSubIndex, &ulDeceleration, p_pErrorInfo);

                        if (oResult)
                        {
                            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_VELOCITY_ACTUAL_AVERAGE_VALUE, EPOS4_SUBINDEX_VELOCITY_ACTUAL_AVERAGE_VALUE, &ulActualSpeed, p_pErrorInfo);
                        }

                        if (oResult)
                        {
                            p_rulDecelerationTime = ulActualSpeed * MILLISECONDS_PER_SECOND / ulDeceleration;
                        }
                    }
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetStandstillTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTransition, WORD& p_rusStandstillWindowTimeout, CErrorInfo* p_pErrorInfo)
{
    const WORD EPOS4_STANDSTILL_WINDOW_TIMEOUT_DISABLE_VALUE = 0xFFFF;

    BOOL oResult(FALSE);
    WORD usStandstillWindowTimeout(0);

    p_rusStandstillWindowTimeout = 0;

    if(m_pCia402StateMachine->IsStandstillWindowRelevantForTransition(p_usTransition))
    {
        if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_STANDSTILL_WINDOW_CONFIGURATION, EPOS4_SUBINDEX_STANDSTILL_WINDOW_TIMEOUT, &usStandstillWindowTimeout, p_pErrorInfo))
        {
            if (usStandstillWindowTimeout != EPOS4_STANDSTILL_WINDOW_TIMEOUT_DISABLE_VALUE)
            {
                p_rusStandstillWindowTimeout = usStandstillWindowTimeout;
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetBrakeRiseTime(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTransition, WORD& p_rusHoldingBrakeRiseTime, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usHoldingBrakeRiseTime(0);

    p_rusHoldingBrakeRiseTime = 0;

    if(m_pCia402StateMachine->IsBrakeRiseTimeRelevantForTransition(p_usTransition))
    {
        if(IsHoldingBrakeConfigured(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo))
        {
            if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_HOLDING_BRAKE, EPOS4_SUBINDEX_HOLDING_BRAKE_RISE_TIME, &usHoldingBrakeRiseTime, p_pErrorInfo))
            {
                p_rusHoldingBrakeRiseTime = usHoldingBrakeRiseTime;
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetBrakeFallTime(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTransition, WORD& p_rusHoldingBrakeFallTime, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usHoldingBrakeFallTime(0);

    p_rusHoldingBrakeFallTime = 0;

    if(m_pCia402StateMachine->IsBrakeFallTimeRelevantForTransition(p_usTransition))
    {
        if(IsHoldingBrakeConfigured(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo))
        {
            if(ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_HOLDING_BRAKE, EPOS4_SUBINDEX_HOLDING_BRAKE_FALL_TIME, &usHoldingBrakeFallTime, p_pErrorInfo))
            {
                p_rusHoldingBrakeFallTime = usHoldingBrakeFallTime;
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsHoldingBrakeConfigured(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubNumberOfDigitalOutputs(0);
    BYTE ubDigitalOutputConfiguration(0);

    if(ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIGITAL_OUTPUT_CONFIGURATION, SUBINDEX_DIGITAL_OUTPUT_CONFIGURATION, &ubNumberOfDigitalOutputs, p_pErrorInfo))
    {
        for (BYTE ubDigitalOutputNumber = 1; ubDigitalOutputNumber <= ubNumberOfDigitalOutputs; ubDigitalOutputNumber++)
        {
            if(ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIGITAL_OUTPUT_CONFIGURATION, ubDigitalOutputNumber, &ubDigitalOutputConfiguration, p_pErrorInfo))
            {
                if (ubDigitalOutputConfiguration == EPOS4_DOC_HOLDING_BRAKE)
                {
                    oResult = TRUE;
                    break;
                }
            }
        }
    }

    return oResult;
}

CGatewayVCStoEpos4& CGatewayVCStoEpos4::operator=(CGatewayVCStoEpos4& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayVCStoDCS*)this) = *((CGatewayVCStoDCS*)&p_rOther);
    }

    return *this;
}

void CGatewayVCStoEpos4::InitCommands_ObjectDictionary()
{
    DeleteCommands_ObjectDictionary();

    m_pCommand_WriteObject = new CCommand_DCS_Epos4();
    m_pCommand_WriteObject->InitCommand(EPOS4_WRITE_OBJECT);

    m_pCommand_ReadObject = new CCommand_DCS_Epos4();
    m_pCommand_ReadObject->InitCommand(EPOS4_READ_OBJECT);

    m_pCommand_InitiateSegmentedWrite = new CCommand_DCS_Epos4();
    m_pCommand_InitiateSegmentedWrite->InitCommand(EPOS4_INITIATE_SEGMENT_WRITE);

    m_pCommand_InitiateSegmentedRead = new CCommand_DCS_Epos4();
    m_pCommand_InitiateSegmentedRead->InitCommand(EPOS4_INITIATE_SEGMENT_READ);

    m_pCommand_SegmentedWrite = new CCommand_DCS_Epos4();
    m_pCommand_SegmentedWrite->InitCommand(EPOS4_SEGMENT_WRITE);

    m_pCommand_SegmentedRead = new CCommand_DCS_Epos4();
    m_pCommand_SegmentedRead->InitCommand(EPOS4_SEGMENT_READ);

    m_pCommand_AbortTransfer = new CCommand_DCS_Epos4();
    m_pCommand_AbortTransfer->InitCommand(EPOS4_ABORT_TRANSFER);
}

void CGatewayVCStoEpos4::DeleteCommands_ObjectDictionary()
{
    if(m_pCommand_WriteObject)
    {
        delete m_pCommand_WriteObject;
        m_pCommand_WriteObject = 0;
    }
    if(m_pCommand_InitiateSegmentedWrite)
    {
        delete m_pCommand_InitiateSegmentedWrite;
        m_pCommand_InitiateSegmentedWrite = 0;
    }
    if(m_pCommand_SegmentedWrite)
    {
        delete m_pCommand_SegmentedWrite;
        m_pCommand_SegmentedWrite = 0;
    }
    if(m_pCommand_ReadObject)
    {
        delete m_pCommand_ReadObject;
        m_pCommand_ReadObject = 0;
    }
    if(m_pCommand_InitiateSegmentedRead)
    {
        delete m_pCommand_InitiateSegmentedRead;
        m_pCommand_InitiateSegmentedRead = 0;
    }
    if(m_pCommand_SegmentedRead)
    {
        delete m_pCommand_SegmentedRead;
        m_pCommand_SegmentedRead = 0;
    }
    if(m_pCommand_AbortTransfer)
    {
        delete m_pCommand_AbortTransfer;
        m_pCommand_AbortTransfer = 0;
    }
}

void CGatewayVCStoEpos4::InitCommands_NetworkManagement()
{
    DeleteCommands_NetworkManagement();

    m_pCommand_SendNMTService = new CCommand_DCS_Epos4();
    m_pCommand_SendNMTService->InitCommand(EPOS4_SEND_NMT_SERVICE);
}

void CGatewayVCStoEpos4::DeleteCommands_NetworkManagement()
{
    if(m_pCommand_SendNMTService)
    {
        delete m_pCommand_SendNMTService;
        m_pCommand_SendNMTService = 0;
    }
}

void CGatewayVCStoEpos4::InitCommands_GeneralGateway()
{
    DeleteCommands_GeneralGateway();

    m_pCommand_RequestCANFrame = new CCommand_DCS_Epos4();
    m_pCommand_RequestCANFrame->InitCommand(EPOS4_REQUEST_CAN_FRAME);

    m_pCommand_SendCANFrame = new CCommand_DCS_Epos4();
    m_pCommand_SendCANFrame->InitCommand(EPOS4_SEND_CAN_FRAME);

    m_pCommand_ReadCANFrame = new CCommand_DCS_Epos4();
    m_pCommand_ReadCANFrame->InitCommand(EPOS4_READ_CAN_FRAME);
}

void CGatewayVCStoEpos4::InitCommands_LayerSettingServices()
{
    DeleteCommands_LayerSettingServices();

    //Init Command Send
    m_pCommand_SendLSSFrame = new CCommand_DCS_Epos4();
    m_pCommand_SendLSSFrame->InitCommand(EPOS4_SEND_LSS_FRAME);
    //Init Command Read
    m_pCommand_ReadLSSFrame = new CCommand_DCS_Epos4();
    m_pCommand_ReadLSSFrame->InitCommand(EPOS4_READ_LSS_FRAME);
}

void CGatewayVCStoEpos4::InitCommands_Register()
{
    DeleteCommands_Register();

    //Init command read ESC register
    m_pCommand_ReadEscRegister = new CCommand_DCS_Epos4();
    m_pCommand_ReadEscRegister->InitCommand(EPOS4_READ_ESC_MEMORY);

    //Init command write ESC register
    m_pCommand_WriteEscRegister = new CCommand_DCS_Epos4();
    m_pCommand_WriteEscRegister->InitCommand(EPOS4_WRITE_ESC_MEMORY);
}

void CGatewayVCStoEpos4::InitCommands_EtherCatStateMachine()
{
    DeleteCommands_EtherCatStateMachine();

    //Init command init ESM
    m_pCommand_InitEsm = new CCommand_DCS_Epos4();
    m_pCommand_InitEsm->InitCommand(EPOS4_INIT_ESM);

    //Init command change ESM state
    m_pCommand_ChangeEsmState = new CCommand_DCS_Epos4();
    m_pCommand_ChangeEsmState->InitCommand(EPOS4_CHANGE_ESM_STATE);

    //Init command get ESM state
    m_pCommand_GetEsmState = new CCommand_DCS_Epos4();
    m_pCommand_GetEsmState->InitCommand(EPOS4_GET_ESM_STATE);
}

void CGatewayVCStoEpos4::DeleteCommands_GeneralGateway()
{
    if(m_pCommand_RequestCANFrame)
    {
        delete m_pCommand_RequestCANFrame;
        m_pCommand_RequestCANFrame = 0;
    }
    if(m_pCommand_SendCANFrame)
    {
        delete m_pCommand_SendCANFrame;
        m_pCommand_SendCANFrame = 0;
    }
    if(m_pCommand_ReadCANFrame)
    {
        delete m_pCommand_ReadCANFrame;
        m_pCommand_ReadCANFrame = 0;
    }
}

void CGatewayVCStoEpos4::DeleteCommands_LayerSettingServices()
{
    if(m_pCommand_SendLSSFrame)
    {
        delete m_pCommand_SendLSSFrame;
        m_pCommand_SendLSSFrame = 0;
    }
    if(m_pCommand_ReadLSSFrame)
    {
        delete m_pCommand_ReadLSSFrame;
        m_pCommand_ReadLSSFrame = 0;
    }
}

void CGatewayVCStoEpos4::DeleteCommands_Register()
{
    if(m_pCommand_WriteEscRegister)
    {
        delete m_pCommand_WriteEscRegister;
        m_pCommand_WriteEscRegister = 0;
    }

    if(m_pCommand_ReadEscRegister)
    {
        delete m_pCommand_ReadEscRegister;
        m_pCommand_ReadEscRegister = 0;
    }
}

void CGatewayVCStoEpos4::DeleteCommands_EtherCatStateMachine()
{
    if(m_pCommand_InitEsm)
    {
        delete m_pCommand_InitEsm;
        m_pCommand_InitEsm = 0;
    }

    if(m_pCommand_ChangeEsmState)
    {
        delete m_pCommand_ChangeEsmState;
        m_pCommand_ChangeEsmState = 0;
    }

    if(m_pCommand_GetEsmState)
    {
        delete m_pCommand_GetEsmState;
        m_pCommand_GetEsmState = 0;
    }
}

BOOL CGatewayVCStoEpos4::DCS_SendCANFrame(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usCobID, WORD p_usLength, void* p_pData, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;

    if (m_pCommand_SendCANFrame && p_pManager)
    {
        //Set parameter data
        m_pCommand_SendCANFrame->ResetStatus();
        m_pCommand_SendCANFrame->SetParameterData(0, &p_usCobID, sizeof(p_usCobID));
        m_pCommand_SendCANFrame->SetParameterData(1, &p_usLength, sizeof(p_usLength));
        m_pCommand_SendCANFrame->SetParameterData(2, p_pData, p_usLength);

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_SendCANFrame, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SendCANFrame->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_SendCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadCANFrame(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usCobID, WORD p_usLength, void* p_pData, DWORD p_ulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_ReadCANFrame && p_pManager)
    {
        //Set parameter data
        m_pCommand_ReadCANFrame->ResetStatus();
        m_pCommand_ReadCANFrame->SetParameterData(0, &p_usCobID, sizeof(p_usCobID));
        m_pCommand_ReadCANFrame->SetParameterData(1, &p_usLength, sizeof(p_usLength));
        m_pCommand_ReadCANFrame->SetParameterData(2, &p_ulTimeout, sizeof(p_ulTimeout));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_ReadCANFrame, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadCANFrame->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_ReadCANFrame->GetReturnParameterData(1, p_pData, p_usLength);

        //Get error code
        m_pCommand_ReadCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_RequestCANFrame(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usCobID, WORD p_usLength, void* p_pData, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_RequestCANFrame && p_pManager)
    {
        //Set parameter data
        m_pCommand_RequestCANFrame->ResetStatus();
        m_pCommand_RequestCANFrame->SetParameterData(0, &p_usCobID, sizeof(p_usCobID));
        m_pCommand_RequestCANFrame->SetParameterData(1, &p_usLength, sizeof(p_usLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_RequestCANFrame, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_RequestCANFrame->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_RequestCANFrame->GetReturnParameterData(1, p_pData, p_usLength);

        //Get error code
        m_pCommand_RequestCANFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_SendNMTService(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usNodeIdentifier, WORD p_usCommandSpecifier, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pCommand_SendNMTService && p_pManager)
    {
        //Set parameter data
        m_pCommand_SendNMTService->ResetStatus();
        m_pCommand_SendNMTService->SetParameterData(0, &p_usNodeIdentifier, sizeof(p_usNodeIdentifier));
        m_pCommand_SendNMTService->SetParameterData(1, &p_usCommandSpecifier, sizeof(p_usCommandSpecifier));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_SendNMTService, p_hDCS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_SendNMTService->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_SendLSSFrame(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_SendLSSFrame && p_pManager)
    {
        //Set parameter data
        m_pCommand_SendLSSFrame->ResetStatus();
        m_pCommand_SendLSSFrame->SetParameterData(0, p_pDataBuffer, p_ulDataBufferLength);

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_SendLSSFrame, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SendLSSFrame->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_SendLSSFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadLSSFrame(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTimeout, void* p_pDataBuffer, DWORD p_ulDataBufferLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_ReadLSSFrame && p_pManager)
    {
        //Set parameter data
        m_pCommand_ReadLSSFrame->ResetStatus();
        m_pCommand_ReadLSSFrame->SetParameterData(0, &p_usTimeout, sizeof(p_usTimeout));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_ReadLSSFrame, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadLSSFrame->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_ReadLSSFrame->GetReturnParameterData(1, p_pDataBuffer, p_ulDataBufferLength);

        //Get error code
        m_pCommand_ReadLSSFrame->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_WriteEscRegister(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usAdo, WORD p_usLength, void* p_pData, WORD* p_pusLengthWritten, CErrorInfo* p_pErrorInfo)
{
    DWORD ulDeviceErrorCode = 0;
    BOOL oResult = FALSE;
    CMmcSingleLock lock(m_SyncObject.get(), true);

    if(m_pCommand_WriteEscRegister && p_pManager)
    {
        //Set parameter data
        m_pCommand_WriteEscRegister->ResetStatus();
        m_pCommand_WriteEscRegister->SetParameterData(0, &p_usAdo, sizeof(p_usAdo));
        m_pCommand_WriteEscRegister->SetParameterData(1, &p_usLength, sizeof(p_usLength));
        m_pCommand_WriteEscRegister->SetParameterData(2, p_pData, p_usLength);

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_WriteEscRegister, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_WriteEscRegister->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_WriteEscRegister->GetReturnParameterData(1, p_pusLengthWritten, sizeof(*p_pusLengthWritten));

        //Get error code
        m_pCommand_WriteEscRegister->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadEscRegister(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usAdo, WORD p_usLength, void* p_pData, WORD* p_pusLengthRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;
    DWORD ulDeviceErrorCode = 0;
    CMmcSingleLock lock(m_SyncObject.get(), true);

    if(m_pCommand_ReadEscRegister && p_pManager)
    {
        //Set parameter data
        m_pCommand_ReadEscRegister->ResetStatus();
        m_pCommand_ReadEscRegister->SetParameterData(0, &p_usAdo, sizeof(p_usAdo));
        m_pCommand_ReadEscRegister->SetParameterData(1, &p_usLength, sizeof(p_usLength));

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_ReadEscRegister, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadEscRegister->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_ReadEscRegister->GetReturnParameterData(1, p_pusLengthRead, sizeof(*p_pusLengthRead));
        m_pCommand_ReadEscRegister->GetReturnParameterData(2, p_pData, p_usLength);

        //Get error code
        m_pCommand_ReadEscRegister->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadSiiParameter(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, ESiiParameterAddress p_eParameterAddress, void* p_pData, DWORD p_ulNbOfBytesToRead, CErrorInfo* p_pErrorInfo)
{
    const DWORD TIMEOUT = 100;

    BOOL oResult(TRUE);
    SSiiControlStatusAddress controlAddress;
    SSiiControlStatus control;
    SSiiControlStatus status;
    DWORD ulAddressOffset(0);
    DWORD ulSiiData(0);
    WORD usTotalLengthToRead(0);
    WORD usLengthToRead(0);
    BOOL oReadOperationFinished;
    DWORD ulTime(0);
    DWORD ulStartTime(0);

    //Init
    memset(&control, 0, sizeof(control));
    memset(&status, 0, sizeof(status));
    memset(&controlAddress, 0, sizeof(controlAddress));
    control.ReadOperation = 0;
    controlAddress.sControlStatus.ReadOperation = 1;

    usTotalLengthToRead = (WORD)p_ulNbOfBytesToRead;

    while(oResult && (usTotalLengthToRead > 0))
    {
        //Reset SII control (ReadOperation = 0)
        if(WriteEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRSSS_ESI_EEPROM_CONTROL, sizeof(control), (char*)&control, p_pErrorInfo))
        {
            //Update address & size
            controlAddress.ulAddress = (DWORD)p_eParameterAddress + (ulAddressOffset / sizeof(WORD));
            WORD usSiiDataSize = sizeof(ulSiiData);
            if(usTotalLengthToRead > usSiiDataSize) usLengthToRead = usSiiDataSize; else usLengthToRead = usTotalLengthToRead;

            //Write SII Control and address (ReadOperation = 1)
            if(WriteEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRSSS_ESI_EEPROM_CONTROL, sizeof(controlAddress), (char*)&controlAddress, p_pErrorInfo))
            {
                //Wait until ReadOperation finished
                ulStartTime = MmcGetTickCount();
                ulTime = 0;
                oReadOperationFinished = FALSE;
                while(oResult && !oReadOperationFinished)
                {
                    if(ReadEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRSSS_ESI_EEPROM_CONTROL, sizeof(status), (char*)&status, p_pErrorInfo))
                    {
                        //Evaluate status
                        if(status.CommandError || status.DeviceInfoError || status.CheckSErrDLu)
                        {
                            //Error
                            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                            oResult = FALSE;
                        }
                        else if(status.Busy || (status.ReadOperation == 1))
                        {
                            //Data not available yet -> Check timeout
                            ulTime = MmcGetTickCount() - ulStartTime;
                            if(ulTime > TIMEOUT)
                            {
                                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                                oResult = FALSE;
                            }
                        }
                        else
                        {
                            //Data available
                            oReadOperationFinished = TRUE;
                        }
                    }
                    else
                    {
                        //Failed to read status
                        oResult = FALSE;
                    }
                }

                //Read SII data
                if(oResult)
                {
                    if(ReadEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRSSS_ESI_EEPROM_DATA, usLengthToRead, ((char*)p_pData + ulAddressOffset), p_pErrorInfo))
                    {
                        //Update address and size
                        usTotalLengthToRead -= usLengthToRead;
                        ulAddressOffset += usLengthToRead;
                        oResult = TRUE;
                    }
                    else
                    {
                        //Failed to read data
                        oResult = FALSE;
                    }
                }
            }
            else
            {
                //Failed to write control register
                oResult = FALSE;
            }
        }
        else
        {
            //Failed to reset read operation
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadSiiCategorySize(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, ESiiCategory p_eCategory, DWORD* p_pulCategorySize, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BOOL oReadNextCategory(TRUE);
    DWORD ulSiiSize(0);
    int lAddressOffset(0);
    SSiiCatagoryHeader categoryHeader;

    if(p_pulCategorySize)
    {
        //Read SII size
        if(ReadSiiSize(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulSiiSize, p_pErrorInfo))
        {
            while(oReadNextCategory)
            {
                //Check SII size
                if((((DWORD)(SII_CATEGORIES + lAddressOffset)*  sizeof(WORD)) + sizeof(categoryHeader)) <= ulSiiSize)
                {
                    //Read category header
                    if(DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ESiiParameterAddress((int)SII_CATEGORIES + lAddressOffset), &categoryHeader, sizeof(categoryHeader), p_pErrorInfo))
                    {
                        //Evaluate header
                        if(p_eCategory == categoryHeader.usType)
                        {
                            *p_pulCategorySize = (categoryHeader.usDataSize * sizeof(WORD));
                            oReadNextCategory = FALSE;
                            oResult = TRUE;
                        }
                        else if(SII_CAT_END == categoryHeader.usType)
                        {
                            //Last Category
                            if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_CATEGORY);
                            oReadNextCategory = FALSE;
                            oResult = FALSE;
                        }
                        else
                        {
                            //Calculate address Offset for next category
                            lAddressOffset += (sizeof(categoryHeader)/sizeof(WORD) + categoryHeader.usDataSize);
                            oReadNextCategory = TRUE;
                        }
                    }
                    else
                    {
                        //Communication Error
                        oReadNextCategory = FALSE;
                        oResult = FALSE;
                    }
                }
                else
                {
                    //SII size Error -> Category not found
                    if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_CATEGORY);
                    oReadNextCategory = FALSE;
                    oResult = FALSE;
                }
            }
        }
    }
    else
    {
        if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_INTERNAL);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadSiiCategoryData(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, ESiiCategory p_eCategory, void* p_pData, DWORD p_ulNbOfBytesToRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    int lAddressOffset(0);
    SSiiCatagoryHeader categoryHeader;
    DWORD ulSizeToRead(0);
    BOOL oReadNextCategory(TRUE);
    DWORD ulSiiSize(0);

    if(p_pData && (p_ulNbOfBytesToRead > 0))
    {
        //Read SII size
        if(ReadSiiSize(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulSiiSize, p_pErrorInfo))
        {
            while(oReadNextCategory)
            {
                //Check SII size
                if((((DWORD)(SII_CATEGORIES + lAddressOffset)*  sizeof(WORD)) + sizeof(categoryHeader)) <= ulSiiSize)
                {
                    //Read Category Header
                    if(DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ESiiParameterAddress((int)SII_CATEGORIES + lAddressOffset), &categoryHeader, sizeof(categoryHeader), p_pErrorInfo))
                    {
                        //Evaluate header
                        if(p_eCategory == categoryHeader.usType)
                        {
                            //Check size
                            ulSizeToRead = p_ulNbOfBytesToRead;
                            if(ulSizeToRead > categoryHeader.usDataSize*  sizeof(WORD))
                            {
                                ulSizeToRead = categoryHeader.usDataSize*  sizeof(WORD);
                            }

                            //Check SII size
                            if((((DWORD)(SII_CATEGORIES + lAddressOffset)*  sizeof(WORD)) + sizeof(categoryHeader) + ulSizeToRead) <= ulSiiSize)
                            {
                                //Read category data
                                oResult = DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ESiiParameterAddress((int)SII_CATEGORIES + lAddressOffset + sizeof(categoryHeader)/sizeof(WORD)), p_pData, ulSizeToRead, p_pErrorInfo);
                                oReadNextCategory = FALSE;
                            }
                            else
                            {
                                //SII size error -> Category not found
                                if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_CATEGORY);
                                oReadNextCategory = FALSE;
                                oResult = FALSE;
                            }
                        }
                        else if(SII_CAT_END == categoryHeader.usType)
                        {
                            //Last Category
                            if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_CATEGORY);
                            oReadNextCategory = FALSE;
                            oResult = FALSE;
                        }
                        else
                        {
                            //Calculate address offset for next category
                            lAddressOffset += (sizeof(categoryHeader)/sizeof(WORD) + categoryHeader.usDataSize);
                            oReadNextCategory = TRUE;
                        }
                    }
                    else
                    {
                        //Communication error
                        oReadNextCategory = FALSE;
                        oResult = FALSE;
                    }
                }
                else
                {
                    //SII size Error -> Category not found
                    if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_CATEGORY);
                    oReadNextCategory = FALSE;
                    oResult = FALSE;
                }
            }
        }
    }
    else
    {
        if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_INTERNAL);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_InitEsm(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulVendorId, DWORD p_ulProductCode, DWORD p_ulRevisionNb, __uint64 p_ullSyncManager0, __uint64 p_ullSyncManager1, __uint64 p_ullSyncManager2, __uint64 p_ullSyncManager3, void* p_pFmmu0, void* p_pFmmu1, __uint64 p_ullBootStrap, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_InitEsm && p_pManager)
    {
        //Set parameter data
        m_pCommand_InitEsm->ResetStatus();
        m_pCommand_InitEsm->SetParameterData(0, &p_ulVendorId, sizeof(p_ulVendorId));
        m_pCommand_InitEsm->SetParameterData(1, &p_ulProductCode, sizeof(p_ulProductCode));
        m_pCommand_InitEsm->SetParameterData(2, &p_ulRevisionNb, sizeof(p_ulRevisionNb));
        m_pCommand_InitEsm->SetParameterData(3, &p_ullSyncManager0, sizeof(p_ullSyncManager0));
        m_pCommand_InitEsm->SetParameterData(4, &p_ullSyncManager1, sizeof(p_ullSyncManager1));
        m_pCommand_InitEsm->SetParameterData(5, &p_ullSyncManager2, sizeof(p_ullSyncManager2));
        m_pCommand_InitEsm->SetParameterData(6, &p_ullSyncManager3, sizeof(p_ullSyncManager3));
        m_pCommand_InitEsm->SetParameterData(7, p_pFmmu0, LENGTH_FMMU);
        m_pCommand_InitEsm->SetParameterData(8, p_pFmmu1, LENGTH_FMMU);
        m_pCommand_InitEsm->SetParameterData(9, &p_ullBootStrap, sizeof(p_ullBootStrap));

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_InitEsm, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitEsm->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_InitEsm->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ChangeEsmState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usState, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_ChangeEsmState && p_pManager)
    {
        //Set parameter data
        m_pCommand_ChangeEsmState->ResetStatus();
        m_pCommand_ChangeEsmState->SetParameterData(0, &p_usState, sizeof(p_usState));

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_ChangeEsmState, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ChangeEsmState->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_ChangeEsmState->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_GetEsmState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusState, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult = FALSE;
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_GetEsmState && p_pManager)
    {
        //Set parameter data
        m_pCommand_GetEsmState->ResetStatus();

        //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_GetEsmState, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_GetEsmState->GetReturnParameterData(0, &p_rusState, sizeof(p_rusState));
        m_pCommand_GetEsmState->GetReturnParameterData(1, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_GetEsmState->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_InitiateSegmentedRead(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pubSegmentBuffer, DWORD p_ulSegmentBufferLength, DWORD* p_pulSegmentLengthRead, DWORD* p_pulCompleteLength, CErrorInfo* p_pErrorInfo)
{
    const int INDEX_DATA_INDEX = 0;
    const int INDEX_DATA_SUB_INDEX = 1;
    const int INDEX_DATA_SIZE = 2;

    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode;
    BYTE ubSubIndex = (BYTE)p_usSubIndex;

    if(m_pCommand_InitiateSegmentedRead && p_pManager)
    {
        //Set parameter data
        m_pCommand_InitiateSegmentedRead->ResetStatus();
        m_pCommand_InitiateSegmentedRead->SetParameterData(INDEX_DATA_INDEX, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSegmentedRead->SetParameterData(INDEX_DATA_SUB_INDEX, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_InitiateSegmentedRead->SetParameterData(INDEX_DATA_SIZE, &p_ulSegmentBufferLength, sizeof(p_ulSegmentBufferLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_InitiateSegmentedRead, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSegmentedRead->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_InitiateSegmentedRead->GetReturnParameterData(1, p_pulCompleteLength, sizeof(*p_pulCompleteLength));
        m_pCommand_InitiateSegmentedRead->GetReturnParameterData(2, p_pulSegmentLengthRead, sizeof(*p_pulSegmentLengthRead));
        m_pCommand_InitiateSegmentedRead->GetReturnParameterData(3, p_pubSegmentBuffer, p_ulSegmentBufferLength);

        //Get error code
        m_pCommand_InitiateSegmentedRead->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_InitiateSegmentedWrite(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, DWORD p_ulObjectLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubSubIndex = (BYTE)p_usSubIndex;
    DWORD ulDeviceErrorCode;

    if(m_pCommand_InitiateSegmentedWrite && p_pManager)
    {
        //Set parameter data
        m_pCommand_InitiateSegmentedWrite->ResetStatus();
        m_pCommand_InitiateSegmentedWrite->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_InitiateSegmentedWrite->SetParameterData(1, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_InitiateSegmentedWrite->SetParameterData(2, &p_ulObjectLength, sizeof(p_ulObjectLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_InitiateSegmentedWrite, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_InitiateSegmentedWrite->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_InitiateSegmentedWrite->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_ReadObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pObjectBuffer, DWORD p_ulObjectLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubSubIndex = (BYTE)p_usSubIndex;
    DWORD ulDeviceErrorCode;

    if(m_pCommand_ReadObject && p_pManager)
    {
        //Set parameter data
        m_pCommand_ReadObject->ResetStatus();
        m_pCommand_ReadObject->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_ReadObject->SetParameterData(1, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_ReadObject->SetParameterData(2, &p_ulObjectLength, sizeof(p_ulObjectLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_ReadObject, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_ReadObject->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_ReadObject->GetReturnParameterData(1, p_pObjectBuffer, p_ulObjectLength);

        //Get error code
        m_pCommand_ReadObject->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_SegmentedRead(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL p_oToggle, BOOL* p_poLastSegment, BYTE* p_pSegmentBuffer, DWORD p_ulSegmentBufferLength, DWORD* p_pulSegmentLengthRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;
    DWORD ulDataSize = 0;
    UEpos4ControlByte controlByte;

    if(m_pCommand_SegmentedRead && p_pManager)
    {
        //Prepare control byte
        controlByte.bValue = 0;
        controlByte.structure.bToggle = (BYTE)p_oToggle;

        //Set parameter data
        m_pCommand_SegmentedRead->ResetStatus();
        m_pCommand_SegmentedRead->SetParameterData(0, &controlByte, sizeof(controlByte));
        m_pCommand_SegmentedRead->SetParameterData(1, &p_ulSegmentBufferLength, sizeof(p_ulSegmentBufferLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_SegmentedRead, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SegmentedRead->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_SegmentedRead->GetReturnParameterData(1, &controlByte, sizeof(controlByte));
        m_pCommand_SegmentedRead->GetReturnParameterData(2, &ulDataSize, sizeof(ulDataSize));
        m_pCommand_SegmentedRead->GetReturnParameterData(3, p_pSegmentBuffer, p_ulSegmentBufferLength);

        //Get error code
        m_pCommand_SegmentedRead->GetErrorInfo(p_pErrorInfo);

        //Restore control byte
        if(p_pulSegmentLengthRead) *p_pulSegmentLengthRead = ulDataSize;
        if(p_poLastSegment) *p_poLastSegment = controlByte.structure.bLastSegment;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_SegmentedWrite(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL p_oToggle, BOOL p_oLastSegment, BYTE* p_pubSegmentBuffer, DWORD p_ulSegmentLength, DWORD* p_pulSegmentLengthWritten, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceErrorCode = 0;
    DWORD ulSegmentLengthWritten = 0;
    UEpos4ControlByte controlByte;

    if(m_pCommand_SegmentedWrite && p_pManager)
    {
        //Prepare control byte
        controlByte.bValue = 0;
        controlByte.structure.bToggle = (BYTE)p_oToggle;
        controlByte.structure.bLastSegment = (BYTE)p_oLastSegment;

        //Set parameter data
        m_pCommand_SegmentedWrite->ResetStatus();
        m_pCommand_SegmentedWrite->SetParameterData(0, &controlByte, sizeof(controlByte));
        m_pCommand_SegmentedWrite->SetParameterData(1, p_pubSegmentBuffer, p_ulSegmentLength);

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_SegmentedWrite, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_SegmentedWrite->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
        m_pCommand_SegmentedWrite->GetReturnParameterData(1, &controlByte, sizeof(controlByte));
        m_pCommand_SegmentedWrite->GetReturnParameterData(2, &ulSegmentLengthWritten, sizeof(ulSegmentLengthWritten));

        //Restore control byte
        if(p_pulSegmentLengthWritten) *p_pulSegmentLengthWritten = ulSegmentLengthWritten;

        //Get error code
        m_pCommand_SegmentedWrite->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_WriteObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pObjectBuffer, DWORD p_ulObjectLength, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubSubIndex = (BYTE)p_usSubIndex;
    DWORD ulDeviceErrorCode = 0;

    if(m_pCommand_WriteObject && p_pManager)
    {
        //Set parameter data
        m_pCommand_WriteObject->ResetStatus();
        m_pCommand_WriteObject->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_WriteObject->SetParameterData(1, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_WriteObject->SetParameterData(2, p_pObjectBuffer, p_ulObjectLength);
        m_pCommand_WriteObject->SetParameterData(3, &p_ulObjectLength, sizeof(p_ulObjectLength));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_WriteObject, p_hDCS_Handle, p_hTransactionHandle);

        //Get return parameter data
        m_pCommand_WriteObject->GetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));

        //Get error code
        m_pCommand_WriteObject->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::DCS_AbortTransfer(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, DWORD p_ulAbortCode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubSubIndex = (BYTE)p_usSubIndex;

    if(m_pCommand_AbortTransfer && p_pManager)
    {
        //Set parameter data
        m_pCommand_AbortTransfer->ResetStatus();
        m_pCommand_AbortTransfer->SetParameterData(0, &p_usIndex, sizeof(p_usIndex));
        m_pCommand_AbortTransfer->SetParameterData(1, &ubSubIndex, sizeof(ubSubIndex));
        m_pCommand_AbortTransfer->SetParameterData(2, &p_ulAbortCode, sizeof(p_ulAbortCode));

       //Execute command
        oResult = p_pManager->ExecuteCommand(m_pCommand_AbortTransfer, p_hDCS_Handle, p_hTransactionHandle);

        //Get error code
        m_pCommand_AbortTransfer->GetErrorInfo(p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadSegmentedObject(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, WORD p_usSubIndex, BYTE* p_pubDataBuffer, DWORD p_ulNbOfBytesToRead, DWORD* p_pulNbOfBytesRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BOOL oToggle = FALSE;
    BOOL oLastSegment = FALSE;
    BYTE* pSegmentBuffer = (BYTE*)p_pubDataBuffer;
    CErrorInfo errorInfo;
    BYTE ubSubIndex = (BYTE)p_usSubIndex;
    DWORD ulAbortCode = (DWORD)-1;
    DWORD ulCompleteSize = 0;
    DWORD ulSegmentLength = p_ulNbOfBytesToRead;
    DWORD ulSegmentLengthRead = 0;
    DWORD ulTotalNbOfBytesRead = 0;

    if(DCS_InitiateSegmentedRead(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, ubSubIndex, pSegmentBuffer, ulSegmentLength, &ulSegmentLengthRead, &ulCompleteSize, &errorInfo))
    {
        //Initialize variables
        oResult = TRUE;
        ulTotalNbOfBytesRead += ulSegmentLengthRead;
        ulSegmentLength -= ulSegmentLengthRead;
        pSegmentBuffer += ulSegmentLengthRead;

        while(oResult && !oLastSegment && (ulSegmentLength > 0) && !IsCompleteSizeRead(ulCompleteSize, ulTotalNbOfBytesRead))
        {
            // Abort transfer
            if(m_oAbortCommands || (!oLastSegment && ((ulSegmentLength == 0) || IsCompleteSizeRead(ulCompleteSize, ulTotalNbOfBytesRead))))
            {
                if(DCS_AbortTransfer(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, ubSubIndex, ulAbortCode))
                {
                    if(m_pErrorHandling)
                    {
                        m_pErrorHandling->GetError(ERROR_COMMAND_ABORTED_BY_USER, &errorInfo);
                    }

                    oResult = FALSE;
                    break;
                }
            }

            // Do transfer
            oResult = DCS_SegmentedRead(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oToggle, &oLastSegment, pSegmentBuffer, ulSegmentLength, &ulSegmentLengthRead, &errorInfo);
            if(oResult)
            {
                // Prepare next segment
                if(ulSegmentLengthRead > ulSegmentLength)
                {
                    ulTotalNbOfBytesRead += ulSegmentLength;
                    pSegmentBuffer += ulSegmentLength;
                    ulSegmentLength = 0;
                }
                else
                {
                    ulTotalNbOfBytesRead += ulSegmentLengthRead;
                    pSegmentBuffer += ulSegmentLengthRead;
                    ulSegmentLength -= ulSegmentLengthRead;
                }

                oToggle = !oToggle;
            }
        }

        if(!ulCompleteSize)
        {
            // Abort transfer
            if(!oLastSegment && ((ulSegmentLength > 0) || (oLastSegment && !IsCompleteSizeRead(ulCompleteSize, ulTotalNbOfBytesRead))))
            {
                DCS_AbortTransfer(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usIndex, ubSubIndex, ulAbortCode);
            }
        }

        if(p_pulNbOfBytesRead) *p_pulNbOfBytesRead = ulTotalNbOfBytesRead;
    }

    if(p_pErrorInfo != 0) *p_pErrorInfo = errorInfo;

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsCompleteSizeRead(DWORD p_ulCompleteSize, DWORD p_ulTotalNbOfBytesRead)
{
    BOOL oResult(FALSE);

    if(p_ulCompleteSize > 0)
    {
        oResult = (p_ulTotalNbOfBytesRead >= p_ulCompleteSize);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetAddressParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE& p_rubNodeId, CErrorInfo* p_pErrorInfo)
{
    const CStdString NODE_ID(_T("NodeId"));

    BOOL oResult(FALSE);

    if (p_pCommand && p_pManager)
    {
        //Get node id from command
        if (!oResult && p_pCommand)
        {
            oResult = p_pCommand->GetLayerParameter(DEVICE_COMMAND_SET_LAYER, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
        }

        //Get node id from device
        if (!oResult)
        {
            oResult = p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
        }

        //Check for undefined value
        if (oResult && (p_rubNodeId == NODE_ID_UNDEFINED))
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetAddressParameter(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE& p_rubNodeId, CErrorInfo* p_pErrorInfo)
{
    const CStdString NODE_ID(_T("NodeId"));

    BOOL oResult(FALSE);

    if (p_pManager)
    {
        //Get node id from device
        if (!oResult)
        {
            oResult = p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo);
        }

        //Check for undefined value
        if (oResult && (p_rubNodeId == NODE_ID_UNDEFINED))
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::InitLayerParameterStack(CCommandRoot* p_pCommand)
{
    CLayerParameterStack layerParameterStack;
    CLayerParameterSet layerParameterSet;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        if(p_pCommand->GetLayerParameterStack(layerParameterStack))
        {
            //Pop VCS layer Parameter Set
            layerParameterStack.PopLayer(VIRTUAL_COMMAND_SET_LAYER, layerParameterSet);

            //Set DCS layer Commands
            if(m_pCommand_WriteObject) m_pCommand_WriteObject->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_InitiateSegmentedWrite) m_pCommand_InitiateSegmentedWrite->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_SegmentedWrite) m_pCommand_SegmentedWrite->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ReadObject) m_pCommand_ReadObject->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_InitiateSegmentedRead) m_pCommand_InitiateSegmentedRead->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_SegmentedRead) m_pCommand_SegmentedRead->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_AbortTransfer) m_pCommand_AbortTransfer->SetLayerParameterStack(layerParameterStack);

            if(m_pCommand_SendCANFrame) m_pCommand_SendCANFrame->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_RequestCANFrame) m_pCommand_RequestCANFrame->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ReadCANFrame) m_pCommand_ReadCANFrame->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_SendNMTService) m_pCommand_SendNMTService->SetLayerParameterStack(layerParameterStack);

            if(m_pCommand_WriteEscRegister) m_pCommand_WriteEscRegister->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ReadEscRegister) m_pCommand_ReadEscRegister->SetLayerParameterStack(layerParameterStack);

            if(m_pCommand_SendLSSFrame) m_pCommand_SendLSSFrame->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ReadLSSFrame) m_pCommand_ReadLSSFrame->SetLayerParameterStack(layerParameterStack);

            if(m_pCommand_InitEsm) m_pCommand_InitEsm->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_GetEsmState) m_pCommand_GetEsmState->SetLayerParameterStack(layerParameterStack);
            if(m_pCommand_ChangeEsmState) m_pCommand_ChangeEsmState->SetLayerParameterStack(layerParameterStack);

            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDeviceTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, DWORD& p_rulTimeout)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));
    CStdString interfaceName(_T(""));
    WORD usTimeout(0);

    if(p_pManager)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hHandle, &protocolStackName) && p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            if((PROTOCOL_MAXON_SERIAL_V1 == protocolStackName) || (PROTOCOL_MAXON_SERIAL_V2 == protocolStackName))
            {
                if(interfaceName == INTERFACE_RS232)
                {
                    //RS232 frame timeout
                    oResult = ReadUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_RS232_FRAME_TIMEOUT, SUBINDEX_RS232_FRAME_TIMEOUT, &usTimeout);
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    //USB frame timeout
                    oResult = ReadUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_USB_FRAME_TIMEOUT, SUBINDEX_USB_FRAME_TIMEOUT, &usTimeout);
                }
            }

            //Result
            if(oResult) p_rulTimeout = usTimeout;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::SetDeviceTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, DWORD p_ulTimeout)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));
    CStdString interfaceName(_T(""));
    WORD usTimeout(0);

    if(p_pManager)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hHandle, &protocolStackName) && p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            //Init
            usTimeout = (WORD)p_ulTimeout;

            if((PROTOCOL_MAXON_SERIAL_V1 == protocolStackName) || (PROTOCOL_MAXON_SERIAL_V2 == protocolStackName))
            {
                if(interfaceName == INTERFACE_RS232)
                {
                    //RS232 frame timeout
                    oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_RS232_FRAME_TIMEOUT, SUBINDEX_RS232_FRAME_TIMEOUT, usTimeout);
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    //USB frame timeout
                    oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_USB_FRAME_TIMEOUT, SUBINDEX_USB_FRAME_TIMEOUT, usTimeout);
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::InitGatewayTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));
    CStdString interfaceName(_T(""));
    DWORD ulBaudrate(0);
    DWORD ulTimeout(0);
    WORD usTimeout(0);

    if(p_pManager)
    {
        oResult = p_pManager->DCS_GetProtocolStackSettings(p_hHandle, &ulBaudrate, &ulTimeout, p_pErrorInfo)
                  && p_pManager->DCS_GetProtocolStackName(p_hHandle, &protocolStackName, p_pErrorInfo)
                  && p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName, p_pErrorInfo);

        if(oResult)
        {
            if((PROTOCOL_MAXON_SERIAL_V1 == protocolStackName) || (PROTOCOL_MAXON_SERIAL_V2 == protocolStackName))
            {
                usTimeout = CalculateDeviceTimeout(ulTimeout);

                if(interfaceName == INTERFACE_RS232)
                {
                    oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_RS232_FRAME_TIMEOUT, SUBINDEX_RS232_FRAME_TIMEOUT, usTimeout, p_pErrorInfo);
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_USB_FRAME_TIMEOUT, SUBINDEX_USB_FRAME_TIMEOUT, usTimeout, p_pErrorInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetDeviceTransferRate(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, DWORD& p_rulTransferRate, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));
    CStdString interfaceName(_T(""));
    WORD usTransferRateIndex(0);

    if(p_pManager)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hHandle, &protocolStackName) && p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            if(protocolStackName == PROTOCOL_CAN_OPEN)
            {
                //CAN bitrate
                if(ReadUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, &usTransferRateIndex, p_pErrorInfo))
                {
                    if(CheckBitrateIndexCiaStandard(p_pManager, p_hHandle, p_hTransactionHandle, usTransferRateIndex))
                    {
                        //Use CiA standard for conversion
                        p_rulTransferRate = ConvertBitrateIndexToValue_CiaStandard(usTransferRateIndex);
                        oResult = TRUE;
                    }
                    else
                    {
                        //Use old maxon standard for conversion
                        p_rulTransferRate = ConvertBitrateIndexToValue_Maxon(usTransferRateIndex);
                        oResult = TRUE;
                    }
                }
            }
            else
            {
                if(interfaceName == INTERFACE_RS232)
                {
                    //RS232 baudrate
                    if(ReadUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_BAUDRATE, SUBINDEX_BAUDRATE, &usTransferRateIndex, p_pErrorInfo))
                    {
                        p_rulTransferRate = ConvertBaudrateIndexToValue(usTransferRateIndex);
                        oResult = TRUE;
                    }
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    //No USB transfer rate
                    oResult = FALSE;
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::SetDeviceTransferRate(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, DWORD p_ulTransferRate, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));
    CStdString interfaceName(_T(""));
    WORD usTransferRateIndex(0);

    if(p_pManager)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hHandle, &protocolStackName) && p_pManager->DCS_GetInterfaceName(p_hHandle, &interfaceName))
        {
            if(protocolStackName == PROTOCOL_CAN_OPEN)
            {
                if(CheckBitrateValueCiaStandard(p_pManager, p_hHandle, p_hTransactionHandle, p_ulTransferRate))
                {
                    //Use CiA standard for conversion
                    usTransferRateIndex = ConvertBitrateValueToIndex_CiaStandard(p_ulTransferRate);
                }
                else
                {
                    //Use old maxon standard for conversion
                    usTransferRateIndex = ConvertBitrateValueToIndex_Maxon(p_ulTransferRate);
                }

                //CAN bitrate
                oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, usTransferRateIndex, p_pErrorInfo);
            }
            else
            {
                if(interfaceName == INTERFACE_RS232)
                {
                    //RS232 baudrate
                    usTransferRateIndex = ConvertBaudrateValueToIndex(p_ulTransferRate);
                    oResult = WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_BAUDRATE, SUBINDEX_BAUDRATE, usTransferRateIndex, p_pErrorInfo);
                }
                else if(interfaceName == INTERFACE_USB)
                {
                    //No USB transfer rate
                    oResult = FALSE;
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::StoreParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const DWORD TIMEOUT = 5000;

    BOOL oResult(FALSE);
    DWORD ulBackupTimeout = 0;
    DWORD ulPassword = 0;
    DWORD ulTransferRate = 0;

    if(p_pManager)
    {
        //Init
        oResult = TRUE;

        //Backup timeout
        p_pManager->DCS_GetProtocolStackSettings(p_hHandle, &ulTransferRate, &ulBackupTimeout);

        //Increase timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, TIMEOUT, TRUE);

        //Write save all parameters
        ulPassword = 'e';
        ulPassword *= 256;
        ulPassword += 'v';
        ulPassword *= 256;
        ulPassword += 'a';
        ulPassword *= 256;
        ulPassword += 's';

        if(oResult) oResult = WriteUInt32Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_STORE_PARAMETERS, SUBINDEX_SAVE_ALL_PARAMETERS, ulPassword, p_pErrorInfo);

        //Restore timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, ulBackupTimeout, TRUE);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::RemoveStationId(CCommandRoot* p_pCommand, CLayerParameterStack& p_rBackupLayerParameterStack)
{
    BOOL oResult(FALSE);
    CLayerParameterStack layerParameterStack;
    CLayerParameterSet layerParameterSet;

    if (p_pCommand != 0)
    {
        if (m_pCommand_ReadEscRegister->GetLayerParameterStack(p_rBackupLayerParameterStack))
        {
            layerParameterStack = p_rBackupLayerParameterStack;
            if (layerParameterStack.PopLayer(DEVICE_COMMAND_SET_LAYER, layerParameterSet))
            {
                layerParameterSet.DeleteParameter(_T("StationId"));
                layerParameterStack.PushLayer(layerParameterSet);

                m_pCommand_ReadEscRegister->SetLayerParameterStack(layerParameterStack);
                oResult = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::RestoreLayerParameterStack(CCommandRoot* p_pCommand, CLayerParameterStack& p_rBackupLayerParameterStack)
{
    BOOL oResult(FALSE);

    if (p_pCommand)
    {
        p_pCommand->SetLayerParameterStack(p_rBackupLayerParameterStack);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WaitForReboot(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const DWORD REBOOT_TIMEOUT = 8000;
    const DWORD SLEEP_TIME = 10;
    const int NB_OF_NMT_FRAMES = 10;

    BOOL oResult(FALSE);
    BOOL oCANopenInterface(FALSE);
    BOOL oEtherCatInterface(FALSE);
    CLayerParameterStack backupLayerParameterStack;
    BYTE uErrorRegister = 0;
    DWORD ulStartTime = 0;
    DWORD ulTime = 0;
    WORD usStatusCode = 0;
    WORD usLengthRead = 0;
    BYTE ubNodeId = 0;

    //Init
    RemoveStationId(m_pCommand_ReadEscRegister, backupLayerParameterStack);
    oCANopenInterface = IsCANopenInterface(p_pManager, p_hHandle);
    oEtherCatInterface = IsEtherCatInterface(p_pManager, p_hHandle);
    ulStartTime = MmcGetTickCount();
    ulTime = 0;

    while (ulTime < REBOOT_TIMEOUT)
    {
        //Reset
        if (p_pErrorInfo) p_pErrorInfo->Reset();

        //Check communication
        if (oEtherCatInterface)
        {
            //Read register
            oResult = DCS_ReadEscRegister(p_pManager, p_hHandle, p_hTransactionHandle, ADDRESS_AL_STATUS_CODE, sizeof(usStatusCode), (void*)&usStatusCode, &usLengthRead, p_pErrorInfo);
        }
        else
        {
            //Read object
            oResult = DCS_ReadObject(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_ERROR_REGISTER, SUBINDEX_ERROR_REGISTER, &uErrorRegister, sizeof(uErrorRegister), p_pErrorInfo);
        }

        if (oResult)
        {
            if (oEtherCatInterface)
            {
                DCS_ChangeEsmState(p_pManager, p_hHandle, p_hTransactionHandle, ALS_PRE_OPERATIONAL);
            }

            break;
        }
        else
        {
            if (oCANopenInterface)
            {
                //Force autobitrate detection
                GetAddressParameter(p_pCommand, p_pManager, p_hHandle, ubNodeId, 0);
                for (int i = 0; i < NB_OF_NMT_FRAMES; i++)
                {
                    DCS_SendNMTService(p_pManager, p_hHandle, p_hTransactionHandle, ubNodeId, (WORD)NSI_ENTER_PRE_OPERATIONAL);
                }
            }

            //Update time
            ulTime = MmcGetTickCount() - ulStartTime;
            MmcSleep(SLEEP_TIME);
        }
    }

    //Restore
    RestoreLayerParameterStack(m_pCommand_ReadEscRegister, backupLayerParameterStack);

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetMainSensorType(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusSensorType, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubMainSensor(0);
    WORD usSensorType(0);

    if (GetMainSensor(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubMainSensor, p_pErrorInfo))
    {
        if (GetSensorType(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubMainSensor, usSensorType, p_pErrorInfo))
        {
            p_rusSensorType = usSensorType;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetAuxiliarySensorType(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusSensorType, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubAuxiliarySensor(0);
    WORD usSensorType(0);

    if (GetAuxiliarySensor(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAuxiliarySensor, p_pErrorInfo))
    {
        if (GetSensorType(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAuxiliarySensor, usSensorType, p_pErrorInfo))
        {
            p_rusSensorType = usSensorType;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetMainSensor(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE& p_rubMainSensor, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulSensorControlStructure;
    CErrorInfo errorInfo;

    // Read sensor type
    if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_CONTROL_STRUCTURE, &ulSensorControlStructure, &errorInfo))
    {
        // Extract bit 16 - 19 to get mainsensor-nibble
        p_rubMainSensor = (BYTE)((ulSensorControlStructure >> 16) & 0xF);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetAuxiliarySensor(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE& p_rubAuxiliarySensor, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulSensorControlStructure;
    CErrorInfo errorInfo;

    // Read sensor
    if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_CONTROL_STRUCTURE, &ulSensorControlStructure, &errorInfo))
    {
        // Extract bit 20 - 23 to get auxiliarysensor-nibble
        p_rubAuxiliarySensor = (BYTE)((ulSensorControlStructure >> 20) & 0xF);
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetSensorType(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubSensor, WORD& p_rusSensorType, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if (p_ubSensor == SENSOR_NONE)
    {
        p_rusSensorType = SENSOR_NONE;
        oResult = TRUE;
    }
    else if(p_ubSensor == SENSOR_1)
    {
        oResult = GetSensor1Type(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_ubSensor, p_rusSensorType, p_pErrorInfo);
    }
    else if (p_ubSensor == SENSOR_2)
    {
        oResult = GetSensor2Type(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_ubSensor, p_rusSensorType, p_pErrorInfo);
    }
    else if (p_ubSensor == SENSOR_3)
    {
        oResult = GetSensor3Type(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_ubSensor, p_rusSensorType, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetSensor1Type(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubSensor, WORD& p_rusSensorType, CErrorInfo* p_pErrorInfo)
{
    const WORD WITH_INDEX = 0x0001;

    BOOL oResult(FALSE);
    DWORD ulSensorConfiguration(0);
    BYTE ubSensor1Type(0);
    WORD usIncrementalEncoder1Type(0);
    CErrorInfo errorInfo;

    // Read sensor type
    if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_SENSORS_CONFIGURATION, &ulSensorConfiguration, &errorInfo))
    {
        // Extract bit 0 - 7 to get sensor 1
        ubSensor1Type = (BYTE)((ulSensorConfiguration >> 0) & 0xFF);

        if (ubSensor1Type == SENSOR1_TYPE_NONE)
        {
            p_rusSensorType = SENSOR_TYPE_NONE;
            oResult = TRUE;
        }
        else if (ubSensor1Type == SENSOR1_TYPE_DIGITAL_INC_ENCODER_1)
        {
            // Get incremental encoder 1 type
            if (ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_DIGITAL_INC_ENCODER_1, EPOS4_SUBINDEX_ENCODER_1_TYPE, &usIncrementalEncoder1Type, &errorInfo))
            {
                // Extract 2/3 channel
                if ((usIncrementalEncoder1Type & WITH_INDEX) > 0)
                {
                    // 3-channel encoder (with index)
                    p_rusSensorType = SENSOR_TYPE_INC_ENCODER_3_CHANNEL;
                    oResult = TRUE;
                }
                else
                {
                    // 2-channel encoder (without index)
                    p_rusSensorType = SENSOR_TYPE_INC_ENCODER_2_CHANNEL;
                    oResult = TRUE;
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetSensor2Type(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubSensor, WORD& p_rusSensorType, CErrorInfo* p_pErrorInfo)
{
    const WORD WITH_INDEX = 0x0001;
    const WORD GRAY_CODED_DATA = 0x0001;

    BOOL oResult(FALSE);
    DWORD ulSensorConfiguration(0);
    BYTE ubSensor2Type(0);
    WORD usIncrementalEncoder2Type(0);
    WORD usAnalogEncoderType(0);
    WORD usSsiEncodingType(0);
    CErrorInfo errorInfo;

    // Read sensor type
    if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_SENSORS_CONFIGURATION, &ulSensorConfiguration, &errorInfo))
    {
        // Extract bit 8 - 15 to get sensor 2
        ubSensor2Type = (BYTE)((ulSensorConfiguration >> 8) & 0xFF);

        if (ubSensor2Type == SENSOR2_TYPE_NONE)
        {
            p_rusSensorType = SENSOR_TYPE_NONE;
            oResult = TRUE;
        }
        else if (ubSensor2Type == SENSOR2_TYPE_DIGITAL_INC_ENCODER_2)
        {
            // Get incremental encoder 2 type
            if (ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_DIGITAL_INC_ENCODER_2, EPOS4_SUBINDEX_ENCODER_2_TYPE, &usIncrementalEncoder2Type, &errorInfo))
            {
                // Extract 2/3 channel
                if ((usIncrementalEncoder2Type & WITH_INDEX) > 0)
                {
                    // 3-channel encoder (with index)
                    p_rusSensorType = SENSOR_TYPE_INC_ENCODER2_3_CHANNEL;
                    oResult = TRUE;
                }
                else
                {
                    // 2-channel encoder (without index)
                    p_rusSensorType = SENSOR_TYPE_INC_ENCODER2_2_CHANNEL;
                    oResult = TRUE;
                }
            }
        }
        else if (ubSensor2Type == SENSOR2_TYPE_ANALOG_INC_ENCODER)
        {
            // Get analog encoder type
            if (ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_ANALOG_INC_ENCODER, EPOS4_SUBINDEX_ANALOG_ENCODER_TYPE, &usAnalogEncoderType, &errorInfo))
            {
                // Extract 2/3 channel
                if ((usAnalogEncoderType & WITH_INDEX) > 0)
                {
                    // 3-channel encoder (with index)
                    p_rusSensorType = SENSOR_TYPE_ANALOG_INC_ENCODER_3_CHANNEL;
                    oResult = TRUE;
                }
                else
                {
                    // 2-channel encoder (without index)
                    p_rusSensorType = SENSOR_TYPE_ANALOG_INC_ENCODER_2_CHANNEL;
                    oResult = TRUE;
                }
            }
        }
        else if (ubSensor2Type == SENSOR2_TYPE_SSI_ENCODER)
        {
            // Get SSI encoding type
            if (ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SSI_ABSOLUTE_ENCODER, EPOS4_SUBINDEX_SSI_ENCODING_TYPE, &usSsiEncodingType, &errorInfo))
            {
                // Extract encoding type
                if ((usSsiEncodingType & GRAY_CODED_DATA) > 0)
                {
                    p_rusSensorType = SENSOR_TYPE_SSI_ABS_ENCODER_GREY;
                    oResult = TRUE;
                }
                else
                {
                    p_rusSensorType = SENSOR_TYPE_SSI_ABS_ENCODER_BINARY;
                    oResult = TRUE;
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::GetSensor3Type(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubSensor, WORD& p_rusSensor3Type, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulSensorConfiguration(0);
    BYTE ubSensor3Type(0);
    CErrorInfo errorInfo;

    // Read sensor type
    if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_SENSORS_CONFIGURATION, &ulSensorConfiguration, &errorInfo))
    {
        // Extract bit 16 - 23 to get sensor 1
        ubSensor3Type = (BYTE)((ulSensorConfiguration >> 16) & 0xFF);

        if (ubSensor3Type == SENSOR3_TYPE_NONE)
        {
            p_rusSensor3Type = SENSOR_TYPE_NONE;
            oResult = TRUE;
        }
        else if (ubSensor3Type == SENSOR3_TYPE_DIGITAL_HALL_SENSOR)
        {
            p_rusSensor3Type = SENSOR_TYPE_HALL_SENSORS;
            oResult = TRUE;
        }
    }

    return oResult;
}

DWORD CGatewayVCStoEpos4::ConvertBaudrateIndexToValue(WORD p_usBaudrateIndex)
{
    switch(p_usBaudrateIndex)
    {
        case 0: return 9600;
        case 1: return 14400;
        case 2: return 19200;
        case 3: return 38400;
        case 4: return 57600;
        case 5: return 115200;
        default: return 115200;
    }
}

WORD CGatewayVCStoEpos4::ConvertBaudrateValueToIndex(DWORD p_ulBaudrate)
{
    switch(p_ulBaudrate)
    {
        case 9600: return 0;
        case 14400: return 1;
        case 19200: return 2;
        case 38400: return 3;
        case 57600: return 4;
        case 115200: return 5;
        default: return 0;
    }
}

BOOL CGatewayVCStoEpos4::CheckBitrateIndexCiaStandard(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, WORD p_usBaudrateIndex)
{
    const WORD COMPATIBILITY_LIMIT = 4;

    BOOL oResult(FALSE);

    if(p_usBaudrateIndex <= COMPATIBILITY_LIMIT)
    {
        //Compatible to Cia standard
        oResult = TRUE;
    }
    else
    {
        //Check by trying to write bitrate index
        oResult = CheckBitrateCiaStandard(p_pManager, p_hHandle, p_hTransactionHandle);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CheckBitrateValueCiaStandard(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, DWORD p_ulBaudrate)
{
    const DWORD COMPATIBILITY_LIMIT = 125000;

    BOOL oResult(FALSE);

    if(p_ulBaudrate >= COMPATIBILITY_LIMIT)
    {
        //Compatible to CiA standard
        oResult = TRUE;
    }
    else
    {
        //Check by trying to write bitrate index
        oResult = CheckBitrateCiaStandard(p_pManager, p_hHandle, p_hTransactionHandle);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CheckBitrateCiaStandard(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    const WORD BITRATE_INDEX_CHECK = 7;

    BOOL oResult(FALSE);
    WORD usBitrateIndex = 0;

    //Backup bitrate index
    if(ReadUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, &usBitrateIndex))
    {
        //Check bitrate index
        if(WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, BITRATE_INDEX_CHECK))
        {
            //Compatible to CiA standard
            oResult = TRUE;
        }

        //Restore bitrate index
        WriteUInt16Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, usBitrateIndex);
    }

    return oResult;
}

DWORD CGatewayVCStoEpos4::ConvertBitrateIndexToValue_CiaStandard(WORD p_usBaudrateIndex)
{
    switch(p_usBaudrateIndex)
    {
        case 0:     return 1000000;
        case 1:     return 800000;
        case 2:     return 500000;
        case 3:     return 250000;
        case 4:     return 125000;
        case 5:     return 0;
        case 6:     return 50000;
        case 7:     return 20000;
        case 9:     return TRANSFER_RATE_AUTO;
        default:    return 1000000;
    }
}

WORD CGatewayVCStoEpos4::ConvertBitrateValueToIndex_CiaStandard(DWORD p_ulBaudrate)
{
    switch(p_ulBaudrate)
    {
        case 1000000:               return 0;
        case 800000:                return 1;
        case 500000:                return 2;
        case 250000:                return 3;
        case 125000:                return 4;
        case 50000:                 return 6;
        case 20000:                 return 7;
        case TRANSFER_RATE_AUTO:    return 9;
        default:                    return 0;
    }
}

BOOL CGatewayVCStoEpos4::ReadSiiSize(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD& p_rulSize, CErrorInfo* p_pErrorInfo)
{
    const DWORD KBIT_TO_BYTE = 1024/8;

    BOOL oResult(FALSE);
    WORD usKBitsSize(0);

    //Read SII size
    if(DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_SIZE, &usKBitsSize, sizeof(usKBitsSize), p_pErrorInfo))
    {
        //Calculate number of bytes
        p_rulSize = (usKBitsSize + 1)*  KBIT_TO_BYTE;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadSiiCategoryGeneral(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, SSiiCatagoryGeneral& p_sCategoryGeneral, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    oResult = DCS_ReadSiiCategoryData(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_CAT_GENERAL, &p_sCategoryGeneral, sizeof(p_sCategoryGeneral), p_pErrorInfo);

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadSiiCategoryString(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_uStringIndex, CStdString& p_rStringValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulCategorySize(0);
    BYTE* pDataBuffer(0);
    char* pStringBuffer(0);
    DWORD ulStringBufferSize(0);
    BYTE uNbOfStrings(0);
    BYTE* pStringLen(0);
    BYTE* pStringData(0);

    if(DCS_ReadSiiCategorySize(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_CAT_STRINGS, &ulCategorySize, p_pErrorInfo))
    {
        //Allocate memory
        pDataBuffer = (BYTE*)malloc(ulCategorySize);

        //Read category strings
        if(DCS_ReadSiiCategoryData(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_CAT_STRINGS, pDataBuffer, ulCategorySize, p_pErrorInfo))
        {
            //Init first string
            uNbOfStrings = *pDataBuffer;
            pStringLen = pDataBuffer + sizeof(uNbOfStrings);
            pStringData = pStringLen + sizeof(*pStringLen);

            //Search string
            for(BYTE uIndex = 0; uIndex < uNbOfStrings; uIndex++)
            {
                //Check pointers
                if((pStringLen < pDataBuffer + ulCategorySize) && (pStringData < pDataBuffer + ulCategorySize))
                {
                    //Check string data size
                    if((pStringData +* pStringLen) <= (pDataBuffer + ulCategorySize))
                    {
                        if(uIndex == p_uStringIndex - 1) //Start index is 1
                        {
                            //Allocate string buffer
                            ulStringBufferSize = (*pStringLen) + sizeof(char);
                            pStringBuffer = (char*)malloc(ulStringBufferSize);
                            memset(pStringBuffer, 0, ulStringBufferSize);

                            //Copy string
                            memcpy(pStringBuffer, pStringData, ulStringBufferSize - 1);
                            p_rStringValue = pStringBuffer;

                            //Free string buffer
                            if(pStringBuffer) free(pStringBuffer);

                            oResult = TRUE;
                            break;
                        }
                        else
                        {
                            //Init next string
                            pStringLen = pStringData +* pStringLen;
                            pStringData = pStringLen + sizeof(*pStringLen);
                        }
                    }
                    else
                    {
                        if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_STRING_TABLE);
                        oResult = FALSE;
                        break;
                    }
                }
                else
                {
                    if(p_pErrorInfo) p_pErrorInfo->Init(ERROR_VIRTUAL_DEVICE_BAD_SII_STRING_TABLE);
                    oResult = FALSE;
                    break;
                }
            }
        }

        //Free memory
        if(pDataBuffer) free(pDataBuffer);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::WriteEscRegister(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usAdo, WORD p_usLength, void* p_pData, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usLengthWritten(0);

    if(DCS_WriteEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usAdo, p_usLength, p_pData, &usLengthWritten, p_pErrorInfo))
    {
        //Check size
        if(p_usLength != usLengthWritten)
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }
        else
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ReadEscRegister(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usAdo, WORD p_usLength, void* p_pData, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usLengthRead(0);

    if(DCS_ReadEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usAdo, p_usLength, p_pData, &usLengthRead, p_pErrorInfo))
    {
        //Check size
        if(p_usLength != usLengthRead)
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }
        else
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

DWORD CGatewayVCStoEpos4::ConvertBitrateIndexToValue_Maxon(WORD p_usBaudrateIndex)
{
    switch(p_usBaudrateIndex)
    {
        case 0: return 1000000;
        case 1: return 800000;
        case 2: return 500000;
        case 3: return 250000;
        case 4: return 125000;
        case 5: return 50000;
        case 6: return 20000;
        case 9: return TRANSFER_RATE_AUTO;
        default: return 1000000;
    }
}

WORD CGatewayVCStoEpos4::ConvertBitrateValueToIndex_Maxon(DWORD p_ulBaudrate)
{
    switch(p_ulBaudrate)
    {
        case 1000000: return 0 ;
        case 800000: return 1 ;
        case 500000: return 2;
        case 250000: return 3;
        case 125000: return 4;
        case 50000: return 5;
        case 20000: return 6;
        case TRANSFER_RATE_AUTO: return 9;
        default: return 0;
    }
}

BOOL CGatewayVCStoEpos4::RestoreParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        oResult = RestoreDevice(p_pManager, p_hHandle, p_hTransactionHandle, FALSE, p_pErrorInfo);

        if (oResult)
        {
            oResult = ResetDevice(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle, FALSE, TRUE, p_pErrorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::RestoreFactoryParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        oResult = RestoreDevice(p_pManager, p_hHandle, p_hTransactionHandle, TRUE, p_pErrorInfo);

        if (oResult)
        {
            oResult = ResetDevice(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle, FALSE, TRUE, p_pErrorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::RestoreDevice(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, BOOL p_oFactory, CErrorInfo* p_pErrorInfo)
{
    const DWORD TIMEOUT = 5000;

    BOOL oResult(FALSE);
    DWORD ulBackupDeviceTimeout = 0;
    DWORD ulBackupTimeout = 0;
    DWORD ulTransferRate = 0;
    DWORD ulPassword = 0;

    if (p_pManager)
    {
        //Backup timeout
        p_pManager->DCS_GetProtocolStackSettings(p_hHandle, &ulTransferRate, &ulBackupTimeout);
        GetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, ulBackupDeviceTimeout);

        //Increase timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, TIMEOUT, TRUE);
        SetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, TIMEOUT);

        //Write load all parameters
        ulPassword = 'd';
        ulPassword *= 256;
        ulPassword += 'a';
        ulPassword *= 256;
        ulPassword += 'o';
        ulPassword *= 256;
        ulPassword += 'l';

        //Restore device
        WORD usIndex = INDEX_RESTORE_DEFAULT_PARAMETERS;
        WORD usSubIndex;
        if (p_oFactory)
        {
            usSubIndex = SUBINDEX_RESTORE_FACTORY_DEFAULT_PARAMETERS;
        }
        else
        {
            usSubIndex = SUBINDEX_RESTORE_ALL_DEFAULT_PARAMETERS;
        }
        oResult = WriteUInt32Object(p_pManager, p_hHandle, p_hTransactionHandle, usIndex, usSubIndex, ulPassword, p_pErrorInfo);

        //Restore timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, ulBackupTimeout, TRUE);
        SetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, ulBackupDeviceTimeout);
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::ResetDevice(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle, BOOL p_oBroadcast, BOOL p_oWaitForReboot, CErrorInfo* p_pErrorInfo)
{
    const DWORD RESET_TIMEOUT(2000);
    const BYTE PROGRAM_CONTROL_RESET(0x02);

    BOOL oResult(FALSE);
    DWORD ulBackupDeviceTimeout = 0;
    DWORD ulBackupTimeout = 0;
    DWORD ulTransferRate = 0;
    BYTE ubBroadcastNodeId = 0;
    CErrorInfo errorInfo;

    if (p_pManager)
    {
        //Backup timeout
        p_pManager->DCS_GetProtocolStackSettings(p_hHandle, &ulTransferRate, &ulBackupTimeout);
        GetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, ulBackupDeviceTimeout);

        //Decrease timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, RESET_TIMEOUT, TRUE);
        SetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, RESET_TIMEOUT);

        //Reset
        if (p_oBroadcast)
        {
            //SendNMTService
            oResult = DCS_SendNMTService(p_pManager, p_hHandle, p_hTransactionHandle, ubBroadcastNodeId, (WORD)NSI_RESET_NODE, &errorInfo);
        }
        else
        {
            //Program control
            oResult = WriteUInt8Object(p_pManager, p_hHandle, p_hTransactionHandle, INDEX_PROGRAM_CONTROL, SUBINDEX_PROGRAM_CONTROL_PRG_NB_1, PROGRAM_CONTROL_RESET, &errorInfo);
        }

        //Wait before changing transfer rate
        MmcSleep(RESET_TIMEOUT);

        if (!CErrorHandling::IsDeviceErrorCode(errorInfo) || IsGatewayErrorCode(errorInfo))
        {
            //No response for reset
            oResult = TRUE;
            errorInfo.Reset();

            if (oResult && p_oWaitForReboot)
            {
                //Wait for reboot
                WaitForReboot(p_pCommand, p_pManager, p_hHandle, 0, 0);
            }
        }

        //Restore timeout
        p_pManager->DCS_SetProtocolStackSettings(p_hHandle, ulTransferRate, ulBackupTimeout, TRUE);
        SetDeviceTimeout(p_pManager, p_hHandle, p_hTransactionHandle, ulBackupDeviceTimeout);

        //Return error info
        if(p_pErrorInfo != 0)
        {
            *p_pErrorInfo = errorInfo;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CalculateMainSensorResolution(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD& p_rulMainSensorResolution, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usSensorType(0);
    DWORD ulSensorResolution(0);
    CErrorInfo errorInfo;

    if(GetMainSensorType(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usSensorType, &errorInfo))
    {
        if (CalculateSensorResolution(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usSensorType, ulSensorResolution, &errorInfo))
        {
            p_rulMainSensorResolution = ulSensorResolution;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CalculateAuxiliarySensorResolution(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD& p_rulMainSensorResolution, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usSensorType(0);
    DWORD ulSensorResolution(0);
    CErrorInfo errorInfo;

    if (GetAuxiliarySensorType(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usSensorType, &errorInfo))
    {
        if (CalculateSensorResolution(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usSensorType, ulSensorResolution, &errorInfo))
        {
            p_rulMainSensorResolution = ulSensorResolution;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::CalculateSensorResolution(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usSensorType, DWORD& p_rulSensorResolution, CErrorInfo* p_pErrorInfo)
{
    const int FACTOR_PULSE_TO_INCREMENTS = 4;
    const int HALLSENSOR_INCREMENTS = 6;

    BOOL oResult(FALSE);
    BYTE ubPolePair(0);
    DWORD ulAnalogIncEncoderResolution(0);
    DWORD ulNumberOfPeriods(0);
    DWORD ulSsiNumberOfDataBits(0);
    BYTE ubInterpolationBits(0);
    BYTE ubSingleTurnBits(0);
    CErrorInfo errorInfo;
    DWORD ulEncoderResolution(0);

    switch (p_usSensorType)
    {
        case SENSOR_TYPE_INC_ENCODER_3_CHANNEL:
        case SENSOR_TYPE_INC_ENCODER_2_CHANNEL:
        {
            if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_DIGITAL_INC_ENCODER_1, EPOS4_SUBINDEX_ENCODER_1_NUMBER_OF_PULSES, &ulEncoderResolution, &errorInfo))
            {
                p_rulSensorResolution = ulEncoderResolution * FACTOR_PULSE_TO_INCREMENTS;
                oResult = TRUE;
            }
        }
        break;

        case SENSOR_TYPE_INC_ENCODER2_3_CHANNEL:
        case SENSOR_TYPE_INC_ENCODER2_2_CHANNEL:
        {
            if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_DIGITAL_INC_ENCODER_2, EPOS4_SUBINDEX_ENCODER_2_NUMBER_OF_PULSES, &ulEncoderResolution, &errorInfo))
            {
                p_rulSensorResolution = ulEncoderResolution * FACTOR_PULSE_TO_INCREMENTS;
                oResult = TRUE;
            }
        }
        break;

        case SENSOR_TYPE_ANALOG_INC_ENCODER_3_CHANNEL:
        case SENSOR_TYPE_ANALOG_INC_ENCODER_2_CHANNEL:
        {
            if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_ANALOG_INC_ENCODER, EPOS4_SUBINDEX_ANALOG_ENCODER_RESOLUTION, &ulAnalogIncEncoderResolution, &errorInfo))
            {
                ulNumberOfPeriods = DWORD((ulAnalogIncEncoderResolution >> 8) & 0xFFFFFF);
                ubInterpolationBits = BYTE((ulAnalogIncEncoderResolution >> 0) & 0xFF);
                p_rulSensorResolution = (ulNumberOfPeriods << ubInterpolationBits);
                oResult = TRUE;
            }
        }
        break;

        case SENSOR_TYPE_SSI_ABS_ENCODER_BINARY:
        case SENSOR_TYPE_SSI_ABS_ENCODER_GREY:
        {
            if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SSI_ABSOLUTE_ENCODER, EPOS4_SUBINDEX_SSI_NUMBER_OF_DATABITS, &ulSsiNumberOfDataBits, &errorInfo))
            {
                ubSingleTurnBits = BYTE((ulSsiNumberOfDataBits >> 8) & 0xFF);
                p_rulSensorResolution = (DWORD)pow(2, ubSingleTurnBits);
                oResult = TRUE;
            }
        }
        break;

        case SENSOR_TYPE_HALL_SENSORS:
        {
            if (ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_MOTOR_DATA, EPOS4_SUBINDEX_POLE_PAIR_NUMBER, &ubPolePair, &errorInfo))
            {
                p_rulSensorResolution = ubPolePair * HALLSENSOR_INCREMENTS;
                oResult = TRUE;
            }
        }
        break;
    }

    return oResult;
}

void CGatewayVCStoEpos4::MapBitFromDwordToWord(DWORD p_ulSource, BYTE p_ubSourceBit, WORD& p_rusDestination, BYTE p_ubDestinationBit)
{
    if ((p_ulSource & (0x01 << p_ubSourceBit)) > 0)
    {
        p_rusDestination |= (0x01 << p_ubDestinationBit);
    }
}

void CGatewayVCStoEpos4::MapBitFromWordToDword(WORD p_usSource, BYTE p_ubSourceBit, DWORD& p_rulDestination, BYTE p_ubDestinationBit)
{
    if ((p_usSource & (0x01 << p_ubSourceBit)) > 0)
    {
        p_rulDestination |= (0x01 << p_ubDestinationBit);
    }
}

bool CGatewayVCStoEpos4::IsBitSetInDword(DWORD p_ulSource, BYTE p_ubBitNumber)
{
    if ((p_ulSource & (0x01 << p_ubBitNumber)) > 0)
        return true;
    else
        return false;
}

bool CGatewayVCStoEpos4::IsBitSetInWord(WORD p_usSource, BYTE p_ubBitNumber)
{
    if ((p_usSource & (0x01 << p_ubBitNumber)) > 0)
        return true;
    else
        return false;
}

WORD CGatewayVCStoEpos4::CalculateDeviceTimeout(DWORD p_ulTimeout)
{
    const float TIMEOUT_REDUCTION(0.95f);
    const WORD MIN_TIMEOUT_REDUCTION(50);

    WORD ulReducedTimeout = 0;

    //Timeout reduction
    ulReducedTimeout = WORD(float(p_ulTimeout) * TIMEOUT_REDUCTION);

    if(ulReducedTimeout < MIN_TIMEOUT_REDUCTION)
    {
        ulReducedTimeout = MIN_TIMEOUT_REDUCTION;
    }

    return ulReducedTimeout;
}

BOOL CGatewayVCStoEpos4::IsGatewayErrorCode(CErrorInfo& p_rErrorInfo)
{
    BOOL oResult(FALSE);

    switch(p_rErrorInfo.GetErrorCode())
    {
        case ERROR_DEVICE_CAN_SDO_TIME_OUT:
            {
                oResult = TRUE;
                break;
            }

        default:
            oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsEtherCatInterface(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));

    if (p_pManager != 0)
    {
        if (p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName.Compare(PROTOCOL_ETHERCAT_COE) == 0)
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::IsCANopenInterface(CDeviceCommandSetManagerBase * p_pManager, HANDLE p_hDCS_Handle)
{
    BOOL oResult(FALSE);
    CStdString protocolStackName(_T(""));

    if (p_pManager != 0)
    {
        if (p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName.Compare(PROTOCOL_CAN_OPEN) == 0)
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayVCStoEpos4::HandleBackwardsCompatiblity(BOOL p_oResult, CErrorInfo& p_rErrorInfo)
{
    BOOL oResult = p_oResult;

    if (!p_oResult && p_rErrorInfo.GetErrorCode() == ERROR_DEVICE_CAN_OBJECT_NOT_EXIST)
    {
        p_rErrorInfo.Reset();
        oResult = TRUE;
    }

    return oResult;
}