
#include "stdafx.h"

#include <algorithm>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/CommandRoot.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/Command/LayerParameter/LayerParameterStack.h>
#include <CommunicationModel/DeviceCommandSet/BaseClasses/DeviceCommandSetManagerBase.h>
#include <CommunicationModel/VirtualCommandSet/Classes/Handle/HandleRegistration_VCS.h>
#include <CommunicationModel/VirtualCommandSet/Classes/VirtualDevices/BaseClasses/VirtualDeviceBase.h>
#include <CommunicationModel/VirtualCommandSet/Factory/VirtualCommandSetFactory.h>
#include <Thread/MmcSingleLock.h>

#include <CommunicationModel/VirtualCommandSet/VirtualCommandSet_Manager.h>

using namespace std;

CVirtualCommandSet_Manager::CVirtualCommandSet_Manager():
    m_pHandleRegistrationMap_VCS(0)
{
    InitRegistrationMap();
    InitErrorHandling();
}

CVirtualCommandSet_Manager::CVirtualCommandSet_Manager(int p_lInstanceValue):
    CVirtualCommandSet_ManagerBase(p_lInstanceValue),
    m_pHandleRegistrationMap_VCS(0)
{
    InitRegistrationMap();
    InitErrorHandling();
}

CVirtualCommandSet_Manager::CVirtualCommandSet_Manager(const CVirtualCommandSet_Manager& p_rObject)
    : CVirtualCommandSet_ManagerBase(p_rObject)
    , m_pHandleRegistrationMap_VCS(0)
{
    InitRegistrationMap();
    InitErrorHandling();
}

CVirtualCommandSet_ManagerBase* CVirtualCommandSet_Manager::Clone()
{
    CVirtualCommandSet_Manager* pManager;

    pManager = new CVirtualCommandSet_Manager(*this);

    return pManager;
}

CVirtualCommandSet_Manager::~CVirtualCommandSet_Manager()
{
    ReleaseRegistrationMap();
    DeleteVirtualDeviceList();
}

BOOL CVirtualCommandSet_Manager::InitRegistrationMap()
{
    BOOL oResult(FALSE);

    if(!m_pHandleRegistrationMap_VCS)
    {
        m_pHandleRegistrationMap_VCS = CHandleRegistrationMap_VCS::GetInstance(m_lInstanceValue, this);
        if(m_pHandleRegistrationMap_VCS)
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::ReleaseRegistrationMap()
{
    BOOL oResult(TRUE);

    if(m_pHandleRegistrationMap_VCS)
    {
        CHandleRegistrationMap_VCS::ReleaseInstance(m_lInstanceValue, this);
        m_pHandleRegistrationMap_VCS = 0;
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetVirtualDevice(CStdString p_VirtualDeviceName, CVirtualDeviceBase*& p_rpVirtualDevice, BOOL& p_roNewCreated, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    //Init
    p_rpVirtualDevice = 0;
    p_roNewCreated = FALSE;

    //Create virtual device
    if(p_VirtualDeviceName.size() == 0)
    {
        //Don't create a virtual device
        oResult = TRUE;
    }
    else
    {
         //Virtual device is already existing
        if(!FindVirtualDevice(p_VirtualDeviceName, &p_rpVirtualDevice))
        {
            //Init
            p_roNewCreated = TRUE;
            oResult = TRUE;

            oResult = CVirtualCommandSetFactory::CreateInstance(p_VirtualDeviceName, p_rpVirtualDevice);

            if(!oResult)
            {
                //Virtual device not found
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_VIRTUAL_DEVICE_NAME, p_pErrorInfo);
                p_roNewCreated = FALSE;
                oResult = FALSE;
            }
        }
        else
        {
            //Init
            p_roNewCreated = FALSE;
            oResult = TRUE;
        }
    }

    //Init virtual device
    if(oResult && p_rpVirtualDevice)
    {
        p_rpVirtualDevice->InitInstanceValue(m_lInstanceValue);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_InitVirtualDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;
    BOOL oNewCreated = FALSE;
    BOOL oResult(FALSE);

    //Get virtual device
    if(GetVirtualDevice(p_virtualDeviceName, pVirtualDevice, oNewCreated, p_pErrorInfo) && pVirtualDevice)
    {
        //Init virtual device
        if(pVirtualDevice->InitVirtualDevice(p_deviceName, p_protocolStackName, p_interfaceName, p_pErrorInfo))
        {
            if(oNewCreated) m_VirtualDeviceList.push_back(pVirtualDevice);
            oResult = TRUE;
        }
        else
        {
            if(oNewCreated) delete pVirtualDevice;
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_InitVirtualDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CStdString p_driverName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oNewCreated(FALSE);
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice(0);

    //Get virtual device
    if(GetVirtualDevice(p_virtualDeviceName, pVirtualDevice, oNewCreated, p_pErrorInfo) && pVirtualDevice)
    {
        //Init virtual device
        if(pVirtualDevice->InitVirtualDevice(p_deviceName, p_protocolStackName, p_interfaceName, p_driverName, p_pErrorInfo))
        {
            if(oNewCreated)
            {
                m_VirtualDeviceList.push_back(pVirtualDevice);
            }

            oResult = TRUE;
        }
        else
        {
            if(oNewCreated)
            {
                delete pVirtualDevice;
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_InitVirtualGatewayDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_gatewayDeviceName, CStdString p_gatewayProtocolStackName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oNewCreated(FALSE);
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice(0);

    //Get virtual device
    if(GetVirtualDevice(p_virtualDeviceName, pVirtualDevice, oNewCreated, p_pErrorInfo) && pVirtualDevice)
    {
        //Init virtual device
        if(pVirtualDevice->InitVirtualGatewayDevice(p_deviceName, p_protocolStackName, p_gatewayDeviceName, p_gatewayProtocolStackName, p_pErrorInfo))
        {
            if(oNewCreated)
            {
                m_VirtualDeviceList.push_back(pVirtualDevice);
            }

            oResult = TRUE;
        }
        else
        {
            if(oNewCreated)
            {
                delete pVirtualDevice;
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_InitBaudrateSelection(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CStdDWordArray& p_ulBaudrateSel, CErrorInfo* p_pErrorInfo)
{
    CVirtualDeviceBase* pVirtualDevice = 0;
    BOOL oResult(TRUE);

    if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice))
    {
        if(pVirtualDevice)
        {
            if(!pVirtualDevice->InitBaudrateSelection(p_deviceName, p_protocolStackName, p_interfaceName, p_ulBaudrateSel))
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                oResult = FALSE;
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_InitDefaultProtocolStackSettings(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, DWORD p_ulBaudrate, DWORD p_ulTimeout, CErrorInfo* p_pErrorInfo)
{
    CVirtualDeviceBase* pVirtualDevice = 0;
    BOOL oResult(TRUE);

    if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice))
    {
        if(pVirtualDevice)
        {
            if(!pVirtualDevice->InitDefaultProtocolStackSettings(p_deviceName, p_protocolStackName, p_interfaceName, p_ulBaudrate, p_ulTimeout))
            {
                if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                oResult = FALSE;
            }
        }
    }

    return oResult;
}

void CVirtualCommandSet_Manager::DeleteVirtualDeviceList()
{
    for(list<CVirtualDeviceBase*>::iterator it = m_VirtualDeviceList.begin(); it != m_VirtualDeviceList.end(); it++)
    {
        CVirtualDeviceBase* pVirtualDevice = (*it);
        delete pVirtualDevice;
    }

    m_VirtualDeviceList.clear();
}

BOOL CVirtualCommandSet_Manager::FindCorrectManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CDeviceCommandSetManagerBase** p_ppManager, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
    {
        if(pVirtualDevice->FindCorrectManager(p_deviceName, p_ppManager))
        {
            oResult = TRUE;
        }
        else
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_DEVICE_NAME, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_VIRTUAL_DEVICE_NAME, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::FindVirtualDevice(CStdString p_virtualDeviceName, CVirtualDeviceBase** p_ppVirtualDevice)
{
    BOOL oResult(FALSE);

    for(list<CVirtualDeviceBase*>::iterator it = m_VirtualDeviceList.begin(); it != m_VirtualDeviceList.end(); it++)
    {
        CVirtualDeviceBase* pVirtualDevice = (*it);
        CStdString strName;
        if(pVirtualDevice && pVirtualDevice->VCS_GetVirtualDeviceName(&strName))
        {
            if(p_virtualDeviceName.CompareNoCase(strName) == 0)
            {
                if(p_ppVirtualDevice) *p_ppVirtualDevice = pVirtualDevice;

                oResult = TRUE;
                break;
            }
        }
    }

    return oResult;
}

HANDLE CVirtualCommandSet_Manager::VCS_OpenVirtualDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CStdString p_portName, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle;
    HANDLE hReturn = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Find associated manager
        if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, p_pErrorInfo) && pManager)
        {
            //Device command set handle
            hDCS_Handle = pManager->DCS_OpenDevice(p_deviceName, p_protocolStackName, p_interfaceName, p_portName, p_pErrorInfo);
            if(hDCS_Handle)
            {
                if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
                {
                    if(pVirtualDevice->InitGateway(p_deviceName))
                    {
                        //Device command set handle
                        hReturn = m_pHandleRegistrationMap_VCS->RegisterHandle(pVirtualDevice, pManager, hDCS_Handle);
                    }
                }

                if(hReturn == 0)
                {
                    pManager->DCS_CloseDevice(hDCS_Handle);

                    if(m_pErrorHandling)
                    {
                        m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    }
                }
            }
        }
    }
    else
    {
        if(m_pErrorHandling)
        {
            m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
        }
    }

    return hReturn;
}

BOOL CVirtualCommandSet_Manager::VCS_CloseVirtualDevice(HANDLE p_hVCS_Handle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CMmcSingleLock lock(&m_Sync, TRUE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CHandleRegistration_VCS* pRegistration = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Close only if no other handles are registered
        if(oResult && m_pHandleRegistrationMap_VCS->IsLastRegistration(p_hVCS_Handle, hDCS_Handle))
        {
            //Close Device
            if(!pManager->DCS_CloseDevice(hDCS_Handle, p_pErrorInfo))
            {
                if(pRegistration) pRegistration->DoUnlock();
                oResult = FALSE;
            }
        }
        //Delete registration
        if(oResult && !m_pHandleRegistrationMap_VCS->DeleteMapRegistration(p_hVCS_Handle))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_CloseAllVirtualDevices(CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    HANDLE hVCSHandle;
    tHandleRegistrationVCS_List* pRegistrationList;

    if(m_pHandleRegistrationMap_VCS)
    {
        if(m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            oResult = TRUE;
            list<CHandleRegistration_VCS*>::iterator it = pRegistrationList->begin();
            while(it != pRegistrationList->end())
            {
                CHandleRegistration_VCS* pRegistration = (*it);
                if(pRegistration != 0)
                {
                    hVCSHandle = pRegistration->GetKeyHandle();
                    if(hVCSHandle)
                    {
                        if(!VCS_CloseVirtualDevice(hVCSHandle, p_pErrorInfo))
                        {
                            oResult = FALSE;
                            break;
                        }
                    }
                    pRegistrationList->remove(pRegistration);
                    delete pRegistration;
                }

                it = pRegistrationList->begin();
            }

            if(oResult)
            {
                pRegistrationList->clear();
            }
        }
    }

    return oResult;
}

HANDLE CVirtualCommandSet_Manager::VCS_OpenVirtualDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, HANDLE p_hPS_BaseLayerHandle, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle;
    HANDLE hReturn = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Find associated manager
        if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, p_pErrorInfo) && pManager)
        {
            //Device command set handle
            hDCS_Handle = pManager->DCS_OpenDevice(p_deviceName, p_hPS_BaseLayerHandle, p_pErrorInfo);
            if(hDCS_Handle)
            {
                if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
                {
                    if(pVirtualDevice->InitGateway(p_deviceName))
                    {
                        //Device command set handle
                        hReturn = m_pHandleRegistrationMap_VCS->RegisterHandle(pVirtualDevice, pManager, hDCS_Handle);
                    }
                }

                if(hReturn == 0)
                {
                    pManager->DCS_CloseDevice(hDCS_Handle, p_hPS_BaseLayerHandle);
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                }
            }
        }
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    }

    return hReturn;
}

BOOL CVirtualCommandSet_Manager::VCS_CloseVirtualDevice(HANDLE p_hVCS_Handle, HANDLE p_hPS_BaseLayerHandle, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CHandleRegistration_VCS* pRegistration = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Close device
        if(oResult && !pManager->DCS_CloseDevice(hDCS_Handle, p_hPS_BaseLayerHandle, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Delete registration
        if(oResult && !m_pHandleRegistrationMap_VCS->DeleteMapRegistration(p_hVCS_Handle))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_CloseAllVirtualDevices(HANDLE p_hPS_BaseLayerHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    tHandleRegistrationVCS_List* pRegistrationList;
    HANDLE hVCSHandle;

    if(m_pHandleRegistrationMap_VCS)
    {
        if(m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            oResult = TRUE;
            tHandleRegistrationVCS_List::iterator it = pRegistrationList->begin();
            while(it != pRegistrationList->end())
            {
                CHandleRegistration_VCS* pRegistration = (*it);
                if(pRegistration)
                {
                    hVCSHandle = pRegistration->GetKeyHandle();
                    if(hVCSHandle)
                    {
                        if(!VCS_CloseVirtualDevice(hVCSHandle, p_hPS_BaseLayerHandle, p_pErrorInfo))
                        {
                            oResult = FALSE;
                            break;
                        }
                    }

                    pRegistrationList->remove(pRegistration);

                    //Delete registration
                    delete pRegistration;

                    it = pRegistrationList->begin();
                }
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_AreAllVirtualDevicesClosed()
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = m_pHandleRegistrationMap_VCS->IsRegistrationMapEmpty();
    }

    return oResult;
}

HANDLE CVirtualCommandSet_Manager::VCS_OpenVirtualDevice(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, HANDLE p_hVirtualCommandSet, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);

    BOOL oResult(TRUE);
    CDeviceCommandSetManagerBase* pManager(0);
    CVirtualDeviceBase* pVirtualDevice(0);
    CVirtualDeviceBase* pVirtualDeviceRegistered(0);
    HANDLE hDeviceCommandSetRegistered(0);
    HANDLE hReturn(0);

    if(m_pHandleRegistrationMap_VCS)
    {
        if(oResult && !FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, p_pErrorInfo) && pManager)
        {
            oResult = FALSE;
        }

        if(oResult && !m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVirtualCommandSet, &pVirtualDeviceRegistered, &pManager, &hDeviceCommandSetRegistered))
        {
            oResult = FALSE;
        }

        if(oResult && !FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
        {
            oResult = FALSE;
        }

        if(oResult)
        {
            hReturn = OpenDevice(pVirtualDevice, pManager, p_deviceName, p_protocolStackName, hDeviceCommandSetRegistered, p_pErrorInfo);
        }
    }
    else
    {
        if(m_pErrorHandling)
        {
            m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
        }
    }

    return hReturn;
}

BOOL CVirtualCommandSet_Manager::VCS_RemoveAllVirtualDevices(HANDLE p_hVirtualCommandSet, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    BOOL oResult(FALSE);
    CHandleRegistration_VCS* pRegistration(0);
    list<HANDLE> ahVirtualCommandSet;
    HANDLE hVirtualCommandSet;

    if (m_pHandleRegistrationMap_VCS != 0)
    {
        if (m_pHandleRegistrationMap_VCS->GetRegistration(p_hVirtualCommandSet, &pRegistration))
        {
            oResult = TRUE;
            ahVirtualCommandSet = GetGatewayHandles(p_hVirtualCommandSet);
            for (list<HANDLE>::iterator it = ahVirtualCommandSet.begin(); it != ahVirtualCommandSet.end(); it++)
            {
                hVirtualCommandSet = (*it);

                if (hVirtualCommandSet)
                {
                    if (!VCS_CloseVirtualDevice(hVirtualCommandSet, p_pErrorInfo))
                    {
                        oResult = FALSE;
                        break;
                    }
                }
            }
        }
        else
        {
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::SetParameter(EParameterType p_eParameterType, HANDLE p_hVCS_Handle, CStdString p_Name, BYTE* p_pValue, DWORD p_ulSize, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = m_pHandleRegistrationMap_VCS->SetParameter(p_eParameterType, p_hVCS_Handle, p_Name, p_pValue, p_ulSize, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::SetParameter(EParameterType p_eParameterType, HANDLE p_hVCS_Handle, CStdString p_Name, CStdString p_Value, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = m_pHandleRegistrationMap_VCS->SetParameter(p_eParameterType, p_hVCS_Handle, p_Name, p_Value, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetParameter(EParameterType p_eParameterType, HANDLE p_hVCS_Handle, CStdString p_Name, BYTE* p_pValue, DWORD p_ulSize, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = m_pHandleRegistrationMap_VCS->GetParameter(p_eParameterType, p_hVCS_Handle, p_Name, p_pValue, p_ulSize, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetParameter(EParameterType p_eParameterType, HANDLE p_hVCS_Handle, CStdString p_Name, CStdString& p_rValue, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = m_pHandleRegistrationMap_VCS->GetParameter(p_eParameterType, p_hVCS_Handle, p_Name, p_rValue, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetKeyHandle(CStdString p_strVirtualDeviceName, CStdString p_strDeviceName, CStdString p_strProtocolStackName, CStdString p_strInterfaceName, CStdString p_strPortName, HANDLE* p_pVCSKeyHandle, CErrorInfo* p_pErrorInfo)
{
    tHandleRegistrationVCS_List* pRegistrationList = 0;

    CStdString strVirtualDeviceNameReg;
    CStdString strDeviceNameReg;
    CStdString strProtocolStackNameReg;
    CStdString strInterfaceNameReg;
    CStdString strPortNameReg;

    BOOL oVirtualDeviceNameFound = FALSE;
    BOOL oDeviceNameFound = FALSE;
    BOOL oProtocolStackNameFound = FALSE;
    BOOL oInterfaceNameFound = FALSE;
    BOOL oPortNameFound = FALSE;

    CVirtualDeviceBase* pVirtualDevice = 0;
    CDeviceCommandSetManagerBase* pManager = 0;
    HANDLE hDeviceCommandSetHandle = 0;
    HANDLE hKeyHandle = 0;
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;

        if(m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            for(tHandleRegistrationVCS_List::iterator it = pRegistrationList->begin(); it != pRegistrationList->end(); it++)
            {
                CHandleRegistration_VCS* pRegistration = (*it);

                pRegistration->GetRegistrationValues(&pVirtualDevice, &pManager, &hDeviceCommandSetHandle);

                if(!pVirtualDevice || !pManager)
                {
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    oResult = FALSE;
                    break;
                }

                //Compare virtual deviceName
                if(pVirtualDevice->VCS_GetVirtualDeviceName(&strVirtualDeviceNameReg))
                {
                    if(p_strVirtualDeviceName.CompareNoCase(strVirtualDeviceNameReg) == 0)
                    {
                        oVirtualDeviceNameFound = TRUE;

                        //Compare DeviceName
                        if(pManager->DCS_GetDeviceName(hDeviceCommandSetHandle, &strDeviceNameReg))
                        {
                            if(p_strDeviceName.CompareNoCase(strDeviceNameReg) == 0)
                            {
                                oDeviceNameFound = TRUE;

                                //Compare ProtocolStackName
                                if(pManager->DCS_GetProtocolStackName(hDeviceCommandSetHandle, &strProtocolStackNameReg))
                                {
                                    if(p_strProtocolStackName.CompareNoCase(strProtocolStackNameReg) == 0)
                                    {
                                        oProtocolStackNameFound = TRUE;

                                        //Compare InterfaceName
                                        if(pManager->DCS_GetInterfaceName(hDeviceCommandSetHandle, &strInterfaceNameReg))
                                        {
                                            if(p_strInterfaceName.CompareNoCase(strInterfaceNameReg) == 0)
                                            {
                                                oInterfaceNameFound = TRUE;

                                                //Compare PortName
                                                if(pManager->DCS_GetPortName(hDeviceCommandSetHandle, &strPortNameReg))
                                                {
                                                    if(p_strPortName.CompareNoCase(strPortNameReg) == 0)
                                                    {
                                                        oPortNameFound = TRUE;
                                                        hKeyHandle = pRegistration->GetKeyHandle();
                                                        break;
                                                    }
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        //Error handling
        if(oResult && hKeyHandle == 0)
        {
            if(m_pErrorHandling)
            {
                if(!oVirtualDeviceNameFound)        m_pErrorHandling->GetError(ERROR_BAD_VIRTUAL_DEVICE_NAME, p_pErrorInfo);
                else if(!oDeviceNameFound)          m_pErrorHandling->GetError(ERROR_BAD_DEVICE_NAME, p_pErrorInfo);
                else if(!oProtocolStackNameFound)   m_pErrorHandling->GetError(ERROR_BAD_PROTOCOL_STACK_NAME, p_pErrorInfo);
                else if(!oInterfaceNameFound)       m_pErrorHandling->GetError(ERROR_BAD_INTERFACE_NAME, p_pErrorInfo);
                else if(!oPortNameFound)            m_pErrorHandling->GetError(ERROR_BAD_PORT_NAME, p_pErrorInfo);
                else if(!hKeyHandle)                m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            }
        }
    }

    if(p_pVCSKeyHandle)
    {
        *p_pVCSKeyHandle = hKeyHandle;
    }
    if(oResult)
    {
        oResult = (hKeyHandle != 0);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetKeyHandle(CStdString p_strVirtualDeviceName, CStdString p_strDeviceName, CLayerParameterStack& p_rAddressParameter, HANDLE* p_pVCSKeyHandle, CErrorInfo* p_pErrorInfo)
{
    tHandleRegistrationVCS_List* pRegistrationList = 0;

    CStdString virtualDeviceNameReg;
    CStdString deviceNameReg;

    BOOL oVirtualDeviceNameFound(FALSE);
    BOOL oDeviceNameFound(FALSE);
    BOOL oAddressSettingsFound(FALSE);

    CVirtualDeviceBase* pVirtualDevice(0);
    CDeviceCommandSetManagerBase* pManager(0);
    HANDLE hDeviceCommandSetHandle(0);
    HANDLE hKeyHandle(0);
    BYTE ubNodeIdLayer(0);
    BYTE ubNodeId(0);
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        oResult = TRUE;

        if(m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            for(tHandleRegistrationVCS_List::iterator it = pRegistrationList->begin(); it != pRegistrationList->end(); it++)
            {
                CHandleRegistration_VCS* pRegistration = (*it);
                pRegistration->GetRegistrationValues(&pVirtualDevice, &pManager, &hDeviceCommandSetHandle);

                if(!pVirtualDevice || !pManager)
                {
                    if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                    oResult = FALSE;
                    break;
                }

                //Compare virtual device name
                if(pVirtualDevice->VCS_GetVirtualDeviceName(&virtualDeviceNameReg))
                {
                    if(p_strVirtualDeviceName.CompareNoCase(virtualDeviceNameReg) == 0)
                    {
                        oVirtualDeviceNameFound = TRUE;

                        //Compare device name
                        if(pManager->DCS_GetDeviceName(hDeviceCommandSetHandle, &deviceNameReg))
                        {
                            if(p_strDeviceName.CompareNoCase(deviceNameReg) == 0)
                            {
                                oDeviceNameFound = TRUE;

                                if(p_rAddressParameter.GetLayerParameter(DEVICE_COMMAND_SET_LAYER, _T("NodeId"), &ubNodeId, sizeof(ubNodeId)))
                                {
                                    if(pManager->GetParameter(PT_DEVICE, hDeviceCommandSetHandle, _T("NodeId"), &ubNodeIdLayer, sizeof(ubNodeIdLayer), p_pErrorInfo))
                                    {
                                        if(ubNodeId == ubNodeIdLayer)
                                        {
                                            oAddressSettingsFound = TRUE;
                                            hKeyHandle = pRegistration->GetKeyHandle();
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        //Error handling
        if(oResult && hKeyHandle == 0)
        {
            if(m_pErrorHandling)
            {
                if(!oVirtualDeviceNameFound)    m_pErrorHandling->GetError(ERROR_BAD_VIRTUAL_DEVICE_NAME, p_pErrorInfo);
                else if(!oDeviceNameFound)      m_pErrorHandling->GetError(ERROR_BAD_DEVICE_NAME, p_pErrorInfo);
                else if(!oAddressSettingsFound) m_pErrorHandling->GetError(ERROR_BAD_ADDRESS_PARAMETER, p_pErrorInfo);
                else if(!hKeyHandle)            m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                else;
            }
        }
    }

    if(p_pVCSKeyHandle)
    {
        *p_pVCSKeyHandle = hKeyHandle;
    }
    if(oResult)
    {
        oResult = (hKeyHandle != 0);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetGatewayKeyHandle(HANDLE p_hVCS_Handle, HANDLE* p_phVCS_GatewayHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager(0);
    CHandleRegistration_VCS* pRegistration(0);
    CVirtualDeviceBase* pVirtualDevice(0);
    HANDLE hHandle(0);
    HANDLE hDCS_GatewayHandle(0);

    if (m_pHandleRegistrationMap_VCS && p_phVCS_GatewayHandle)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if (oResult && !pRegistration->DoLock())
        {
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if (oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hHandle) || (pManager == 0) || (pVirtualDevice == 0)))
        {
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if (pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get Gateway KeyHandle
        if (oResult && !pManager->DCS_GetGatewayKeyHandle(hHandle, &hDCS_GatewayHandle))
        {
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if (pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }

        //Get Parent KeyHandle
        if (oResult)
        {
            *p_phVCS_GatewayHandle = GetParentVcsHandle(hDCS_GatewayHandle);
            oResult = (*p_phVCS_GatewayHandle != 0);
        }

        //Unlock
        if (oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetBaseLayerKeyHandle(HANDLE p_hVCS_Handle, HANDLE* p_phPS_BaseLayerKeyHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager(0);
    CHandleRegistration_VCS* pRegistration(0);
    CVirtualDeviceBase* pVirtualDevice(0);
    HANDLE hHandle(0);

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hHandle) || (pManager == 0) || (pVirtualDevice == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get BaseLayer ProtocolStack KeyHandle
        if(oResult && !pManager->DCS_GetBaseLayerKeyHandle(hHandle, p_phPS_BaseLayerKeyHandle))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetVirtualDeviceName(HANDLE p_hVCS_Handle, CStdString* p_pVirtualDeviceName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pVirtualDevice == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get virtual device name
        if(oResult && !pVirtualDevice->VCS_GetVirtualDeviceName(p_pVirtualDeviceName))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetDeviceName(HANDLE p_hVCS_Handle, CStdString* p_pDeviceName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }

        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }

        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get device name
        if(oResult && !pManager->DCS_GetDeviceName(hDCS_Handle, p_pDeviceName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetProtocolStackName(HANDLE p_hVCS_Handle, CStdString* p_pProtocolStackName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get protocol stack name
        if(oResult && !pManager->DCS_GetProtocolStackName(hDCS_Handle, p_pProtocolStackName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetInterfaceName(HANDLE p_hVCS_Handle, CStdString* p_pInterfaceName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get device name
        if(oResult && !pManager->DCS_GetInterfaceName(hDCS_Handle, p_pInterfaceName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetPortName(HANDLE p_hVCS_Handle, CStdString* p_pPortName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Get device name
        if(oResult && !pManager->DCS_GetPortName(hDCS_Handle, p_pPortName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetVirtualDeviceNameSelection(CStdStringArray* p_pVirtualDeviceNameSel, CErrorInfo* p_pErrorInfo)
{
    CMmcSingleLock lock(&m_Sync, TRUE);
    BOOL oResult(FALSE);
    CStdString strNameSel;

    if(p_pVirtualDeviceNameSel)
    {
        p_pVirtualDeviceNameSel->clear();

        for(list<CVirtualDeviceBase*>::iterator it = m_VirtualDeviceList.begin(); it != m_VirtualDeviceList.end(); it++)
        {
            CVirtualDeviceBase* pVirtualDevice = (*it);
            if(pVirtualDevice->VCS_GetVirtualDeviceName(&strNameSel))
            {
                p_pVirtualDeviceNameSel->push_back(strNameSel);
            }
        }

        oResult = TRUE;
    }
    else
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_NULL_POINTER, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetDeviceNameSelection(CStdString p_virtualDeviceName, CStdStringArray* p_pDeviceNameSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }

    //Get devicename selection
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_GetDeviceNameSelection(p_pDeviceNameSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetProtocolStackNameSelection(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdStringArray* p_pProtocolStackNameSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }

    //Get protocol stack name selection
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_GetProtocolStackNameSelection(p_deviceName, p_pProtocolStackNameSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetInterfaceNameSelection(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdStringArray* p_pInterfaceNameSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }

    //Get interface name selection
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_GetInterfaceNameSelection(p_deviceName, p_protocolStackName, p_pInterfaceNameSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetPortNameSelection(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CStdStringArray* p_pPortSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }

    //Get port name election
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_GetPortNameSelection(p_deviceName, p_protocolStackName, p_interfaceName, p_pPortSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetBaudrateSelection(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CStdString p_portName, CStdDWordArray* p_pulBaudrateSel, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }

    //Get baudrate selection
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_GetBaudrateSelection(p_deviceName, p_protocolStackName, p_interfaceName, p_portName, p_pulBaudrateSel, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_ResetPortEnumeration(CStdString p_strVirtualDeviceName, CStdString p_strDeviceName, CStdString p_strProtocolStackName, CStdString p_strInterfaceName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(TRUE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    //Find virtual device
    if(!FindVirtualDevice(p_strVirtualDeviceName, &pVirtualDevice) || (pVirtualDevice == 0))
    {
        if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
        oResult = FALSE;
    }
    //Enumeration
    if(oResult)
    {
        oResult = pVirtualDevice->VCS_ResetPortEnumeration(p_strDeviceName, p_strProtocolStackName, p_strInterfaceName, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetProtocolStackSettings(HANDLE p_hVCS_Handle, DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //GetProtocolStackSettings
        if(oResult && !pManager->DCS_GetProtocolStackSettings(hDCS_Handle, p_pulBaudrate, p_pulTimeout, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_SetProtocolStackSettings(HANDLE p_hVCS_Handle, DWORD p_ulBaudrate, DWORD p_ulTimeout, BOOL p_oChangeOnly, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Try to write gateway timeout
        if (oResult && pVirtualDevice)
        {
            pVirtualDevice->SetGatewayDeviceTimeout(pManager, hDCS_Handle, p_ulTimeout);
        }
        //SetProtocolStackSettings
        if(oResult && !pManager->DCS_SetProtocolStackSettings(hDCS_Handle, p_ulBaudrate, p_ulTimeout, p_oChangeOnly, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration)
        {
            pRegistration->DoUnlock();
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetCommands(HANDLE p_Handle, ELayer p_eLayer, CStdString* p_pCommandInfo)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;
    CDeviceCommandSetManagerBase* pManager = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;

        //Get registration entries
        if(!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pVirtualDevice == 0))
        {
            oResult = FALSE;
        }

        if(oResult)
        {
            if(CheckLayer(p_eLayer))
            {
                //Get commands
                oResult = pVirtualDevice->GetCommands(p_pCommandInfo);
            }
            else
            {
                //Forward get commands to next layer
                oResult = pManager->GetCommands(hDCS_Handle, p_eLayer, p_pCommandInfo);
            }
        }
    }

    return oResult;
}

void CVirtualCommandSet_Manager::InitJournalManager(CJournalManagerBase* p_pJournalManager)
{
    for(list<CVirtualDeviceBase*>::iterator it = m_VirtualDeviceList.begin();it != m_VirtualDeviceList.begin();it++)
    {
        CVirtualDeviceBase* pVirtualDevice = (*it);

        if(pVirtualDevice != 0)
        {
            pVirtualDevice->InitJournalManager(p_pJournalManager);
        }
    }
}

void CVirtualCommandSet_Manager::ResetJournalManager()
{
    for(list<CVirtualDeviceBase*>::iterator it = m_VirtualDeviceList.begin();it != m_VirtualDeviceList.begin();it++)
    {
        CVirtualDeviceBase* pVirtualDevice = (*it);
        if(pVirtualDevice != 0)
        {
            pVirtualDevice->ResetJournalManager();
        }
    }
}

BOOL CVirtualCommandSet_Manager::InitJournalManager(HANDLE p_hHandle, CJournalManagerBase* p_pJournalManager)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;
    CDeviceCommandSetManagerBase* pManager = 0;
    CHandleRegistration_VCS* pRegistration = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Get registration
        if(m_pHandleRegistrationMap_VCS->GetRegistration(p_hHandle, &pRegistration))
        {
            //Lock
            if(pRegistration && !pRegistration->IsLocked())
            {
                pRegistration->DoLock();
            }

            //Get registration entries
            if(m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hHandle, &pVirtualDevice, &pManager, &hDCS_Handle))
            {
                if(pVirtualDevice) pVirtualDevice->InitJournalManager(p_pJournalManager);
                if(pManager) pManager->InitJournalManager(hDCS_Handle, p_pJournalManager);

                oResult = TRUE;
            }

            //Unlock
            if(pRegistration)
            {
                pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::ResetJournalManager(HANDLE p_hHandle)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;
    CDeviceCommandSetManagerBase* pManager = 0;
    CHandleRegistration_VCS* pRegistration = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Get registration
        if(m_pHandleRegistrationMap_VCS->GetRegistration(p_hHandle, &pRegistration))
        {
            //Lock
            if(pRegistration && !pRegistration->IsLocked())
            {
                pRegistration->DoLock();
            }

            //Get registration entries
            if(m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hHandle, &pVirtualDevice, &pManager, &hDCS_Handle))
            {
                if(pVirtualDevice) pVirtualDevice->ResetJournalManager();
                if(pManager) pManager->ResetJournalManager(hDCS_Handle);

                oResult = TRUE;
            }

            //Unlock
            if(pRegistration)
            {
                pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::ExecuteCommand(CCommandRoot* p_pCommand, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oKeepLock = 0;
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;
    CHandleRegistration_VCS* pRegistration = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;

        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hHandle, &pRegistration))) || (pRegistration == 0))
        {
            if(p_pCommand)
            {
                errorInfo.Init(ERROR_HANDLE_NOT_VALID);
                p_pCommand->SetStatus(oResult, &errorInfo);
            }
            oResult = FALSE;
        }

        //Lock (Only if registration is not locked from last command)
        if(oResult && !pRegistration->IsLocked())
        {
            oResult = pRegistration->DoLock();
        }

        //Get registration entries
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hHandle, &pVirtualDevice, &pManager, &hDCS_Handle))) ||(pVirtualDevice == 0))
        {
            oResult = FALSE;
        }

        if(oResult)
        {
            if(CheckLayer(p_pCommand))
            {
                //Execute command
                oResult = pVirtualDevice->ExecuteCommand(p_pCommand, pManager, hDCS_Handle, p_hTransactionHandle);

                //Check lock state
                oKeepLock = pVirtualDevice->IsLocked();
            }
            else
            {
                //Forward command to next layer
                oResult = pManager->ExecuteCommand(p_pCommand, hDCS_Handle, p_hTransactionHandle);

                //Check lock state
                oKeepLock = p_pCommand->IsSubLayerLocked();
            }
        }

        //Unlock (Only if virtual device does not remain locked until next command)
        if(!oKeepLock)
        {
            //Unlock
            if(pRegistration) pRegistration->DoUnlock();
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::AbortCommands(HANDLE p_hHandle, BOOL p_oActive)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    HANDLE hDCS_Handle = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;
        //Get registration
        if ((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hHandle, &pRegistration))) || (pRegistration == 0))
        {
            oResult = FALSE;
        }
        //Get registration entries
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hHandle, &pVirtualDevice, &pManager, &hDCS_Handle))) ||(pVirtualDevice == 0))
        {
            oResult = FALSE;
        }
        //Abort commands
        if(oResult && !pVirtualDevice->AbortCommands(p_oActive))
        {
            oResult = FALSE;
        }
        //Close only if no other handles are registered
        if(oResult && m_pHandleRegistrationMap_VCS->IsLastRegistration(p_hHandle, hDCS_Handle))
        {
            //Forward AbortCommands to next layer
            if(!pManager->AbortCommands(hDCS_Handle, p_oActive))
            {
                oResult = FALSE;
            }
        }
        //Abort lock
        if(oResult) pRegistration->AbortLock(p_oActive);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetLayerManager(HANDLE p_Handle, ELayer p_eLayer, CLayerManagerBase** p_ppLayerManager, HANDLE* p_pLayerHandle)
{
    CDeviceCommandSetManagerBase* pManager = 0;
    HANDLE hDCS_Handle = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    CHandleRegistration_VCS* pRegistration = 0;
    BOOL oResult(FALSE);

    if(m_pHandleRegistrationMap_VCS)
    {
        if(CheckLayer(p_eLayer))
        {
            if(p_ppLayerManager) *p_ppLayerManager = this;
            if(p_pLayerHandle) *p_pLayerHandle = p_Handle;
            oResult = TRUE;
        }
        else
        {
            //Init
            oResult = TRUE;
            //Get registration
            if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_Handle, &pRegistration))) || (pRegistration == 0))
            {
                oResult = FALSE;
            }
            //Lock
            if(oResult && !pRegistration->DoLock())
            {
                oResult = FALSE;
            }
            //Get registration entries
            if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
            {
                //Unlock
                pRegistration->DoUnlock();
                oResult = FALSE;
            }
            if(oResult)
            {
                //Forward to next layer
                oResult = pManager->GetLayerManager(hDCS_Handle, p_eLayer, p_ppLayerManager, p_pLayerHandle);

                //Unlock
                pRegistration->DoUnlock();
            }
        }
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::SetInterfaceManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CInterfaceManagerBase* p_pInterfaceManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->SetInterfaceManager(p_deviceName, p_protocolStackName, p_pInterfaceManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetInterfaceManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, CInterfaceManagerBase** p_ppInterfaceManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->GetInterfaceManager(p_deviceName, p_protocolStackName, p_interfaceName, p_ppInterfaceManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::RemoveInterfaceManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CInterfaceManagerBase* p_pInterfaceManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->RemoveInterfaceManager(p_deviceName, p_protocolStackName, p_pInterfaceManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::SetProtocolStackManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CProtocolStackManagerBase* p_pProtocolStackManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->SetProtocolStackManager(p_deviceName, p_pProtocolStackManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetProtocolStackManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CProtocolStackManagerBase** p_ppProtocolStackManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->GetProtocolStackManager(p_deviceName, p_protocolStackName, p_ppProtocolStackManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::RemoveProtocolStackManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CProtocolStackManagerBase* p_pProtocolStackManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->RemoveProtocolStackManager(p_deviceName, p_pProtocolStackManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::SetDeviceCommandSetManager(CStdString p_virtualDeviceName, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
    {
        oResult = pVirtualDevice->AddManager(p_pDeviceCommandSetManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::GetDeviceCommandSetManager(CStdString p_virtualDeviceName, CStdString p_deviceName, CDeviceCommandSetManagerBase** p_ppDeviceCommandSetManager)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        if(p_ppDeviceCommandSetManager) *p_ppDeviceCommandSetManager = pManager;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::RemoveDeviceCommandSetManager(CStdString p_virtualDeviceName, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager)
{
    BOOL oResult(FALSE);
    CVirtualDeviceBase* pVirtualDevice = 0;

    if(FindVirtualDevice(p_virtualDeviceName, &pVirtualDevice) && pVirtualDevice)
    {
        oResult = pVirtualDevice->RemoveManager(p_pDeviceCommandSetManager);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::VCS_GetDefaultProtocolStackSettings(CStdString p_virtualDeviceName, CStdString p_deviceName, CStdString p_protocolStackName, CStdString p_interfaceName, DWORD* p_pulBaudrate, DWORD* p_pulTimeout, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CErrorInfo errorInfo;

    //Find associated manager
    if(FindCorrectManager(p_virtualDeviceName, p_deviceName, &pManager, &errorInfo) && pManager)
    {
        oResult = pManager->DCS_GetDefaultProtocolStackSettings(p_deviceName, p_protocolStackName, p_interfaceName, p_pulBaudrate, p_pulTimeout, p_pErrorInfo);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className = "VirtualCommandSetManager";

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(VIRTUAL_COMMAND_SET_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::EnableTracing(HANDLE p_hVCS_Handle, CStdString p_TracingFileName, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;

        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Enable tracing
        if(oResult && !pManager->EnableTracing(hDCS_Handle, p_TracingFileName, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

BOOL CVirtualCommandSet_Manager::DisableTracing(HANDLE p_hVCS_Handle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CDeviceCommandSetManagerBase* pManager = 0;
    CVirtualDeviceBase* pVirtualDevice = 0;
    HANDLE hDCS_Handle = 0;
    CHandleRegistration_VCS* pRegistration = 0;

    if(m_pHandleRegistrationMap_VCS)
    {
        //Init
        oResult = TRUE;

        //Get registration
        if((oResult && (!m_pHandleRegistrationMap_VCS->GetRegistration(p_hVCS_Handle, &pRegistration))) || (pRegistration == 0))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Lock
        if(oResult && !pRegistration->DoLock())
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_HANDLE_NOT_VALID, p_pErrorInfo);
            oResult = FALSE;
        }
        //Get registration values
        if(oResult && (!m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVCS_Handle, &pVirtualDevice, &pManager, &hDCS_Handle) || (pManager == 0)))
        {
            if(m_pErrorHandling) m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Disable tracing
        if(oResult && !pManager->DisableTracing(hDCS_Handle, p_pErrorInfo))
        {
            if(pRegistration) pRegistration->DoUnlock();
            oResult = FALSE;
        }
        //Unlock
        if(oResult && pRegistration) pRegistration->DoUnlock();
    }

    return oResult;
}

list<HANDLE> CVirtualCommandSet_Manager::GetGatewayHandles(HANDLE p_hVirtualCommandSet)
{
    CDeviceCommandSetManagerBase* pManager(0);
    CVirtualDeviceBase* pVirtualDevice(0);
    HANDLE hDeviceCommandSet(0);
    HANDLE hVirtualCommandSet(0);
    HANDLE hVirtualCommandSetGateway(0);
    list<HANDLE> ahVirtualCommandSet;
    tHandleRegistrationVCS_List* pRegistrationList;

    if(m_pHandleRegistrationMap_VCS)
    {
        if(m_pHandleRegistrationMap_VCS->GetRegistrationValues(p_hVirtualCommandSet, &pVirtualDevice, &pManager, &hVirtualCommandSetGateway) && hVirtualCommandSetGateway)
        {
            if(m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
            {
                for(list<CHandleRegistration_VCS*>::iterator it = pRegistrationList->begin(); it != pRegistrationList->end(); it++)
                {
                    CHandleRegistration_VCS* pHandleRegistration = (*it);

                    hVirtualCommandSet = pHandleRegistration->GetKeyHandle();

                    if(hVirtualCommandSet != p_hVirtualCommandSet)
                    {
                        if(m_pHandleRegistrationMap_VCS->GetRegistrationValues(hVirtualCommandSet, &pVirtualDevice, &pManager, &hDeviceCommandSet) && hDeviceCommandSet)
                        {
                            if(pManager->IsGatewayDevice(hVirtualCommandSetGateway, hDeviceCommandSet))
                            {
                                ahVirtualCommandSet.push_back(hVirtualCommandSet);
                            }
                        }
                    }
                }
            }
        }
    }

    return ahVirtualCommandSet;
}

HANDLE CVirtualCommandSet_Manager::GetParentVcsHandle(HANDLE p_hDCS_Handle)
{
    HANDLE hParentVcsHandle(0);
    CDeviceCommandSetManagerBase* pManager(0);
    CVirtualDeviceBase* pVirtualDevice(0);
    HANDLE hDeviceCommandSet(0);
    tHandleRegistrationVCS_List* pRegistrationList;

    if (m_pHandleRegistrationMap_VCS)
    {
        if (m_pHandleRegistrationMap_VCS->GetRegistrationList(&pRegistrationList) && pRegistrationList)
        {
            for (list<CHandleRegistration_VCS*>::iterator it = pRegistrationList->begin(); it != pRegistrationList->end(); it++)
            {
                CHandleRegistration_VCS* pHandleRegistration = (*it);

                if (pHandleRegistration->GetRegistrationValues(&pVirtualDevice, &pManager, &hDeviceCommandSet))
                {
                    if (p_hDCS_Handle == hDeviceCommandSet)
                    {
                        hParentVcsHandle = pHandleRegistration->GetKeyHandle();
                        break;
                    }
                }
            }
        }
    }

    return hParentVcsHandle;
}

HANDLE CVirtualCommandSet_Manager::OpenDevice(CVirtualDeviceBase* p_pVirtualDevice, CDeviceCommandSetManagerBase* p_pDeviceCommandSetManager, CStdString p_deviceName, CStdString p_protocolStackName, HANDLE p_hDeviceCommandSet, CErrorInfo* p_pErrorInfo)
{
    HANDLE hDeviceCommandSet(0);
    HANDLE hReturn(0);

    if(m_pHandleRegistrationMap_VCS && p_pVirtualDevice && p_pDeviceCommandSetManager)
    {
        hDeviceCommandSet = p_pDeviceCommandSetManager->DCS_OpenDevice(p_deviceName, p_protocolStackName, p_hDeviceCommandSet, p_pErrorInfo);
        if(hDeviceCommandSet)
        {
            if(p_pVirtualDevice->InitGateway(p_deviceName))
            {
                hReturn = m_pHandleRegistrationMap_VCS->RegisterHandle(p_pVirtualDevice, p_pDeviceCommandSetManager, hDeviceCommandSet);
            }

            if(hReturn == 0)
            {
                p_pDeviceCommandSetManager->DCS_CloseDevice(hDeviceCommandSet);

                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                }
            }
        }
    }

    return hReturn;
}
