#include "stdafx.h"

#include <Process/MmcProcess.h>
#include <Thread/MmcCriticalSection.h>
#include <Thread/MmcSingleLock.h>
#include <Drv/Helper/UsbTree/UsbHostController.h>
#include <Drv/Helper/UsbTree/UsbDevices.h>
#include <Drv/Helper/UsbTree/UsbDevice.h>
#include <Drv/Helper/UsbTree/UsbDeviceBase.h>
#include <Drv/Helper/UsbTree/UsbTreeFactory.h>

#include <Drv/Helper/UsbTree/UsbTree.h>

using namespace std;

CUsbTree CUsbTree::m_Singleton;

CUsbTree::CUsbTree(void)
    : m_ulInhibitStartCount(0)
{
    InitInhibitTime();
}

CUsbTree::~CUsbTree(void)
{
    Lock();
    DeleteHostControllers();
    Unlock();
}

CUsbTree* CUsbTree::GetInstance()
{
    return &m_Singleton;
}

void CUsbTree::InitInhibitTime()
{
    m_ulInhibitStartCount = 0;
}

void CUsbTree::ResetInhibitTime()
{
    m_ulInhibitStartCount = MmcGetTickCount();
}

BOOL CUsbTree::IsInhibitTimeElapsed()
{
    const DWORD INHIBIT_TIME = 100;

    BOOL oResult(TRUE);
    DWORD ulTime(0);

    if(m_ulInhibitStartCount > 0)
    {
        ulTime = MmcGetTickCount() - m_ulInhibitStartCount;
        if(ulTime <= INHIBIT_TIME)
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CUsbTree::Load()
{
    BOOL oResult(TRUE);

    if(IsInhibitTimeElapsed())
    {
        oResult = LoadHostControllers();

        ResetInhibitTime();
    }

    return oResult;
}

BOOL CUsbTree::Reset()
{
    Lock();

    InitInhibitTime();

    Unlock(); 

    return TRUE;
}

BOOL CUsbTree::LoadHostControllers()
{
    BOOL oResult(FALSE);
    CUsbTreeFactory factory;

    Lock();
    DeleteHostControllers();

    if(factory.LoadHostControllers(m_UsbHostControllers))
    {
        oResult = TRUE;
    }

    Unlock();

    return oResult;
}

void CUsbTree::DeleteHostControllers()
{
    for(list<CUsbHostController*>::iterator it = m_UsbHostControllers.begin(); it != m_UsbHostControllers.end(); it++)
    {
        CUsbHostController* pHostController = *(it);
        delete pHostController;
    }

    m_UsbHostControllers.clear();
}

BOOL CUsbTree::GetPortChain(CStdString p_devicePath, CStdString& p_rPortChain)
{
    BOOL oResult(FALSE);

    Lock();

    for(list<CUsbHostController*>::iterator it = m_UsbHostControllers.begin(); it != m_UsbHostControllers.end(); it++)
    {
        CUsbHostController* pHostController = *(it);
        if(pHostController->GetPortChain(p_devicePath, p_rPortChain))
        {
            oResult = TRUE;
            break;
        }
    }

    Unlock();

    return oResult;
}

BOOL CUsbTree::GetPortChain(WORD p_usVendorId, WORD p_usProductId, CStdString p_serialNumber, CStdString& p_rPortChain)
{
    BOOL oResult(FALSE);

    Lock();

    for(list<CUsbHostController*>::iterator it = m_UsbHostControllers.begin(); it != m_UsbHostControllers.end(); it++)
    {
        CUsbHostController* pHostController = *(it);
        if(pHostController->GetPortChain(p_usVendorId, p_usProductId, p_serialNumber, p_rPortChain))
        {
            oResult = TRUE;
            break;
        }
    }

    Unlock();

    return oResult;
}

BOOL CUsbTree::GetPortChains(WORD p_usVendorId, WORD p_usProductId, CStdStringArray& p_rPortChains)
{
    BOOL oResult(TRUE);

    Lock();

    p_rPortChains.clear();
    for(list<CUsbHostController*>::iterator it = m_UsbHostControllers.begin(); it != m_UsbHostControllers.end(); it++)
    {
        CUsbHostController* pHostController = *(it);
        if(!pHostController->GetPortChains(p_usVendorId, p_usProductId, p_rPortChains))
        {
            oResult = FALSE;
        }
    }

    Unlock();

    return oResult;
}

BOOL CUsbTree::GetInfos(CStdString p_portChain, CStdString& p_rDevicePath, WORD& p_rusVendorId, WORD& p_rusProductId, CStdString& p_rSerialNumber)
{
    BOOL oResult(FALSE);

    Lock();

    for(list<CUsbHostController*>::iterator it = m_UsbHostControllers.begin(); it != m_UsbHostControllers.end(); it++)
    {
        CUsbHostController* pHostController = *(it);
        if(pHostController->GetInfos(p_portChain, p_rDevicePath, p_rusVendorId, p_rusProductId, p_rSerialNumber))
        {
            oResult = TRUE;
            break;
        }
    }

    Unlock();

    return oResult;
}

void CUsbTree::Lock()
{
    m_CriticalSection.Lock(-1, _T("UsbTree"));
}

void CUsbTree::Unlock()
{
    m_CriticalSection.Unlock(_T("UsbTree"));
}