#define _XOPEN_SOURCE 700

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include "StdString.h"
#include <list>
#include <ftw.h>
#include "../../../../include/Drv/Ftd2xx/ftd2xx.h"
#include "../../../../include/Drv/Helper/UsbTree/LinuxUsbDefinitions.h"
#include "../../../../include/Drv/Helper/UsbTree/LinuxRootHub.h"
#include <Drv/Helper/UsbApi/UsbApiLinux.h>
#include <Thread/MmcSingleLock.h>

#include "../../../../include/Drv/Helper/UsbTree/LinuxUsbTree.h"

#ifndef USE_FDS
#define USE_FDS 15
#endif

std::list<CLinuxUsbDirectory*> CLinuxUsbTree::m_Directories;
CMmcCriticalSection CLinuxUsbTree::m_Sync;

CLinuxUsbTree::CLinuxUsbTree()
{
}

CLinuxUsbTree::~CLinuxUsbTree()
{
	UsbFreeEnumeration(m_DeviceInfoList);
	DeleteRootHubEnumeration();
}

void CLinuxUsbTree::DeleteDirectories()
{
    std::list<CLinuxUsbDirectory*>::iterator it=m_Directories.begin();
    
    while(it!=m_Directories.end())
    {
        CLinuxUsbDirectory* pDir = *it;        
        delete pDir;
        m_Directories.erase(it);
        it=m_Directories.begin();
    }
}

void CLinuxUsbTree::AddDirectory(const char *p_pName, dirent *p_pEntry)
{
	const int MAX_DIR_LENGTH = 1024;
	char path[MAX_DIR_LENGTH];
	int len = 0;

	path[len] = 0;

	if(p_pName && p_pEntry)
	{
		len = snprintf(path, sizeof(path)-1, "%s/%s", p_pName, p_pEntry->d_name);

		if (strcmp(p_pEntry->d_name, ".") != 0 && strcmp(p_pEntry->d_name, "..") != 0)
		{
			m_Directories.push_back(new CLinuxUsbDirectory(path));

			EnumerateDirectory(path);
		}
	}
}

void CLinuxUsbTree::AddLinkedDirectory(const char *p_pName, dirent *p_pEntry)
{
	char   *pTarget = 0;
	const int MAX_DIR_LENGTH = 1024;
	char path[MAX_DIR_LENGTH];
	size_t  maxlen = MAX_DIR_LENGTH-1;
	ssize_t len;

	if(p_pName && p_pEntry)
	{
		len = snprintf(path, sizeof(path)-1, "%s/%s", p_pName, p_pEntry->d_name);

		path[len] = 0;

		while (1)
		{
			pTarget = new char[maxlen + 1];

			if (pTarget == NULL)
			{
				break;
			}

			len = readlink(path, pTarget, maxlen);

			if (len == (ssize_t)-1)
			{
				delete []pTarget;
				break;
			}

			if (len >= (ssize_t)maxlen)
			{
				delete []pTarget;
				maxlen += 1024;
				continue;
			}

			pTarget[len] = '\0';
			break;
		}

		m_Directories.push_back(new CLinuxUsbDirectory(path, pTarget));

		delete []pTarget;

		EnumerateDirectory(pTarget);
	}
}

int CLinuxUsbTree::AddDirectory(const char *p_pFilePath, const struct stat *p_pStatInfo, const int p_lTypeFlag, struct FTW *p_pPathInfo)
{
    if(p_lTypeFlag == FTW_SL)
    {        
        char   *target;
        size_t  maxlen = 1023;
        ssize_t len;

        while (1) {

            target = (char*)malloc(maxlen + 1);
            if (target == NULL)
			{
                return ENOMEM;
			}

            len = readlink(p_pFilePath, target, maxlen);
			
            if (len == (ssize_t)-1) 
			{
                const int saved_errno = errno;
                free(target);
                return saved_errno;
            }
            if (len >= (ssize_t)maxlen) 
			{
                free(target);
                maxlen += 1024;
                continue;
            }

            target[len] = '\0';
            break;
        }

        m_Directories.push_back(new CLinuxUsbDirectory(p_pFilePath, target));
        
        free(target);
    }
    if (p_lTypeFlag == FTW_D || p_lTypeFlag == FTW_DP)
    {   
        m_Directories.push_back(new CLinuxUsbDirectory(p_pFilePath));
    }
    
    return 0;
}

void CLinuxUsbTree::EnumerateDirectory(const char *p_pName)
{
	nftw(p_pName, AddDirectory, USE_FDS, FTW_PHYS | FTW_DEPTH);
}

int CLinuxUsbTree::EnumerateDirectoryTree(CStdString p_DirectoryRoot)
{
    BOOL oResult = TRUE;

    DeleteDirectories();

    EnumerateDirectory(p_DirectoryRoot.c_str());

    return oResult;
}

bool CLinuxUsbTree::EnumerateDevices()
{
    bool oResult(false);

	if(UsbEnumerate(0, 0, m_DeviceInfoList))
	{
		oResult = true;
	}

	return oResult;
}

bool CLinuxUsbTree::DirectorySort(CLinuxUsbDirectory* p_pLeft, CLinuxUsbDirectory* p_pRight) 
{ 
    CStdString leftTargetFileName = p_pLeft->GetTargetFileName();
    CStdString leftFileName = p_pLeft->GetFileName();
    CStdString rightTargetFileName = p_pRight->GetTargetFileName();
    CStdString rightFileName = p_pRight->GetFileName();
    bool oResult = true;
    
    if(leftTargetFileName.GetLength() > 0 && rightTargetFileName.GetLength()>0)
    {       
        leftFileName = leftTargetFileName;
        rightFileName = rightTargetFileName;
    }
        
    if(leftFileName.GetLength() == rightFileName.GetLength())
    {
        if(leftFileName < rightFileName)
        {
            oResult = false;
        }
    }
    else
    {
        if(leftFileName.GetLength() > rightFileName.GetLength())
        {
            oResult = false;
        }
    }   
    
    return oResult;
}

void CLinuxUsbTree::DeleteRootHubEnumeration()
{
	std::list<CLinuxRootHub*>::iterator it = m_RootHubs.begin();

	if(it != m_RootHubs.end())
	{
		CLinuxRootHub* pRootHub = (*it);

		delete pRootHub;

		m_RootHubs.erase(it);

		it = m_RootHubs.begin();
	}

	DeleteDirectories();
}

bool CLinuxUsbTree::GetRootHubs(std::list<CLinuxRootHub*>& p_rRootHubs)
{
	bool oResult = false;

	oResult = EnumerateRootHubs();

	if(oResult)
	{
		p_rRootHubs = m_RootHubs;
	}

	return oResult;
}

bool CLinuxUsbTree::DeviceExists(int p_lBusNum, int p_lPortNum, CStdString p_DevicePath, int p_lIdProduct, int p_lIdVendor)
{
	bool oResult = false;

	for(std::list<CLinuxRootHub*>::iterator it = m_RootHubs.begin(); it!=m_RootHubs.end(); it++)
	{
		CLinuxRootHub* pRootHub = (*it);

		oResult = pRootHub->DeviceExists(p_lBusNum, p_lPortNum, p_DevicePath, p_lIdProduct, p_lIdVendor);

		if(oResult)
		{
			break;
		}
	}

	return oResult;
}

bool CLinuxUsbTree::EnumerateRootHubs()
{
    bool oResult = false;
    CMmcSingleLock lock(&m_Sync, TRUE);
    
    EnumerateDirectoryTree(SYS_BUS_USB_DEVICES);
    
    m_Directories.sort(DirectorySort);
    
    EnumerateDevices();

    for(std::list<CLinuxUsbDirectory*>::iterator it = m_Directories.begin(); it!=m_Directories.end(); it++)
    {
        CLinuxUsbDirectory* pDirectory = (*it);
        
        if(!pDirectory->IsParsed() && CLinuxRootHub::IsRootHubDirectory(pDirectory->GetFileName()))
        {        
            CLinuxRootHub* pRootHub = new CLinuxRootHub();
            
            pRootHub->Init(this, m_Directories, m_DeviceInfoList, pDirectory);
            
            if(pRootHub->Load())
            {            
            	m_RootHubs.push_back(pRootHub);
                
                oResult = true;
            }
        }
    }

    return oResult;
}
