
#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

/* C */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <locale.h>
#include <errno.h>

/* Unix */
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/utsname.h>
#include <fcntl.h>
#include <pthread.h>
#include <wchar.h>

#include <../../Linux/include/Drv/Ftd2xx/ftd2xx.h>
#include <iconv.h>

#include "../../../include/Drv/Hid/HidApiLinux.h"
#include <Drv/Helper/UsbApi/UsbApiLinux.h>

#ifdef DEBUG_PRINTF
#define LOG(...) fprintf(stderr, __VA_ARGS__)
#else
#define LOG(...) do {} while (0)
#endif

#ifndef __FreeBSD__
#define DETACH_KERNEL_DRIVER
#endif

#include "../../../include/Drv/Hid/HidApiLinux.h"

CHidDevice *CHidApi::NewDevice()
{
    CHidDevice *dev = new CHidDevice();

    dev->blocking = 1;

    pthread_mutex_init(&dev->mutex, NULL);
    pthread_cond_init(&dev->condition, NULL);
    pthread_barrier_init(&dev->barrier, NULL, 2);

    return dev;
}

void CHidApi::FreeDevice(CHidDevice *dev)
{
    pthread_barrier_destroy(&dev->barrier);
    pthread_cond_destroy(&dev->condition);
    pthread_mutex_destroy(&dev->mutex);

    delete dev;
}

int CHidApi::Init()
{
    return UsbInit();
}

int CHidApi::Exit()
{
	return UsbExit();
}

bool CHidApi::Enumerate(unsigned short vendor_id, unsigned short product_id, std::list<CHidDeviceDescription*>& p_rDeviceInfoList)
{
    libusb_device **devs;
    libusb_device *dev;
    libusb_device_handle *handle;
    ssize_t num_devs;
    int i = 0;
    bool oResult = true;

    if(Init() < 0)
    {
    	oResult = false;
    }

    num_devs = libusb_get_device_list(UsbContext, &devs);

    if (num_devs < 0)
    {
    	oResult = false;
    }
    while ((dev = devs[i++]) != NULL)
    {
        struct libusb_device_descriptor desc;
        struct libusb_config_descriptor *conf_desc = NULL;
        int j, k;
        int interface_num = 0;

        int res = libusb_get_device_descriptor(dev, &desc);

        unsigned short dev_vid = desc.idVendor;
        unsigned short dev_pid = desc.idProduct;

        res = libusb_get_active_config_descriptor(dev, &conf_desc);

        if (res < 0)
        {
            libusb_get_config_descriptor(dev, 0, &conf_desc);
        }

        if (conf_desc)
        {
            for (j = 0; j < conf_desc->bNumInterfaces; j++)
            {
                const struct libusb_interface *intf = &conf_desc->interface[j];

                for (k = 0; k < intf->num_altsetting; k++)
                {
                    const struct libusb_interface_descriptor *intf_desc;
                    intf_desc = &intf->altsetting[k];

                    if (intf_desc->bInterfaceClass == LIBUSB_CLASS_HID)
                    {
                        interface_num = intf_desc->bInterfaceNumber;

                        /* Check the VID/PID against the arguments */
                        if ((vendor_id == 0x0 || vendor_id == dev_vid) &&
                            (product_id == 0x0 || product_id == dev_pid))
                        {
                        	CHidDeviceDescription *pDeviceInfo = new CHidDeviceDescription();

                            pDeviceInfo->path = MakePath(dev, interface_num);

                            res = libusb_open(dev, &handle);

                            if (res >= 0)
                            {
                                if (desc.iSerialNumber > 0)
                                {
                                	pDeviceInfo->serial_number_string =
                                        GetUsbString(handle, desc.iSerialNumber);
                                }


                                if (desc.iManufacturer > 0)
                                {
                                	pDeviceInfo->manufacturer_string =
                                			GetUsbString(handle, desc.iManufacturer);
                                }

                                if (desc.iProduct > 0)
                                {
                                	pDeviceInfo->product_string =
                                			GetUsbString(handle, desc.iProduct);
                                }

                                pDeviceInfo->serial_number = desc.iSerialNumber;

                                libusb_close(handle);
                            }

                            pDeviceInfo->vendor_id = dev_vid;
                            pDeviceInfo->product_id = dev_pid;

                            pDeviceInfo->release_number = desc.bcdDevice;
                            pDeviceInfo->interface_class = intf_desc->bInterfaceClass;
                            pDeviceInfo->interface_number = interface_num;

                            p_rDeviceInfoList.push_back(pDeviceInfo);
                        }
                    }
                }
            }

            libusb_free_config_descriptor(conf_desc);
        }
    }

    libusb_free_device_list(devs, 1);

    return oResult;
}



void CHidApi::FreeEnumeration(std::list<CHidDeviceDescription*>& p_rDeviceInfoList)
{
	std::list<CHidDeviceDescription*>::iterator it = p_rDeviceInfoList.begin();

	while(it!=p_rDeviceInfoList.end())
	{
		CHidDeviceDescription* pHidDeviceInfo = (*it);

		free(pHidDeviceInfo->path);
		free(pHidDeviceInfo->serial_number_string);
		free(pHidDeviceInfo->manufacturer_string);
		free(pHidDeviceInfo->product_string);

		delete pHidDeviceInfo;

		p_rDeviceInfoList.erase(it);

		it = p_rDeviceInfoList.begin();
	}
}

CHidDevice * CHidApi::Open(unsigned short vendor_id, unsigned short product_id, const wchar_t *serial_number)
{
    const char *path_to_open = NULL;
    CHidDevice *handle = NULL;
    std::list<CHidDeviceDescription*> deviceInfoList;

    if(Enumerate(vendor_id, product_id, deviceInfoList))
    {
    	for(std::list<CHidDeviceDescription*>::iterator it = deviceInfoList.begin(); it != deviceInfoList.end(); it++)
		{
			CHidDeviceDescription* pDeviceInfo = (*it);

			if(pDeviceInfo->vendor_id == vendor_id && pDeviceInfo->product_id)
			{
				if (serial_number)
				{
					if (pDeviceInfo->serial_number_string &&
						wcscmp(serial_number, pDeviceInfo->serial_number_string) == 0)
					{
						path_to_open = pDeviceInfo->path;
						break;
					}
				}
				else
				{
					path_to_open = pDeviceInfo->path;
					break;
				}
			}
		}

		if (path_to_open)
		{
			handle = OpenPath(path_to_open);
		}

		FreeEnumeration(deviceInfoList);
    }

    return handle;
}

void CHidApi::ReadCallback(struct libusb_transfer *transfer)
{
    CHidDevice *dev = (CHidDevice *)transfer->user_data;
    int res;

    if (transfer->status == LIBUSB_TRANSFER_COMPLETED) {

        CInputReport *rpt = (CInputReport *)malloc(sizeof(*rpt));

        rpt->data = (uint8_t *)malloc(transfer->actual_length);

        memcpy(rpt->data, transfer->buffer, transfer->actual_length);
        rpt->len = transfer->actual_length;
        rpt->next = NULL;

        pthread_mutex_lock(&dev->mutex);


        if (dev->input_reports == NULL)
        {
            dev->input_reports = rpt;

            pthread_cond_signal(&dev->condition);
        }
        else
        {
            CInputReport *cur = dev->input_reports;
            int num_queued = 0;
            while (cur->next != NULL) {
                cur = cur->next;
                num_queued++;
            }
            cur->next = rpt;

            if (num_queued > 30)
            {
                ReturnData(dev, NULL, 0);
            }
        }
        pthread_mutex_unlock(&dev->mutex);
    }
    else if (transfer->status == LIBUSB_TRANSFER_CANCELLED) {
        dev->shutdown_thread = 1;
        dev->cancelled = 1;
        return;
    }
    else if (transfer->status == LIBUSB_TRANSFER_NO_DEVICE) {
        dev->shutdown_thread = 1;
        dev->cancelled = 1;
        return;
    }
    else if (transfer->status == LIBUSB_TRANSFER_TIMED_OUT) {
        //LOG("Timeout (normal)\n");
    }
    else {
        LOG("Unknown transfer code: %d\n", transfer->status);
    }

    /* Re-submit the transfer object. */
    res = libusb_submit_transfer(transfer);
    if (res != 0) {
        LOG("Unable to submit URB. libusb error code: %d\n", res);
        dev->shutdown_thread = 1;
        dev->cancelled = 1;
    }
}


void *CHidApi::ReadThread(void *p_pParam)
{
    CHidDevice *dev = (CHidDevice *)p_pParam;
    unsigned char *buf;
    const size_t length = dev->input_ep_max_packet_size;

    buf = (unsigned char *)malloc(length);

    dev->transfer = libusb_alloc_transfer(0);

    libusb_fill_interrupt_transfer(dev->transfer,
        dev->device_handle,
        dev->input_endpoint,
        buf,
        length,
        ReadCallback,
        dev,
        5000/*timeout*/);

    /* Make the first submission. Further submissions are made
       from inside read_callback() */
    libusb_submit_transfer(dev->transfer);

    /* Notify the main thread that the read thread is up and running. */
    pthread_barrier_wait(&dev->barrier);

    /* Handle all the events. */
    while (!dev->shutdown_thread) {
        int res;
        res = libusb_handle_events(UsbContext);
        if (res < 0) {
            /* There was an error. */
            LOG("read_thread(): libusb reports error # %d\n", res);

            /* Break out of this loop only on fatal error.*/
            if (res != LIBUSB_ERROR_BUSY &&
                res != LIBUSB_ERROR_TIMEOUT &&
                res != LIBUSB_ERROR_OVERFLOW &&
                res != LIBUSB_ERROR_INTERRUPTED) {
                break;
            }
        }
    }

    /* Cancel any transfer that may be pending. This call will fail
       if no transfers are pending, but that's OK. */
    libusb_cancel_transfer(dev->transfer);

    while (!dev->cancelled)
        libusb_handle_events_completed(UsbContext, &dev->cancelled);

    /* Now that the read thread is stopping, Wake any threads which are
       waiting on data (in hid_read_timeout()). Do this under a mutex to
       make sure that a thread which is about to go to sleep waiting on
       the condition actually will go to sleep before the condition is
       signaled. */
    pthread_mutex_lock(&dev->mutex);
    pthread_cond_broadcast(&dev->condition);
    pthread_mutex_unlock(&dev->mutex);

    /* The dev->transfer->buffer and dev->transfer objects are cleaned up
       in hid_close(). They are not cleaned up here because this thread
       could end either due to a disconnect or due to a user
       call to hid_close(). In both cases the objects can be safely
       cleaned up after the call to pthread_join() (in hid_close()), but
       since hid_close() calls libusb_cancel_transfer(), on these objects,
       they can not be cleaned up here. */

    LOG("read thread completed!");

    return NULL;
}


CHidDevice* CHidApi::OpenPath(const char *path)
{
    CHidDevice *dev = NULL;

    libusb_device **devs;
    libusb_device *usb_dev;
    int res;
    int d = 0;
    int good_open = 0;

    if(Init() < 0)
        return NULL;

    dev = NewDevice();

    libusb_get_device_list(UsbContext, &devs);

    while ((usb_dev = devs[d++]) != NULL)
    {
        struct libusb_device_descriptor desc;
        struct libusb_config_descriptor *conf_desc = NULL;
        int i,j,k;
        libusb_get_device_descriptor(usb_dev, &desc);

        if (libusb_get_active_config_descriptor(usb_dev, &conf_desc) < 0)
            continue;
        for (j = 0; j < conf_desc->bNumInterfaces; j++) {
            const struct libusb_interface *intf = &conf_desc->interface[j];
            for (k = 0; k < intf->num_altsetting; k++) {
                const struct libusb_interface_descriptor *intf_desc;
                intf_desc = &intf->altsetting[k];
                if (intf_desc->bInterfaceClass == LIBUSB_CLASS_HID) {
                    char *dev_path = MakePath(usb_dev, intf_desc->bInterfaceNumber);
                    if (!strcmp(dev_path, path)) {
                        /* Matched Paths. Open this device */

                        /* OPEN HERE */
                        res = libusb_open(usb_dev, &dev->device_handle);
                        if (res < 0) {
                            LOG("can't open device\n");
                            free(dev_path);
                            break;
                        }
                        good_open = 1;
#ifdef DETACH_KERNEL_DRIVER
                        /* Detach the kernel driver, but only if the
                           device is managed by the kernel */
                        if (libusb_kernel_driver_active(dev->device_handle, intf_desc->bInterfaceNumber) == 1) {
                            res = libusb_detach_kernel_driver(dev->device_handle, intf_desc->bInterfaceNumber);
                            if (res < 0) {
                                libusb_close(dev->device_handle);
                                LOG("Unable to detach Kernel Driver\n");
                                free(dev_path);
                                good_open = 0;
                                break;
                            }
                        }
#endif
                        res = libusb_claim_interface(dev->device_handle, intf_desc->bInterfaceNumber);
                        if (res < 0) {
                            LOG("can't claim interface %d: %d\n", intf_desc->bInterfaceNumber, res);
                            free(dev_path);
                            libusb_close(dev->device_handle);
                            good_open = 0;
                            break;
                        }

                        /* Store off the string descriptor indexes */
                        dev->manufacturer_index = desc.iManufacturer;
                        dev->product_index      = desc.iProduct;
                        dev->serial_index       = desc.iSerialNumber;

                        /* Store off the interface number */
                        dev->interface = intf_desc->bInterfaceNumber;

                        /* Find the INPUT and OUTPUT endpoints. An
                           OUTPUT endpoint is not required. */
                        for (i = 0; i < intf_desc->bNumEndpoints; i++) {
                            const struct libusb_endpoint_descriptor *ep
                                = &intf_desc->endpoint[i];

                            /* Determine the type and direction of this
                               endpoint. */
                            int is_interrupt =
                                (ep->bmAttributes & LIBUSB_TRANSFER_TYPE_MASK)
                                  == LIBUSB_TRANSFER_TYPE_INTERRUPT;
                            int is_output =
                                (ep->bEndpointAddress & LIBUSB_ENDPOINT_DIR_MASK)
                                  == LIBUSB_ENDPOINT_OUT;
                            int is_input =
                                (ep->bEndpointAddress & LIBUSB_ENDPOINT_DIR_MASK)
                                  == LIBUSB_ENDPOINT_IN;

                            /* Decide whether to use it for input or output. */
                            if (dev->input_endpoint == 0 &&
                                is_interrupt && is_input) {
                                /* Use this endpoint for INPUT */
                                dev->input_endpoint = ep->bEndpointAddress;
                                dev->input_ep_max_packet_size = ep->wMaxPacketSize;
                            }
                            if (dev->output_endpoint == 0 &&
                                is_interrupt && is_output) {
                                /* Use this endpoint for OUTPUT */
                                dev->output_endpoint = ep->bEndpointAddress;
                            }
                        }

                        pthread_create(&dev->thread, NULL, ReadThread, dev);

                        /* Wait here for the read thread to be initialized. */
                        pthread_barrier_wait(&dev->barrier);

                    }
                    free(dev_path);
                }
            }
        }
        libusb_free_config_descriptor(conf_desc);

    }

    libusb_free_device_list(devs, 1);

    /* If we have a good handle, return it. */
    if (good_open) {
        return dev;
    }
    else {
        /* Unable to open any devices. */
        FreeDevice(dev);
        return NULL;
    }
}

int CHidApi::Write(CHidDevice *dev, const unsigned char *data, size_t length)
{
    int res;
    int report_number = data[0];
    int skipped_report_id = 0;
    int timeout = 30000; /*timeout millis*/

    if (report_number == 0x0)
    {
        data++;
        length--;
        skipped_report_id = 1;
    }

    if (dev->output_endpoint <= 0)
    {
        /* No interrupt out endpoint. Use the Control Endpoint */
        res = libusb_control_transfer(dev->device_handle,
            LIBUSB_REQUEST_TYPE_CLASS|LIBUSB_RECIPIENT_INTERFACE|LIBUSB_ENDPOINT_OUT,
            0x09/*HID Set_Report*/,
            (2/*HID output*/ << 8) | report_number,
            dev->interface,
            (unsigned char *)data, length,
            timeout);

        if (res < 0)
        {
        	LOG( "libusb_control_transfer error, error code = %d\n", res);
        	return -1;
        }

        if (skipped_report_id)
            length++;

        return length;
    }
    else {
        /* Use the interrupt out endpoint */
        int actual_length;
        res = libusb_interrupt_transfer(dev->device_handle,
            dev->output_endpoint,
            (unsigned char*)data,
            length,
            &actual_length, timeout);

        if (res < 0)
            return -1;

        if (skipped_report_id)
            actual_length++;

        return actual_length;
    }
}

int CHidApi::ReturnData(CHidDevice *dev, unsigned char *data, size_t length)
{
    CInputReport *rpt = dev->input_reports;

    size_t len = (length < rpt->len)? length: rpt->len;

    if (len > 0)
    {
        memcpy(data, rpt->data, len);
    }

    dev->input_reports = rpt->next;

    free(rpt->data);
    free(rpt);

    return len;
}

void CHidApi::CleanupMutex(void *p_pParam)
{
    CHidDevice *dev = (CHidDevice *)p_pParam;

    pthread_mutex_unlock(&dev->mutex);
}


int CHidApi::ReadTimeout(CHidDevice *dev, unsigned char *data, size_t length, int milliseconds)
{
    int bytes_read = -1;

    pthread_mutex_lock(&dev->mutex);

    pthread_cleanup_push(&CleanupMutex, dev);

    if (dev->input_reports)
    {
        bytes_read = ReturnData(dev, data, length);
        goto ret;
    }

    if (dev->shutdown_thread)
    {
        bytes_read = -1;
        goto ret;
    }

    if (milliseconds == -1)
    {
        while (!dev->input_reports && !dev->shutdown_thread)
        {
            pthread_cond_wait(&dev->condition, &dev->mutex);
        }
        if (dev->input_reports)
        {
            bytes_read = ReturnData(dev, data, length);
        }
    }
    else if (milliseconds > 0) {

        int res;
        struct timespec ts;
        clock_gettime(CLOCK_REALTIME, &ts);
        ts.tv_sec += milliseconds / 1000;
        ts.tv_nsec += (milliseconds % 1000) * 1000000;

        if (ts.tv_nsec >= 1000000000L)
        {
            ts.tv_sec++;
            ts.tv_nsec -= 1000000000L;
        }

        while (!dev->input_reports && !dev->shutdown_thread)
        {
            res = pthread_cond_timedwait(&dev->condition, &dev->mutex, &ts);

            if (res == 0)
            {
                if (dev->input_reports)
                {
                    bytes_read = ReturnData(dev, data, length);
                    break;
                }
            }
            else if (res == ETIMEDOUT) {
                bytes_read = 0;
                break;
            }
            else
            {
                bytes_read = -1;
                break;
            }
        }
    }
    else
    {
        bytes_read = 0;
    }

ret:
    pthread_mutex_unlock(&dev->mutex);
    pthread_cleanup_pop(0);

    return bytes_read;
}

int CHidApi::Read(CHidDevice *dev, unsigned char *data, size_t length)
{
    return ReadTimeout(dev, data, length, dev->blocking ? -1 : 0);
}

int CHidApi::SetNonblocking(CHidDevice *dev, int nonblock)
{
    dev->blocking = !nonblock;

    return 0;
}

int CHidApi::SendFeatureReport(CHidDevice *dev, const unsigned char *data, size_t length)
{
    int res = -1;
    int skipped_report_id = 0;
    int report_number = data[0];

    if (report_number == 0x0) {
        data++;
        length--;
        skipped_report_id = 1;
    }

    res = libusb_control_transfer(dev->device_handle,
        LIBUSB_REQUEST_TYPE_CLASS|LIBUSB_RECIPIENT_INTERFACE|LIBUSB_ENDPOINT_OUT,
        0x09/*HID set_report*/,
        (3/*HID feature*/ << 8) | report_number,
        dev->interface,
        (unsigned char *)data, length,
        1000/*timeout millis*/);

    if (res < 0)
        return -1;

    /* Account for the report ID */
    if (skipped_report_id)
        length++;

    return length;
}

int CHidApi::GetFeatureReport(CHidDevice *dev, unsigned char *data, size_t length)
{
    int res = -1;
    int skipped_report_id = 0;
    int report_number = data[0];

    if (report_number == 0x0) {
        /* Offset the return buffer by 1, so that the report ID
           will remain in byte 0. */
        data++;
        length--;
        skipped_report_id = 1;
    }
    res = libusb_control_transfer(dev->device_handle,
        LIBUSB_REQUEST_TYPE_CLASS|LIBUSB_RECIPIENT_INTERFACE|LIBUSB_ENDPOINT_IN,
        0x01/*HID get_report*/,
        (3/*HID feature*/ << 8) | report_number,
        dev->interface,
        (unsigned char *)data, length,
        1000/*timeout millis*/);

    if (res < 0)
        return -1;

    if (skipped_report_id)
        res++;

    return res;
}


void CHidApi::Close(CHidDevice *dev)
{
    if (!dev)
        return;

    /* Cause read_thread() to stop. */
    dev->shutdown_thread = 1;
    libusb_cancel_transfer(dev->transfer);

    /* Wait for read_thread() to end. */
    pthread_join(dev->thread, NULL);

    /* Clean up the Transfer objects allocated in read_thread(). */
    free(dev->transfer->buffer);
    libusb_free_transfer(dev->transfer);

    /* release the interface */
    libusb_release_interface(dev->device_handle, dev->interface);

    /* Close the handle */
    libusb_close(dev->device_handle);

    /* Clear out the queue of received reports. */
    pthread_mutex_lock(&dev->mutex);
    while (dev->input_reports) {
        ReturnData(dev, NULL, 0);
    }
    pthread_mutex_unlock(&dev->mutex);

    FreeDevice(dev);
}


int CHidApi::GetManufacturerString(CHidDevice *dev, wchar_t *string, size_t maxlen)
{
    return GetIndexedString(dev, dev->manufacturer_index, string, maxlen);
}

int CHidApi::GetProductString(CHidDevice *dev, wchar_t *string, size_t maxlen)
{
    return GetIndexedString(dev, dev->product_index, string, maxlen);
}

int CHidApi::GetSerialNumberString(CHidDevice *dev, wchar_t *string, size_t maxlen)
{
    return GetIndexedString(dev, dev->serial_index, string, maxlen);
}

int CHidApi::GetIndexedString(CHidDevice *dev, int string_index, wchar_t *string, size_t maxlen)
{
    wchar_t *str;

    str = GetUsbString(dev->device_handle, string_index);

    if (str)
    {
        wcsncpy(string, str, maxlen);
        string[maxlen-1] = L'\0';
        free(str);
        return 0;
    }
    else
        return -1;
}
