#include <memory.h>
#include <Process/MmcProcess.h>
#include <Time/MmcHiResTimer.h>
#include <Drv/Helper/UsbTree/UsbTree.h>
#include <Thread/MmcSingleLock.h>
#include "../../../include/Drv/Hid/HidApiLinux.h"

#include <Drv/Hid/MmcHidHndlLinux.h>

#ifdef DEBUG_PRINTF
#define LOG(...) fprintf(stderr, __VA_ARGS__)
#else
#define LOG(...) do {} while (0)
#endif

CMmcHidHndlBase::CMmcHidHndlBase()
{
}

CMmcHidHndlBase::~CMmcHidHndlBase()
{
}

bool CMmcHidHndlBase::Init()
{
    bool oResult = true;

    CHidApi::Init();

    return oResult;
}

CStdString CMmcHidHndlBase::GetDeviceClassType()
{
    CHidDeviceInfo deviceInfo;

    return deviceInfo.GetClassType();
}

bool CMmcHidHndlBase::GetDeviceInfos(tUsbDeviceInfoList& p_rDeviceInfoList, unsigned short p_vendorId, unsigned short p_productId, BOOL p_oIsHidDevice)
{
	bool oResult(false);
	CMmcSingleLock lock(&m_CriticalSection, TRUE);
	CUsbTree* pUsbTree = CUsbTree::GetInstance();
	CStdStringArray portChains;
	CStdString devicePath(_T(""));
	WORD usVendorId(0);
	WORD usProductId(0);
	CStdString serialNumber(_T(""));
	CUsbDeviceInfo* pDeviceInfo(0);

	if(pUsbTree != 0)
	{
		pUsbTree->Load();

		if(pUsbTree->GetPortChains(p_vendorId, p_productId, portChains))
		{
			for(CStdStringArray::iterator it = portChains.begin(); it != portChains.end(); it++)
			{
				CStdString portChain = *(it);

				//Init
				pUsbTree->GetInfos(portChain, devicePath, usVendorId, usProductId, serialNumber);

				//Add DeviceInfo
				if(p_oIsHidDevice)
				{
					pDeviceInfo = new CHidDeviceInfo();
				}
				else
				{
					pDeviceInfo = new CUsbDeviceInfo();
				}
#ifdef _Logging_UsbEnumeration
				printf("NEW %s device, chain=%s path=%s vendor=%d product=%d serial=%s\n", p_oIsHidDevice > 0 ? "HID" : "USB", portChain.c_str(), devicePath.c_str(), usVendorId, usProductId, serialNumber.c_str());
#endif
				pDeviceInfo->Init(portChain, devicePath, usVendorId, usProductId, serialNumber);

				p_rDeviceInfoList.push_back(pDeviceInfo);
			}

			oResult = true;
		}
	}

	return oResult;
}

bool CMmcHidHndlBase::GetOverlappedResult(DWORD* p_pulNumberOfBytesTransferred, BOOL p_oWait)
{
    return true;
}

bool CMmcHidHndlBase::SetBaudrate(DWORD p_ulBaudrate)
{
    bool oResult = false;

    switch (p_ulBaudrate)
    {
        case USB_HID_BAUDRATE_1M:
            {
                oResult = true;
                break;
            }

        default: oResult = false;
    }

    return oResult;
}

int CMmcHidHndlBase::GetInputReportByteLength()
{
    int lResult = 0;
    CHidDevice *pDevice;

    pDevice = (CHidDevice*)m_hDeviceObject;

    if(m_hDeviceObject)
    {
        lResult = pDevice->input_ep_max_packet_size + 1;
    }

    return lResult;
}

int CMmcHidHndlBase::GetOutputReportByteLength()
{
    int lResult = 0;

    if(m_hDeviceObject)
    {
        lResult = m_hDeviceObject->input_ep_max_packet_size + 1;
    }

    return lResult;
}

bool CMmcHidHndlBase::OpenPort(CHidDeviceInfo& p_rInfo)
{
    bool oResult = false;

    m_hDeviceObject = CHidApi::OpenPath(p_rInfo.GetDevicePath());

    if (m_hDeviceObject)
    {
    	m_Path = p_rInfo.GetDevicePath();

    	CHidApi::SetNonblocking(m_hDeviceObject, 0);

		oResult = true;
    }

    return oResult;
}

bool CMmcHidHndlBase::ClosePort()
{
    bool oResult = false;

    if(m_hDeviceObject)
    {
    	CHidApi::Close(m_hDeviceObject);
        m_hDeviceObject = 0;

        oResult = true;
    }

    return oResult;
}

bool CMmcHidHndlBase::IsValid()
{
    return true;
}

bool CMmcHidHndlBase::IsOpened()
{
    return (m_hDeviceObject != 0);
}

bool CMmcHidHndlBase::PurgeBuffer()
{
    return true;
}

bool CMmcHidHndlBase::WaitForOutReportData(unsigned long p_ulTimeout)
{
    return true;
}

bool CMmcHidHndlBase::CancelIo()
{
	LOG("cancel io");

    return true;
}

unsigned long CMmcHidHndlBase::WaitForOverlapped(unsigned long p_ulTimeout)
{
	LOG("wait for result, timeout = %ld", p_ulTimeout);

    return 0;
}

bool CMmcHidHndlBase::UpdateOutReport()
{
    return true;
}

bool CMmcHidHndlBase::ReadFile(void* p_lpBuffer, DWORD p_nNumberOfBytesToRead, DWORD* p_lpNumberOfBytesRead)
{
    bool oResult = false;
    unsigned char* pBuffer = new unsigned char[p_nNumberOfBytesToRead+1];
    DWORD ulNumberOfBytesRead = 0;

    memset(pBuffer, 0, p_nNumberOfBytesToRead+1);

    if(m_hDeviceObject)
    {
        ulNumberOfBytesRead = CHidApi::Read(m_hDeviceObject, (unsigned char*)pBuffer, p_nNumberOfBytesToRead-1);

        if(ulNumberOfBytesRead == p_nNumberOfBytesToRead - 1)
        {
            //0 report number, library cuts report number byte
            memcpy(((unsigned char*)p_lpBuffer) + 1, pBuffer, ulNumberOfBytesRead);

            *p_lpNumberOfBytesRead = ulNumberOfBytesRead + 1;

            oResult = true;
        }
    }

    delete[] pBuffer;

    return oResult;
}

bool CMmcHidHndlBase::WriteFile(void* p_lpBuffer, DWORD p_nNumberOfBytesToWrite, DWORD *p_lpNumberOfBytesWritten)
{
	bool oResult = false;

    if(m_hDeviceObject && p_lpBuffer)
    {
    	int lNbOfbytesWritten = CHidApi::Write(m_hDeviceObject, (unsigned char*)p_lpBuffer, p_nNumberOfBytesToWrite);

    	if(lNbOfbytesWritten != -1)
    	{
    		*p_lpNumberOfBytesWritten = lNbOfbytesWritten;

    		oResult = (p_nNumberOfBytesToWrite == *p_lpNumberOfBytesWritten);
    	}
    }

    return oResult;
}

CStdString CMmcHidHndlBase::GetLastError()
{
    CStdString lastError;

    return lastError;
}

bool CMmcHidHndlBase::ResetOutReport()
{
    return false;
}
