#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <Thread/MmcCriticalSection.h>
#include <Process/MmcProcess.h>
#include "../../include/Thread/MmcSemaphoreLinux.h"

CMmcSemaphoreLinux::~CMmcSemaphoreLinux()
{
	DestroySemaphore();
}

CMmcSemaphoreLinux::CMmcSemaphoreLinux(int p_lInitialCount, int p_lMaxCount, CStdString p_Name)
	: m_pSemaphore(0)
	, m_Name(p_Name)
	, m_lInitialCount(p_lInitialCount)
{
	CreateSemaphore();
}

void CMmcSemaphoreLinux::CreateSemaphore()
{
	if(m_Name.IsEmpty())
	{
		m_pSemaphore = new sem_t();

		sem_init(m_pSemaphore, 0, m_lInitialCount);
	}
	else
	{
		m_pSemaphore = sem_open(m_Name.c_str(), O_CREAT, 0777, m_lInitialCount);
	}
}

void CMmcSemaphoreLinux::DestroySemaphore()
{
	if(m_pSemaphore)
	{
		if(m_Name.IsEmpty())
		{
			sem_destroy(m_pSemaphore);
			delete m_pSemaphore;
		}
		else
		{
			sem_close(m_pSemaphore);
			sem_unlink(m_Name.c_str());
		}
	}
}

bool CMmcSemaphoreLinux::InitialLock()
{
	bool oResult = false;
	struct timespec abs_time;
	const __time_t MAX_WAIT_TIME_IN_SEC = 10;

	if(m_pSemaphore)
	{
		clock_gettime(CLOCK_REALTIME , &abs_time);

		abs_time.tv_sec += MAX_WAIT_TIME_IN_SEC;

		if (sem_timedwait(m_pSemaphore, &abs_time) != 0)
		{
			oResult = !(sem_post(m_pSemaphore) == 0);
		}
		else
		{
			oResult = true;
		}
	}

	return oResult;
}

bool CMmcSemaphoreLinux::FastLock()
{
    bool oResult = false;

    if(m_pSemaphore)
    {
        oResult = (sem_trywait(m_pSemaphore) == 0);
    }

    return oResult;
}

bool CMmcSemaphoreLinux::TimedLock(unsigned int p_ulTimeout)
{
    bool oResult = false;
	unsigned int ulTimeout = p_ulTimeout;
	const unsigned int MIN_TIMEOUT_IN_MS = 10;
    int lWaitResult = -1;

    if(m_pSemaphore)
    {
        lWaitResult = sem_trywait(m_pSemaphore);

	    while(lWaitResult!=0 && ulTimeout > 0)
	    {
		    lWaitResult = sem_trywait(m_pSemaphore);

            if(lWaitResult == EAGAIN)
            {
		        MmcSleep(MIN_TIMEOUT_IN_MS);
		        ulTimeout -= MIN_TIMEOUT_IN_MS;
            }
            else 
            {
                break;
            }
	    }

        oResult = (lWaitResult == 0);
    }
    
    return oResult;
}

bool CMmcSemaphoreLinux::Lock(unsigned int p_lTimeout)
{
	bool oResult = false;

	if(p_lTimeout != 0)
	{
		oResult = TimedLock(p_lTimeout);
	}
	else
	{
		oResult = FastLock();
	}

	return oResult;
}

bool CMmcSemaphoreLinux::Unlock()
{
	bool oResult = false;

	if(m_pSemaphore)
	{
		oResult = (sem_post(m_pSemaphore) == 0);
	}

	return oResult;
}
