#include "stdafx.h"
#include <bitset>

#include <Classes/MmcDataConversion.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenCommunicationDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsCanOpenDriveDef.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/DeviceCommandSet/DcsEpos4Def.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/ProtocolStack/Command_PS_CANopen.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_CanOpen.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Configuration.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_DataRecorder.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Device.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_ErrorHandling.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Firmware.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Gateway.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_InputsOutputs.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Motion.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_OperationMode.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Parameter.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_Scanning.h>
#include <CommunicationModel/CommonLayer/Classes/Commands/VirtualCommandSet/Command_VCS_StateMachine.h>
#include <CommunicationModel/DeviceCommandSet/BaseClasses/DeviceCommandSetManagerBase.h>
#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/Drive/DataRecorder/RecorderChannelObjects.h>
#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/Drive/DiagnosisHistory/DiagnosisHistory.h>
#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/Drive/Helper/ControllerGainsEpos4.h>
#include <Log/MmcLog/MmcLog.h>
#include <Process/MmcProcess.h>
#include <Time/MmcHiResTimer.h>

#include <CommunicationModel/VirtualCommandSet/Classes/Gateway/Drive/GatewayDriveToEpos4.h>

using namespace std;

// Constants
const CStdString GATEWAY_NODE_ID = _T("GatewayNodeId");
const CStdString NODE_ID = _T("NodeId");
const DWORD RECORDER_BUFFER_SIZE(8000);
const int SAMPLE_SIZE(4);

CGatewayDriveToEpos4::CGatewayDriveToEpos4()
    : m_pDiagnosisHistory(0)
    , m_ulDeviceTimeoutBackup(0)
{
    CreateDiagnosisHistory();
    CreateRecorderChannelObjects(MAX_NB_OF_DATA_RECORDER_VARIABLES);

    InitErrorHandling();
}

CGatewayDriveToEpos4::~CGatewayDriveToEpos4()
{
    DeleteDiagnosisHistory();
}

CGateway* CGatewayDriveToEpos4::Clone()
{
    CGatewayDriveToEpos4* pClonedGateway;

    pClonedGateway = new CGatewayDriveToEpos4();
    *pClonedGateway = *this;

    return (CGatewayVCStoDCS*)pClonedGateway;
}

CGatewayDriveToEpos4& CGatewayDriveToEpos4::operator=(CGatewayDriveToEpos4& p_rOther)
{
    if(this != &p_rOther)
    {
        *((CGatewayVCStoEpos4*)this) = *((CGatewayVCStoEpos4*)&p_rOther);
    }

    return *this;
}

BOOL CGatewayDriveToEpos4::ProcessCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);

    if(CheckLayers(p_pCommand, p_pManager))
    {
        BOOL oLogMessageActive = p_pCommand->IsLogMessageActive();

        LogCmdMessageIn(p_pCommand, oLogMessageActive, _T("CGatewayDriveToEpos4"));

        oResult = ProcessDriveCommand(p_pCommand, p_pManager, p_hHandle, p_hTransactionHandle);

        LogCmdMessageOut(p_pCommand, oLogMessageActive, _T("CGatewayDriveToEpos4"));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ProcessDriveCommand(CCommandRoot* p_pCommand, CLayerManagerBase* p_pManager, HANDLE p_hHandle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CCommand_VCS* pCommand_VCS = (CCommand_VCS*)p_pCommand;
    CDeviceCommandSetManagerBase* pDeviceCommandSetManager = (CDeviceCommandSetManagerBase*)p_pManager;

    if(Lock(p_pCommand))
    {
        //Process command
        switch(p_pCommand->GetCommandId())
        {
            // Controller gains
            case CONF_SET_CONTROLLER_GAIN: oResult = Process_SetControllerGain(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_CONTROLLER_GAIN: oResult = Process_GetControllerGain(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Units
            case CONF_GET_VELOCITY_UNITS: oResult = Process_GetVelocityUnits(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_VELOCITY_UNITS: oResult = Process_SetVelocityUnits(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            //Safety
            case CONF_GET_MAX_FOLLOWING_ERROR: oResult = Process_GetMaxFollowingError(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_MAX_PROFILE_VELOCITY: oResult = Process_GetMaxProfileVelocity(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_MAX_ACCELERATION: oResult = Process_GetMaxAcceleration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_MAX_FOLLOWING_ERROR: oResult = Process_SetMaxFollowingError(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_MAX_PROFILE_VELOCITY: oResult = Process_SetMaxProfileVelocity(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_MAX_ACCELERATION: oResult = Process_SetMaxAcceleration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            //Configuration motor
            case CONF_GET_MOTOR_TYPE: oResult = Process_GetMotorType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_DC_MOTOR_PARAMETER: oResult = Process_GetDcMotorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_EC_MOTOR_PARAMETER: oResult = Process_GetEcMotorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_MOTOR_TYPE: oResult = Process_SetMotorType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_DC_MOTOR_PARAMETER: oResult = Process_SetDcMotorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_EC_MOTOR_PARAMETER: oResult = Process_SetEcMotorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Configuration sensor
            case CONF_GET_SENSOR_TYPE: oResult = Process_GetSensorType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_INC_ENCODER_PARAMETER: oResult = Process_GetIncEncoderParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_HALL_SENSOR_PARAMETER: oResult = Process_GetHallSensorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_SSI_ABS_ENCODER_PARAMETER: oResult = Process_GetSsiAbsEncoderParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_SSI_ABS_ENCODER_PARAMETER_EX: oResult = Process_GetSsiAbsEncoderParameterEx(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_SENSOR_TYPE: oResult = Process_SetSensorType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_INC_ENCODER_PARAMETER: oResult = Process_SetIncEncoderParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_HALL_SENSOR_PARAMETER: oResult = Process_SetHallSensorParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_SSI_ABS_ENCODER_PARAMETER: oResult = Process_SetSsiAbsEncoderParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_SET_SSI_ABS_ENCODER_PARAMETER_EX: oResult = Process_SetSsiAbsEncoderParameterEx(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_MAIN_SENSOR_RESOLUTION: oResult = Process_GetMainSensorResolution(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case CONF_GET_AUXILIARY_SENSOR_RESOLUTION: oResult = Process_GetAuxiliarySensorResolution(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Homing mode
            case MODE_HM_FIND_HOME: oResult = Process_FindHome(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_HM_GET_HOMING_PARAMETER: oResult = Process_GetHomingParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_HM_SET_HOMING_PARAMETER: oResult = Process_SetHomingParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_HM_STOP_HOMING: oResult = Process_StopHoming(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_HM_DEFINE_POSITION: oResult = Process_DefinePosition(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_HM_GET_HOMING_STATE: oResult = Process_GetHomingState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Motion info
            case MOTION_GET_CURRENT_IS: oResult = Process_GetCurrentIs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MOTION_GET_CURRENT_IS_AVERAGED: oResult = Process_GetCurrentIsAveraged(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MOTION_GET_MOVEMENT_STATE: oResult = Process_GetMovementState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MOTION_GET_POSITION_IS: oResult = Process_GetPositionIs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MOTION_GET_VELOCITY_IS: oResult = Process_GetVelocityIs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MOTION_GET_VELOCITY_IS_AVERAGED: oResult = Process_GetVelocityIsAveraged(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Profile position mode
            case MODE_PPM_GET_POSITION_PROFILE: oResult = Process_GetPositionProfile(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PPM_GET_TARGET_POSITION: oResult = Process_GetTargetPosition(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PPM_HALT_POSITION_MOVEMENT: oResult = Process_HaltPositionMovement(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PPM_MOVE_TO_POSITION: oResult = Process_MoveToPosition(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PPM_SET_POSITION_PROFILE: oResult = Process_SetPositionProfile(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Profile velocity mode
            case MODE_PVM_GET_TARGET_VELOCITY: oResult = Process_GetTargetVelocity(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PVM_GET_VELOCITY_PROFILE: oResult = Process_GetVelocityProfile(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PVM_HALT_VELOCITY_MOVEMENT: oResult = Process_HaltVelocityMovement(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PVM_MOVE_WITH_VELOCITY: oResult = Process_MoveWithVelocity(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_PVM_SET_VELOCITY_PROFILE: oResult = Process_SetVelocityProfile(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Current mode
            case MODE_CUM_GET_CURRENT_MUST: oResult = Process_GetCurrentMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_CUM_SET_CURRENT_MUST: oResult = Process_SetCurrentMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Velocity mode
            case MODE_VEM_GET_VELOCITY_MUST: oResult = Process_GetVelocityMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_VEM_SET_VELOCITY_MUST: oResult = Process_SetVelocityMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Position mode
            case MODE_POM_GET_POSITION_MUST: oResult = Process_GetPositionMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_POM_SET_POSITION_MUST: oResult = Process_SetPositionMust(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            //I/O
            case IO_DIGITAL_INPUT_CONFIGURATION: oResult = Process_DigitalInputConfiguration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_DIGITAL_OUTPUT_CONFIGURATION: oResult = Process_DigitalOutputConfiguration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_ANALOG_OUTPUT_CONFIGURATION: oResult = Process_AnalogOutputConfiguration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            case IO_GET_ALL_DIGITAL_INPUTS: oResult = Process_GetAllDigitalInputs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_GET_ALL_DIGITAL_OUTPUTS: oResult = Process_GetAllDigitalOutputs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_GET_ANALOG_INPUT: oResult = Process_GetAnalogInput(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_GET_ANALOG_INPUT_VOLTAGE: oResult = Process_GetAnalogInputVoltage(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_SET_ALL_DIGITAL_OUTPUTS: oResult = Process_SetAllDigitalOutputs(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_SET_ANALOG_OUTPUT: oResult = Process_SetAnalogOutput(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_SET_ANALOG_OUTPUT_VOLTAGE: oResult = Process_SetAnalogOutputVoltage(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case IO_SET_ANALOG_OUTPUT_STATE: oResult = Process_SetAnalogOutputState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // State machine
            case STATE_CLEAR_FAULT: oResult = Process_ClearFault(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_GET_DISABLE_STATE: oResult = Process_GetDisableState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_GET_ENABLE_STATE: oResult = Process_GetEnableState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_GET_FAULT_STATE: oResult = Process_GetFaultState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_GET_QUICK_STOP_STATE: oResult = Process_GetQuickStopState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_SET_DISABLE_STATE: oResult = Process_SetDisableState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_SET_ENABLE_STATE: oResult = Process_SetEnableState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case STATE_SET_QUICK_STOP_STATE: oResult = Process_SetQuickStopState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Operation mode
            case MODE_GET_OPERATION_MODE: oResult = Process_GetOperationMode(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case MODE_SET_OPERATION_MODE: oResult = Process_SetOperationMode(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Error handling
            case EH_RESET_DEVICE: oResult = Process_ResetDevice(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case EH_WAIT_FOR_REBOOT: oResult = Process_WaitForReboot(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case EH_CLEAR_DEVICE_ERRORS: oResult = Process_ClearDeviceErrors(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case EH_GET_NB_OF_DEVICE_ERRORS: oResult = Process_GetNbOfDeviceErrors(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case EH_GET_DEVICE_ERROR_CODE: oResult = Process_GetDeviceErrorCode(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case EH_GOTO_SAFE_STATE: oResult = Process_GotoSafeState(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Version
            case DEVICE_GET_VERSION: oResult = Process_GetVersion(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case DEVICE_GET_TYPE: oResult = Process_GetDeviceType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Persistence
            case PARAMETER_STORE: oResult = Process_Store(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case PARAMETER_RESTORE: oResult = Process_Restore(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Object dictionary
            case VCS_CANOPEN_GET_OBJECT: oResult = Process_GetObject(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_SET_OBJECT: oResult = Process_SetObject(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_RESTORE: oResult = Process_Restore(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_STORE: oResult = Process_Store(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_RESTORE_FACTORY: oResult = Process_RestoreFactory(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // General gateway
            case VCS_CANOPEN_REQUEST_CAN_FRAME: oResult = Process_RequestCANFrame(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_SEND_CAN_FRAME: oResult = Process_SendCANFrame(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_READ_CAN_FRAME: oResult = Process_ReadCANFrame(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_SEND_NMT_SERVICE: oResult = Process_SendNMTService(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            //LSS
            case VCS_CANOPEN_SEND_LSS_FRAME: oResult = Process_SendLSSFrame(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case VCS_CANOPEN_READ_LSS_FRAME: oResult = Process_ReadLSSFrame(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Scanning
            case SCAN_READ_VENDOR_ID: oResult = Process_ReadVendorId(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_PRODUCT_CODE: oResult = Process_ReadProductCode(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_REVISION_NUMBER: oResult = Process_ReadRevisionNumber(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_SERIAL_NUMBER: oResult = Process_ReadSerialNumber(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_DEVICE_TYPE: oResult = Process_ReadDeviceType(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_DEVICE_NAME: oResult = Process_ReadDeviceName(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_IS_IDENTIFY_SUPPORTED: oResult = Process_IsIdentifySupported(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_START_IDENTIFY: oResult = Process_StartIdentify(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_STOP_IDENTIFY: oResult = Process_StopIdentify(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_NODE_ID: oResult = Process_ReadNodeId(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_WRITE_NODE_ID: oResult = Process_WriteNodeId(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_STATION_ID: oResult = Process_ReadStationId(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_WRITE_STATION_ID: oResult = Process_WriteStationId(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_READ_TRANSFER_RATE: oResult = Process_ReadTransferRate(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_WRITE_TRANSFER_RATE: oResult = Process_WriteTransferRate(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_STORE: oResult = Process_ScanStore(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_RESET_DEVICE: oResult = Process_ScanResetDevice(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SCAN_WAIT_FOR_REBOOT: oResult = Process_ScanWaitForReboot(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Firmware download
            case FIRMWARE_GET_SERIAL_NUMBER: oResult = Process_GetSerialNumber(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case FIRMWARE_CLEANUP_CANOPEN_COMMUNICATION: oResult = Process_CleanupCANopenCommunication(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // Gateway
            case GET_GATEWAY_SETTINGS: oResult = Process_GetGatewaySettings(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case SET_GATEWAY_SETTINGS: oResult = Process_SetGatewaySettings(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // DataRecorder setup
            case REC_SET_RECORDER_PARAMETER: oResult = Process_SetDataRecordingParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_RECORDER_PARAMETER: oResult = Process_GetDataRecordingParameter(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_ENABLE_TRIGGER: oResult = Process_EnableTrigger(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_DISABLE_ALL_TRIGGERS: oResult = Process_DisableAllTriggers(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_ACTIVATE_CHANNEL: oResult = Process_ActivateChannel(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_ACTIVE_CHANNEL: oResult = Process_ActiveChannel(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_DEACTIVATE_ALL_CHANNELS: oResult = Process_DeactivateAllChannels(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_VARIABLE: oResult = Process_GetVariable(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_MAX_NUMBER_OF_SAMPLES: oResult = Process_GetMaxNbOfSamples(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_NUMBER_OF_RECORDED_SAMPLES: oResult = Process_GetNbOfRecordedSamples(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_SET_NUMBER_OF_PRECEDING_SAMPLES: oResult = Process_SetNbOfPrecedingSamples(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_NUMBER_OF_PRECEDING_SAMPLES: oResult = Process_GetNbOfPrecedingSamples(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_NUMBER_OF_SAMPLING_VARIABLES: oResult = Process_GetNbOfSamplingVariables(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_SAMPLING_PERIOD: oResult = Process_GetSamplingPeriod(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // DataRecorder status
            case REC_START_RECORDER: oResult = Process_StartRecorder(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_STOP_RECORDER: oResult = Process_StopRecorder(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_FORCE_TRIGGER: oResult = Process_ForceTrigger(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_IS_RECORDER_RUNNING: oResult = Process_IsRecorderRunning(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_IS_RECORDER_TRIGGERED: oResult = Process_IsRecorderTriggered(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_RECORDER_STATUS: oResult = Process_GetRecorderStatus(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_SET_RECORDER_CONFIGURATION: oResult = Process_SetRecorderConfiguration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_RECORDER_CONFIGURATION: oResult = Process_GetRecorderConfiguration(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // DataRecorder data
            case REC_READ_CHANNEL_VECTOR_SIZE: oResult = Process_ReadChannelVectorSize(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
            case REC_GET_VECTOR_START_OFFSET: oResult = Process_GetVectorStartOffset(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;

            // DataRecorder advanced functions
            case REC_READ_DATA_BUFFER: oResult = Process_ReadDataBuffer(pCommand_VCS, pDeviceCommandSetManager, p_hHandle, p_hTransactionHandle); break;
        }

        Unlock();
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::InitErrorHandling()
{
    BOOL oResult(FALSE);
    CErrorProducer errorProducer;
    CStdString className(_T("GatewayDriveToEpos4"));

    if(m_pErrorHandling)
    {
        //Init error producer
        errorProducer.Init(VIRTUAL_COMMAND_SET_LAYER, className);
        oResult = m_pErrorHandling->InitErrorProducer(&errorProducer);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetControllerGain(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CControllerGainsEpos4 controllerGain;
    CErrorInfo errorInfo;
    DWORD ulValue;
    EObjectDataType dataType;
    WORD usController;
    WORD usGain;
    WORD usIndex(0);
    WORD usSubIndex(0);
    WORD usValue;
    __uint64 ullValue(0);

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usController, sizeof(usController));
        p_pCommand->GetParameterData(1, &usGain, sizeof(usGain));

        if (controllerGain.GetControllerGainIndex(usController, usGain, usIndex, usSubIndex, dataType, errorInfo))
        {
            if (dataType == ODT_UINT32)
            {
                if (ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulValue, &errorInfo))
                {
                    ullValue = ulValue;
                    oResult = TRUE;
                }
            }
            else if (dataType == ODT_UINT16)
            {
                if (ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usValue, &errorInfo))
                {
                    ullValue = usValue;
                    oResult = TRUE;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ullValue, sizeof(ullValue));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetControllerGain(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CControllerGainsEpos4 controllerGain;
    CErrorInfo errorInfo;
    DWORD ulValue;
    EObjectDataType dataType;
    WORD usController;
    WORD usGain;
    WORD usIndex;
    WORD usSubIndex;
    WORD usValue;
    __uint64 ullValue(0);

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usController, sizeof(usController));
        p_pCommand->GetParameterData(1, &usGain, sizeof(usGain));
        p_pCommand->GetParameterData(2, &ullValue, sizeof(ullValue));

        if (controllerGain.GetControllerGainIndex(usController, usGain, usIndex, usSubIndex, dataType, errorInfo))
        {
            if (dataType == ODT_UINT32)
            {
                ulValue = (DWORD)ullValue;
                oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulValue, &errorInfo);
            }
            else if (dataType == ODT_UINT16)
            {
                usValue = (WORD)ullValue;
                oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usValue, &errorInfo);
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMaxFollowingError(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulMaxFollowingError = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        // Init
        oResult = TRUE;

        // Max following error
        if(oResult)
        {
            usIndex = INDEX_MAX_FOLLOWING_ERROR;
            usSubIndex = SUBINDEX_MAX_FOLLOWING_ERROR;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulMaxFollowingError, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulMaxFollowingError, sizeof(ulMaxFollowingError));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetMaxFollowingError(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulMaxFollowingError = 0;

    if (p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Get parameter data
        p_pCommand->GetParameterData(0, &ulMaxFollowingError, sizeof(ulMaxFollowingError));

        //Max following error
        if (oResult)
        {
            usIndex = INDEX_MAX_FOLLOWING_ERROR;
            usSubIndex = SUBINDEX_MAX_FOLLOWING_ERROR;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulMaxFollowingError, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetMaxProfileVelocity(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulMaxProfileVelocity = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Get parameter data
        p_pCommand->GetParameterData(0, &ulMaxProfileVelocity, sizeof(ulMaxProfileVelocity));

        if(oResult)
        {
            usIndex = INDEX_MAX_PROFILE_VELOCITY;
            usSubIndex = SUBINDEX_MAX_PROFILE_VELOCITY;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulMaxProfileVelocity, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMaxProfileVelocity(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulMaxProfileVelocity = 0;

    if(p_pCommand)
    {
        usIndex = INDEX_MAX_PROFILE_VELOCITY;
        usSubIndex = SUBINDEX_MAX_PROFILE_VELOCITY;
        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulMaxProfileVelocity, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulMaxProfileVelocity, sizeof(ulMaxProfileVelocity));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetMaxAcceleration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulMaxAcceleration = 0;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ulMaxAcceleration, sizeof(ulMaxAcceleration));

        usIndex = INDEX_MAX_ACCELERATION;
        usSubIndex = SUBINDEX_MAX_ACCELERATION;
        oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulMaxAcceleration, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMaxAcceleration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulMaxAcceleration = 0;

    if(p_pCommand)
    {
        usIndex = INDEX_MAX_ACCELERATION;
        usSubIndex = SUBINDEX_MAX_ACCELERATION;
        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulMaxAcceleration, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulMaxAcceleration, sizeof(ulMaxAcceleration));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMotorType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    WORD usMotorType = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        usIndex = INDEX_MOTOR_TYPE;
        usSubIndex = SUBINDEX_MOTOR_TYPE;
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usMotorType, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usMotorType, sizeof(usMotorType));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetDcMotorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulContinuousCurrent = 0;
    DWORD ulPeakCurrent = 0;
    WORD usContinuousCurrent = 0;
    WORD usPeakCurrent = 0;
    WORD usThermalTimeConstant = 0;

    if (p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Read continuous current
        if (oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulContinuousCurrent, &errorInfo);
        }

        //Read peak current
        if (oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_PEAK_CURRENT;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulPeakCurrent, &errorInfo);
        }

        //Read thermal time constant
        if (oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_THERMAL_TIME_CONSTANT;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usThermalTimeConstant, &errorInfo);
        }

        //Convert to command datatypes
        usContinuousCurrent = (WORD)ulContinuousCurrent;
        usPeakCurrent = (WORD)ulPeakCurrent;

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usContinuousCurrent, sizeof(usContinuousCurrent));
        p_pCommand->SetReturnParameterData(1, &usPeakCurrent, sizeof(usPeakCurrent));
        p_pCommand->SetReturnParameterData(2, &usThermalTimeConstant, sizeof(usThermalTimeConstant));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetEcMotorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulContinuousCurrent = 0;
    DWORD ulPeakCurrent = 0;
    WORD usContinuousCurrent = 0;
    WORD usPeakCurrent = 0;
    BYTE ubPolePair = 0;
    WORD usThermalTimeConstant = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Read continuous current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulContinuousCurrent, &errorInfo);
        }

        //Read peak current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_PEAK_CURRENT;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulPeakCurrent, &errorInfo);
        }

        //Read thermal time constant
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_THERMAL_TIME_CONSTANT;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usThermalTimeConstant, &errorInfo);
        }

        //Read pole pairs
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_POLE_PAIR_NUMBER;
            oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ubPolePair, &errorInfo);
        }

        //Convert to command datatypes
        usContinuousCurrent = (WORD)ulContinuousCurrent;
        usPeakCurrent = (WORD)ulPeakCurrent;

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usContinuousCurrent, sizeof(usContinuousCurrent));
        p_pCommand->SetReturnParameterData(1, &usPeakCurrent, sizeof(usPeakCurrent));
        p_pCommand->SetReturnParameterData(2, &usThermalTimeConstant, sizeof(usThermalTimeConstant));
        p_pCommand->SetReturnParameterData(3, &ubPolePair, sizeof(ubPolePair));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetMotorType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD SENSOR_3_HALL_SENSOR = 0x00100000;
    const DWORD SENSOR_3_BITMASK = 0x00FF0000;
    const DWORD COMMUTATION_SENSOR_ABSOLUTE_BITMASK = 0x000000F0;
    const DWORD COMMUTATION_SENSOR_RELATIVE_BITMASK = 0x0000000F;

    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    WORD usMotorType;
    BYTE ubCommutationSensors;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usMotorType, sizeof(usMotorType));

        //Init
        oResult = TRUE;

        //Write motor type
        if(oResult)
        {
            usIndex = INDEX_MOTOR_TYPE;
            usSubIndex = SUBINDEX_MOTOR_TYPE;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usMotorType, &errorInfo);
        }

        if (usMotorType != MOTOR_TYPE_DC_MOTOR)
        {
            if (oResult)
            {
                usIndex = EPOS4_INDEX_SENSOR_DATA;
                usSubIndex = EPOS4_SUBINDEX_SENSORS_CONFIGURATION;
                oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SENSOR_3_HALL_SENSOR, SENSOR_3_BITMASK, &errorInfo);
            }
        }

        oResult = FindCommutationSensors(usMotorType, ubCommutationSensors, &errorInfo);

        // Write commutation sensors
        if (oResult)
        {
            DWORD ulCommutationSensorsNewPart = (DWORD)(ubCommutationSensors);
            DWORD ulCommutationSensorsBitmask = (DWORD)(COMMUTATION_SENSOR_ABSOLUTE_BITMASK | COMMUTATION_SENSOR_RELATIVE_BITMASK);
            usIndex = EPOS4_INDEX_SENSOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_COMMUTATIONS_SENSORS;
            oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulCommutationSensorsNewPart, ulCommutationSensorsBitmask, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::FindCommutationSensors(WORD p_usMotorType, BYTE& p_rubCommutationSensors, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    switch (p_usMotorType)
    {
        case MOTOR_TYPE_DC_MOTOR:
        {
            BuildCommutationSensorByte(SENSOR_NONE, SENSOR_NONE, p_rubCommutationSensors);
            oResult = TRUE;
            break;
        }
        case MOTOR_TYPE_EC_BLOCK:
        {
            BuildCommutationSensorByte(SENSOR_3, SENSOR_NONE, p_rubCommutationSensors);
            oResult = TRUE;
            break;
        }
        case MOTOR_TYPE_EC_SINUS:
        {
            BuildCommutationSensorByte(SENSOR_3, SENSOR_1, p_rubCommutationSensors);
            oResult = TRUE;
            break;
        }
        default:
        {
            // Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_BAD_PARAMETER, p_pErrorInfo);
            oResult = FALSE;
            break;
        }
    }

    return oResult;
}

void CGatewayDriveToEpos4::BuildCommutationSensorByte(BYTE p_ubAbsoluteSensor, BYTE p_ubRelativeSensor, BYTE& p_rubCommutationSensors)
{
    const BYTE COMMUTATION_ABSOLUTE_FIRST_BIT_POSITION = 4;
    p_rubCommutationSensors = (BYTE)(p_ubAbsoluteSensor << COMMUTATION_ABSOLUTE_FIRST_BIT_POSITION) | p_ubRelativeSensor;
}

BOOL CGatewayDriveToEpos4::Process_SetDcMotorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulContinuousCurrent;
    DWORD ulPeakCurrent;
    WORD usContinuousCurrent;
    WORD usPeakCurrent;
    WORD usThermalTimeConstant;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usContinuousCurrent, sizeof(usContinuousCurrent));
        p_pCommand->GetParameterData(1, &usPeakCurrent, sizeof(usPeakCurrent));
        p_pCommand->GetParameterData(2, &usThermalTimeConstant, sizeof(usThermalTimeConstant));

        //Convert to object datatypes
        ulContinuousCurrent = (DWORD)usContinuousCurrent;
        ulPeakCurrent = (DWORD)usPeakCurrent;

        //Init
        oResult = TRUE;

        //Write continuous current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulContinuousCurrent, &errorInfo);
        }

        //Write peak current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_PEAK_CURRENT;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulPeakCurrent, &errorInfo);
        }

        //Write thermal time constant
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_THERMAL_TIME_CONSTANT;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usThermalTimeConstant, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetEcMotorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usIndex;
    WORD usSubIndex;
    DWORD ulContinuousCurrent;
    DWORD ulPeakCurrent;
    WORD usContinuousCurrent;
    WORD usPeakCurrent;
    BYTE ubPolePair;
    WORD usThermalTimeConstant;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Get parameter data
        p_pCommand->GetParameterData(0, &usContinuousCurrent, sizeof(usContinuousCurrent));
        p_pCommand->GetParameterData(1, &usPeakCurrent, sizeof(usPeakCurrent));
        p_pCommand->GetParameterData(2, &usThermalTimeConstant, sizeof(usThermalTimeConstant));
        p_pCommand->GetParameterData(3, &ubPolePair, sizeof(ubPolePair));

        //Convert to object datatypes
        ulContinuousCurrent = (DWORD)usContinuousCurrent;
        ulPeakCurrent = (DWORD)usPeakCurrent;

        //Write continuous current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulContinuousCurrent, &errorInfo);
        }

        //Write peak current
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_MAX_PEAK_CURRENT;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulPeakCurrent, &errorInfo);
        }

        //Write thermal time constant
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_THERMAL_TIME_CONSTANT;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usThermalTimeConstant, &errorInfo);
        }

        //Write pole paar
        if(oResult)
        {
            usIndex = EPOS4_INDEX_MOTOR_DATA;
            usSubIndex = EPOS4_SUBINDEX_POLE_PAIR_NUMBER;
            oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ubPolePair, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetSensorType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usSensorType(0);

    if(p_pCommand)
    {
        if (GetMainSensorType(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usSensorType, &errorInfo))
        {
            // Limit for SetSensorType compatibility
            switch (usSensorType)
            {
                case SENSOR_TYPE_NONE:
                case SENSOR_TYPE_INC_ENCODER_3_CHANNEL:
                case SENSOR_TYPE_INC_ENCODER_2_CHANNEL:
                case SENSOR_TYPE_HALL_SENSORS:
                case SENSOR_TYPE_SSI_ABS_ENCODER_BINARY:
                case SENSOR_TYPE_SSI_ABS_ENCODER_GREY:
                case SENSOR_TYPE_INC_ENCODER2_3_CHANNEL:
                case SENSOR_TYPE_INC_ENCODER2_2_CHANNEL:
                case SENSOR_TYPE_ANALOG_INC_ENCODER_3_CHANNEL:
                case SENSOR_TYPE_ANALOG_INC_ENCODER_2_CHANNEL:
                    {
                        oResult = TRUE;
                    }
                    break;

                default:
                    {
                        usSensorType = SENSOR_TYPE_NONE;
                        if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_SENSOR_CONFIG_UNKNOWN, &errorInfo);
                        oResult = FALSE;
                    }
                    break;
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usSensorType, sizeof(usSensorType));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetIncEncoderParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulEncoderResolution = 0;
    WORD usInvertedPolarity = 0;
    DWORD ulInvertedPolarity = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        // Init
        oResult = TRUE;

        // Read sensor resolution
        if(oResult)
        {
            usIndex = EPOS4_INDEX_DIGITAL_INC_ENCODER_1;
            usSubIndex = EPOS4_SUBINDEX_ENCODER_1_NUMBER_OF_PULSES;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulEncoderResolution, &errorInfo);
        }

        // Read inverted polarity
        if(oResult)
        {
            usIndex = EPOS4_INDEX_DIGITAL_INC_ENCODER_1;
            usSubIndex = EPOS4_SUBINDEX_ENCODER_1_TYPE;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usInvertedPolarity, &errorInfo);

            if(oResult)
            {
                //Extract polarity
                if ((usInvertedPolarity & 0x10) > 0)
                {
                    //Inverted
                    ulInvertedPolarity = 1;
                }
                else
                {
                    //maxon
                    ulInvertedPolarity = 0;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulEncoderResolution, sizeof(ulEncoderResolution));
        p_pCommand->SetReturnParameterData(1, &ulInvertedPolarity, sizeof(ulInvertedPolarity));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetHallSensorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    WORD usInvertedPolarity = 0;
    DWORD ulInvertedPolarity = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        // Init
        oResult = TRUE;

        // Read inverted polarity
        if(oResult)
        {
            usIndex = EPOS4_INDEX_DIGITAL_HALL_SENSOR;
            usSubIndex = EPOS4_SUBINDEX_HALL_SENSOR_TYPE;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usInvertedPolarity, &errorInfo);

            if(oResult)
            {
                //Extract polarity (bit 0)
                if ((usInvertedPolarity & 0x1) > 0)
                {
                    //Inverted
                    ulInvertedPolarity = 1;
                }
                else
                {
                    //maxon
                    ulInvertedPolarity = 0;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulInvertedPolarity, sizeof(ulInvertedPolarity));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetSsiAbsEncoderParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BYTE NB_OF_MULTITURN_DATABITS_BIT_POSITION = 16;
    const DWORD NB_OF_MULTITURN_DATABITS_BITMASK = 0x00FF0000;
    const BYTE NB_OF_SINGLETURN_DATABITS_BIT_POSITION = 8;
    const DWORD NB_OF_SINGLETURN_DATABITS_BITMASK = 0x0000FF00;

    BOOL oResult(FALSE);
    WORD usDataRate(0);
    DWORD ulDataBits(0);
    WORD usNbOfMultiTurnDataBits(0);
    WORD usNbOfSingleTurnDataBits(0);
    WORD usSsiPolarity(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = TRUE;

        if (oResult)
        {
            GetSsiDataRate(usDataRate, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            GetSsiDatabits(ulDataBits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
            usNbOfMultiTurnDataBits = (ulDataBits & NB_OF_MULTITURN_DATABITS_BITMASK) >> NB_OF_MULTITURN_DATABITS_BIT_POSITION;
            usNbOfSingleTurnDataBits = (ulDataBits & NB_OF_SINGLETURN_DATABITS_BITMASK) >> NB_OF_SINGLETURN_DATABITS_BIT_POSITION;
        }

        if (oResult)
        {
            oResult = GetSsiPolarity(usSsiPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usDataRate, sizeof(usDataRate));
        p_pCommand->SetReturnParameterData(1, &usNbOfMultiTurnDataBits, sizeof(usNbOfMultiTurnDataBits));
        p_pCommand->SetReturnParameterData(2, &usNbOfSingleTurnDataBits, sizeof(usNbOfSingleTurnDataBits));
        p_pCommand->SetReturnParameterData(3, &usSsiPolarity, sizeof(usSsiPolarity));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetSsiAbsEncoderParameterEx(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BYTE NB_OF_MULTITURN_DATABITS_BIT_POSITION = 16;
    const DWORD NB_OF_MULTITURN_DATABITS_BITMASK = 0x00FF0000;
    const BYTE NB_OF_SINGLETURN_DATABITS_BIT_POSITION = 8;
    const DWORD NB_OF_SINGLETURN_DATABITS_BITMASK = 0x0000FF00;
    const BYTE NB_OF_SPECIAL_DATABITS_BIT_POSITION = 0;
    const DWORD NB_OF_SPECIAL_DATABITS_BITMASK = 0x000000FF;

    BOOL oResult(FALSE);
    WORD usDataRate(0);
    DWORD ulDataBits(0);
    WORD usNbOfMultiTurnDataBits(0);
    WORD usNbOfSingleTurnDataBits(0);
    WORD usNbOfSpecialDataBits(0);
    WORD usSsiPolarity(0);
    WORD usTimeout(0);
    WORD usPowerupTime(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = TRUE;

        if (oResult)
        {
            GetSsiDataRate(usDataRate, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            GetSsiDatabits(ulDataBits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
            usNbOfMultiTurnDataBits = (ulDataBits & NB_OF_MULTITURN_DATABITS_BITMASK) >> NB_OF_MULTITURN_DATABITS_BIT_POSITION;
            usNbOfSingleTurnDataBits = (ulDataBits & NB_OF_SINGLETURN_DATABITS_BITMASK) >> NB_OF_SINGLETURN_DATABITS_BIT_POSITION;
            usNbOfSpecialDataBits = (ulDataBits & NB_OF_SPECIAL_DATABITS_BITMASK) >> NB_OF_SPECIAL_DATABITS_BIT_POSITION;
        }

        if (oResult)
        {
            oResult = GetSsiPolarity(usSsiPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
            WORD usSubIndex = EPOS4_SUBINDEX_SSI_TIMEOUT_TIME;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usTimeout, &errorInfo);
        }

        if (oResult)
        {
            WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
            WORD usSubIndex = EPOS4_SUBINDEX_SSI_POWERUP_TIME;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usPowerupTime, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usDataRate, sizeof(usDataRate));
        p_pCommand->SetReturnParameterData(1, &usNbOfMultiTurnDataBits, sizeof(usNbOfMultiTurnDataBits));
        p_pCommand->SetReturnParameterData(2, &usNbOfSingleTurnDataBits, sizeof(usNbOfSingleTurnDataBits));
        p_pCommand->SetReturnParameterData(3, &usNbOfSpecialDataBits, sizeof(usNbOfSpecialDataBits));
        p_pCommand->SetReturnParameterData(4, &usSsiPolarity, sizeof(usSsiPolarity));
        p_pCommand->SetReturnParameterData(5, &usTimeout, sizeof(usTimeout));
        p_pCommand->SetReturnParameterData(6, &usPowerupTime, sizeof(usPowerupTime));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetSensorType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usParameterSensorType;
    BYTE ubSensorNumber;
    BYTE ubSensorType;
    BYTE ubEncoderIndex;
    BYTE ubEncodingType;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        p_pCommand->GetParameterData(0, &usParameterSensorType, sizeof(usParameterSensorType));

        oResult = FindSensorTypeAndIndex(usParameterSensorType, ubSensorNumber, ubSensorType, ubEncoderIndex, ubEncodingType, &errorInfo);

        if (oResult)
        {
            switch (usParameterSensorType)
            {
                case SENSOR_TYPE_NONE:
                {
                    oResult = SetSensorTypeNone(ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
                case SENSOR_TYPE_INC_ENCODER_3_CHANNEL:
                case SENSOR_TYPE_INC_ENCODER_2_CHANNEL:
                {
                    oResult = SetSensorTypeIncrementalEncoder1(ubSensorNumber, ubSensorType, ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
                case SENSOR_TYPE_HALL_SENSORS:
                {
                    oResult = SetSensorTypeHallSensors(ubSensorNumber, ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
                case SENSOR_TYPE_SSI_ABS_ENCODER_BINARY:
                case SENSOR_TYPE_SSI_ABS_ENCODER_GREY:
                {
                    oResult = SetSensorTypeSsiAbsoluteEncoder(ubSensorNumber, ubSensorType, ubEncodingType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
                case SENSOR_TYPE_INC_ENCODER2_2_CHANNEL:
                case SENSOR_TYPE_INC_ENCODER2_3_CHANNEL:
                {
                    oResult = SetSensorTypeIncrementalEncoder2(ubSensorNumber, ubSensorType, ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
                case SENSOR_TYPE_ANALOG_INC_ENCODER_2_CHANNEL:
                case SENSOR_TYPE_ANALOG_INC_ENCODER_3_CHANNEL:
                {
                    oResult = SetSensorTypeAnalogIncrementalEncoder(ubSensorNumber, ubSensorType, ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
                    break;
                }
            }
        }
        else
        {
            if (m_pErrorHandling)
            {
                m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_SENSOR_CONFIG_NOT_SUPPORTED, &errorInfo);
            }
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeNone(BYTE p_ubSensorNumber, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeIncrementalEncoder1(BYTE p_ubSensorNumber, BYTE p_ubSensorType, BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usMotorType;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_MOTOR_TYPE, SUBINDEX_MOTOR_TYPE, &usMotorType, p_pErrorInfo);

    if (oResult)
    {
        oResult = ModifySensorConfiguration(p_ubSensorNumber, p_ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult & (usMotorType == MOTOR_TYPE_EC_SINUS))
    {
        oResult = ModifyCommutationSensorRelative(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetIncrementalEncoder1TypeNoOfChannels(p_ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeHallSensors(BYTE p_ubSensorNumber, BYTE p_ubSensorType, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usMotorType;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_MOTOR_TYPE, SUBINDEX_MOTOR_TYPE, &usMotorType, p_pErrorInfo);

    if (oResult)
    {
        oResult = ModifySensorConfiguration(p_ubSensorNumber, p_ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult & (usMotorType == MOTOR_TYPE_EC_SINUS))
    {
        oResult = ModifyCommutationSensorAbsolute(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeSsiAbsoluteEncoder(BYTE p_ubSensorNumber, BYTE p_ubSensorType, BYTE p_ubEncodingType, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usMotorType;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_MOTOR_TYPE, SUBINDEX_MOTOR_TYPE, &usMotorType, p_pErrorInfo);

    if (oResult)
    {
        oResult = ModifySensorConfiguration(p_ubSensorNumber, p_ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult & (usMotorType == MOTOR_TYPE_EC_SINUS))
    {
        oResult = ModifyCommutationSensorRelative(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetSsiAbsoluteEncodingType(p_ubEncodingType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeIncrementalEncoder2(BYTE p_ubSensorNumber, BYTE p_ubSensorType, BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usMotorType;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_MOTOR_TYPE, SUBINDEX_MOTOR_TYPE, &usMotorType, p_pErrorInfo);

    if (oResult)
    {
        oResult = ModifySensorConfiguration(p_ubSensorNumber, p_ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult & (usMotorType == MOTOR_TYPE_EC_SINUS))
    {
        oResult = ModifyCommutationSensorRelative(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetIncrementalEncoder2TypeNoOfChannels(p_ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSensorTypeAnalogIncrementalEncoder(BYTE p_ubSensorNumber, BYTE p_ubSensorType, BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usMotorType;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_MOTOR_TYPE, SUBINDEX_MOTOR_TYPE, &usMotorType, p_pErrorInfo);

    if (oResult)
    {
        oResult = ModifySensorConfiguration(p_ubSensorNumber, p_ubSensorType, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = ModifyControlStructure(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult & (usMotorType == MOTOR_TYPE_EC_SINUS))
    {
        oResult = ModifyCommutationSensorRelative(p_ubSensorNumber, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetAnalogEncoderTypeNoOfChannels(p_ubEncoderIndex, p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ModifySensorConfiguration(BYTE p_ubSensorNumber, BYTE p_ubSensorType, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    DWORD ulSensorsConfigurationData = 0x00000000;
    DWORD ulSensorsConfigurationBitMask = 0x00000000;

    SetSensorsConfiguration(p_ubSensorNumber, p_ubSensorType, ulSensorsConfigurationData, ulSensorsConfigurationBitMask);

    oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_SENSORS_CONFIGURATION, ulSensorsConfigurationData, ulSensorsConfigurationBitMask, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::ModifyControlStructure(BYTE p_ubSensorNumber, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE ON_MOTOR = 0x00;

    BOOL oResult(FALSE);

    DWORD ulControlStructureData = 0x00000000;
    DWORD ulControlStructureBitMask = 0x00000000;

    SetControlStructureProcessValueReference(ON_MOTOR, ulControlStructureData, ulControlStructureBitMask);
    SetControlStructureMainSensor(p_ubSensorNumber, ulControlStructureData, ulControlStructureBitMask);
    SetControlStructureMountingPosition(p_ubSensorNumber, ON_MOTOR, ulControlStructureData, ulControlStructureBitMask);

    oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_CONTROL_STRUCTURE, ulControlStructureData, ulControlStructureBitMask, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::ModifyCommutationSensorRelative(BYTE p_ubMainSensor, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulCommutationSensorsData = 0x00000000;
    DWORD ulCommutationSensorsBitMask = 0x00000000;

    SetCommutationSensorRelative(p_ubMainSensor, ulCommutationSensorsData, ulCommutationSensorsBitMask);

    oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_COMMUTATIONS_SENSORS, ulCommutationSensorsData, ulCommutationSensorsBitMask, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::ModifyCommutationSensorAbsolute(BYTE p_ubMainSensor, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulCommutationSensorsData = 0x00000000;
    DWORD ulCommutationSensorsBitMask = 0x00000000;

    SetCommutationSensorAbsolute(p_ubMainSensor, ulCommutationSensorsData, ulCommutationSensorsBitMask);

    oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_COMMUTATIONS_SENSORS, ulCommutationSensorsData, ulCommutationSensorsBitMask, p_pErrorInfo);

    return oResult;
}

void CGatewayDriveToEpos4::SetSensorsConfiguration(BYTE p_ubSensorNumber, BYTE p_ubSensorType, DWORD& p_rulSensorsConfigurationData, DWORD& p_rulSensorsConfigurationBitMask)
{
    const BYTE SENSOR_CONFIGURATION_BITCOUNT = 8;
    const BYTE SENSOR_CONFIGURATION_BITMASK = 0x00000003;

    DWORD ulSensorType = (DWORD)(p_ubSensorType << (SENSOR_CONFIGURATION_BITCOUNT * (p_ubSensorNumber - 1)));
    DWORD ulSensorTypeBitMask = SENSOR_CONFIGURATION_BITMASK << (SENSOR_CONFIGURATION_BITCOUNT * (p_ubSensorNumber - 1));

    p_rulSensorsConfigurationData = (p_rulSensorsConfigurationData & (ulSensorTypeBitMask ^ 0xFFFFFFFF)) | (ulSensorType & ulSensorTypeBitMask);
    p_rulSensorsConfigurationBitMask = ulSensorTypeBitMask;
}

void CGatewayDriveToEpos4::SetControlStructureVelocity(BYTE p_ubController, DWORD& p_rulControlStructureData, DWORD& p_rulControlStructureBitMask)
{
    const DWORD VELOCITY_CONTORL_STRUCTURE_FIRST_BIT_POSITION = 4;
    const DWORD VELOCITY_CONTORL_STRUCTURE_BITMASK = 0x000000F0;

    DWORD ulVelocityControlStructure = (DWORD)(p_ubController << VELOCITY_CONTORL_STRUCTURE_FIRST_BIT_POSITION);

    p_rulControlStructureData = (p_rulControlStructureData & (VELOCITY_CONTORL_STRUCTURE_BITMASK ^ 0xFFFFFFFF)) | (ulVelocityControlStructure & VELOCITY_CONTORL_STRUCTURE_BITMASK);
    p_rulControlStructureBitMask |= VELOCITY_CONTORL_STRUCTURE_BITMASK;
}

void CGatewayDriveToEpos4::SetControlStructurePosition(BYTE p_ubController, DWORD& p_rulControlStructureData, DWORD& p_rulControlStructureBitMask)
{
    const DWORD POSITION_CONTORL_STRUCTURE_FIRST_BIT_POSITION = 8;
    const DWORD POSITION_CONTORL_STRUCTURE_BITMASK = 0x00000F00;

    DWORD ulPositionControlStructure = (DWORD)(p_ubController << POSITION_CONTORL_STRUCTURE_FIRST_BIT_POSITION);

    p_rulControlStructureData = (p_rulControlStructureData & (POSITION_CONTORL_STRUCTURE_BITMASK ^ 0xFFFFFFFF)) | (ulPositionControlStructure & POSITION_CONTORL_STRUCTURE_BITMASK);
    p_rulControlStructureBitMask |= POSITION_CONTORL_STRUCTURE_BITMASK;
}

void CGatewayDriveToEpos4::SetControlStructureProcessValueReference(BYTE p_ubReference, DWORD& p_rulControlStructureData, DWORD& p_rulControlStructureBitMask)
{
    const DWORD PROCESS_VALUE_REFERENCE_FIRST_BIT_POSITION = 14;
    const DWORD PROCESS_VALUE_REFERENCE_BITMASK = 0x0000C000;

    DWORD ulProcessValueReference = (DWORD)(p_ubReference << PROCESS_VALUE_REFERENCE_FIRST_BIT_POSITION);

    p_rulControlStructureData = (p_rulControlStructureData & (PROCESS_VALUE_REFERENCE_BITMASK ^ 0xFFFFFFFF)) | (ulProcessValueReference & PROCESS_VALUE_REFERENCE_BITMASK);
    p_rulControlStructureBitMask |= PROCESS_VALUE_REFERENCE_BITMASK;
}

void CGatewayDriveToEpos4::SetControlStructureMainSensor(BYTE p_ubSensorNumber, DWORD& p_rulControlStructureData, DWORD& p_rulControlStructureBitMask)
{
    const BYTE MAIN_SENSOR_FIRST_BIT_POSITION = 16;
    const DWORD MAIN_SENSOR_BITMASK = 0x000F0000;

    DWORD ulMainSensor = (DWORD)(p_ubSensorNumber << MAIN_SENSOR_FIRST_BIT_POSITION);

    p_rulControlStructureData = (p_rulControlStructureData & (MAIN_SENSOR_BITMASK ^ 0xFFFFFFFF)) | (ulMainSensor & MAIN_SENSOR_BITMASK);
    p_rulControlStructureBitMask |= MAIN_SENSOR_BITMASK;
}

void CGatewayDriveToEpos4::SetControlStructureMountingPosition(BYTE p_ubSensorNumber, BYTE p_ubMountingPosition, DWORD& p_rulControlStructureData, DWORD& p_rulControlStructureBitMask)
{
    const BYTE MOUNTING_POSITION_FIRST_BIT_POSITION = 24;
    const BYTE MOUNTING_POSITION_BITCOUNT = 2;
    const DWORD MOUNTING_POSITION_BITMASK = 0x03000000;

    if (p_ubSensorNumber != SENSOR_NONE)
    {
        DWORD ulMountingPosition = (DWORD)(p_ubMountingPosition << (MOUNTING_POSITION_FIRST_BIT_POSITION + MOUNTING_POSITION_BITCOUNT * (p_ubSensorNumber - 1)));
        DWORD ulMountingPositionBitMask = (DWORD)(MOUNTING_POSITION_BITMASK << (MOUNTING_POSITION_BITCOUNT * (p_ubSensorNumber - 1)));

        p_rulControlStructureData = (p_rulControlStructureData & (ulMountingPositionBitMask ^ 0xFFFFFFFF)) | (ulMountingPosition & ulMountingPositionBitMask);
        p_rulControlStructureBitMask |= ulMountingPositionBitMask;
    }
}

void CGatewayDriveToEpos4::SetCommutationSensorRelative(BYTE p_ubSensorNumber, DWORD& p_rulCommutationSensorsData, DWORD& p_rulCommutationSensorsBitMask)
{
    const BYTE COMMUTATION_SENSOR_RELATIVE_FIRST_BIT_POSITION = 0;
    const DWORD COMMUTATION_SENSOR_RELATIVE_BITMASK = 0x0000000F;

    DWORD ulCommutationSensorRelative = (DWORD)(p_ubSensorNumber << COMMUTATION_SENSOR_RELATIVE_FIRST_BIT_POSITION);

    p_rulCommutationSensorsData = (p_rulCommutationSensorsData & (COMMUTATION_SENSOR_RELATIVE_BITMASK ^ 0xFFFFFFFF)) | (ulCommutationSensorRelative & COMMUTATION_SENSOR_RELATIVE_BITMASK);
    p_rulCommutationSensorsBitMask |= COMMUTATION_SENSOR_RELATIVE_BITMASK;
}

void CGatewayDriveToEpos4::SetCommutationSensorAbsolute(BYTE p_ubSensorNumber, DWORD& p_rulCommutationSensorsData, DWORD& p_rulCommutationSensorsBitMask)
{
    const BYTE COMMUTATION_SENSOR_ABSOLUTE_FIRST_BIT_POSITION = 4;
    const DWORD COMMUTATION_SENSOR_ABSOLUTE_BITMASK = 0xF0;

    DWORD ulCommutationSensorAbsolute = (DWORD)(p_ubSensorNumber << COMMUTATION_SENSOR_ABSOLUTE_FIRST_BIT_POSITION);

    p_rulCommutationSensorsData = (p_rulCommutationSensorsData & (COMMUTATION_SENSOR_ABSOLUTE_BITMASK ^ 0xFFFFFFFF)) | (ulCommutationSensorAbsolute & COMMUTATION_SENSOR_ABSOLUTE_BITMASK);
    p_rulCommutationSensorsBitMask |= COMMUTATION_SENSOR_ABSOLUTE_BITMASK;
}

BOOL CGatewayDriveToEpos4::SetIncrementalEncoder1TypeNoOfChannels(BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const WORD SENSOR_INDEX_BITMASK = 0x0001;

    BOOL oResult(FALSE);

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_DIGITAL_INC_ENCODER_1, EPOS4_SUBINDEX_ENCODER_1_TYPE, p_ubEncoderIndex, SENSOR_INDEX_BITMASK, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetIncrementalEncoder2TypeNoOfChannels(BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const WORD SENSOR_INDEX_BITMASK = 0x0001;

    BOOL oResult(FALSE);
    WORD usIndex = EPOS4_INDEX_DIGITAL_INC_ENCODER_2;
    WORD usSubIndex = EPOS4_SUBINDEX_ENCODER_2_TYPE;

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, p_ubEncoderIndex, SENSOR_INDEX_BITMASK, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetAnalogEncoderTypeNoOfChannels(BYTE p_ubEncoderIndex, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const WORD SENSOR_INDEX_BITMASK = 0x0001;

    BOOL oResult(FALSE);
    WORD usIndex = EPOS4_INDEX_ANALOG_INC_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_ANALOG_ENCODER_TYPE;

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, p_ubEncoderIndex, SENSOR_INDEX_BITMASK, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSsiAbsoluteEncodingType(WORD p_ubEncodingType, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const WORD SSI_SENSOR_ENCODING_BITMASK = 0x0001;

    BOOL oResult(FALSE);

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SSI_ABSOLUTE_ENCODER, EPOS4_SUBINDEX_SSI_ENCODING_TYPE, p_ubEncodingType, SSI_SENSOR_ENCODING_BITMASK, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::FindSensorTypeAndIndex(WORD p_usSensorType, BYTE& p_rubSensorNumber, BYTE& p_rubSensorType, BYTE& p_rubEncoderIndex, BYTE& p_rubEncodingType, CErrorInfo* p_pErrorInfo)
{
    const BYTE SENSOR_WITHOUT_INDEX = 0x00;
    const BYTE SENSOR_WITH_INDEX = 0x01;
    const BYTE BINARY_CODED = 0x00;
    const BYTE GRAY_CODED = 0x01;

    p_rubSensorNumber = SENSOR_NONE;
    p_rubSensorType = SENSOR_TYPE_NONE;
    p_rubEncoderIndex = SENSOR_WITHOUT_INDEX;
    p_rubEncodingType = BINARY_CODED;
    BOOL oResult(FALSE);

    switch (p_usSensorType)
    {
        case SENSOR_TYPE_NONE:
        {
            p_rubSensorNumber = SENSOR_NONE;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_INC_ENCODER_3_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_1;
            p_rubSensorType = SENSOR1_TYPE_DIGITAL_INC_ENCODER_1;
            p_rubEncoderIndex = SENSOR_WITH_INDEX;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_INC_ENCODER_2_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_1;
            p_rubSensorType = SENSOR1_TYPE_DIGITAL_INC_ENCODER_1;
            p_rubEncoderIndex = SENSOR_WITHOUT_INDEX;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_HALL_SENSORS:
        {
            p_rubSensorNumber = SENSOR_3;
            p_rubSensorType = SENSOR3_TYPE_DIGITAL_HALL_SENSOR;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_SSI_ABS_ENCODER_BINARY:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_SSI_ENCODER;
            p_rubEncodingType = BINARY_CODED;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_SSI_ABS_ENCODER_GREY:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_SSI_ENCODER;
            p_rubEncodingType = GRAY_CODED;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_INC_ENCODER2_3_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_DIGITAL_INC_ENCODER_2;
            p_rubEncoderIndex = SENSOR_WITH_INDEX;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_INC_ENCODER2_2_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_DIGITAL_INC_ENCODER_2;
            p_rubEncoderIndex = SENSOR_WITHOUT_INDEX;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_ANALOG_INC_ENCODER_3_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_ANALOG_INC_ENCODER;
            p_rubEncoderIndex = SENSOR_WITH_INDEX;
            oResult = TRUE;
            break;
        }
        case SENSOR_TYPE_ANALOG_INC_ENCODER_2_CHANNEL:
        {
            p_rubSensorNumber = SENSOR_2;
            p_rubSensorType = SENSOR2_TYPE_ANALOG_INC_ENCODER;
            p_rubEncoderIndex = SENSOR_WITHOUT_INDEX;
            oResult = TRUE;
            break;
        }
        default:
        {
            // Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_SENSOR_CONFIG_NOT_SUPPORTED, p_pErrorInfo);
            oResult = FALSE;
            break;
        }
    }
    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetIncEncoderParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);

    DWORD ulEncoderResolution;
    DWORD ulInvertedPolarity;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &ulEncoderResolution, sizeof(ulEncoderResolution));
        p_pCommand->GetParameterData(1, &ulInvertedPolarity, sizeof(ulInvertedPolarity));

        // Init
        oResult = TRUE;

        // Write sensor resolution
        if(oResult)
        {
            usIndex = EPOS4_INDEX_DIGITAL_INC_ENCODER_1;
            usSubIndex = EPOS4_SUBINDEX_ENCODER_1_NUMBER_OF_PULSES;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulEncoderResolution, &errorInfo);
        }

        // Write sensor inverted polarity
        if(oResult)
        {
            oResult = SetIncEncoderPolarity(ulInvertedPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetIncEncoderPolarity(DWORD ulInvertedPolarity, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE SENSOR_CONFIGURATION_DIRECTION_BIT = 4;

    BOOL oResult(FALSE);
    WORD usIndex;
    WORD usSubIndex;
    WORD usSensorsConfigurationNewPart;
    DWORD ulBitMask;

    usSensorsConfigurationNewPart = (WORD)(ulInvertedPolarity << SENSOR_CONFIGURATION_DIRECTION_BIT);
    ulBitMask = (DWORD)(1 << SENSOR_CONFIGURATION_DIRECTION_BIT);

    usIndex = EPOS4_INDEX_DIGITAL_INC_ENCODER_1;
    usSubIndex = EPOS4_SUBINDEX_ENCODER_1_TYPE;

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usSensorsConfigurationNewPart, (WORD)ulBitMask, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetHallSensorParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usInvertedPolarity;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usInvertedPolarity, sizeof(usInvertedPolarity));

        // Init
        oResult = TRUE;

        // Write sensor inverted polarity
        if(oResult)
        {
            oResult = SetHallSensorPolarity(usInvertedPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetHallSensorPolarity(WORD ulInvertedPolarity, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE SENSOR_CONFIGURATION_DIRECTION_BIT = 0;

    BOOL oResult(FALSE);
    WORD usSensorsConfigurationNewPart;
    DWORD ulBitMask;
    WORD usIndex;
    WORD usSubIndex;

    usSensorsConfigurationNewPart = (WORD)(ulInvertedPolarity << SENSOR_CONFIGURATION_DIRECTION_BIT);
    ulBitMask = (DWORD)(1 << SENSOR_CONFIGURATION_DIRECTION_BIT);

    usIndex = EPOS4_INDEX_DIGITAL_HALL_SENSOR;
    usSubIndex = EPOS4_SUBINDEX_HALL_SENSOR_TYPE;

    oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usSensorsConfigurationNewPart, (WORD)ulBitMask, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetSsiAbsEncoderParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD NB_OF_SPECIAL_DATABITS_BITMASK = 0x000000FF;
    const BYTE MULTITURN_DATABITS_MIN = 0;
    const BYTE SINGLETURN_DATABITS_MIN = 6;

    BOOL oResult(FALSE);
    WORD usDataRate;
    WORD usInvertedPolarity;
    WORD usNbOfMultiTurnDataBits;
    WORD usNbOfSingleTurnDataBits;
    WORD usNbOfSpecialDatabits(0);

    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usDataRate, sizeof(usDataRate));
        p_pCommand->GetParameterData(1, &usNbOfMultiTurnDataBits, sizeof(usNbOfMultiTurnDataBits));
        p_pCommand->GetParameterData(2, &usNbOfSingleTurnDataBits, sizeof(usNbOfSingleTurnDataBits));
        p_pCommand->GetParameterData(3, &usInvertedPolarity, sizeof(usInvertedPolarity));

        oResult = SetSsiDaterate(usDataRate, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        DWORD ulSsiDataBits(0);
        if (oResult)
        {
            oResult = GetSsiDatabits(ulSsiDataBits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            usNbOfSpecialDatabits = ulSsiDataBits & NB_OF_SPECIAL_DATABITS_BITMASK;

            oResult = SetSsiDatabits((BYTE)MULTITURN_DATABITS_MIN, (BYTE)SINGLETURN_DATABITS_MIN, (BYTE)usNbOfSpecialDatabits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetSsiDatabits((BYTE)usNbOfMultiTurnDataBits, (BYTE)usNbOfSingleTurnDataBits, (BYTE)usNbOfSpecialDatabits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetSsiPolarity(usInvertedPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetSsiAbsEncoderParameterEx(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usDataRate;
    WORD usNbOfMultiTurnDataBits;
    WORD usNbOfSingleTurnDataBits;
    WORD usNbOfSpecialDatabits;
    WORD usInvertedPolarity;
    WORD usTimeout;
    WORD usPowerupTime;

    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usDataRate, sizeof(usDataRate));
        p_pCommand->GetParameterData(1, &usNbOfMultiTurnDataBits, sizeof(usNbOfMultiTurnDataBits));
        p_pCommand->GetParameterData(2, &usNbOfSingleTurnDataBits, sizeof(usNbOfSingleTurnDataBits));
        p_pCommand->GetParameterData(3, &usNbOfSpecialDatabits, sizeof(usNbOfSpecialDatabits));
        p_pCommand->GetParameterData(4, &usInvertedPolarity, sizeof(usInvertedPolarity));
        p_pCommand->GetParameterData(5, &usTimeout, sizeof(usTimeout));
        p_pCommand->GetParameterData(6, &usPowerupTime, sizeof(usPowerupTime));

        // Init
        oResult = TRUE;

        // Write
        if (oResult)
        {
            oResult = SetSsiDaterate(usDataRate, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetSsiDatabits((BYTE)usNbOfMultiTurnDataBits, (BYTE)usNbOfSingleTurnDataBits, (BYTE)usNbOfSpecialDatabits, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetSsiPolarity(usInvertedPolarity, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if (oResult)
        {
            WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
            WORD usSubIndex = EPOS4_SUBINDEX_SSI_TIMEOUT_TIME;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usTimeout, &errorInfo);
        }

        if (oResult)
        {
            WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
            WORD usSubIndex = EPOS4_SUBINDEX_SSI_POWERUP_TIME;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usPowerupTime, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSsiDaterate(WORD p_usDataRate, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_DATARATE;

    BOOL oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, p_usDataRate, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::GetSsiDatabits(DWORD& p_ulSsiDatabits, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_NUMBER_OF_DATABITS;

    BOOL oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &p_ulSsiDatabits, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSsiDatabits(BYTE p_ubNbOfMultiturnDatabits, BYTE p_ubNbOfSingleturnDatabits, BYTE p_ubNbOfSpecialDatabits, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE MULTIBITS_BIT_OFFSET = 16;
    const BYTE SINGLEBITS_BIT_OFFSET = 8;
    const BYTE SPECIALBITS_BIT_OFFSET = 0;

    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_NUMBER_OF_DATABITS;
    DWORD ulSsiDataBits = (p_ubNbOfMultiturnDatabits << MULTIBITS_BIT_OFFSET) | (p_ubNbOfSingleturnDatabits << SINGLEBITS_BIT_OFFSET) | (p_ubNbOfSpecialDatabits << SPECIALBITS_BIT_OFFSET);

    BOOL oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulSsiDataBits, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetSsiPolarity(WORD p_usInvertedPolarity, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE INVERTED_POLARITY_BIT_POSITION = 4;
    const WORD SSI_ENCODING_INVERTED_POLARITY_BITMASK = 0x0010;

    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_ENCODING_TYPE;
    WORD usSsiEncodingType = p_usInvertedPolarity << INVERTED_POLARITY_BIT_POSITION;

    BOOL oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usSsiEncodingType, SSI_ENCODING_INVERTED_POLARITY_BITMASK, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::GetSsiPolarity(WORD& p_usInvertedPolarity, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE INVERTED_POLARITY_BIT_POSITION = 4;

    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_ENCODING_TYPE;
    WORD usSsiEncodingType;

    BOOL oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usSsiEncodingType, p_pErrorInfo);

    p_usInvertedPolarity = usSsiEncodingType >> INVERTED_POLARITY_BIT_POSITION;
    return oResult;
}

BOOL CGatewayDriveToEpos4::GetSsiDataRate(WORD& p_usDataRate, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    WORD usIndex = EPOS4_INDEX_SSI_ABSOLUTE_ENCODER;
    WORD usSubIndex = EPOS4_SUBINDEX_SSI_DATARATE;

    BOOL oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &p_usDataRate, p_pErrorInfo);
    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMainSensorResolution(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulMainSensorResolution(0);

    if(p_pCommand)
    {
        //Read from firmware
        if(ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EPOS4_INDEX_SENSOR_DATA, EPOS4_SUBINDEX_MAIN_SENSOR_RESOLUTION, &ulMainSensorResolution, &errorInfo))
        {
            oResult = TRUE;
        }
        else
        {
            //Calculate in software
            oResult = CalculateMainSensorResolution(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulMainSensorResolution, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulMainSensorResolution, sizeof(ulMainSensorResolution));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetAuxiliarySensorResolution(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulAuxiliarySensorResolution(0);

    if (p_pCommand)
    {
        //Calculate in software
        oResult = CalculateAuxiliarySensorResolution(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulAuxiliarySensorResolution, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulAuxiliarySensorResolution, sizeof(ulAuxiliarySensorResolution));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetPositionProfile(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulProfileVelocity;
    DWORD ulProfileAcceleration;
    DWORD ulProfileDeceleration;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->ResetStatus();

        usIndex = INDEX_PROFILE_VELOCITY;
        usSubIndex = SUBINDEX_PROFILE_VELOCITY;
        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulProfileVelocity, &errorInfo);

        if(oResult)
        {
            usIndex = INDEX_PROFILE_ACCELERATION;
            usSubIndex = SUBINDEX_PROFILE_ACCELERATION;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulProfileAcceleration, &errorInfo);
        }

        if(oResult)
        {
            usIndex = INDEX_PROFILE_DECELERATION;
            usSubIndex = SUBINDEX_PROFILE_DECELERATION;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulProfileDeceleration, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulProfileVelocity, sizeof(ulProfileVelocity));
        p_pCommand->SetReturnParameterData(1, &ulProfileAcceleration, sizeof(ulProfileAcceleration));
        p_pCommand->SetReturnParameterData(2, &ulProfileDeceleration, sizeof(ulProfileDeceleration));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetTargetPosition(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    int lTargetPosition;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        usIndex = INDEX_TARGET_POSITION;
        usSubIndex = SUBINDEX_TARGET_POSITION;
        oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lTargetPosition, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lTargetPosition, sizeof(lTargetPosition));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_HaltPositionMovement(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;
        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, STOP_PROFILE, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::LeaveQuickStopState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usStatusWord = 0;
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    BOOL oQuickStopActive = FALSE;

    oResult = ReadStatusWord(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usStatusWord, p_ErrorInfo);

    if(oResult)
    {
        oQuickStopActive = IsQuickStopActive(usStatusWord);
    }

    if(oResult && oQuickStopActive)
    {
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;

        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SWITCH_ON_AND_ENABLE_OPERATION, p_ErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::LeaveHalt(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usControlWord = 0;
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    BOOL oHaltActive = FALSE;

    oResult = ReadControlWord(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usControlWord, p_ErrorInfo);

    if(oResult)
    {
        oHaltActive = IsHaltActive(usControlWord);
    }

    if(oResult && oHaltActive)
    {
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;

        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SWITCH_ON_AND_ENABLE_OPERATION, p_ErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetTargetPosition(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, int p_lTargetPosition, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usIndex = 0;
    WORD usSubIndex = 0;

    usIndex = INDEX_TARGET_POSITION;
    usSubIndex = SUBINDEX_TARGET_POSITION;

    oResult = WriteInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, p_lTargetPosition, p_ErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetNewSetPoint(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL p_oAbsolute, BOOL p_oImmediately, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usIndex = 0;
    WORD usSubIndex = 0;

    usIndex = INDEX_CONTROL_WORD;
    usSubIndex = SUBINDEX_CONTROL_WORD;

    if(p_oAbsolute)
    {
        if(p_oImmediately)
        {
            //Send absolute and immediately
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, IMMEDIATE_ABS_POSITION, p_ErrorInfo);
        }
        else
        {
            //Send absolute
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ABSOLUTE_POSITION, p_ErrorInfo);
        }
    }
    else
    {
        if(p_oImmediately)
        {
            //Send relative and immediately
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, IMMEDIATE_REL_POSITION, p_ErrorInfo);
        }
        else
        {
            //Send relative
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, RELATIVE_POSITION, p_ErrorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ResetNewSetPoint(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult = FALSE;
    WORD usIndex = 0;
    WORD usSubIndex = 0;

    usIndex = INDEX_CONTROL_WORD;
    usSubIndex = SUBINDEX_CONTROL_WORD;

    oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, START_PROFILE, p_ErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_MoveToPosition(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BOOL oAbsolute = FALSE;
    BOOL oImmediately = FALSE;
    CErrorInfo errorInfo;
    int lTargetPosition = 0;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &lTargetPosition, sizeof(lTargetPosition));
        p_pCommand->GetParameterData(1, &oAbsolute, sizeof(oAbsolute));
        p_pCommand->GetParameterData(2, &oImmediately, sizeof(oImmediately));

        oResult = SetTargetPosition(p_pManager, p_hDCS_Handle, p_hTransactionHandle, lTargetPosition, &errorInfo);

        if(oResult)
        {
            oResult = LeaveQuickStopState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if(oResult)
        {
            oResult = LeaveHalt(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if(oResult)
        {
            oResult = SetNewSetPoint(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oAbsolute, oImmediately, &errorInfo);
        }

        if(oResult)
        {
            WaitForSetPointAcknowledge(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        if(oResult)
        {
            oResult = ResetNewSetPoint(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsSetPointAckActive(WORD p_usStatusWord)
{
    BOOL oResult = FALSE;
    const int SET_POINT_ACK_ACTIVE_BIT = 12;
    bitset<sizeof(WORD)*8> statusWordBits(p_usStatusWord);

    oResult = statusWordBits.test(SET_POINT_ACK_ACTIVE_BIT);

    return oResult;
}

BOOL CGatewayDriveToEpos4::WaitForSetPointAcknowledge(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_ErrorInfo)
{
    BOOL oResult = FALSE;
    BOOL oNoTimeout = TRUE;
    const DWORD MINIMAL_WAIT_TIME = 1;
    WORD usStatusWord = 0;
    DWORD ulTransferRate = 0;
    DWORD ulTimeout = 0;
    CMmcHiResTimer timer;

    if(p_pManager && p_hDCS_Handle)
    {
        p_pManager->DCS_GetProtocolStackSettings(p_hDCS_Handle, &ulTransferRate, &ulTimeout);

        timer.start();

        while((oResult = ReadStatusWord(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usStatusWord, p_ErrorInfo)) && (oNoTimeout = timer.getElapsedTimeInMilliSec() < ulTimeout))
        {
            if(IsSetPointAckActive(usStatusWord))
            {
                break;
            }

            MmcSleep(MINIMAL_WAIT_TIME);
        }

        if(!oNoTimeout)
        {
            oResult = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetPositionProfile(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulProfileVelocity;
    DWORD ulProfileAcceleration;
    DWORD ulProfileDeceleration;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->ResetStatus();
        p_pCommand->GetParameterData(0, &ulProfileVelocity, sizeof(ulProfileVelocity));
        p_pCommand->GetParameterData(1, &ulProfileAcceleration, sizeof(ulProfileAcceleration));
        p_pCommand->GetParameterData(2, &ulProfileDeceleration, sizeof(ulProfileDeceleration));

        //Send velocity
        usIndex = INDEX_PROFILE_VELOCITY;
        usSubIndex = SUBINDEX_PROFILE_VELOCITY;
        oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulProfileVelocity, &errorInfo);

        if(oResult)
        {
            //Send acceleration
            usIndex = INDEX_PROFILE_ACCELERATION;
            usSubIndex = SUBINDEX_PROFILE_ACCELERATION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulProfileAcceleration, &errorInfo);
        }

        if(oResult)
        {
            //Send deceleration
            usIndex = INDEX_PROFILE_DECELERATION;
            usSubIndex = SUBINDEX_PROFILE_DECELERATION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulProfileDeceleration, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetTargetVelocity(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    int lTargetVelocity = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Read target velocity
        if(oResult)
        {
            usIndex = INDEX_TARGET_VELOCITY;
            usSubIndex = SUBINDEX_TARGET_VELOCITY;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lTargetVelocity, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lTargetVelocity, sizeof(lTargetVelocity));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVelocityProfile(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulProfileAcceleration = 0;
    DWORD ulProfileDeceleration = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Read profile acceleration
        if(oResult)
        {
            usIndex = INDEX_PROFILE_ACCELERATION;
            usSubIndex = SUBINDEX_PROFILE_ACCELERATION;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulProfileAcceleration, &errorInfo);
        }

        //Read profile deceleration
        if(oResult)
        {
            usIndex = INDEX_PROFILE_DECELERATION;
            usSubIndex = SUBINDEX_PROFILE_DECELERATION;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulProfileDeceleration, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulProfileAcceleration, sizeof(ulProfileAcceleration));
        p_pCommand->SetReturnParameterData(1, &ulProfileDeceleration, sizeof(ulProfileDeceleration));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_HaltVelocityMovement(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Send stop
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;
        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, STOP_PROFILE, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_MoveWithVelocity(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    int lTargetVelocity;
    BOOL oQuickStopActive = FALSE;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Set parameter data
        p_pCommand->GetParameterData(0, &lTargetVelocity, sizeof(lTargetVelocity));

        //Init
        oResult = TRUE;

        //Send target velocity
        if(oResult)
        {
            usIndex = INDEX_TARGET_VELOCITY;
            usSubIndex = SUBINDEX_TARGET_VELOCITY;
            oResult = WriteInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, lTargetVelocity, &errorInfo);
        }

        oResult = IsQuickStopStateActive(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oQuickStopActive, &errorInfo);

        //Leave quick stop state
        if(oResult && oQuickStopActive)
        {
            usIndex = INDEX_CONTROL_WORD;
            usSubIndex = SUBINDEX_CONTROL_WORD;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SWITCH_ON_AND_ENABLE_OPERATION, &errorInfo);
        }

        //Move
        if(oResult)
        {
            usIndex = INDEX_CONTROL_WORD;
            usSubIndex = SUBINDEX_CONTROL_WORD;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, MOVE_WITH_VELOCITY, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetVelocityProfile(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulProfileAcceleration;
    DWORD ulProfileDeceleration;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ulProfileAcceleration, sizeof(ulProfileAcceleration));
        p_pCommand->GetParameterData(1, &ulProfileDeceleration, sizeof(ulProfileDeceleration));

        //Init
        oResult = TRUE;

        //Set acceleration
        if(oResult)
        {
            usIndex = INDEX_PROFILE_ACCELERATION;
            usSubIndex = SUBINDEX_PROFILE_ACCELERATION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulProfileAcceleration, &errorInfo);
        }

        //Set deceleration
        if(oResult)
        {
            usIndex = INDEX_PROFILE_DECELERATION;
            usSubIndex = SUBINDEX_PROFILE_DECELERATION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulProfileDeceleration, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetCurrentMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const short PER_THOUSAND(1000);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    short sTargetCurrentMust(0);
    short sTargetTorque(0);
    WORD usContinuousCurrent(0);
    WORD usIndex(0);
    WORD usSubIndex(0);

    if(p_pCommand)
    {
        // Init
        oResult = TRUE;

        // Read target torque current parameter
        usIndex = INDEX_TARGET_TORQUE;
        usSubIndex = SUBINDEX_TARGET_TORQUE;
        if(oResult)
        {
            oResult = ReadInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &sTargetTorque, &errorInfo);
        }

        // Read nominal current
        usIndex = EPOS4_INDEX_MOTOR_DATA;
        usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
        if(oResult)
        {
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usContinuousCurrent, &errorInfo);
            sTargetCurrentMust = (short)(sTargetTorque * usContinuousCurrent / PER_THOUSAND);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &sTargetCurrentMust, sizeof(sTargetCurrentMust));
    }
    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetCurrentMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const short PER_THOUSAND(1000);
    const short ROUNDING_DIVISOR(10);
    const short ROUNDING_FACTOR(10);
    const short ROUNDING_VALUE(5);

    BOOL oResult(FALSE);
    WORD usIndex(0);
    WORD usSubIndex(0);
    WORD usContinuousCurrent(0);
    short sTargetCurrentMust(0);
    short sTargetTorque(0);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        // Get parameter data
        p_pCommand->GetParameterData(0, &sTargetCurrentMust, sizeof(sTargetCurrentMust));

        // Read nominal current
        usIndex = EPOS4_INDEX_MOTOR_DATA;
        usSubIndex = EPOS4_SUBINDEX_MAX_CON_CURRENT;
        if(oResult)
        {
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usContinuousCurrent, &errorInfo);
            if (usContinuousCurrent > 0)
            {
                if (sTargetCurrentMust >= 0)
                {
                    sTargetTorque = (short)(((sTargetCurrentMust * (PER_THOUSAND * ROUNDING_FACTOR) / usContinuousCurrent) + ROUNDING_VALUE) / ROUNDING_DIVISOR); // incl. rounding
                }
                else
                {
                    sTargetTorque = (short)(((sTargetCurrentMust * (PER_THOUSAND * ROUNDING_FACTOR) / usContinuousCurrent) - ROUNDING_VALUE) / ROUNDING_DIVISOR); // incl. rounding
                }
            }
        }

        // Write target torque current parameter
        usIndex = INDEX_TARGET_TORQUE;
        usSubIndex = SUBINDEX_TARGET_TORQUE;
        if(oResult)
        {
            oResult = WriteInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, sTargetTorque, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }
    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVelocityMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lVelocityMust(0);
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Target velocity
        usIndex = INDEX_TARGET_VELOCITY;
        usSubIndex = SUBINDEX_TARGET_VELOCITY;
        oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lVelocityMust, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lVelocityMust, sizeof(lVelocityMust));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetVelocityMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lVelocityMust(0);
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &lVelocityMust, sizeof(lVelocityMust));

        // Target velocity
        usIndex = INDEX_TARGET_VELOCITY;
        usSubIndex = SUBINDEX_TARGET_VELOCITY;
        oResult = WriteInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, lVelocityMust, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetPositionMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lPositionMust(0);
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Target position
        usIndex = INDEX_TARGET_POSITION;
        usSubIndex = SUBINDEX_TARGET_POSITION;
        oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lPositionMust, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lPositionMust, sizeof(lPositionMust));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetPositionMust(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lPositionMust(0);
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &lPositionMust, sizeof(lPositionMust));

        // Target position
        usIndex = INDEX_TARGET_POSITION;
        usSubIndex = SUBINDEX_TARGET_POSITION;
        oResult = WriteInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, lPositionMust, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_FindHome(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    __int8 bHomingMethod;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &bHomingMethod, sizeof(bHomingMethod));

        MapHomingMethodActualPositionToValue37(&bHomingMethod);

        //Set homing method
        usIndex = INDEX_HOMING_METHOD;
        usSubIndex = SUBINDEX_HOMING_METHOD;
        oResult = WriteInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, bHomingMethod, &errorInfo);

        //Send 0->1 Bit change
        if(oResult)
        {
            usIndex = INDEX_CONTROL_WORD;
            usSubIndex = SUBINDEX_CONTROL_WORD;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SWITCH_ON_AND_ENABLE_OPERATION, &errorInfo);
        }

        if(oResult)
        {
            usIndex = INDEX_CONTROL_WORD;
            usSubIndex = SUBINDEX_CONTROL_WORD;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, START_HOMING, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetHomingParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulHomingAcceleration = 0;
    DWORD ulSpeedIndex = 0;
    DWORD ulSpeedSwitch = 0;
    int lHomeOffset = 0;
    int lHomePosition = 0;
    WORD usCurrentThreshold = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Homing acceleration
        if(oResult)
        {
            usIndex = INDEX_HOMING_ACCELERATION;
            usSubIndex = SUBINDEX_HOMING_ACCELERATION;
            oResult= ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulHomingAcceleration, &errorInfo);
        }

        //Speed switch
        if(oResult)
        {
            usIndex = INDEX_HOMING_SPEED;
            usSubIndex = SUBINDEX_SPEED_SEARCH_SWITCH;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulSpeedSwitch, &errorInfo);
        }

        //Speed index
        if(oResult)
        {
            usIndex = INDEX_HOMING_SPEED;
            usSubIndex = SUBINDEX_SPEED_SEARCH_ZERO;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulSpeedIndex, &errorInfo);
        }

        //Home offset
        if(oResult)
        {
            usIndex = EPOS4_INDEX_HOME_OFFSET;
            usSubIndex = EPOS4_SUBINDEX_HOME_OFFSET;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, (DWORD*) &lHomeOffset, &errorInfo);
        }

        //CurrentThreshold
        if (oResult)
        {
            usIndex = EPOS4_INDEX_HOME_CURRENT_THRESHOLD;
            usSubIndex = EPOS4_SUBINDEX_HOME_CURRENT_THRESHOLD;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usCurrentThreshold, &errorInfo);

            oResult = HandleBackwardsCompatiblity(oResult, errorInfo);
        }

        //Home position
        if(oResult)
        {
            usIndex = EPOS4_INDEX_HOME_POSITION;
            usSubIndex = EPOS4_SUBINDEX_HOME_POSITION;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, (DWORD*) &lHomePosition, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulHomingAcceleration, sizeof(ulHomingAcceleration));
        p_pCommand->SetReturnParameterData(1, &ulSpeedSwitch, sizeof(ulSpeedSwitch));
        p_pCommand->SetReturnParameterData(2, &ulSpeedIndex, sizeof(ulSpeedIndex));
        p_pCommand->SetReturnParameterData(3, &lHomeOffset, sizeof(lHomeOffset));
        p_pCommand->SetReturnParameterData(4, &usCurrentThreshold, sizeof(usCurrentThreshold));
        p_pCommand->SetReturnParameterData(5, &lHomePosition, sizeof(lHomePosition));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetHomingParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulHomingAcceleration = 0;
    DWORD ulSpeedIndex = 0;
    DWORD ulSpeedSwitch = 0;
    int lHomeOffset = 0;
    int lHomePosition = 0;
    WORD usCurrentThreshold = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Get parameter data
        p_pCommand->ResetStatus();
        p_pCommand->GetParameterData(0, &ulHomingAcceleration, sizeof(ulHomingAcceleration));
        p_pCommand->GetParameterData(1, &ulSpeedSwitch, sizeof(ulSpeedSwitch));
        p_pCommand->GetParameterData(2, &ulSpeedIndex, sizeof(ulSpeedIndex));
        p_pCommand->GetParameterData(3, &lHomeOffset, sizeof(lHomeOffset));
        p_pCommand->GetParameterData(4, &usCurrentThreshold, sizeof(usCurrentThreshold));
        p_pCommand->GetParameterData(5, &lHomePosition, sizeof(lHomePosition));

        //Homing acceleration
        if(oResult)
        {
            usIndex = INDEX_HOMING_ACCELERATION;
            usSubIndex = SUBINDEX_HOMING_ACCELERATION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulHomingAcceleration, &errorInfo);
        }

        //Speed switch
        if(oResult)
        {
            usIndex = INDEX_HOMING_SPEED;
            usSubIndex = SUBINDEX_SPEED_SEARCH_SWITCH;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulSpeedSwitch, &errorInfo);
        }

        //Speed index
        if(oResult)
        {
            usIndex = INDEX_HOMING_SPEED;
            usSubIndex = SUBINDEX_SPEED_SEARCH_ZERO;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulSpeedIndex, &errorInfo);
        }

        //Home offset
        if(oResult)
        {
            usIndex = EPOS4_INDEX_HOME_OFFSET;
            usSubIndex = EPOS4_SUBINDEX_HOME_OFFSET;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, (DWORD)lHomeOffset, &errorInfo);
        }

        //CurrentThreshold
        if (oResult)
        {
            usIndex = EPOS4_INDEX_HOME_CURRENT_THRESHOLD;
            usSubIndex = EPOS4_SUBINDEX_HOME_CURRENT_THRESHOLD;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usCurrentThreshold, &errorInfo);

            oResult = HandleBackwardsCompatiblity(oResult, errorInfo);
        }

        //Home position
        if(oResult)
        {
            usIndex = EPOS4_INDEX_HOME_POSITION;
            usSubIndex = EPOS4_SUBINDEX_HOME_POSITION;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, (DWORD)lHomePosition, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_DefinePosition(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lHomePosition = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        // Get parameter data
        p_pCommand->ResetStatus();
        p_pCommand->GetParameterData(0, &lHomePosition, sizeof(lHomePosition));

        // Home position
        usIndex = EPOS4_INDEX_HOME_POSITION;
        usSubIndex = EPOS4_SUBINDEX_HOME_POSITION;
        oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, (DWORD)lHomePosition, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_StopHoming(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;
        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, STOP_HOMING, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetHomingState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oHomingAttained;
    BOOL oHomingError;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usHomingState(0);
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        //Read Homing State
        usIndex = INDEX_STATUS_WORD;
        usSubIndex = SUBINDEX_STATUS_WORD;
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usHomingState, &errorInfo);
        if(oResult)
        {
            //Check homing attained
            usHomingState &= HS_HOMING_ATTAINED;
            if(usHomingState == HS_HOMING_ATTAINED)
            {
                oHomingAttained = TRUE;
            }
            else
            {
                oHomingAttained = FALSE;
            }

            //Check homing error
            usHomingState &= HS_HOMING_ERROR;
            if(usHomingState == HS_HOMING_ERROR)
            {
                oHomingError = TRUE;
            }
            else
            {
                oHomingError = FALSE;
            }
        }

        //Set return parameter Data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oHomingAttained, sizeof(oHomingAttained));
        p_pCommand->SetReturnParameterData(1, &oHomingError, sizeof(oHomingError));
    }
    else
    {
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetCurrentIs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lCurrentIs;
    short sCurrentIs;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        oResult = TRUE;

        if(oResult)
        {
            usIndex = EPOS4_INDEX_CURRENT_ACTUAL_VALUE;
            usSubIndex = EPOS4_SUBINDEX_CURRENT_ACTUAL_VALUE;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lCurrentIs, &errorInfo);
        }

        //Convert to command datatypes
        sCurrentIs = (short)lCurrentIs;

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &sCurrentIs, sizeof(sCurrentIs));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetCurrentIsAveraged(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usIndex;
    WORD usSubIndex;
    int lCurrentIs;
    short sCurrentIs;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        if(oResult)
        {
            usIndex = EPOS4_INDEX_CURRENT_ACTUAL_VALUE;
            usSubIndex = EPOS4_SUBINDEX_CURRENT_ACTUAL_AVERAGE_VALUE;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lCurrentIs, &errorInfo);
        }

        //Convert to command datatypes
        sCurrentIs = (short)lCurrentIs;

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &sCurrentIs, sizeof(sCurrentIs));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMovementState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    WORD usDeviceState = 0;
    BOOL oResult(FALSE);
    BOOL oTargetReached;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Target reached
        if(oResult)
        {
            //Read device state
            usIndex = INDEX_STATUS_WORD;
            usSubIndex = SUBINDEX_STATUS_WORD;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usDeviceState, &errorInfo);
            if(oResult)
            {
                usDeviceState = usDeviceState & 0x7FF;

                //Check target reached
                usDeviceState &= DS_TARGET_REACHED;
                if(usDeviceState == DS_TARGET_REACHED)
                    oTargetReached = TRUE;
                else
                    oTargetReached = FALSE;
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oTargetReached, sizeof(oTargetReached));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetPositionIs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    int lPositionIs = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        if(oResult)
        {
            usIndex = INDEX_POSITION_ACTUAL_VALUE;
            usSubIndex = SUBINDEX_POSITION_ACTUAL_VALUE;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lPositionIs, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lPositionIs, sizeof(lPositionIs));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVelocityIs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    int lVelocityIs = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        if(oResult)
        {
            usIndex = INDEX_VELOCITY_ACTUAL_VALUE;
            usSubIndex = SUBINDEX_VELOCITY_ACTUAL_VALUE;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lVelocityIs, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lVelocityIs, sizeof(lVelocityIs));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVelocityIsAveraged(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    int lVelocityIs = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        oResult = TRUE;

        if(oResult)
        {
            usIndex = EPOS4_INDEX_VELOCITY_ACTUAL_AVERAGE_VALUE;
            usSubIndex = EPOS4_SUBINDEX_VELOCITY_ACTUAL_AVERAGE_VALUE;
            oResult = ReadInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &lVelocityIs, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lVelocityIs, sizeof(lVelocityIs));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_DigitalInputConfiguration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex, usSubIndex;
    BOOL oResult(FALSE);
    WORD usDigitalInputNb;
    WORD usConfiguration;
    BYTE ubTranslatedConfiguration(0);
    BOOL oMask;
    BOOL oPolarity;
    BOOL oExecutionMask;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usDigitalInputNb, sizeof(usDigitalInputNb));
        p_pCommand->GetParameterData(1, &usConfiguration, sizeof(usConfiguration));
        p_pCommand->GetParameterData(2, &oMask, sizeof(oMask));
        p_pCommand->GetParameterData(3, &oPolarity, sizeof(oPolarity));
        p_pCommand->GetParameterData(4, &oExecutionMask, sizeof(oExecutionMask));

        oResult = TRUE;
        switch (usConfiguration)
        {
        case DIC_NEGATIVE_LIMIT_SWITCH:
        {
            if (oExecutionMask == 1)
            {
                ubTranslatedConfiguration = EPOS4_DIC_NEGATIVE_LIMIT_SWITCH;
            }
            else
            {
                ubTranslatedConfiguration = EPOS4_DIC_NEGATIVE_LIMIT_SWITCH_CHECK;
            }
            break;
        }
        case DIC_POSITIVE_LIMIT_SWITCH:
        {
            if (oExecutionMask == 1)
            {
                ubTranslatedConfiguration = EPOS4_DIC_POSITIVE_LIMIT_SWITCH;
            }
            else
            {
                ubTranslatedConfiguration = EPOS4_DIC_POSITIVE_LIMIT_SWITCH_CHECK;
            }
            break;
        }
        case DIC_HOME_SWITCH:       ubTranslatedConfiguration = EPOS4_DIC_HOME_SWITCH; break;
        case DIC_NO_FUNCTIONALITY:  ubTranslatedConfiguration = EPOS4_DIC_NO_FUNCTIONALITY; break;
        case DIC_GENERAL_PURPOSE_H: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_H; break;
        case DIC_GENERAL_PURPOSE_G: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_G; break;
        case DIC_GENERAL_PURPOSE_F: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_F; break;
        case DIC_GENERAL_PURPOSE_E: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_E; break;
        case DIC_GENERAL_PURPOSE_D: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_D; break;
        case DIC_GENERAL_PURPOSE_C: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_C; break;
        case DIC_GENERAL_PURPOSE_B: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_B; break;
        case DIC_GENERAL_PURPOSE_A: ubTranslatedConfiguration = EPOS4_DIC_GENERAL_PURPOSE_A; break;
        case DIC_TOUCH_PROBE_1:     ubTranslatedConfiguration = EPOS4_DIC_TOUCH_PROBE_1; break;
        default:
        {
            // Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_CONFIG_NOT_SUPPORTED, &errorInfo);
            oResult = FALSE;
            break;
        }
        }

        if ((oResult) && (oMask != 1))
        {
            m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_MASKING_DIGITAL_INPUT_NOT_SUPPORTED, &errorInfo);
            oResult = FALSE;
        }

        //Write polarity to digital input configuration
        if (oResult)
        {
            WORD usDigitalInputPolarityValue = (WORD)((oPolarity & 0x0001) << (usDigitalInputNb-1));
            WORD usDigitalInputPolarityMask = (WORD)(0x01 << (usDigitalInputNb-1));
            usIndex = INDEX_DIGITAL_INPUT_PROPERTIES;
            usSubIndex = SUBINDEX_DIGITAL_INPUT_POLARITY_CONFIGURATION;
            oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usDigitalInputPolarityValue, usDigitalInputPolarityMask, &errorInfo);
        }

        //Write configuration to digital input
        if (oResult)
        {
            usIndex = INDEX_DIGITAL_INPUT_CONFIGURATION;
            usSubIndex = (BYTE)usDigitalInputNb;
            oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ubTranslatedConfiguration, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_DigitalOutputConfiguration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oMask;
    BOOL oPolarity;
    BOOL oResult(FALSE);
    BOOL oState;
    BYTE ubTranslatedConfiguration(0);
    CErrorInfo errorInfo;
    WORD usConfiguration;
    WORD usDigitalOutputNb;
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usDigitalOutputNb, sizeof(usDigitalOutputNb));
        p_pCommand->GetParameterData(1, &usConfiguration, sizeof(usConfiguration));
        p_pCommand->GetParameterData(2, &oState, sizeof(oState));
        p_pCommand->GetParameterData(3, &oMask, sizeof(oMask));
        p_pCommand->GetParameterData(4, &oPolarity, sizeof(oPolarity));

        oResult = TRUE;
        switch (usConfiguration)
        {
        case DOC_READY_FAULT:       ubTranslatedConfiguration = EPOS4_DOC_READY_FAULT; break;
        case DOC_GENERAL_PURPOSE_C: ubTranslatedConfiguration = EPOS4_DOC_GENERAL_PURPOSE_C; break;
        case DOC_GENERAL_PURPOSE_B: ubTranslatedConfiguration = EPOS4_DOC_GENERAL_PURPOSE_B; break;
        case DOC_GENERAL_PURPOSE_A: ubTranslatedConfiguration = EPOS4_DOC_GENERAL_PURPOSE_A; break;
        default:
        {
            //Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_CONFIG_NOT_SUPPORTED, &errorInfo);
            oResult = FALSE;
            break;
        }
        }

        //Write polarity to digital output configuration
        if (oResult)
        {
            WORD usDigitalOutputPolarityValue = (WORD)((oPolarity & 0x0001) << (usDigitalOutputNb - 1));
            WORD usDigitalOutputPolarityMask = (WORD)(0x01 << (usDigitalOutputNb - 1));
            usIndex = INDEX_DIGITAL_OUTPUT_POLARITY_CONFIGURATION;
            usSubIndex = SUBINDEX_DIGITAL_OUTPUT_POLARITY_CONFIGURATION;
            oResult = UpdateBitsInUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usDigitalOutputPolarityValue, usDigitalOutputPolarityMask, &errorInfo);
        }

        //Write configuration to digital output
        if (oResult)
        {
            usIndex = INDEX_DIGITAL_OUTPUT_CONFIGURATION;
            usSubIndex = (BYTE)usDigitalOutputNb;
            oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ubTranslatedConfiguration, &errorInfo);
        }

        //Write state to digital output
        if (oResult)
        {
            DWORD ulDigitalOutputValue = (DWORD)((oState & 0x00000001) << (ubTranslatedConfiguration));
            DWORD ulDigitalOutputMask = (DWORD)(0x00000001 << (ubTranslatedConfiguration));
            usIndex = INDEX_DIGITAL_OUTPUT_STATE;
            usSubIndex = SUBINDEX_DIGITAL_PHYSICAL_OUTPUT_STATE;
            oResult = UpdateBitsInUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulDigitalOutputValue, ulDigitalOutputMask, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_AnalogOutputConfiguration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usAnalogOutputNb(0);
    WORD usConfiguration(0);
    BYTE ubConfiguration(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usAnalogOutputNb, sizeof(usAnalogOutputNb));
        p_pCommand->GetParameterData(1, &usConfiguration, sizeof(usConfiguration));

        //Convert
        ubConfiguration = (BYTE)usConfiguration;

        //Write Analog Output Configuration
        if (WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_ANALOG_OUTPUT_CONFIGURATION, usAnalogOutputNb, ubConfiguration, &errorInfo))
        {
            oResult = TRUE;
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetAllDigitalInputs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex, usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulInputs(0);
    WORD usInputTranslated(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = TRUE;

        //Read state of all digital inputs
        if (oResult)
        {
            usIndex = INDEX_DIGITAL_INPUT_STATE;
            usSubIndex = SUBINDEX_DIGITAL_PHYSICAL_INPUT_STATE;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulInputs, &errorInfo);
        }

        if (oResult)
        {
            // Translate
            if (IsBitSetInDword(ulInputs, EPOS4_DIC_NEGATIVE_LIMIT_SWITCH) || (IsBitSetInDword(ulInputs, EPOS4_DIC_NEGATIVE_LIMIT_SWITCH_CHECK)))
            {
                usInputTranslated |= (0x01 << DIC_NEGATIVE_LIMIT_SWITCH);
            }

            if (IsBitSetInDword(ulInputs, EPOS4_DIC_POSITIVE_LIMIT_SWITCH) || (IsBitSetInDword(ulInputs, EPOS4_DIC_POSITIVE_LIMIT_SWITCH_CHECK)))
            {
                usInputTranslated |= (0x01 << DIC_POSITIVE_LIMIT_SWITCH);
            }

            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_HOME_SWITCH, usInputTranslated, DIC_HOME_SWITCH);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_H, usInputTranslated, DIC_GENERAL_PURPOSE_H);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_G, usInputTranslated, DIC_GENERAL_PURPOSE_G);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_F, usInputTranslated, DIC_GENERAL_PURPOSE_F);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_E, usInputTranslated, DIC_GENERAL_PURPOSE_E);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_D, usInputTranslated, DIC_GENERAL_PURPOSE_D);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_C, usInputTranslated, DIC_GENERAL_PURPOSE_C);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_B, usInputTranslated, DIC_GENERAL_PURPOSE_B);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_GENERAL_PURPOSE_A, usInputTranslated, DIC_GENERAL_PURPOSE_A);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_TOUCH_PROBE_1, usInputTranslated, DIC_TOUCH_PROBE_1);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_DRIVE_ENABLE, usInputTranslated, DIC_DRIVE_ENABLE);
            MapBitFromDwordToWord(ulInputs, EPOS4_DIC_QUICK_STOP, usInputTranslated, DIC_QUICK_STOP);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usInputTranslated, sizeof(usInputTranslated));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetAllDigitalOutputs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    DWORD ulOutputs(0);
    WORD usOutputsTranslated(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = TRUE;

        //Read state of all digital outputs
        if (oResult)
        {
            usIndex = INDEX_DIGITAL_OUTPUT_STATE;
            usSubIndex = SUBINDEX_DIGITAL_PHYSICAL_OUTPUT_STATE;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulOutputs, &errorInfo);
        }

        if (oResult)
        {
            MapBitFromDwordToWord(ulOutputs, EPOS4_DOC_READY_FAULT, usOutputsTranslated, DOC_READY_FAULT);
            MapBitFromDwordToWord(ulOutputs, EPOS4_DOC_GENERAL_PURPOSE_A, usOutputsTranslated, DOC_GENERAL_PURPOSE_A);
            MapBitFromDwordToWord(ulOutputs, EPOS4_DOC_GENERAL_PURPOSE_B, usOutputsTranslated, DOC_GENERAL_PURPOSE_B);
            MapBitFromDwordToWord(ulOutputs, EPOS4_DOC_GENERAL_PURPOSE_C, usOutputsTranslated, DOC_GENERAL_PURPOSE_C);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usOutputsTranslated, sizeof(usOutputsTranslated));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetAnalogInput(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    WORD usNumber;
    WORD usAnalog = 0;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = TRUE;

        //Get parameter data
        p_pCommand->GetParameterData(0, &usNumber, sizeof(usNumber));

        //Read value of one analog input
        if (oResult)
        {
            usIndex = INDEX_ANALOG_INPUTS;
            usSubIndex = usNumber;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usAnalog, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usAnalog, sizeof(usAnalog));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetAnalogInputVoltage(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usNumber;
    short sVoltageValue(0);
    long lVoltageValue(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usNumber, sizeof(usNumber));

        //Read value of analog input
        if (ReadInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_ANALOG_INPUTS, usNumber, &sVoltageValue, &errorInfo))
        {
            lVoltageValue = (long)sVoltageValue;
            oResult = TRUE;
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &lVoltageValue, sizeof(lVoltageValue));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetAllDigitalOutputs(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    WORD usOutputs(0);
    DWORD ulOutputsTranslated(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usOutputs, sizeof(usOutputs));

        oResult = TRUE;
        MapBitFromWordToDword(usOutputs, DOC_READY_FAULT, ulOutputsTranslated, EPOS4_DOC_READY_FAULT);
        MapBitFromWordToDword(usOutputs, DOC_GENERAL_PURPOSE_A, ulOutputsTranslated, EPOS4_DOC_GENERAL_PURPOSE_A);
        MapBitFromWordToDword(usOutputs, DOC_GENERAL_PURPOSE_B, ulOutputsTranslated, EPOS4_DOC_GENERAL_PURPOSE_B);
        MapBitFromWordToDword(usOutputs, DOC_GENERAL_PURPOSE_C, ulOutputsTranslated, EPOS4_DOC_GENERAL_PURPOSE_C);

        //Write state to all digital outputs
        if (oResult)
        {
            usIndex = INDEX_DIGITAL_OUTPUT_STATE;
            usSubIndex = SUBINDEX_DIGITAL_PHYSICAL_OUTPUT_STATE;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulOutputsTranslated, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVelocityUnits(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubDenominator(0);
    BYTE ubNumeration(0);
    BYTE ubVelocityDimension = 0;
    CErrorInfo errorInfo;
    char bVelocityNotation = 0;
    DWORD ulVelocityUnitValue(0);
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        oResult = TRUE;

        //Velocity dimension
        if (oResult)
        {
            usIndex = INDEX_VELOCITY_UNIT;
            usSubIndex = SUBINDEX_VELOCITY_UNIT;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulVelocityUnitValue, &errorInfo);
        }

        //Find notation standard, deci, centi, milli
        bVelocityNotation = ulVelocityUnitValue >> 24;

        //Find dimension rpm
        ubNumeration = ulVelocityUnitValue >> 16 & 0xFF;
        ubDenominator = ulVelocityUnitValue >> 8 & 0xFF;

        if ((ubNumeration == EPOS4_UNIT_REVOLUTION) && (ubDenominator == EPOS4_UNIT_MINUT))
        {
            ubVelocityDimension = VD_RPM;
        }
        else
        {
            //Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_CONFIG_NOT_SUPPORTED, &errorInfo);
            oResult = FALSE;
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ubVelocityDimension, sizeof(ubVelocityDimension));
        p_pCommand->SetReturnParameterData(1, &bVelocityNotation, sizeof(bVelocityNotation));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetVelocityUnits(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubDenominator(0);
    BYTE ubNumeration(0);
    BYTE ubVelocityDimension = 0;
    CErrorInfo errorInfo;
    char bVelocityNotation = 0;
    WORD usIndex;
    WORD usSubIndex;

    if (p_pCommand)
    {
        oResult = TRUE;

        //Get Parameter Data
        p_pCommand->GetParameterData(0, &ubVelocityDimension, sizeof(ubVelocityDimension));
        p_pCommand->GetParameterData(1, &bVelocityNotation, sizeof(bVelocityNotation));

        switch (ubVelocityDimension)
        {
        case VD_RPM:
        {
            ubNumeration = EPOS4_UNIT_REVOLUTION;
            ubDenominator = EPOS4_UNIT_MINUT;
            break;
        }
        default:
        {
            //Generate error
            if (m_pErrorHandling) m_pErrorHandling->GetError(ERROR_VIRTUAL_DEVICE_CONFIG_NOT_SUPPORTED, &errorInfo);
            oResult = FALSE;
            break;
        }
        }

        //Velocity dimension
        if (oResult)
        {
            DWORD ulVelocityUnitValue = (bVelocityNotation << 24) | (ubNumeration << 16) | (ubDenominator << 8);
            usIndex = INDEX_VELOCITY_UNIT;
            usSubIndex = SUBINDEX_VELOCITY_UNIT;
            oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, ulVelocityUnitValue, &errorInfo);
        }

        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ClearLSSFrameQueue(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usTimeout, CErrorInfo* p_pErrorInfo)
{
    const int MAX_COUNT = 127;
    const int DATA_SIZE = 8;

    BOOL oResult(TRUE);

    BYTE data[DATA_SIZE];
    int lCount = 0;

    while (DCS_ReadLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_usTimeout, data, DATA_SIZE, p_pErrorInfo))
    {
        MmcSleep(0);
        lCount++;
        if (lCount > MAX_COUNT)
        {
            oResult = false;
            break;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ClearAllDeviceErrors(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE FIELDBUS_ETHERCAT(0x02);
    const WORD BIT_MASK_FAULT(0x004F);

    BOOL oResult(FALSE);
    BOOL oFault(TRUE);
    WORD usDeviceState(0);
    WORD usIndex(0);
    WORD usSubIndex(0);

    if(p_pManager && p_pErrorInfo)
    {
        //Check fault state
        oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, p_pErrorInfo);
        if(oResult)
        {
            usDeviceState &= BIT_MASK_FAULT;
            oFault = (usDeviceState == DS_FAULT);
        }

        if(oResult && oFault)
        {
            if(IsCorrectFieldbus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, FIELDBUS_ETHERCAT, p_pErrorInfo))
            {
                usIndex = INDEX_DIAGNOSIS_HISTORY;
                usSubIndex = SUBINDEX_DIAGNOSIS_HISTORY_NEWEST_ACK_MESSAGE;
                oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, CLEAR_ALL_MESSAGES, p_pErrorInfo);
            }
        }

        usIndex = INDEX_CONTROL_WORD;
        usSubIndex = SUBINDEX_CONTROL_WORD;

        if(oResult && oFault)
        {
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, DISABLE_VOLTAGE, p_pErrorInfo);
        }

        if(oResult && oFault)
        {
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, FAULT_RESET, p_pErrorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::UpdateDiagnosisHistory(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubMaximumMessages(0);
    BYTE ubNewestMessage(0);
    BYTE ubNewestAckMessage(0);
    DWORD ulNbOfBytesRead(0);

    if(m_pDiagnosisHistory)
    {
        //Maximum messages
        oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIAGNOSIS_HISTORY, SUBINDEX_DIAGNOSIS_HISTORY_MAX_MESSAGES, &ubMaximumMessages, p_pErrorInfo);

        //Newest message
        if(oResult && !ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIAGNOSIS_HISTORY, SUBINDEX_DIAGNOSIS_HISTORY_NEWEST_MESSAGE, &ubNewestMessage, p_pErrorInfo))
        {
            oResult = FALSE;
        }

        //Newest ack message
        if(oResult && !ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIAGNOSIS_HISTORY, SUBINDEX_DIAGNOSIS_HISTORY_NEWEST_ACK_MESSAGE, &ubNewestAckMessage, p_pErrorInfo))
        {
            oResult = FALSE;
        }

        if(oResult)
        {
            if(m_pDiagnosisHistory->InitDiagnosisHistory(ubMaximumMessages, ubNewestMessage, ubNewestAckMessage))
            {
                //Read messages
                for(BYTE i = 0; i < ubMaximumMessages; i++)
                {
                    SDiagnosisLongMessage message;

                    if(ReadDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIAGNOSIS_HISTORY, SUBINDEX_DIAGNOSIS_HISTORY_FIRST_MESSAGE + i, &message, sizeof(message), &ulNbOfBytesRead, p_pErrorInfo))
                    {
                        if(!m_pDiagnosisHistory->AddDiagnoseMessage(message))
                        {
                            if(m_pErrorHandling)
                            {
                                m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                            }

                            oResult = FALSE;
                            break;
                        }
                    }
                    else
                    {
                        oResult = FALSE;
                        break;
                    }
                }
            }
            else
            {
                if(m_pErrorHandling)
                {
                    m_pErrorHandling->GetError(ERROR_INTERNAL, p_pErrorInfo);
                }

                oResult = FALSE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ClearFault(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = ClearAllDeviceErrors(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetDisableState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD BIT_MASK_DISABLE(0x004F);

    BOOL oIsDisabled;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usDeviceState = 0;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Read disable state
        if(oResult)
        {
            oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, &errorInfo);

            if(oResult)
            {
                usDeviceState &= BIT_MASK_DISABLE;
                if(usDeviceState == DS_SWITCH_ON_DISABLED)
                {
                    oIsDisabled = TRUE;
                }
                else
                {
                    oIsDisabled = FALSE;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oIsDisabled, sizeof(oIsDisabled));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetEnableState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD BIT_MASK_ENABLE(0x006F);

    WORD usDeviceState = 0;
    BOOL oResult(FALSE);
    BOOL oIsEnabled;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Read enable state
        if(oResult)
        {
            oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, &errorInfo);

            if(oResult)
            {
                usDeviceState &= BIT_MASK_ENABLE;
                if(usDeviceState == DS_OPERATION_ENABLED)
                {
                    oIsEnabled = TRUE;
                }
                else
                {
                    oIsEnabled = FALSE;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oIsEnabled, sizeof(oIsEnabled));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetFaultState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD BIT_MASK_FAULT(0x004F);

    WORD usDeviceState = 0;
    BOOL oResult(FALSE);
    BOOL oIsInFault;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Read fault state
        if(oResult)
        {
            oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, &errorInfo);

            if(oResult)
            {
                usDeviceState &= BIT_MASK_FAULT;
                if(usDeviceState == DS_FAULT)
                    oIsInFault = TRUE;
                else
                    oIsInFault = FALSE;
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oIsInFault, sizeof(oIsInFault));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsHaltActive(WORD p_usControlWord)
{
    BOOL oResult = FALSE;
    const int HALT_ACTIVE_BIT = 8;
    bitset<sizeof(WORD)*8> controlWordBits(p_usControlWord);

    oResult = controlWordBits.test(HALT_ACTIVE_BIT);

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsQuickStopActive(WORD p_usStatusWord)
{
    BOOL oResult = FALSE;
    const int QUICK_STOP_ACTIVE_BIT = 5;
    bitset<sizeof(WORD)*8> statusWordBits(p_usStatusWord);

    oResult = !statusWordBits.test(QUICK_STOP_ACTIVE_BIT);

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsQuickStopStateActive(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL& p_oQuickStopActive, CErrorInfo* p_pErrorInfo)
{
    const WORD BIT_MASK_QUICK_STOP(0x006F);

    BOOL oResult = FALSE;
    WORD usDeviceState = 0;

    p_oQuickStopActive = FALSE;

    oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, p_pErrorInfo);

    if(oResult)
    {
        usDeviceState &= BIT_MASK_QUICK_STOP;
        if(usDeviceState == DS_QUICK_STOP_ACTIV)
        {
            p_oQuickStopActive = TRUE;
        }
        else
        {
            p_oQuickStopActive = FALSE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetQuickStopState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BOOL oIsQuickStopActive = FALSE;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = IsQuickStopStateActive(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oIsQuickStopActive, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oIsQuickStopActive, sizeof(oIsQuickStopActive));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetDisableState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = SetDisableState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetEnableState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    vector<EDeviceState> eDeviceStates(2);
    WORD usDeviceState = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, &errorInfo);
        if(oResult)
        {
            usIndex = INDEX_CONTROL_WORD;
            usSubIndex = SUBINDEX_CONTROL_WORD;

            switch(usDeviceState)
            {
                case DS_OPERATION_ENABLED:
                    break;

                case DS_SWITCH_ON_DISABLED:
                    if(oResult)
                        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SHUTDOWN, &errorInfo);
                    if(oResult)
                        oResult = CheckDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DS_READY_TO_SWITCH_ON, &errorInfo);

                case DS_READY_TO_SWITCH_ON:
                case DS_SWITCHED_ON:
                case DS_QUICK_STOP_ACTIV:
                    if (oResult)
                    {
                        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, SWITCH_ON_AND_ENABLE_OPERATION, &errorInfo);
                    }
                    if (oResult)
                    {
                        eDeviceStates[0] = DS_OPERATION_ENABLED;
                        eDeviceStates[1] = DS_QUICK_STOP_ACTIV;

                        oResult = CheckDeviceStates(p_pManager, p_hDCS_Handle, p_hTransactionHandle, eDeviceStates, &errorInfo);
                    }
                    break;

                default:
                    oResult = FALSE;
            }
        }

        //set error
        if (!oResult && errorInfo.GetErrorCode() == NO_ERROR_DETECTED)
        {
            errorInfo.Init(ERROR_EXECUTING_COMMAND);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetQuickStopState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    WORD usDeviceState = 0;
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Read device state
        oResult = ReadDeviceState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &usDeviceState, &errorInfo);
        if(oResult)
        {
            if(((usDeviceState & DS_QUICK_STOP_ACTIV) || (usDeviceState & DS_OPERATION_ENABLED))&&(oResult))
            {
                if((usDeviceState & DS_OPERATION_ENABLED) == DS_OPERATION_ENABLED)
                {
                    //Set quick stop
                    usIndex = INDEX_CONTROL_WORD;
                    usSubIndex = SUBINDEX_CONTROL_WORD;
                    oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, QUICK_STOP, &errorInfo);
                }
                else
                {
                    oResult = TRUE;
                }
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetOperationMode(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    __int8 bMode;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &bMode, sizeof(bMode));

        SimulateCurrentModeThroughCyclicSyncTorqueMode(&bMode);
        SimulatePositionModeThroughCyclicSyncPositionMode(&bMode);
        SimulateVelocityModeThroughCyclicSyncVelocityMode(&bMode);

        usIndex = INDEX_OPERATION_MODE;
        usSubIndex = SUBINDEX_OPERATION_MODE;
        oResult = WriteInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, bMode, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

void CGatewayDriveToEpos4::SimulateCurrentModeThroughCyclicSyncTorqueMode(__int8* p_pbMode)
{
    if(p_pbMode != 0)
    {
        if (*p_pbMode == OM_CURRENT_MODE)
        {
            *p_pbMode = OM_CYCLIC_SYNC_TORQUE_MODE;
        }
    }
}

void CGatewayDriveToEpos4::SimulatePositionModeThroughCyclicSyncPositionMode(__int8* p_pbMode)
{
    if (p_pbMode != 0)
    {
        if (*p_pbMode == OM_POSITION_MODE)
        {
            *p_pbMode = OM_CYCLIC_SYNC_POSITION_MODE;
        }
    }
}

void CGatewayDriveToEpos4::SimulateVelocityModeThroughCyclicSyncVelocityMode(__int8* p_pbMode)
{
    if (p_pbMode != 0)
    {
        if (*p_pbMode == OM_VELOCITY_MODE)
        {
            *p_pbMode = OM_CYCLIC_SYNC_VELOCITY_MODE;
        }
    }
}

void CGatewayDriveToEpos4::MapHomingMethodActualPositionToValue37(__int8* p_pbMode)
{
    const __int8 HOMING_METHOD_ACTUAL_POSITION(37);
    const __int8 HOMING_METHOD_ACTUAL_POSITION_OBSOLETE(35);

    if(p_pbMode != 0)
    {
        if (*p_pbMode == HOMING_METHOD_ACTUAL_POSITION_OBSOLETE)
        {
            *p_pbMode = HOMING_METHOD_ACTUAL_POSITION;
        }
    }
}

BOOL CGatewayDriveToEpos4::Process_GetOperationMode(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    WORD usIndex;
    WORD usSubIndex;
    BOOL oResult(FALSE);
    __int8 bMode = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        usIndex = INDEX_OPERATION_MODE_DISPLAY;
        usSubIndex = SUBINDEX_OPERATION_MODE_DISPLAY;
        oResult = ReadInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &bMode, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &bMode, sizeof(bMode));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ResetDevice(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oBroadcast = FALSE;
    BOOL oWaitForReboot = FALSE;
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &oWaitForReboot, sizeof(oWaitForReboot));

        oResult = ResetDevice(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, oBroadcast, oWaitForReboot, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_WaitForReboot(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        //Wait for reboot
        oResult = WaitForReboot(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ClearDeviceErrors(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = ClearAllDeviceErrors(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetNbOfDeviceErrors(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BYTE FIELDBUS_ETHERCAT(0x02);

    BOOL oResult(FALSE);
    BYTE ubNbOfDeviceError = 0;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        if(IsCorrectFieldbus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, FIELDBUS_ETHERCAT, &errorInfo))
        {
            oResult = UpdateDiagnosisHistory(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

            //Number of diagnosis messages
            if(oResult && m_pDiagnosisHistory)
            {
                ubNbOfDeviceError = m_pDiagnosisHistory->GetNbOfDiagnoseMessages();
            }
        }
        else
        {
            //Number of device errors
            oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_ERROR_HISTORY, SUBINDEX_NB_OF_ERRORS, &ubNbOfDeviceError, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ubNbOfDeviceError, sizeof(ubNbOfDeviceError));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetDeviceErrorCode(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BYTE FIELDBUS_ETHERCAT(0x02);

    BOOL oResult(FALSE);
    BYTE ubErrorNumber(0);
    DWORD ulDeviceErrorCode(0);
    CErrorInfo errorInfo;
    SDiagnosisMessage message;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ubErrorNumber, sizeof(ubErrorNumber));

        if(IsCorrectFieldbus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, FIELDBUS_ETHERCAT, &errorInfo))
        {
            if(m_pDiagnosisHistory)
            {
                if(m_pDiagnosisHistory->GetDiagnoseMessage(ubErrorNumber, message))
                {
                    ulDeviceErrorCode = message.ulDiagCode_ErrorCode;

                    oResult = TRUE;
                }
            }
        }
        else
        {
            //Code from device error
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_ERROR_HISTORY, ubErrorNumber, &ulDeviceErrorCode, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulDeviceErrorCode, sizeof(ulDeviceErrorCode));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GotoSafeState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = SetDisableState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetDataRecordingParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD CORRECTION_10_MICRO_SECONDS_TO_100_MICRO_SECONDS(10);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usPrecedingSamples(0);
    WORD usSamplingPeriod(0);
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usSamplingPeriod, sizeof(usSamplingPeriod));
        p_pCommand->GetParameterData(1, &usPrecedingSamples, sizeof(usPrecedingSamples));

        // Sampling period
        usIndex = INDEX_INTERNAL_RECORDER_CONFIGURATION;
        usSubIndex = SUBINDEX_INTERNAL_RECORDER_SAMPLING_PERIOD;
        usSamplingPeriod *= CORRECTION_10_MICRO_SECONDS_TO_100_MICRO_SECONDS;

        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usSamplingPeriod, &errorInfo);

        // Number of preceding samples
        if (oResult)
        {
            usIndex = INDEX_INTERNAL_RECORDER_TRIGGER;
            usSubIndex = SUBINDEX_INTERNAL_RECORDER_PRECEDING_SAMPLES;
            oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, usPrecedingSamples, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetDataRecordingParameter(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD CORRECTION_100_MICRO_SECONDS_TO_10_MICRO_SECONDS(10);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usIndex;
    WORD usPrecedingSamples(0);
    WORD usSamplingPeriod(0);
    WORD usSubIndex;

    if (p_pCommand)
    {
        // Sampling period
        usIndex = INDEX_INTERNAL_RECORDER_CONFIGURATION;
        usSubIndex = SUBINDEX_INTERNAL_RECORDER_SAMPLING_PERIOD;
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usSamplingPeriod, &errorInfo);

        if(oResult)
        {
            usSamplingPeriod /= CORRECTION_100_MICRO_SECONDS_TO_10_MICRO_SECONDS;
        }

        // Preceding samples
        if (oResult)
        {
            usIndex = INDEX_INTERNAL_RECORDER_TRIGGER;
            usSubIndex = SUBINDEX_INTERNAL_RECORDER_PRECEDING_SAMPLES;
            oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &usPrecedingSamples, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usSamplingPeriod, sizeof(usSamplingPeriod));
        p_pCommand->SetReturnParameterData(1, &usPrecedingSamples, sizeof(usPrecedingSamples));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_EnableTrigger(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubTriggerType(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &ubTriggerType, sizeof(ubTriggerType));

        oResult = ConfigureTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubTriggerType, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_DisableAllTriggers(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD RESET_VALUE(0);
    const WORD RECORDER_TRIGGERED(0x0002);
    const WORD STOP_RECORDER(0x0000);
    const WORD START_RECORDER(0x0001);
    const WORD TRIGGER_OFF(0x0000);

    BOOL oResult(FALSE);
    BOOL oRecorderTriggered(FALSE);
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());

    if (p_pCommand)
    {
        oResult = SetRecorderTriggerVariableIndex(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RESET_VALUE, RESET_VALUE, &errorInfo);

        if (oResult)
        {
            oResult = SetRecorderTriggerMode(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RESET_VALUE, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetRecorderTriggerHighValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RESET_VALUE, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetRecorderTriggerLowValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RESET_VALUE, &errorInfo);
        }

        if (oResult)
        {
            oResult = SetRecorderTriggerMask(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RESET_VALUE, &errorInfo);
        }

        if (oResult)
        {
            if (CheckRecorderStatus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RECORDER_TRIGGERED, oRecorderTriggered, &errorInfo) && oRecorderTriggered)
            {
                if (SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, STOP_RECORDER, &errorInfo) && SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, START_RECORDER, &errorInfo))
                {
                    oResult = SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, TRIGGER_OFF, &errorInfo);
                }
            }
        }

        if (oResult && pRecorderChannelObjects)
        {
            pRecorderChannelObjects->ResetTrigger();
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ActivateChannel(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubChannelNumber;
    BYTE ubObjectSize;
    BYTE ubObjectSubIndex;
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    DWORD ulRecorderVariable(0);
    WORD usObjectIndex;

    if (p_pCommand && pRecorderChannelObjects)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &ubChannelNumber, sizeof(ubChannelNumber));
        p_pCommand->GetParameterData(1, &usObjectIndex, sizeof(usObjectIndex));
        p_pCommand->GetParameterData(2, &ubObjectSubIndex, sizeof(ubObjectSubIndex));
        p_pCommand->GetParameterData(3, &ubObjectSize, sizeof(ubObjectSize));

        ulRecorderVariable = usObjectIndex;
        ulRecorderVariable <<= 8;
        ulRecorderVariable += ubObjectSubIndex;
        ulRecorderVariable <<= 8;

        oResult = WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_VARIABLE_INDEX, ubChannelNumber, ulRecorderVariable, &errorInfo);


        if (oResult)
        {
            oResult = pRecorderChannelObjects->SetChannel(ubChannelNumber - 1, usObjectIndex, ubObjectSubIndex, ubObjectSize);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ActiveChannel(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubChannelNumber;
    BYTE ubObjectSize(0);
    BYTE ubObjectSubIndex;
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    DWORD ulRecorderVariable(0);
    WORD usObjectIndex;

    if (p_pCommand && pRecorderChannelObjects)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &ubChannelNumber, sizeof(ubChannelNumber));

        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_VARIABLE_INDEX, ubChannelNumber, &ulRecorderVariable, &errorInfo);

        if (oResult)
        {
            ulRecorderVariable >>= 8;
            ubObjectSubIndex = (BYTE)ulRecorderVariable;
            ulRecorderVariable >>= 8;
            usObjectIndex = (WORD)ulRecorderVariable;

            if (!pRecorderChannelObjects->IsEqual(ubChannelNumber - 1, usObjectIndex, ubObjectSubIndex))
            {
                oResult = pRecorderChannelObjects->SetChannel(ubChannelNumber - 1, usObjectIndex, ubObjectSubIndex, ubObjectSize);
            }
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usObjectIndex, sizeof(usObjectIndex));
        p_pCommand->SetReturnParameterData(1, &ubObjectSubIndex, sizeof(ubObjectSubIndex));
        p_pCommand->SetReturnParameterData(2, &ubObjectSize, sizeof(ubObjectSize));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_DeactivateAllChannels(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD RESET_VALUE(0);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    WORD usDefaultValue(0);

    if (p_pCommand && pRecorderChannelObjects)
    {
        oResult = TRUE;

        for (WORD usSubIndex(1); usSubIndex <= MAX_NB_OF_DATA_RECORDER_VARIABLES; usSubIndex++)
        {
            if (!WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_VARIABLE_INDEX, usSubIndex, RESET_VALUE, &errorInfo))
            {
                oResult = FALSE;
                break;
            }

            if (oResult)
            {
                oResult = pRecorderChannelObjects->SetChannel(usSubIndex - 1, usDefaultValue, (BYTE)usDefaultValue, (BYTE)usDefaultValue);
            }
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVariable(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    DWORD ulRecorderVariable(0);
    WORD usChannelIndex;
    WORD usObjectIndex;
    WORD usObjectSize(2);
    WORD usObjectSubIndex;

    if (p_pCommand && pRecorderChannelObjects)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usChannelIndex, sizeof(usChannelIndex));

        oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_VARIABLE_INDEX, usChannelIndex + 1, &ulRecorderVariable, &errorInfo);

        if (oResult)
        {
            ulRecorderVariable >>= 8;
            usObjectSubIndex = (BYTE)ulRecorderVariable;
            ulRecorderVariable >>= 8;
            usObjectIndex = (WORD)ulRecorderVariable;

            if (!pRecorderChannelObjects->IsEqual(usChannelIndex, usObjectIndex, (BYTE)usObjectSubIndex))
            {
                oResult = pRecorderChannelObjects->SetChannel(usChannelIndex, usObjectIndex, (BYTE)usObjectSubIndex, (BYTE)usObjectSize);
            }
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usObjectIndex, sizeof(usObjectIndex));
        p_pCommand->SetReturnParameterData(1, &usObjectSubIndex, sizeof(usObjectSubIndex));
        p_pCommand->SetReturnParameterData(2, &usObjectSize, sizeof(usObjectSize));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetMaxNbOfSamples(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usMaxNbOfSamples(0);

    if (p_pCommand)
    {
        oResult = CalculateMaxNbOfSamples(usMaxNbOfSamples, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usMaxNbOfSamples, sizeof(usMaxNbOfSamples));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetNbOfRecordedSamples(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usNbOfRecordedSamples(0);

    if (p_pCommand)
    {
        oResult = CalculateNbOfRecordedSamples(usNbOfRecordedSamples, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usNbOfRecordedSamples, sizeof(usNbOfRecordedSamples));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetNbOfPrecedingSamples(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usNbOfPrecedingSamples(0);

    if (p_pCommand)
    {
        p_pCommand->GetParameterData(0, &usNbOfPrecedingSamples, sizeof(usNbOfPrecedingSamples));

        oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_PRECEDING_SAMPLES, usNbOfPrecedingSamples, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetNbOfPrecedingSamples(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usNbOfPrecedingSamples(0);

    if (p_pCommand)
    {
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_PRECEDING_SAMPLES, &usNbOfPrecedingSamples, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usNbOfPrecedingSamples, sizeof(usNbOfPrecedingSamples));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetNbOfSamplingVariables(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    WORD usNbOfSamplingVariables(0);

    if (p_pCommand && pRecorderChannelObjects)
    {
        oResult = TRUE;
        usNbOfSamplingVariables = pRecorderChannelObjects->GetNbOfConfiguredChannels();

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usNbOfSamplingVariables, sizeof(usNbOfSamplingVariables));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetSamplingPeriod(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usSamplingPeriod(0);

    if (p_pCommand)
    {
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_CONFIGURATION, SUBINDEX_INTERNAL_RECORDER_SAMPLING_PERIOD, &usSamplingPeriod, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usSamplingPeriod, sizeof(usSamplingPeriod));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_StartRecorder(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD START_RECORDER(0x0001);
    const WORD STOP_RECORDER(0xFFFE);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Send 0->1 Bit change
        if (SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, STOP_RECORDER, &errorInfo))
        {
            oResult = SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, START_RECORDER, &errorInfo);
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_StopRecorder(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD STOP_RECORDER(0xFFFE);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, STOP_RECORDER, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ForceTrigger(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD FORCE_TRIGGER(0x0003);

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = SetRecorderControlword(p_pManager, p_hDCS_Handle, p_hTransactionHandle, FORCE_TRIGGER, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_IsRecorderRunning(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD RECORDER_RUNNING(0x0001);

    BOOL oResult(FALSE);
    BOOL oRecorderRunning(FALSE);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = CheckRecorderStatus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RECORDER_RUNNING, oRecorderRunning, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oRecorderRunning, sizeof(oRecorderRunning));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_IsRecorderTriggered(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD RECORDER_TRIGGERED(0x0002);

    BOOL oResult(FALSE);
    BOOL oRecorderTriggered(FALSE);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        oResult = CheckRecorderStatus(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RECORDER_TRIGGERED, oRecorderTriggered, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &oRecorderTriggered, sizeof(oRecorderTriggered));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetRecorderStatus(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usRecorderStatus(0);

    if (p_pCommand)
    {
        oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_CONTROL, SUBINDEX_INTERNAL_RECORDER_STATUSWORD, &usRecorderStatus, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usRecorderStatus, sizeof(usRecorderStatus));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetRecorderConfiguration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubTriggerType(0);
    CErrorInfo errorInfo;
    WORD usConfiguration(0);

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &usConfiguration, sizeof(usConfiguration));

        ubTriggerType = (BYTE)usConfiguration;
        oResult = ConfigureTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubTriggerType, &errorInfo);

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetRecorderConfiguration(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    WORD usConfiguration(0);

    if (p_pCommand)
    {
        if (pRecorderChannelObjects)
        {
            usConfiguration = pRecorderChannelObjects->GetTrigger();

            oResult = TRUE;
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usConfiguration, sizeof(usConfiguration));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadChannelVectorSize(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulVectorSize(0);
    WORD usNumberOfRecordedSamples(0);

    if (p_pCommand)
    {
        if (CalculateNbOfRecordedSamples(usNumberOfRecordedSamples, &errorInfo))
        {
            ulVectorSize = usNumberOfRecordedSamples;

            oResult = TRUE;
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulVectorSize, sizeof(ulVectorSize));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVectorStartOffset(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usVectorStartOffset(0);

    if (p_pCommand)
    {
        oResult = TRUE;

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usVectorStartOffset, sizeof(usVectorStartOffset));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadDataBuffer(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE* pBuffer(0);
    CErrorInfo errorInfo;
    DWORD ulBufferSizeRead(0);
    DWORD ulBufferSizeToRead;
    WORD usMaxNbOfSamples(0);
    WORD usNbOfRecordedSamples(0);
    WORD usVectorStartOffset(0);

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(0, &ulBufferSizeToRead, sizeof(ulBufferSizeToRead));

        // Read data buffer
        pBuffer = (BYTE*)malloc(ulBufferSizeToRead);

        oResult = ReadDataBuffer(p_pManager, p_hDCS_Handle, p_hTransactionHandle, pBuffer, ulBufferSizeToRead, &ulBufferSizeRead, usMaxNbOfSamples, usNbOfRecordedSamples, &errorInfo);

        // Vector start offset
        if (oResult)
        {
            usVectorStartOffset = 0;
        }

        // Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, pBuffer, ulBufferSizeToRead);
        p_pCommand->SetReturnParameterData(1, &ulBufferSizeRead, sizeof(ulBufferSizeRead));
        p_pCommand->SetReturnParameterData(2, &usVectorStartOffset, sizeof(usVectorStartOffset));
        p_pCommand->SetReturnParameterData(3, &usMaxNbOfSamples, sizeof(usMaxNbOfSamples));
        p_pCommand->SetReturnParameterData(4, &usNbOfRecordedSamples, sizeof(usNbOfRecordedSamples));

        if (pBuffer)
        {
            free(pBuffer);
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetVersion(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulValue = 0;
    WORD usIndex;
    WORD usSubIndex;
    WORD usSoftwareVersion = 0;
    WORD usHardwareVersion = 0;
    WORD usApplicationNumber = 0;
    WORD usApplicationVersion = 0;

    if(p_pCommand)
    {
        //Init
        oResult = TRUE;

        //Read product code
        usIndex = INDEX_DEVICE_IDENTIFICATION;
        usSubIndex = SUBINDEX_PRODUCT_CODE;
        if(oResult && !ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulValue, &errorInfo))
        {
            oResult = FALSE;
        }
        else
        {
            usHardwareVersion = HIWORD(ulValue);
            usApplicationNumber = LOWORD(ulValue);
        }

        //Read revision number
        usIndex = INDEX_DEVICE_IDENTIFICATION;
        usSubIndex = SUBINDEX_REVISION_NO;
        if(oResult && !ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulValue, &errorInfo))
        {
            oResult = FALSE;
        }
        else
        {
            usSoftwareVersion = HIWORD(ulValue);
            usApplicationVersion = LOWORD(ulValue);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usHardwareVersion, sizeof(usHardwareVersion));
        p_pCommand->SetReturnParameterData(1, &usSoftwareVersion, sizeof(usSoftwareVersion));
        p_pCommand->SetReturnParameterData(2, &usApplicationNumber, sizeof(usApplicationNumber));
        p_pCommand->SetReturnParameterData(3, &usApplicationVersion, sizeof(usApplicationVersion));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetDeviceType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulDeviceType = 0;
    WORD usIndex;
    WORD usSubIndex;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Device type
        if(oResult)
        {
            usIndex = INDEX_DEVICE_TYPE;
            usSubIndex = SUBINDEX_DEVICE_TYPE;
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ulDeviceType, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulDeviceType, sizeof(ulDeviceType));

    }
    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetObject(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usObjectIndex;
    BYTE ubObjectSubIndex;
    DWORD ulNbOfBytesToRead;
    void* pData = 0;
    DWORD ulNbOfBytesRead = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        oResult = TRUE;

        //Get parameter data
        p_pCommand->GetParameterData(0, &usObjectIndex, sizeof(usObjectIndex));
        p_pCommand->GetParameterData(1, &ubObjectSubIndex, sizeof(ubObjectSubIndex));
        p_pCommand->GetParameterData(2, &ulNbOfBytesToRead, sizeof(ulNbOfBytesToRead));

        //Get object
        if(oResult)
        {
            pData = malloc(ulNbOfBytesToRead);
            memset(pData, 0, ulNbOfBytesToRead);

            oResult = ReadDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usObjectIndex, ubObjectSubIndex, pData, ulNbOfBytesToRead, &ulNbOfBytesRead, &errorInfo);
        }

        //Set return parameter
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, pData, ulNbOfBytesToRead);
        p_pCommand->SetReturnParameterData(1, &ulNbOfBytesRead, sizeof(ulNbOfBytesRead));
        if(pData) free(pData);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_Restore(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand && p_pManager)
    {
        oResult = RestoreParameter(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_RestoreFactory(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;

    if(p_pCommand && p_pManager)
    {
        oResult = RestoreFactoryParameter(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetObject(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usObjectIndex;
    BYTE ubObjectSubIndex;
    void* pData = 0;
    DWORD ulNbOfBytesToWrite;
    DWORD ulNbOfBytesWritten = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usObjectIndex, sizeof(usObjectIndex));
        p_pCommand->GetParameterData(1, &ubObjectSubIndex, sizeof(ubObjectSubIndex));
        p_pCommand->GetParameterData(3, &ulNbOfBytesToWrite, sizeof(ulNbOfBytesToWrite));
        pData = malloc(ulNbOfBytesToWrite);
        p_pCommand->GetParameterData(2, pData, ulNbOfBytesToWrite);

        //Set object
        oResult = WriteDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usObjectIndex, ubObjectSubIndex, pData, ulNbOfBytesToWrite, &ulNbOfBytesWritten, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulNbOfBytesWritten, sizeof(ulNbOfBytesWritten));
        if(pData) free(pData);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_Store(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand && p_pManager)
    {
        //Store
        oResult = StoreParameter(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_RequestCANFrame(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usCobID;
    WORD usLength;
    void* pData = NULL;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usCobID, sizeof(usCobID));
        p_pCommand->GetParameterData(1, &usLength, sizeof(usLength));
        pData = malloc(usLength);

        oResult = DCS_RequestCANFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usCobID, usLength, pData, &errorInfo);

        //Set ReturnParameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, pData, usLength);
        if (pData) free(pData);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SendCANFrame(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usCobID;
    WORD usLength;
    void* pData = NULL;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usCobID, sizeof(usCobID));
        p_pCommand->GetParameterData(1, &usLength, sizeof(usLength));
        pData = malloc(usLength);
        memset(pData, 0, usLength);
        p_pCommand->GetParameterData(2, pData, usLength);

        oResult = DCS_SendCANFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usCobID, usLength, pData, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        if (pData) free(pData);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadCANFrame(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usCobID;
    WORD usLength;
    DWORD ulTimeout;
    void* pData = NULL;
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usCobID, sizeof(usCobID));
        p_pCommand->GetParameterData(1, &usLength, sizeof(usLength));
        p_pCommand->GetParameterData(2, &ulTimeout, sizeof(ulTimeout));

        pData = malloc(usLength);
        oResult = DCS_ReadCANFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usCobID, usLength, pData, ulTimeout, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, pData, usLength);
        if (pData) free(pData);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SendNMTService(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usCommandSpecifier;
    WORD usNodeIdentifer;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usNodeIdentifer, sizeof(usNodeIdentifer));
        p_pCommand->GetParameterData(1, &usCommandSpecifier, sizeof(usCommandSpecifier));

        oResult = DCS_SendNMTService(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usNodeIdentifer, usCommandSpecifier, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadVendorId(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    DWORD ulVendorId(0);

    if(p_pCommand && p_pManager)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            //Read from EtherCAT Slave Information Interface (SII)
            oResult = DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_VENDOR_ID, &ulVendorId, sizeof(ulVendorId), &errorInfo);
        }
        else
        {
            //Read from object dictionary
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_IDENTIFICATION, SUBINDEX_VENDOR_ID, &ulVendorId, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulVendorId, sizeof(ulVendorId));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadDeviceType(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    DWORD ulDeviceType(0);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            //Not supported in ESC registers without CoE
            oResult = FALSE;
        }
        else
        {
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_TYPE, SUBINDEX_DEVICE_TYPE, &ulDeviceType, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulDeviceType, sizeof(ulDeviceType));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadDeviceName(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CStdString deviceName(_T(""));
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    SSiiCatagoryGeneral categoryGeneral;

    if(p_pCommand)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            memset(&categoryGeneral, 0, sizeof(categoryGeneral));

            oResult = ReadSiiCategoryGeneral(p_pManager, p_hDCS_Handle, p_hTransactionHandle, categoryGeneral, &errorInfo);

            if(oResult)
            {
                //Read SII string
                oResult = ReadSiiCategoryString(p_pManager, p_hDCS_Handle, p_hTransactionHandle, categoryGeneral.uOrderIdx, deviceName, &errorInfo);
            }
        }
        else
        {
            oResult = ReadStringObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_NAME, SUBINDEX_DEVICE_NAME, &deviceName, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, deviceName.GetBuffer(), deviceName.GetLength());
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_IsIdentifySupported(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BYTE IDENTIFY_NOT_SUPPORTED = 0;
    const BYTE IDENTIFY_SUPPORTED = 1;
    const DWORD FIRST_FIRMWARE_SOFTWARE_VERSION_SUPPORTING_LSS = 0x01600000;
    const DWORD FIRMWARE_SOFTWARE_VERSION_MASK = 0xFFFF0000;

    CErrorInfo errorInfo;
    BOOL oResult(TRUE);
    BYTE ubIdentifySupported(IDENTIFY_NOT_SUPPORTED);
    DWORD ulRevisionNumber(0);

    errorInfo.Init(0);

    if (p_pCommand)
    {
        p_pCommand->GetParameterData(0, &ulRevisionNumber, sizeof(ulRevisionNumber));
    }

    ulRevisionNumber &= FIRMWARE_SOFTWARE_VERSION_MASK;

    if (ulRevisionNumber >= FIRST_FIRMWARE_SOFTWARE_VERSION_SUPPORTING_LSS)
    {
        ubIdentifySupported = IDENTIFY_SUPPORTED;
    }

    p_pCommand->SetStatus(oResult, &errorInfo);
    p_pCommand->SetReturnParameterData(0, &ubIdentifySupported, sizeof(ubIdentifySupported));

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_StartIdentify(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const WORD CLEAR_TIMEOUT = 10;
    const int DATA_SIZE = 8;
    const BYTE ALL_NODES = 0;

    BOOL oResult(FALSE);
    DWORD ulVendorId(0);
    DWORD ulProductCode(0);
    DWORD ulRevisionNumber(0);
    DWORD ulSerialNumber(0);
    WORD usTimeout(0);
    BYTE ubNodeId(0);
    BYTE data[DATA_SIZE];

    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ulVendorId, sizeof(ulVendorId));
        p_pCommand->GetParameterData(1, &ulProductCode, sizeof(ulProductCode));
        p_pCommand->GetParameterData(2, &ulRevisionNumber, sizeof(ulRevisionNumber));
        p_pCommand->GetParameterData(3, &ulSerialNumber, sizeof(ulSerialNumber));
        p_pCommand->GetParameterData(4, &usTimeout, sizeof(usTimeout));
        p_pCommand->GetParameterData(5, &ubNodeId, sizeof(ubNodeId));

        //Stop remote node
        DCS_SendNMTService(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ALL_NODES, NSI_STOP_REMOTE_NODE, NULL);

        //Clear LSS Queue
        oResult = ClearLSSFrameQueue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, CLEAR_TIMEOUT, &errorInfo);

        //Send LSS Frame VendorID
        if (oResult)
        {
            memset(data, 0, DATA_SIZE);
            data[0] = CIA305_SWITCH_STATE_SELECTIVE_PROTOCOL_COMMAND_SPECIFIER_VENDOR_ID;
            *((DWORD*)&data[1]) = ulVendorId;
            oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, data, DATA_SIZE, &errorInfo);
        }

        //Send LSS Frame ProductCode
        if (oResult)
        {
            memset(data, 0, DATA_SIZE);
            data[0] = CIA305_SWITCH_STATE_SELECTIVE_PROTOCOL_COMMAND_SPECIFIER_PRODUCT_CODE;
            *((DWORD*)&data[1]) = ulProductCode;
            oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, data, DATA_SIZE, &errorInfo);
        }

        //Send LSS Frame RevisionNumber
        if (oResult)
        {
            memset(data, 0, DATA_SIZE);
            data[0] = CIA305_SWITCH_STATE_SELECTIVE_PROTOCOL_COMMAND_SPECIFIER_REVISION_NUMBER;
            *((DWORD*)&data[1]) = ulRevisionNumber;
            oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, data, DATA_SIZE, &errorInfo);
        }

        //Send LSS Frame SerialNumber
        if (oResult)
        {
            memset(data, 0, DATA_SIZE);
            data[0] = CIA305_SWITCH_STATE_SELECTIVE_PROTOCOL_COMMAND_SPECIFIER_SERIAL_NUMBER;
            *((DWORD*)&data[1]) = ulSerialNumber;
            oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, data, DATA_SIZE, &errorInfo);
        }

        if (oResult)
        {
            MmcSleep(100);

            //Read LSS Frame
            oResult = DCS_ReadLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usTimeout, data, DATA_SIZE, &errorInfo);
        }

        if (oResult)
        {
            oResult = (data[0] == CIA305_SWITCH_STATE_SELECTIVE_PROTOCOL_COMMAND_SPECIFIER_RESPONSE);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_StopIdentify(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const int DATA_SIZE = 8;
    const BYTE ALL_NODES = 0;

    BOOL oResult(FALSE);
    BYTE ubMode(0);
    BYTE data[DATA_SIZE];
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ubMode, sizeof(ubMode));

        memset(data, 0, DATA_SIZE);
        data[0] = CIA305_SWITCH_STATE_GLOBAL_PROTOCOL_COMMAND_SPECIFIER_MODE;
        data[1] = ubMode;

        oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, data, DATA_SIZE, &errorInfo);

        //Enter NMT State pre-operational
        if (oResult)
        {
            oResult = DCS_SendNMTService(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ALL_NODES, NSI_ENTER_PRE_OPERATIONAL, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadProductCode(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    DWORD ulProductCode(0);

    if(p_pCommand)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            //Read from EtherCAT Slave Information Interface (SII)
            oResult = DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_PRODUCT_CODE, &ulProductCode, sizeof(ulProductCode), &errorInfo);
        }
        else
        {
            //Read from object dictionary
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_IDENTIFICATION, SUBINDEX_PRODUCT_CODE, &ulProductCode, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulProductCode, sizeof(ulProductCode));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadRevisionNumber(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    DWORD ulRevisionNumber(0);

    if(p_pCommand)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            //Read from EtherCAT Slave Information Interface (SII)
            oResult = DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_REVISION_NUMBER, &ulRevisionNumber, sizeof(ulRevisionNumber), &errorInfo);
        }
        else
        {
            //Read from object dictionary
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_IDENTIFICATION, SUBINDEX_REVISION_NO, &ulRevisionNumber, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulRevisionNumber, sizeof(ulRevisionNumber));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadSerialNumber(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));
    DWORD ulSerialNumber(0);

    if(p_pCommand)
    {
        if(p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName) && protocolStackName == PROTOCOL_ETHERCAT_COE)
        {
            //Read from EtherCAT Slave Information Interface (SII)
            oResult = DCS_ReadSiiParameter(p_pManager, p_hDCS_Handle, p_hTransactionHandle, SII_SERIAL_NUMBER, &ulSerialNumber, sizeof(ulSerialNumber), &errorInfo);
        }
        else
        {
            //Read from object dictionary
            oResult = ReadUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DEVICE_IDENTIFICATION, SUBINDEX_SERIAL_NO, &ulSerialNumber, &errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulSerialNumber, sizeof(ulSerialNumber));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadNodeId(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubNodeIdValue(0);
    CErrorInfo errorInfo;
    DWORD ulNodeIdValue(0);

    if(p_pCommand)
    {
        oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_NODE_ID, SUBINDEX_NODE_ID, &ubNodeIdValue, &errorInfo);

        //Convert type
        ulNodeIdValue = (DWORD)ubNodeIdValue;

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulNodeIdValue, sizeof(ulNodeIdValue));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_WriteNodeId(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubNodeIdValue(0);
    CErrorInfo errorInfo;
    DWORD ulNodeIdValue(0);

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ulNodeIdValue, sizeof(ulNodeIdValue));

        //Convert type
        ubNodeIdValue = (BYTE)ulNodeIdValue;

        //Write object
        oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_NODE_ID, SUBINDEX_NODE_ID, ubNodeIdValue, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadStationId(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usStationIdValue(0);
    WORD usLengthRead(0);

    if(p_pCommand)
    {
        //Read station id
        oResult = DCS_ReadEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRESS_CONFIGURED_STATION_ADDRESS, sizeof(usStationIdValue), (char*)&usStationIdValue, &usLengthRead, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &usStationIdValue, sizeof(usStationIdValue));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_WriteStationId(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    WORD usStationIdValue(0);
    WORD usLengthWritten(0);

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &usStationIdValue, sizeof(usStationIdValue));

        //Write station id
        oResult = DCS_WriteEscRegister(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ADDRESS_CONFIGURED_STATION_ADDRESS, sizeof(usStationIdValue), (char*)&usStationIdValue, &usLengthWritten, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadTransferRate(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    DWORD ulTransferRate(0);
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        oResult = GetDeviceTransferRate(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulTransferRate, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulTransferRate, sizeof(ulTransferRate));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_WriteTransferRate(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    DWORD ulTransferRate(0);
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &ulTransferRate, sizeof(ulTransferRate));

        oResult = SetDeviceTransferRate(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulTransferRate, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ScanStore(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand && p_pManager)
    {
        oResult = StoreParameter(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ScanResetDevice(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oBroadcast(FALSE);
    BOOL oWaitForReboot(FALSE);
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        //Get parameter data
        p_pCommand->GetParameterData(0, &oBroadcast, sizeof(oBroadcast));
        p_pCommand->GetParameterData(1, &oWaitForReboot, sizeof(oWaitForReboot));

        //Reset
        oResult = ResetDevice(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, oBroadcast, oWaitForReboot, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ScanWaitForReboot(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    CErrorInfo errorInfo;
    BOOL oResult(FALSE);

    if(p_pCommand)
    {
        oResult = WaitForReboot(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetSerialNumber(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    __uint64 llSerialNumber = 0;
    CErrorInfo errorInfo;

    if(p_pCommand)
    {
        //Read serial number
        oResult = ReadUInt64Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_SERIAL_NUMBER, SUBINDEX_SERIAL_NUMBER, &llSerialNumber, &errorInfo);

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &llSerialNumber, sizeof(llSerialNumber));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::UpdateNodeId(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE p_ubNodeId, BYTE& p_rubNodeIdBackup)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        if (p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_rubNodeIdBackup, sizeof(p_rubNodeIdBackup)))
        {
            oResult = p_pManager->SetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_ubNodeId, sizeof(p_ubNodeId));
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ResetNodeId(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE p_ubNodeId)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        oResult = p_pManager->SetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_ubNodeId, sizeof(p_ubNodeId));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_CleanupCANopenCommunication(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const int PARAMETER_LAST_RESULT(0);

    BOOL oResult(FALSE);
    BOOL oCleanupCanPassive(FALSE);
    BYTE ubGatewayNodeId(0);
    BYTE ubNodeIdBackup(0);
    DWORD ulErrorCode(0);

    if (p_pCommand)
    {
        // Get parameter data
        p_pCommand->GetParameterData(PARAMETER_LAST_RESULT, &oCleanupCanPassive, sizeof(oCleanupCanPassive));

        if(IsGatewayTopology(p_pManager, p_hDCS_Handle))
        {
            GetGatewayNodeId(p_pManager, p_hDCS_Handle, ubGatewayNodeId);

            if (UpdateNodeId(p_pManager, p_hDCS_Handle, ubGatewayNodeId, ubNodeIdBackup))
            {
                if (oCleanupCanPassive)
                {
                    ResetGatewayCommunication(p_pManager, p_hDCS_Handle);

                    if (IsGatewayInCanPassiveState(p_pManager, p_hDCS_Handle))
                    {
                        ClearGatewayErrors(p_pManager, p_hDCS_Handle);
                    }
                }

                ResetNodeId(p_pManager, p_hDCS_Handle, ubNodeIdBackup);

                oResult = TRUE;
            }
        }
        else
        {
            oResult = CleanupCanOpenCommunication(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oCleanupCanPassive, &ulErrorCode);
        }

        // Get return parameter data
        p_pCommand->GetErrorCode(&ulErrorCode);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_GetGatewaySettings(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    BYTE usCANBitRateIndex(0);
    DWORD ulCANBitRate(0);

    if(p_pCommand)
    {
        oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, &usCANBitRateIndex, &errorInfo);
        ulCANBitRate = ConvertBitrateIndexToValue_CiaStandard(usCANBitRateIndex);

        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulCANBitRate, sizeof(ulCANBitRate));
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetGatewaySettings(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const BOOL BROADCAST(FALSE);
    const BOOL WAIT_FOR_REBOOT(TRUE);

    BOOL oResult(FALSE);
    BOOL oUpdateCanBitrate(FALSE);
    BYTE ubCANBitRateIndex(0);
    CErrorInfo errorInfo;
    DWORD ulCANBitRate(0);
    BOOL oUpdateTimeout(FALSE);

    if(p_pManager)
    {
        oResult = IsGatewaySupported(p_pCommand, p_pManager, p_hDCS_Handle, &errorInfo);

        if(oResult && p_pCommand)
        {
            p_pCommand->GetParameterData(0, &ulCANBitRate, sizeof(ulCANBitRate));
            ubCANBitRateIndex = (BYTE)ConvertBitrateValueToIndex_CiaStandard(ulCANBitRate);

            oResult = CheckUpdateCanBitrateRequired(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubCANBitRateIndex, oUpdateCanBitrate, &errorInfo);

            if (oResult && oUpdateCanBitrate)
            {
                oResult = WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, ubCANBitRateIndex, &errorInfo);
            }

            oResult = CheckUpdateTimeoutRequired(p_pManager, p_hDCS_Handle, p_hTransactionHandle, oUpdateTimeout, &errorInfo);

            if(oResult && oUpdateTimeout)
            {
                oResult = InitGatewayTimeout(p_pManager,p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
            }

            if(oResult && (oUpdateTimeout || oUpdateCanBitrate))
            {
                oResult = StoreParameter(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, &errorInfo);
            }

            if(oResult && (oUpdateCanBitrate || oUpdateTimeout))
            {
                oResult = ResetDevice(p_pCommand, p_pManager, p_hDCS_Handle, p_hTransactionHandle, BROADCAST, WAIT_FOR_REBOOT, &errorInfo);
            }

            p_pCommand->SetStatus(oResult, &errorInfo);
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetGatewayDeviceTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulGatewayTimeout)
{
    BOOL oResult(FALSE);
    BYTE ubGatewayNodeId(0);
    BYTE ubNodeIdBackup(0);

    if (IsGatewayTopology(p_pManager, p_hDCS_Handle))
    {
        GetGatewayNodeId(p_pManager, p_hDCS_Handle, ubGatewayNodeId);

        if (UpdateNodeId(p_pManager, p_hDCS_Handle, ubGatewayNodeId, ubNodeIdBackup))
        {
            // Execute
            oResult = SetGatewayTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_ulGatewayTimeout);

            ResetNodeId(p_pManager, p_hDCS_Handle, ubNodeIdBackup);
        }
    }
    else
    {
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsGatewaySupported(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    CStdString protocolStackName(_T(""));

    if(p_pManager)
    {
        oResult = p_pManager->DCS_GetProtocolStackName(p_hDCS_Handle, &protocolStackName, p_pErrorInfo);

        if(oResult)
        {
            if((PROTOCOL_MAXON_SERIAL_V1 == protocolStackName) || (PROTOCOL_MAXON_SERIAL_V2 == protocolStackName))
            {
                oResult = TRUE;
            }
            else
            {
                oResult = FALSE;
                errorInfo.Init(ERROR_FUNCTION_NOT_SUPPORTED);

                if(p_pCommand)
                {
                    p_pCommand->SetStatus(oResult, &errorInfo);
                }
            }
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsCorrectFieldbus(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubFielbusMode, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    BYTE ubFielbusMode(0);
    WORD usIndex(INDEX_ACTIVE_FIELD_BUS);
    WORD usSubIndex(SUBINDEX_ACTIVE_FIELD_BUS);

    if(p_pManager)
    {
        if(ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &ubFielbusMode, &errorInfo) && p_ubFielbusMode == ubFielbusMode)
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CheckUpdateCanBitrateRequired(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubCANBitRateIndex, BOOL& p_roUpdateCanBitrate, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE ubCANBitRateIndex(0);

    if (ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_CAN_BITRATE, SUBINDEX_CAN_BITRATE, &ubCANBitRateIndex, p_pErrorInfo))
    {
        if (ubCANBitRateIndex != p_ubCANBitRateIndex)
        {
            p_roUpdateCanBitrate = TRUE;
        }

        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CheckUpdateTimeoutRequired(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL& p_roRequired, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    DWORD ulBaudrate(0);
    DWORD ulRequestedTimeout(0);
    DWORD ulCurrentTimeout(0);
    WORD usRequestedTimeout(0);

    p_roRequired = FALSE;

    if(p_pManager)
    {
        oResult = p_pManager->DCS_GetProtocolStackSettings(p_hDCS_Handle, &ulBaudrate, &ulRequestedTimeout, p_pErrorInfo);

        if(oResult)
        {
            usRequestedTimeout = CalculateDeviceTimeout(ulRequestedTimeout);
        }

        oResult = GetDeviceTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulCurrentTimeout);

        if(oResult)
        {
            if((DWORD)usRequestedTimeout != ulCurrentTimeout)
            {
                p_roRequired = TRUE;
            }
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CreateDiagnosisHistory()
{
    BOOL oResult(FALSE);

    if(m_pDiagnosisHistory == 0)
    {
        m_pDiagnosisHistory = new CDiagnosisHistory();
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::DeleteDiagnosisHistory()
{
    BOOL oResult(FALSE);

    if(m_pDiagnosisHistory != 0)
    {
        delete m_pDiagnosisHistory;
        m_pDiagnosisHistory = 0;
        oResult = TRUE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ReadAnalogOutputConfigurationsCount(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE& p_ubAnalogOutputConfigurationCount, CErrorInfo& p_rErrorInfo)
{
    BOOL oResult = FALSE;
    WORD usIndex = INDEX_ANALOG_OUTPUT_CONFIGURATION;
    WORD usSubIndex = SUBINDEX_ANALOG_OUTPUT_CONFIGURATION;

    p_ubAnalogOutputConfigurationCount = 0;

    oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &p_ubAnalogOutputConfigurationCount, &p_rErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::ReadAnalogOutputConfiguration(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usAnalogOutputNumber, BYTE& p_ubAnalogOutputConfiguration, CErrorInfo& p_rErrorInfo)
{
    const BYTE FUNCTIONALITY_NONE = 255;

    BOOL oResult = FALSE;
    WORD usIndex = 0;
    WORD usSubIndex = 0;
    BYTE ubAnalogOutputConfigurationCount = 0;

    p_ubAnalogOutputConfiguration = FUNCTIONALITY_NONE;

    if (ReadAnalogOutputConfigurationsCount(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAnalogOutputConfigurationCount, p_rErrorInfo) && ubAnalogOutputConfigurationCount > 0)
    {
        if (p_usAnalogOutputNumber > 0 && p_usAnalogOutputNumber <= ubAnalogOutputConfigurationCount)
        {
            usIndex = INDEX_ANALOG_OUTPUT_CONFIGURATION;
            usSubIndex = p_usAnalogOutputNumber;

            oResult = ReadUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, &p_ubAnalogOutputConfiguration, &p_rErrorInfo);
        }
        else
        {
            p_rErrorInfo.Init(ERROR_DEVICE_CAN_SUB_OBJECT);
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::WriteAnalogOutputGeneralPurposeValue(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubAnalogOutputConfiguration, long p_lValue, CErrorInfo& p_rErrorInfo)
{
    const BYTE GENERAL_PURPOSE_A = 0;
    const BYTE GENERAL_PURPOSE_B = 1;

    BOOL oResult = FALSE;
    WORD usSubIndex = 0;

    if (p_ubAnalogOutputConfiguration >= GENERAL_PURPOSE_A && p_ubAnalogOutputConfiguration <= GENERAL_PURPOSE_B)
    {
        switch (p_ubAnalogOutputConfiguration)
        {
        case GENERAL_PURPOSE_A:
            usSubIndex = SUBINDEX_ANALOG_OUTPUT_GENERAL_PURPOSE_A;
            break;
        case GENERAL_PURPOSE_B:
            usSubIndex = SUBINDEX_ANALOG_OUTPUT_GENERAL_PURPOSE_B;
            break;
        }

        oResult = WriteInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_ANALOG_OUTPUT_GENERAL_PURPOSE, usSubIndex, p_lValue, &p_rErrorInfo);
    }
    else
    {
        p_rErrorInfo.Init(ERROR_DEVICE_CAN_SUB_OBJECT);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsAnalogOutputGeneralPurposeConfiguration(BYTE p_ubAnalogOutputConfiguration, CErrorInfo& p_rErrorInfo)
{
    const BYTE GENERAL_PURPOSE_A = 0;
    const BYTE GENERAL_PURPOSE_B = 1;

    BOOL oResult = FALSE;

    if (p_ubAnalogOutputConfiguration >= GENERAL_PURPOSE_A && p_ubAnalogOutputConfiguration <= GENERAL_PURPOSE_B)
    {
        oResult = TRUE;
    }
    else
    {
        p_rErrorInfo.Init(ERROR_VIRTUAL_DEVICE_CONFIG_NOT_SUPPORTED);
        oResult = FALSE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetAnalogOutput(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    BYTE ubAnalogOutputConfiguration(255);
    WORD usAnalogOutputNumber(0);
    WORD usOutputValue(0);
    long lOutputValue(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        p_pCommand->GetParameterData(0, &usAnalogOutputNumber, sizeof(usAnalogOutputNumber));
        p_pCommand->GetParameterData(1, &usOutputValue, sizeof(usOutputValue));

        //Convert
        lOutputValue = (long)usOutputValue;

        //Write
        if (ReadAnalogOutputConfiguration(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usAnalogOutputNumber, ubAnalogOutputConfiguration, errorInfo))
        {
            if (IsAnalogOutputGeneralPurposeConfiguration(ubAnalogOutputConfiguration, errorInfo))
            {
                oResult = WriteAnalogOutputGeneralPurposeValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAnalogOutputConfiguration, lOutputValue, errorInfo);
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetAnalogOutputVoltage(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usAnalogOutputNumber(0);
    BYTE ubAnalogOutputConfiguration(255);
    long lVoltageValue(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        p_pCommand->GetParameterData(0, &usAnalogOutputNumber, sizeof(usAnalogOutputNumber));
        p_pCommand->GetParameterData(1, &lVoltageValue, sizeof(lVoltageValue));

        if (ReadAnalogOutputConfiguration(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usAnalogOutputNumber, ubAnalogOutputConfiguration, errorInfo))
        {
            if (IsAnalogOutputGeneralPurposeConfiguration(ubAnalogOutputConfiguration, errorInfo))
            {
                oResult = WriteAnalogOutputGeneralPurposeValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAnalogOutputConfiguration, lVoltageValue, errorInfo);
            }
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SetAnalogOutputState(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);
    WORD usConfiguration(0);
    BYTE ubAnalogOutputConfiguration(0);
    long lVoltageValue(0);
    CErrorInfo errorInfo;

    if (p_pCommand)
    {
        p_pCommand->GetParameterData(0, &usConfiguration, sizeof(usConfiguration));
        p_pCommand->GetParameterData(1, &lVoltageValue, sizeof(lVoltageValue));

        //Convert
        ubAnalogOutputConfiguration = (BYTE)usConfiguration;

        //Write
        if (IsAnalogOutputGeneralPurposeConfiguration((BYTE)usConfiguration, errorInfo))
        {
            oResult = WriteAnalogOutputGeneralPurposeValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ubAnalogOutputConfiguration, lVoltageValue, errorInfo);
        }

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetGatewayTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulTimeout)
{
    BOOL oResult(FALSE);
    DWORD ulDeviceTimeout(0);

    if (GetDeviceTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ulDeviceTimeout))
    {
        m_ulDeviceTimeoutBackup = ulDeviceTimeout;
        if (ulDeviceTimeout != p_ulTimeout)
        {
            oResult = SetDeviceTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_ulTimeout);
        }
        else
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::RestoreGatewayTimeout(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    BOOL oResult(FALSE);

    if (m_ulDeviceTimeoutBackup != 0)
    {
        if (SetDeviceTimeout(p_pManager, p_hDCS_Handle, p_hTransactionHandle, m_ulDeviceTimeoutBackup))
        {
            m_ulDeviceTimeoutBackup = 0;
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::GetNodeId(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE& p_rubNodeId, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        if (p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, NODE_ID, (BYTE*)&p_rubNodeId, sizeof(p_rubNodeId), p_pErrorInfo))
        {
            oResult = TRUE;
        }
        else
        {
            p_rubNodeId = 0;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::GetGatewayNodeId(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, BYTE& p_rubGatewayNodeId, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    if (p_pManager)
    {
        if (p_pManager->GetParameter(PT_DEVICE, p_hDCS_Handle, GATEWAY_NODE_ID, (BYTE*)&p_rubGatewayNodeId, sizeof(p_rubGatewayNodeId), p_pErrorInfo))
        {
            oResult = TRUE;
        }
        else
        {
            p_rubGatewayNodeId = 0;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CleanupCanOpenCommunication(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BOOL p_oCleanupCanPassive, DWORD* p_pulErrorCode)
{
    BOOL oResult(FALSE);
    CCommand_PS_CANopen command(CANOPEN_CLEANUP_COMMUNICATION);

    if (p_pManager)
    {
        // Set parameter
        command.ResetStatus();
        command.SetParameterData(0, &p_oCleanupCanPassive, sizeof(p_oCleanupCanPassive));

        // Execute command
        oResult = p_pManager->ExecuteCommand(&command, p_hDCS_Handle, p_hTransactionHandle);

        // Get return parameter data
        command.GetErrorCode(p_pulErrorCode);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::IsGatewayTopology(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle)
{
    BOOL oResult(FALSE);
    BYTE ubGatewayNodeId(0);
    BYTE ubNodeId(0);

    if (GetGatewayNodeId(p_pManager, p_hDCS_Handle, ubGatewayNodeId) && GetNodeId(p_pManager, p_hDCS_Handle, ubNodeId))
    {
        if (ubGatewayNodeId && ubNodeId && (ubGatewayNodeId != ubNodeId))
        {
            oResult = TRUE;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ConfigureTrigger(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubTriggerType, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());

    switch (p_ubTriggerType)
    {
        case RTT_MOVEMENT_START_TRIGGER:
        {
            oResult = ConfigureMovementStartTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
            break;
        }
        case RTT_ERROR_TRIGGER:
        {
            oResult = ConfigureSetErrorTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
            break;
        }
        case RTT_DIGITAL_INPUT_TRIGGER:
        {
            oResult = ConfigureDigitalInputTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
            break;
        }
        case RTT_MOVEMENT_END_TRIGGER:
        {
            oResult = ConfigureMovementEndTrigger(p_pManager, p_hDCS_Handle, p_hTransactionHandle, p_pErrorInfo);
            break;
        }
    }

    if (oResult && pRecorderChannelObjects)
    {
        pRecorderChannelObjects->SetTrigger(p_ubTriggerType);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ConfigureMovementStartTrigger(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE RISING_EDGE = 9;
    const DWORD DRIVE_FOLLOWS_COMMAND_STATE = 0x00001000;
    const DWORD DEFAULT_VALUE = 0x00000000;
    const DWORD DRIVE_FOLLOWS_COMMAND_MASK = 0x00001000;

    BOOL oResult = SetRecorderTriggerVariableIndex(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_STATUS_WORD, SUBINDEX_STATUS_WORD, p_pErrorInfo);

    if (oResult)
    {
        oResult = SetRecorderTriggerMode(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RISING_EDGE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerHighValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DRIVE_FOLLOWS_COMMAND_STATE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerLowValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DEFAULT_VALUE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerMask(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DRIVE_FOLLOWS_COMMAND_MASK, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ConfigureSetErrorTrigger(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE RISING_TO_VALUE_OR_FALLING_TO_VALUE = 8;
    const DWORD ERROR_STATE = 0x00000008;
    const DWORD DEFAULT_VALUE = 0x00000000;
    const DWORD ERROR_MASK = 0x00000008;

    BOOL oResult = SetRecorderTriggerVariableIndex(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_STATUS_WORD, SUBINDEX_STATUS_WORD, p_pErrorInfo);

    if (oResult)
    {
        oResult = SetRecorderTriggerMode(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RISING_TO_VALUE_OR_FALLING_TO_VALUE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerHighValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ERROR_STATE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerLowValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DEFAULT_VALUE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerMask(p_pManager, p_hDCS_Handle, p_hTransactionHandle, ERROR_MASK, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ConfigureDigitalInputTrigger(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE EXIT_OF_RANGE = 13;
    const DWORD DEFAULT_VALUE = 0x00000000;
    const DWORD LOGIC_STATE_MASK = 0xFFFFFFFF;

    BOOL oResult = SetRecorderTriggerVariableIndex(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIGITAL_INPUT_PROPERTIES, SUBINDEX_DIGITAL_INPUT_LOGIC_STATE, p_pErrorInfo);
    WORD usLogicState(0);

    if (oResult)
    {
        oResult = SetRecorderTriggerMode(p_pManager, p_hDCS_Handle, p_hTransactionHandle, EXIT_OF_RANGE, p_pErrorInfo);
    }

    if(oResult)
    {
        oResult = GetDigitalInputLogicState(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usLogicState, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerHighValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usLogicState, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerLowValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DEFAULT_VALUE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerMask(p_pManager, p_hDCS_Handle, p_hTransactionHandle, LOGIC_STATE_MASK, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ConfigureMovementEndTrigger(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, CErrorInfo* p_pErrorInfo)
{
    const BYTE RISING_EDGE = 9;
    const DWORD TARGET_REACHED_STATE = 0x00000400;
    const DWORD DEFAULT_VALUE = 0x00000000;
    const DWORD TARGET_REACHED_MASK = 0x00000400;

    BOOL oResult = SetRecorderTriggerVariableIndex(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_STATUS_WORD, SUBINDEX_STATUS_WORD, p_pErrorInfo);

    if (oResult)
    {
        oResult = SetRecorderTriggerMode(p_pManager, p_hDCS_Handle, p_hTransactionHandle, RISING_EDGE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerHighValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, TARGET_REACHED_STATE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerLowValue(p_pManager, p_hDCS_Handle, p_hTransactionHandle, DEFAULT_VALUE, p_pErrorInfo);
    }

    if (oResult)
    {
        oResult = SetRecorderTriggerMask(p_pManager, p_hDCS_Handle, p_hTransactionHandle, TARGET_REACHED_MASK, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::SetRecorderTriggerVariableIndex(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usIndex, BYTE p_ubSubIndex, CErrorInfo* p_pErrorInfo)
{
    DWORD ulTriggerVariableIndex = p_usIndex;
    ulTriggerVariableIndex <<= 8;
    ulTriggerVariableIndex += p_ubSubIndex;
    ulTriggerVariableIndex <<= 8;

    return WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_TRIGGER_VARIABLE_INDEX, ulTriggerVariableIndex, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::SetRecorderTriggerMode(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE p_ubMode, CErrorInfo* p_pErrorInfo)
{
    return WriteUInt8Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_TRIGGER_MODE, p_ubMode, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::SetRecorderTriggerHighValue(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulValue, CErrorInfo* p_pErrorInfo)
{
    return WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_TRIGGER_HIGH_VALUE, p_ulValue, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::SetRecorderTriggerLowValue(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulValue, CErrorInfo* p_pErrorInfo)
{
    return WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_TRIGGER_LOW_VALUE, p_ulValue, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::SetRecorderTriggerMask(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, DWORD p_ulMask, CErrorInfo* p_pErrorInfo)
{
    return WriteUInt32Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_TRIGGER, SUBINDEX_INTERNAL_RECORDER_TRIGGER_MASK, p_ulMask, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::SetRecorderControlword(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usControlword, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);

    oResult = WriteUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_CONTROL, SUBINDEX_INTERNAL_RECORDER_CONTROLWORD, p_usControlword, p_pErrorInfo);

    return oResult;
}

BOOL CGatewayDriveToEpos4::GetDigitalInputLogicState(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD& p_rusLogicState, CErrorInfo* p_pErrorInfo)
{
    return ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_DIGITAL_INPUT_PROPERTIES, SUBINDEX_DIGITAL_INPUT_LOGIC_STATE, &p_rusLogicState, p_pErrorInfo);
}

BOOL CGatewayDriveToEpos4::CheckRecorderStatus(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, WORD p_usStatus, BOOL& p_roState, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    WORD usRecorderStatus;

    oResult = ReadUInt16Object(p_pManager, p_hDCS_Handle, p_hTransactionHandle, INDEX_INTERNAL_RECORDER_CONTROL, SUBINDEX_INTERNAL_RECORDER_STATUSWORD, &usRecorderStatus, p_pErrorInfo);

    if (oResult && usRecorderStatus & p_usStatus)
    {
        p_roState = TRUE;
    }
    else
    {
        p_roState = FALSE;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_ReadLSSFrame(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_FRAME_SIZE = 8;

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulDataBufferLength;
    DWORD ulErrorCode = 0;
    void* pDataBuffer = NULL;
    WORD usTimeout;

    if (p_pCommand)
    {
        //Prepare data buffer
        ulDataBufferLength = MAX_FRAME_SIZE;
        pDataBuffer = malloc(ulDataBufferLength);
        memset(pDataBuffer, 0, ulDataBufferLength);

        //Get parameter data
        p_pCommand->GetParameterData(0, &usTimeout, sizeof(usTimeout));

        //ReadLSSFrame
        oResult = DCS_ReadLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usTimeout, pDataBuffer, ulDataBufferLength, &errorInfo);

        if (!oResult) ulErrorCode = errorInfo.GetErrorCode();

        //Set ReturnParameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulErrorCode, sizeof(ulErrorCode));
        p_pCommand->SetReturnParameterData(1, pDataBuffer, ulDataBufferLength);

        //Free data buffer
        if (pDataBuffer) free(pDataBuffer);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::Process_SendLSSFrame(CCommand_VCS* p_pCommand, CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle)
{
    const DWORD MAX_FRAME_SIZE = 8;

    BOOL oResult(FALSE);
    CErrorInfo errorInfo;
    DWORD ulDataBufferLength = 0;
    DWORD ulErrorCode = 0;
    void* pDataBuffer = NULL;

    if (p_pCommand)
    {
        //Prepare data buffer
        ulDataBufferLength = p_pCommand->GetParameterLength(0);

        if (ulDataBufferLength > MAX_FRAME_SIZE) ulDataBufferLength = MAX_FRAME_SIZE;

        pDataBuffer = malloc(ulDataBufferLength);

        memset(pDataBuffer, 0, ulDataBufferLength);

        //Get parameter data
        p_pCommand->GetParameterData(0, pDataBuffer, ulDataBufferLength);

        //SendLSSFrame
        oResult = DCS_SendLSSFrame(p_pManager, p_hDCS_Handle, p_hTransactionHandle, pDataBuffer, ulDataBufferLength, &errorInfo);

        if (!oResult) ulErrorCode = errorInfo.GetErrorCode();

        //Set return parameter data
        p_pCommand->SetStatus(oResult, &errorInfo);
        p_pCommand->SetReturnParameterData(0, &ulErrorCode, sizeof(ulErrorCode));
        if (pDataBuffer) free(pDataBuffer);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ReadDataBuffer(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE* p_pDataBuffer, DWORD p_ulDataBufferSizeToRead, DWORD* p_pulDataBufferSizeRead, WORD& p_rusMaxNbOfSamples, WORD& p_rusNbOfRecordedSamples, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE* pBuffer((BYTE*)malloc(RECORDER_BUFFER_SIZE));
    DWORD ulBufferSizeRead(0);
    DWORD ulDataBufferSizeRead(0);
    DWORD ulErrorCode(0);

    if (ReadDataBuffer(p_pManager, p_hDCS_Handle, p_hTransactionHandle, pBuffer, RECORDER_BUFFER_SIZE, &ulBufferSizeRead, p_pErrorInfo))
    {
        if (ExtractVariables(p_pDataBuffer, p_ulDataBufferSizeToRead, ulDataBufferSizeRead, pBuffer, ulBufferSizeRead, &ulErrorCode))
        {
            if (CalculateMaxNbOfSamples(p_rusMaxNbOfSamples, p_pErrorInfo) && CalculateNbOfRecordedSamples(p_rusNbOfRecordedSamples, p_pErrorInfo))
            {

                if (p_pulDataBufferSizeRead)
                {
                    *p_pulDataBufferSizeRead = ulDataBufferSizeRead;
                }

                oResult = TRUE;
            }
        }
        else
        {
            if (m_pErrorHandling)
            {
                m_pErrorHandling->GetError(ulErrorCode, p_pErrorInfo);
            }
        }
    }

    if (pBuffer)
    {
        free(pBuffer);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ReadDataBuffer(CDeviceCommandSetManagerBase* p_pManager, HANDLE p_hDCS_Handle, HANDLE p_hTransactionHandle, BYTE* p_pDataBuffer, DWORD p_ulDataBufferSizeToRead, DWORD* p_pulDataBufferSizeRead, CErrorInfo* p_pErrorInfo)
{
    BOOL oResult(FALSE);
    BYTE* pDataBuffer(0);
    DWORD ulNbOfBytesRead(0);
    DWORD ulNbOfBytesReadTotal(0);
    DWORD ulNbOfBytesToRead;
    DWORD ulDataBufferSizeToRead(p_ulDataBufferSizeToRead);
    WORD usIndex;;
    WORD usSubIndex;

    ulNbOfBytesToRead = ulDataBufferSizeToRead / MAX_NB_OF_DATA_RECORDER_VARIABLES;
    pDataBuffer = p_pDataBuffer;

    usIndex = INDEX_INTERNAL_RECORDER_BUFFER_VARIABLE;
    usSubIndex = SUBINDEX_INTERNAL_RECORDER_BUFFER_VARIABLE_COMPLETE;

    for (BYTE ubVariableNb(0); ubVariableNb < MAX_NB_OF_DATA_RECORDER_VARIABLES; ubVariableNb++, usIndex++)
    {
        oResult = ReadDataBufferObject(p_pManager, p_hDCS_Handle, p_hTransactionHandle, usIndex, usSubIndex, pDataBuffer, ulNbOfBytesToRead, &ulNbOfBytesRead, p_pErrorInfo);

        if (oResult)
        {
            ulNbOfBytesReadTotal += ulNbOfBytesRead;
            pDataBuffer += ulNbOfBytesRead;
        }
        else
        {
            ulNbOfBytesReadTotal = 0;
            break;
        }
    }

    if (p_pulDataBufferSizeRead)
    {
        *p_pulDataBufferSizeRead = ulNbOfBytesReadTotal;
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ExtractVariables(BYTE* p_pDataBuffer, DWORD p_ulDataBufferSizeToRead, DWORD& p_rulDataBufferSizeRead, BYTE* p_pBuffer, DWORD p_ulBufferSizeRead, DWORD* p_pulErrorCode)
{
    BOOL oResult(FALSE);
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());

    if (pRecorderChannelObjects)
    {
        if (pRecorderChannelObjects->GetNbOfConfiguredChannels() > 0)
        {
            oResult = TRUE;

            for (int i = 0; i < MAX_NB_OF_DATA_RECORDER_VARIABLES; i++)
            {
                if (pRecorderChannelObjects->IsSizeAvailable(i))
                {
                    if (!ExtractVariable(p_pDataBuffer, p_ulDataBufferSizeToRead, p_rulDataBufferSizeRead, p_pBuffer, p_ulBufferSizeRead, i, p_pulErrorCode))
                    {
                        oResult = FALSE;
                        break;
                    }
                }
            }
        }
        else
        {
            if (p_pulErrorCode)
            {
                *p_pulErrorCode = ERROR_DATA_RECORDER_NOT_CONFIGURED;
            }
        }
    }
    else
    {
        if (p_pulErrorCode)
        {
            *p_pulErrorCode = ERROR_DATA_RECORDER_NOT_CONFIGURED;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ExtractVariable(BYTE* p_pDataBuffer, DWORD p_ulDataBufferSizeToRead, DWORD& p_rulDataBufferSizeRead, BYTE* p_pBuffer, DWORD p_ulBufferSizeRead, int p_lVariableIndex, DWORD* p_pulErrorCode)
{
    const int TWO_BYTES(2);
    const int FOUR_BYTES(4);
    const WORD ONE_WORD(1);
    const WORD TWO_WORDS(2);

    BOOL oResult(FALSE);
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    DWORD ulBufferSize(0);
    DWORD ulDataBufferSize(0);
    DWORD ulDataBufferSizeRead(0);
    WORD usObjectWordSize;
    WORD usActiveChannels;
    WORD usObjectWordSizeTotal;

    if (p_pDataBuffer && p_pBuffer && pRecorderChannelObjects)
    {
        oResult = TRUE;

        usActiveChannels = pRecorderChannelObjects->GetNbOfConfiguredChannels();
        usObjectWordSizeTotal = pRecorderChannelObjects->GetTotalSizeInWord();
        usObjectWordSize = pRecorderChannelObjects->GetSizeInWord(p_lVariableIndex);

        ulBufferSize = p_ulBufferSizeRead / MAX_NB_OF_DATA_RECORDER_VARIABLES;
        ulDataBufferSize = p_ulDataBufferSizeToRead / (usObjectWordSizeTotal * usObjectWordSize);

        if ((p_rulDataBufferSizeRead + ulDataBufferSize * FOUR_BYTES) > p_ulDataBufferSizeToRead)
        {
            if (p_pulErrorCode)
            {
                *p_pulErrorCode = ERROR_BUFFER_TOO_SMALL;
            }

            oResult = FALSE;
        }

        p_pBuffer += ulBufferSize * p_lVariableIndex;
        p_pDataBuffer += p_rulDataBufferSizeRead;

        switch (usObjectWordSize)
        {
            case ONE_WORD:
            {
                oResult = ExtractVariable(p_pBuffer, ulBufferSize, TWO_BYTES, p_pDataBuffer, ulDataBufferSize * TWO_BYTES, ulDataBufferSizeRead, p_pulErrorCode);
                break;
            }
            case TWO_WORDS:
            {
                oResult = ExtractVariable(p_pBuffer, ulBufferSize, FOUR_BYTES, p_pDataBuffer, ulDataBufferSize * FOUR_BYTES, ulDataBufferSizeRead, p_pulErrorCode);
                break;
            }
            default:
            {
                if (p_pulErrorCode)
                {
                    *p_pulErrorCode = ERROR_VIRTUAL_DEVICE_BAD_OBJECT_SIZE_EXPECTED;
                }

                oResult = FALSE;
                break;
            }
        }

        p_rulDataBufferSizeRead += ulDataBufferSizeRead;
    }
    else
    {
        if (p_pulErrorCode)
        {
            *p_pulErrorCode = ERROR_DATA_RECORDER_NOT_CONFIGURED;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::ExtractVariable(BYTE* p_pBuffer, DWORD p_ulBufferSize, int p_lNbOfBytes, BYTE* p_pDataBuffer, DWORD p_ulDataBufferSize, DWORD& p_rulDataBufferSize, DWORD* p_pulErrorCode)
{
    BOOL oResult(FALSE);
    DWORD ulBufferIndex(0);
    DWORD ulSample(0);

    if (p_pBuffer && p_pDataBuffer)
    {
        oResult = TRUE;

        while (ulBufferIndex < p_ulDataBufferSize && ulBufferIndex < p_ulBufferSize && oResult)
        {
            memcpy((BYTE*)&ulSample, p_pBuffer, SAMPLE_SIZE);

            memcpy(p_pDataBuffer, (BYTE*)&ulSample, p_lNbOfBytes);
            p_pDataBuffer += p_lNbOfBytes;

            p_pBuffer += SAMPLE_SIZE;
            ulBufferIndex += SAMPLE_SIZE;
        }

        p_rulDataBufferSize = ulBufferIndex;
    }
    else
    {
        if (p_pulErrorCode)
        {
            *p_pulErrorCode = ERROR_DATA_RECORDER_NOT_CONFIGURED;
        }
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CalculateMaxNbOfSamples(WORD& p_rMaxNbOfSamples, CErrorInfo* p_pErrorInfo)
{
    const WORD MAX_RINGBUFFER_SIZE(1024);
    const WORD WORD_TO_BYTE_FACTOR(2);

    BOOL oResult(FALSE);
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    WORD usTotalSize(0);

    if (pRecorderChannelObjects)
    {
        usTotalSize = pRecorderChannelObjects->GetTotalSizeInWord() * WORD_TO_BYTE_FACTOR;
        if (usTotalSize)
        {
            p_rMaxNbOfSamples = MAX_RINGBUFFER_SIZE / usTotalSize;

            oResult = TRUE;
        }
    }

    if (!oResult && m_pErrorHandling)
    {
        m_pErrorHandling->GetError(ERROR_DATA_RECORDER_NOT_CONFIGURED, p_pErrorInfo);
    }

    return oResult;
}

BOOL CGatewayDriveToEpos4::CalculateNbOfRecordedSamples(WORD& p_rNbOfRecordedSamples, CErrorInfo* p_pErrorInfo)
{
    const WORD MAX_RINGBUFFER_SIZE(1024);
    const WORD WORD_TO_BYTE_FACTOR(2);

    BOOL oResult(FALSE);
    CRecorderChannelObjects* pRecorderChannelObjects(GetRecorderChannelObjects());
    WORD usTotalSize(0);

    if (pRecorderChannelObjects)
    {
        usTotalSize = pRecorderChannelObjects->GetTotalSizeInWord() * WORD_TO_BYTE_FACTOR;
        if (usTotalSize)
        {
            p_rNbOfRecordedSamples = MAX_RINGBUFFER_SIZE / usTotalSize;

            oResult = TRUE;
        }
    }

    if (!oResult && m_pErrorHandling)
    {
        m_pErrorHandling->GetError(ERROR_DATA_RECORDER_NOT_CONFIGURED, p_pErrorInfo);
    }

    return oResult;
}
